/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ _ZE(MB61VH010/011) ]
 */

#if XM7_VER >= 2

#include <string.h>
#include <stdio.h>
#include "xm7.h"
#include "aluline.h"
#include "event.h"
#include "display.h"
#include "subctrl.h"
#include "multipag.h"
#include "device.h"

/*
 *	O[o [N
 */
BYTE alu_command;						/* _Z R}h */
BYTE alu_color;							/* _Z J[ */
BYTE alu_mask;							/* _Z }XNrbg */
BYTE alu_cmpstat;						/* _Z rXe[^X */
BYTE alu_cmpdat[8];						/* _Z rf[^ */
BYTE alu_disable;						/* _Z ֎~oN */
BYTE alu_tiledat[3];					/* _Z ^Cp^[ */

BOOL line_busy;							/*  BUSY */
WORD line_offset;						/*  AhXItZbg */
WORD line_style;						/*  CX^C */
WORD line_x0;							/*  X0 */
WORD line_y0;							/*  Y0 */
WORD line_x1;							/*  X1 */
WORD line_y1;							/*  Y1 */
BOOL line_boost;						/*  S͕`tO */

/*
 *  X^eBbN [N
 */
static BYTE alu_vram_bank;				/* _Z VRAMoN */
static WORD line_addr_old;				/*  O񏈗VRAMAhX */
static BYTE line_mask;					/*  }XNrbg */
static WORD line_count;					/*  JE^() */
static BYTE line_count_sub;				/*  JE^() */

/*
 *	vg^Cv錾
 */
static void FASTCALL alu_cmp(WORD addr);	/* RyA */

/*
 *	ID
 */
#define	ID	MAKEID('A', 'L', 'U', ' ')
//#define ALU_LOG

/*
 *	_ZE
 *	
 */
BOOL FASTCALL aluline_init(void)
{
	/* ԏBUSYԂƂ */
	line_boost = FALSE;

	return TRUE;
}

/*
 *	_ZE
 *	N[Abv
 */
void FASTCALL aluline_cleanup(void)
{
}

/*
 *	_ZE
 *	Zbg
 */
void FASTCALL aluline_reset(void)
{
	LOG0(Normal, "Zbg");

	/* SẴWX^ */
	alu_command = 0;
	alu_color = 0;
	alu_mask = 0;
	alu_cmpstat = 0;
	memset(alu_cmpdat, 0x80, sizeof(alu_cmpdat));
	alu_disable = 0x00;
	memset(alu_tiledat, 0, sizeof(alu_tiledat));

	line_busy = FALSE;
	line_offset = 0;
	line_style = 0;
	line_x0 = 0;
	line_y0 = 0;
	line_x1 = 0;
	line_y1 = 0;

	alu_vram_bank = 0;
	line_addr_old = 0xffff;
	line_mask = 0xff;
	line_count = 0;
	line_count_sub = 0;
}

/*-[ _Z ]-------------------------------------------------------------*/

/*
 *	_Z
 *	VRAMǂݏo
 */
static BYTE FASTCALL alu_read(WORD addr)
{
	/* ANeBuv[łȂꍇA0xFFԂ */
	if (multi_page & (1 << alu_vram_bank)) {
		return 0xff;
	}

#if XM7_VER >= 3
	if (mode400l) {
		if (addr >= 0x8000) {
			/* 400C0x8000`0xBFFFɂ݂̓Ȃ */
			return 0xff;
		}
		else {
			return vram_ablk[alu_vram_bank * 0x8000 + addr];
		}
	}
	else {
		return vram_aptr[alu_vram_bank * 0x8000 + addr];
	}
#else
	return vram_aptr[alu_vram_bank * 0x4000 + addr];
#endif
}

/*
 *	_Z
 *	VRAMǂݏo(v[w)
 */
static BYTE FASTCALL alu_read_plane(WORD addr, int plane)
{
	/* ANeBuv[łȂꍇA0xFFԂ */
	if (multi_page & (1 << plane)) {
		return 0xff;
	}

#if XM7_VER >= 3
	if (mode400l) {
		if (addr >= 0x8000) {
			/* 400C0x8000`0xBFFFɂ݂̓Ȃ */
			return 0xff;
		}
		else {
			return vram_ablk[plane * 0x8000 + addr];
		}
	}
	else {
		return vram_aptr[plane * 0x8000 + addr];
	}
#else
	return vram_aptr[plane * 0x4000 + addr];
#endif
}

/*
 *	_Z
 *	VRAM
 */
static void FASTCALL alu_write(WORD addr, BYTE dat)
{
	/* ANeBuv[łȂꍇNOP */
	if (multi_page & (1 << alu_vram_bank)) {
		return;
	}

#if XM7_VER >= 3
	if (mode400l) {
		/* 400C0x8000`0xBFFFɂ݂̓Ȃ̂NOP */
		if (addr < 0x8000) {
			if (vram_ablk[alu_vram_bank * 0x8000 + addr] != dat) {
				vram_ablk[alu_vram_bank * 0x8000 + addr] = dat;
				vram_notify(addr, dat);
			}
		}
	}
	else {
		if (vram_aptr[alu_vram_bank * 0x8000 + addr] != dat) {
			vram_aptr[alu_vram_bank * 0x8000 + addr] = dat;
			vram_notify(addr, dat);
		}
	}
#else
	if (vram_aptr[alu_vram_bank * 0x4000 + addr] != dat) {
		vram_aptr[alu_vram_bank * 0x4000 + addr] = dat;
		vram_notify(addr, dat);
	}
#endif
}

/*
 *	_Z
 *	݃Tu(rݕt)
 */
static void FASTCALL alu_writesub(WORD addr, BYTE dat)
{
	BYTE temp;

	/* ɏ݉\ */
	if ((alu_command & 0x40) == 0) {
		alu_write(addr, dat);
		return;
	}

	/* CR[݂ANOTCR[݂ */
	if (alu_command & 0x20) {
		/* NOTCR[ŏ */
		temp = alu_read(addr);
		temp &= alu_cmpstat;
		dat &= (BYTE)(~alu_cmpstat);
	}
	else {
		/* CR[ŏ */
		temp = alu_read(addr);
		temp &= (BYTE)(~alu_cmpstat);
		dat &= alu_cmpstat;
	}

	/* f[^VRAMɏ */
	alu_write(addr, (BYTE)(temp | dat));
}

/*
 *	_Z
 *	PSET
 */
static void FASTCALL alu_pset(WORD addr)
{
	BYTE dat;
	BYTE mask;
	BYTE bit;

	/* VRAMItZbg擾Arbgʒu */
	alu_vram_bank = 0;
	bit = 0x01;

#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif

	/* rݎ́Aɔrf[^擾 */
	if (alu_command & 0x40) {
		alu_cmp(addr);
	}

	/* eoNɑ΂ďs(oN̓XLbv) */
	while (alu_vram_bank < 3) {
		if (!(alu_disable & bit)) {
			/* ZJ[f[^Af[^쐬 */
			if (alu_color & bit) {
				dat = 0xff;
			}
			else {
				dat = 0;
			}

			/* ZȂ(PSET) */
			mask = alu_read(addr);

			/* }XNrbg̏ */
			dat &= (BYTE)(~alu_mask);
			mask &= alu_mask;
			dat |= mask;

			/*  */
			alu_writesub(addr, dat);
		}

		/* ̃oN */
		alu_vram_bank ++;
		bit <<= 1;
	}
}

/*
 *	_Z
 *	R}h1(֎~EF)
 */
static void FASTCALL alu_prohibit(WORD addr)
{
	BYTE dat;
	BYTE mask;
	BYTE bit;

	/* VRAMItZbg擾Arbgʒu */
	alu_vram_bank = 0;
	bit = 0x01;

#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif

	/* rݎ́Aɔrf[^擾 */
	if (alu_command & 0x40) {
		alu_cmp(addr);
	}

	/* eoNɑ΂ďs(oN̓XLbv) */
	while (alu_vram_bank < 3) {
		if (!(alu_disable & bit)) {
			/* ZȂ */
			mask = alu_read(addr);

			/* }XNrbg̏ */
			dat = (mask & alu_mask);

			/*  */
			alu_writesub(addr, dat);
		}

		/* ̃oN */
		alu_vram_bank ++;
		bit <<= 1;
	}
}

/*
 *	_Z
 *	OR
 */
static void FASTCALL alu_or(WORD addr)
{
	BYTE dat;
	BYTE mask;
	BYTE bit;

	/* VRAMItZbg擾Arbgʒu */
	alu_vram_bank = 0;
	bit = 0x01;

#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif


	/* rݎ́Aɔrf[^擾 */
	if (alu_command & 0x40) {
		alu_cmp(addr);
	}

	/* eoNɑ΂ďs(oN̓XLbv) */
	while (alu_vram_bank < 3) {
		if (!(alu_disable & bit)) {
			/* ZJ[f[^Af[^쐬 */
			if (alu_color & bit) {
				dat = 0xff;
			}
			else {
				dat = 0;
			}

			/* Z */
			mask = alu_read(addr);
			dat |= mask;

			/* }XNrbg̏ */
			dat &= (BYTE)(~alu_mask);
			mask &= alu_mask;
			dat |= mask;

			/*  */
			alu_writesub(addr, dat);
		}

		/* ̃oN */
		alu_vram_bank ++;
		bit <<= 1;
	}
}

/*
 *	_Z
 *	AND
 */
static void FASTCALL alu_and(WORD addr)
{
	BYTE dat;
	BYTE mask;
	BYTE bit;

	/* VRAMItZbg擾Arbgʒu */
	alu_vram_bank = 0;
	bit = 0x01;

#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif

	/* rݎ́Aɔrf[^擾 */
	if (alu_command & 0x40) {
		alu_cmp(addr);
	}

	/* eoNɑ΂ďs(oN̓XLbv) */
	while (alu_vram_bank < 3) {
		if (!(alu_disable & bit)) {
			/* ZJ[f[^Af[^쐬 */
			if (alu_color & bit) {
				dat = 0xff;
			}
			else {
				dat = 0;
			}

			/* Z */
			mask = alu_read(addr);
			dat &= mask;

			/* }XNrbg̏ */
			dat &= (BYTE)(~alu_mask);
			mask &= alu_mask;
			dat |= mask;

			/*  */
			alu_writesub(addr, dat);
		}

		/* ̃oN */
		alu_vram_bank ++;
		bit <<= 1;
	}
}

/*
 *	_Z
 *	XOR
 */
static void FASTCALL alu_xor(WORD addr)
{
	BYTE dat;
	BYTE mask;
	BYTE bit;

	/* VRAMItZbg擾Arbgʒu */
	alu_vram_bank = 0;
	bit = 0x01;

#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif

	/* rݎ́Aɔrf[^擾 */
	if (alu_command & 0x40) {
		alu_cmp(addr);
	}

	/* eoNɑ΂ďs(oN̓XLbv) */
	while (alu_vram_bank < 3) {
		if (!(alu_disable & bit)) {
			/* ZJ[f[^Af[^쐬 */
			if (alu_color & bit) {
				dat = 0xff;
			}
			else {
				dat = 0;
			}

			/* Z */
			mask = alu_read(addr);
			dat ^= mask;

			/* }XNrbg̏ */
			dat &= (BYTE)(~alu_mask);
			mask &= alu_mask;
			dat |= mask;

			/*  */
			alu_writesub(addr, dat);
		}

		/* ̃oN */
		alu_vram_bank ++;
		bit <<= 1;
	}
}

/*
 *	_Z
 *	NOT
 */
static void FASTCALL alu_not(WORD addr)
{
	BYTE dat;
	BYTE mask;
	BYTE bit;

	/* VRAMItZbg擾Arbgʒu */
	alu_vram_bank = 0;
	bit = 0x01;

#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif

	/* rݎ́Aɔrf[^擾 */
	if (alu_command & 0x40) {
		alu_cmp(addr);
	}

	/* eoNɑ΂ďs(oN̓XLbv) */
	while (alu_vram_bank < 3) {
		if (!(alu_disable & bit)) {
			/* Z(NOT) */
			mask = alu_read(addr);
			dat = (BYTE)(~mask);

			/* }XNrbg̏ */
			dat &= (BYTE)(~alu_mask);
			mask &= alu_mask;
			dat |= mask;

			/*  */
			alu_writesub(addr, dat);
		}

		/* ̃oN */
		alu_vram_bank ++;
		bit <<= 1;
	}
}

/*
 *	_Z
 *	^CyCg
 */
static void FASTCALL alu_tile(WORD addr)
{
	BYTE bit;
	BYTE mask;
	BYTE dat;

	/* VRAMItZbg擾Arbgʒu */
	alu_vram_bank = 0;
	bit = 0x01;

#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
	 	addr &= 0x3fff;
	}
#else
	 addr &= 0x3fff;
#endif

	/* rݎ́Aɔrf[^擾 */
	if (alu_command & 0x40) {
		alu_cmp(addr);
	}

	/* eoNɑ΂ďs(oN̓XLbv) */
	while (alu_vram_bank < 3) {
		if (!(alu_disable & bit)) {
			/* f[^쐬 */
			dat = alu_tiledat[alu_vram_bank];

			/* }XNrbg̏ */
			dat &= (BYTE)(~alu_mask);
			mask = alu_read(addr);
			mask &= alu_mask;
			dat |= mask;

			/*  */
			alu_writesub(addr, dat);
		}

		/* ̃oN */
		alu_vram_bank ++;
		bit <<= 1;
	}
}

/*
 *	_Z
 *	RyA
 */
static void FASTCALL alu_cmp(WORD addr)
{
	BYTE color;
	BYTE bit;
	int i, j;
	BOOL flag;
	BYTE dat;
	BYTE b, r, g;
	BYTE disflag;

	/* AhX}XN */
#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif

	/* J[f[^擾 */
	b = alu_read_plane(addr, 0);
	r = alu_read_plane(addr, 1);
	g = alu_read_plane(addr, 2);

	/* oNfBZ[ul(_]΍) */
	disflag = (BYTE)((~alu_disable) & 0x07);

	/* rKv */
	dat = 0;
	bit = 0x80;
	for (i=0; i<8; i++) {
		/* F쐬 */
		color = 0;
		if (b & bit) {
			color |= 0x01;
		}
		if (r & bit) {
			color |= 0x02;
		}
		if (g & bit) {
			color |= 0x04;
		}

		/* 8̐FXbg܂āAǂꂩv̂邩 */
		flag = FALSE;
		for (j=0; j<8; j++) {
			if ((alu_cmpdat[j] & 0x80) == 0) {
				if ((alu_cmpdat[j] & disflag) == (color & disflag)) {
					flag = TRUE;
					break;
				}
			}
		}

		/* CR[1ݒ */
		if (flag) {
			dat |= bit;
		}

		/*  */
		bit >>= 1;
	}

	/* f[^ݒ */
	alu_cmpstat = dat;
}

/*-[  ]-------------------------------------------------------------*/

/*
 *	
 *	_ZHN
 */
void FASTCALL aluline_exec(WORD addr)
{
	/* AhX`FbN */
	if (addr >= 0x8000) {
		line_mask = 0xff;
		return;
	}

	/* }XNݒ */
	alu_mask = line_mask;
	line_mask = 0xff;

	/* _Z */
	switch (alu_command & 0x07) {
		/* PSET */
		case 0:
			alu_pset(addr);
			break;
		/* ֎~(F) */
		case 1:
			alu_prohibit(addr);
			break;
		/* OR */
		case 2:
			alu_or(addr);
			break;
		/* AND */
		case 3:
			alu_and(addr);
			break;
		/* XOR */
		case 4:
			alu_xor(addr);
			break;
		/* NOT */
		case 5:
			alu_not(addr);
			break;
		/* ^CyCg */
		case 6:
			alu_tile(addr);
			break;
		/* RyA */
		case 7:
			alu_cmp(addr);
			break;
	}

	/* JEgAbv */
	line_count ++;
}

/*
 *	
 *	_`
 */
static void FASTCALL aluline_pset(int x, int y)
{
	WORD addr;
	static BYTE mask[] = {0x7f, 0xbf, 0xdf, 0xef, 0xf7, 0xfb, 0xfd, 0xfe};

	/* _Z̃f[^oXɓ̂ŁA_ZonKv */
	if (!(alu_command & 0x80)) {
		return;
	}

	/* ʃ[hAAhXZo */
	if (SCR_MODE_ANALOG) {
		addr = (WORD)(y * 40 + (x >> 3));
	}
	else {
		addr = (WORD)(y * 80 + (x >> 3));
	}

	/* ItZbg */
	addr += line_offset;
#if XM7_VER >= 3
	if (mode400l) {
		addr &= 0x7fff;
	}
	else {
		addr &= 0x3fff;
	}
#else
	addr &= 0x3fff;
#endif

	/* ÕAhXƍ̃AhXvȂꍇ_Zs */
	if (line_addr_old != addr) {
		/* _Z𓮂 */
		aluline_exec(line_addr_old);
		line_addr_old = addr;
	}

	/* CX^C */
	/* X^CrbgĂꍇ̂ */
	if (line_style & 0x8000) {
		/* }XNݒ */
		line_mask &= mask[x & 0x07];
	}

	/* CX^CWX^[e[g */
	line_style = (WORD)((line_style << 1) | (line_style >> 15));
}

/*
 *	`(DDA)
 *	Ql:6809@BɂOtBbN̋Z@ 6 (h!FM '86/5)
 */
static void FASTCALL aluline_line(void)
{
	int x1, x2, y1, y2;
	int dx, dy, ux, uy, r;

	/* f[^擾 */
	x1 = (int)line_x0;
	x2 = (int)line_x1;
	y1 = (int)line_y0;
	y2 = (int)line_y1;

	/* JE^ */
	line_count = 0;
	line_addr_old = 0xffff;
	line_mask = 0xff;

	/* ľvZ */
	dx = x2 - x1;
	dy = y2 - y1;
	if (dx < 0) {
		ux = -1;
		dx = -dx;
	}
	else {
		ux = 1;
	}
	if (dy < 0) {
		uy = -1;
		dy = -dy;
	}
	else {
		uy = 1;
	}

	if ((dx == 0) && (dy == 0)) {
		/* P_` */
		aluline_pset(x1, y1);
	}
	else if (dx == 0) {
		/* X̏ꍇ () */
		for (;;) {
			aluline_pset(x1, y1);
			if (y1 == y2) {
				break;
			}
			y1 += uy;
		}
	}
	else if (dy == 0) {
		/* Y̏ꍇ () */
		for (;;) {
			aluline_pset(x1, y1);
			if (x1 == x2) {
				break;
			}
			x1 += ux;
		}
	}
	else if (dx >= dy) {
		/* C[v1 (DX >= DY) */
		r = dx >> 1;
		for (;;) {
			aluline_pset(x1, y1);
			if (x1 == x2) {
				break;
			}

			x1 += ux;
			r -= dy;
			if (r < 0) {
				r += dx;
				y1 += uy;
			}
		}
	}
	else {
		/* C[v2 (DX < DY) */
		r = dy >> 1;
		for (;;) {
			aluline_pset(x1, y1);
			if (y1 == y2) {
				break;
			}

			y1 += uy;
			r -= dx;
			if (r < 0) {
				r += dy;
				x1 += ux;
			}
		}
	}

	/* Ō̃oCg̘_Z */
	aluline_exec(line_addr_old);
}


/*
 *	
 *	Cxg
 */
static BOOL FASTCALL aluline_event(void)
{
	/* ԂREADYɂ */
	line_busy = FALSE;

	schedule_delevent(EVENT_LINE);
	return TRUE;
}

/*-[ }bvhI/O ]----------------------------------------------------*/

/*
 *	_ZE
 *	PoCgǂݏo
 */
BOOL FASTCALL aluline_readb(WORD addr, BYTE *dat)
{
	/* f`FbN */
	if (model_type < MODEL_FM77AV) {
		return FALSE;
	}

	switch (addr) {
		/* _Z R}h */
		case 0xd410:
			*dat = alu_command;
			return TRUE;

		/* _Z J[ */
		case 0xd411:
			*dat = alu_color;
			return TRUE;

		/* _Z }XNrbg */
		case 0xd412:
			*dat = alu_mask;
			return TRUE;

		/* _Z rXe[^X */
		case 0xd413:
			*dat = alu_cmpstat;
			return TRUE;

		/* _Z oN */
		case 0xd41b:
			*dat = alu_disable;
			return TRUE;
	}

	/* _Z rf[^ */
	if ((addr >= 0xd413) && (addr <= 0xd41a)) {
		*dat = 0xff;
		return TRUE;
	}

	/* _Z ^Cp^[ */
	if ((addr >= 0xd41c) && (addr <= 0xd41e)) {
		*dat = 0xff;
		return TRUE;
	}

	/*  */
	if ((addr >= 0xd420) && (addr <= 0xd42b)) {
		*dat = 0xff;
		return TRUE;
	}

	return FALSE;
}

/*
 *	_ZE
 *	PoCg
 */
BOOL FASTCALL aluline_writeb(WORD addr, BYTE dat)
{
	DWORD tmp;

	/* f`FbN */
	if (model_type < MODEL_FM77AV) {
		return FALSE;
	}

	switch (addr) {
		/* _Z R}h */
		case 0xd410:
			alu_command = dat;
			return TRUE;

		/* _Z J[ */
		case 0xd411:
			alu_color = dat;
			return TRUE;

		/* _Z }XNrbg */
		case 0xd412:
			alu_mask = dat;
			return TRUE;

		/* _Z oN */
		case 0xd41b:
			alu_disable = dat;
			return TRUE;

		/*  AhXItZbg(A1璍) */
		case 0xd420:
			line_offset &= 0x01fe;
			line_offset |= (WORD)((dat * 512) & 0x3e00);
			return TRUE;
		case 0xd421:
			line_offset &= 0x3e00;
			line_offset |= (WORD)(dat * 2);
			return TRUE;

		/*  CX^C */
		case 0xd422:
			line_style &= 0xff;
			line_style |= (WORD)(dat * 256);
			return TRUE;
		case 0xd423:
			line_style &= 0xff00;
			line_style |= (WORD)dat;
			return TRUE;

		/*  X0 */
		case 0xd424:
			line_x0 &= 0xff;
			line_x0 |= (WORD)(dat * 256);
			line_x0 &= 0x03ff;
			return TRUE;
		case 0xd425:
			line_x0 &= 0xff00;
			line_x0 |= (WORD)dat;
			return TRUE;

		/*  Y0 */
		case 0xd426:
			line_y0 &= 0xff;
			line_y0 |= (WORD)(dat * 256);
			line_y0 &= 0x01ff;
			return TRUE;
		case 0xd427:
			line_y0 &= 0xff00;
			line_y0 |= (WORD)dat;
			return TRUE;

		/*  X1 */
		case 0xd428:
			line_x1 &= 0xff;
			line_x1 |= (WORD)(dat * 256);
			line_x1 &= 0x03ff;
			return TRUE;
		case 0xd429:
			line_x1 &= 0xff00;
			line_x1 |= (WORD)dat;
			return TRUE;

		/*  Y1 */
		case 0xd42a:
			line_y1 &= 0xff;
			line_y1 |= (WORD)(dat * 256);
			line_y1 &= 0x01ff;
			return TRUE;
		case 0xd42b:
			line_y1 &= 0xff00;
			line_y1 |= (WORD)dat;

			/* ŒԃX^[g */
			aluline_line();

			/* ͈A΂炭BUSYɂĂ */
			/* BUSYԂɂ鎞ԂZo(1oCg`=1/16secƉ) */
			tmp = (DWORD)(line_count >> 4);

			/* BUSYԂ̏s */
			line_count_sub += (BYTE)(line_count & 0x0f);
			if (line_count_sub >= 0x10) {
				tmp ++;
				line_count_sub &= (BYTE)0x0f;
			}

			/* vZʂ1sȏ̏ꍇBUSYɂ */
			if (tmp > 0) {
				line_busy = TRUE;
				schedule_setevent(EVENT_LINE, tmp, aluline_event);
			}

			return TRUE;
	}

	/* _Z rf[^ */
	if ((addr >= 0xd413) && (addr <= 0xd41a)) {
		alu_cmpdat[addr - 0xd413] = dat;
		return TRUE;
	}

	/* _Z ^Cp^[ */
	if ((addr >= 0xd41c) && (addr <= 0xd41e)) {
		alu_tiledat[addr - 0xd41c] = dat;
		return TRUE;
	}

	return FALSE;
}

/*
 *	_ZE
 *	VRAM_~[[h
 */
void FASTCALL aluline_extrb(WORD addr)
{
	/* _ZL */
	if (alu_command & 0x80) {
		/* R}h */
		switch (alu_command & 0x07) {
			/* PSET */
			case 0:
				alu_pset(addr);
				break;
			/* ֎~(F) */
			case 1:
				alu_prohibit(addr);
				break;
			/* OR */
			case 2:
				alu_or(addr);
				break;
			/* AND */
			case 3:
				alu_and(addr);
				break;
			/* XOR */
			case 4:
				alu_xor(addr);
				break;
			/* NOT */
			case 5:
				alu_not(addr);
				break;
			/* ^CyCg */
			case 6:
				alu_tile(addr);
				break;
			/* RyA */
			case 7:
				alu_cmp(addr);
				break;
		}
	}
}

/*-[ t@CI/O ]----------------------------------------------------------*/

/*
 *	_ZE
 *	Z[u
 */
BOOL FASTCALL aluline_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_byte_write(fileh, alu_command)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, alu_color)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, alu_mask)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, alu_cmpstat)) {
		return FALSE;
	}
	if (!file_write(fileh, alu_cmpdat, sizeof(alu_cmpdat))) {
		return FALSE;
	}
	if (!file_byte_write(fileh, alu_disable)) {
		return FALSE;
	}
	if (!file_write(fileh, alu_tiledat, sizeof(alu_tiledat))) {
		return FALSE;
	}

	if (!file_bool_write(fileh, line_busy)) {
		return FALSE;
	}

	if (!file_word_write(fileh, line_offset)) {
		return FALSE;
	}
	if (!file_word_write(fileh, line_style)) {
		return FALSE;
	}
	if (!file_word_write(fileh, line_x0)) {
		return FALSE;
	}
	if (!file_word_write(fileh, line_y0)) {
		return FALSE;
	}
	if (!file_word_write(fileh, line_x1)) {
		return FALSE;
	}
	if (!file_word_write(fileh, line_y1)) {
		return FALSE;
	}

	/* ]o[WƂ̌݊p_~[ */
	if (!file_byte_write(fileh, 0)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	_ZE
 *	[h
 */
BOOL FASTCALL aluline_load(int fileh, int ver)
{
	BYTE tmp;

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_byte_read(fileh, &alu_command)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &alu_color)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &alu_mask)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &alu_cmpstat)) {
		return FALSE;
	}
	if (!file_read(fileh, alu_cmpdat, sizeof(alu_cmpdat))) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &alu_disable)) {
		return FALSE;
	}
	if (!file_read(fileh, alu_tiledat, sizeof(alu_tiledat))) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &line_busy)) {
		return FALSE;
	}

	if (!file_word_read(fileh, &line_offset)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &line_style)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &line_x0)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &line_y0)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &line_x1)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &line_y1)) {
		return FALSE;
	}

	/* ]o[WƂ̌݊p_~[ */
	if (!file_byte_read(fileh, &tmp)) {
		return FALSE;
	}

	/* Cxg */
	schedule_handle(EVENT_LINE, aluline_event);

	return TRUE;
}

#endif /* XM7_VER >= 2 */
