/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ {J[h / gRAMJ[h-64 / gTuVXeROM ]
 */

#if XM7_VER >= 2

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "xm7.h"
#include "device.h"
#include "kanji.h"
#include "jcard.h"

/*
 *	O[o [N
 */
BYTE *dicrom;							/* ROM $40000 */
BYTE *dicram;							/* wKRAM $2000 */
BYTE *extram_b;							/* gRAM $10000 */

BYTE dicrom_bank;						/* ROMoNԍ */
#if XM7_VER >= 3
BOOL extrom_sel;						/* ROM/gROMItO */
#endif
BOOL dicrom_en;							/* ROMANeBu */
BOOL dicram_en;							/* wKRAMANeBu */

#if XM7_VER >= 3
BYTE *extrom;							/* gROM $20000 */
#endif

BOOL jcard_available;					/* {J[hgp\tO */
BYTE jcard_mode;						/* {J[h샂[h */

/*
 *	ID
 */
#define	ID	MAKEID('J', 'C', 'A', 'R')
//#define JCARD_LOG

/*
 *	{J[h
 *	
 */
BOOL FASTCALL jcard_init(void)
{
	/* [NGA */
	extram_b = NULL;
#ifndef JSUB
	dicrom = NULL;
#endif
	dicram = NULL;
#if XM7_VER >= 3
	extrom = NULL;
#endif
#ifndef JSUB
	jcard_available = TRUE;
#endif
	jcard_mode = JCARD_ON;

	/* { gRAM */
	extram_b = (BYTE *)malloc(0x10000);
	if (extram_b == NULL) {
#if XM7_VER == 2
		jcard_available = FALSE;
		return TRUE;
#else
		return FALSE;
#endif
	}

	/* ROM */
#ifdef JSUB
#if XM7_VER >= 3
	if (!jcard_available) {
		return FALSE;
	}
#endif
#else
	dicrom = (BYTE *)malloc(0x40000);
	if (dicrom == NULL) {
#if XM7_VER == 2
		jcard_available = FALSE;
		return TRUE;
#else
		return FALSE;
#endif
	}
	if (!file_load(DICT_ROM, dicrom, 0x40000)) {
#if XM7_VER == 2
		jcard_available = FALSE;
#else
		return FALSE;
#endif
	}
#endif

#if XM7_VER == 2 && !defined(JSUB)
	/* 2ROMȂΎgps */
	if (kanji_rom2_type = KANJI_ROM_NONE) {
		jcard_available = FALSE;
	}
#endif

	/* wKRAMǂݍ */
	dicram = (BYTE *)malloc(0x2000);
	if (dicram == NULL) {
#if XM7_VER == 2
		jcard_available = FALSE;
		return TRUE;
#else
		return FALSE;
#endif
	}
	if (!file_load(DICT_RAM, dicram, 0x2000)) {
		/* t@C݂ȂB */
		memset(dicram, 0xff, 0x2000);
	}

#if XM7_VER >= 3
	/* gROM */
	extrom = (BYTE *)malloc(0x20000);
	if (extrom == NULL) {
		return FALSE;
	}
	memset(extrom, 0xff, 0x20000);
	if (!file_load(EXTSUB_ROM, extrom, 0xc000)) {
		return FALSE;
	}
	else {
		/* oN56`63BASIC ROM̓eRs[ */
		memcpy(&extrom[0x18000], basic_rom, 0x8000);
	}
#endif

	return TRUE;
}

/*
 *	{J[h
 *	N[Abv
 */
void FASTCALL jcard_cleanup(void)
{
#if 0
	ASSERT(extram_b);
	ASSERT(dicram);
	ASSERT(dicrom);
	ASSERT(extrom);
#endif

	/* rŎsꍇl */
#if XM7_VER >= 3
	if (extrom) {
		free(extrom);
	}
#endif
	if (dicram) {
		/* wKRAM̓et@Cɏo */
		file_save(DICT_RAM, dicram, 0x2000);
		free(dicram);
	}
#ifdef JSUB
	dicrom = NULL;
#endif
	if (dicrom) {
		free(dicrom);
	}
	if (extram_b) {
		free(extram_b);
	}
}

/*
 *	{J[h
 *	Zbg
 */
void FASTCALL jcard_reset(void)
{
	LOG0(Normal, "Zbg");

	dicrom_bank = 0;
	dicram_en = FALSE;
	dicrom_en = FALSE;
#if XM7_VER >= 3
	extrom_sel = FALSE;
#endif
}

/*
 *	{J[h
 *	PoCgǂݍ
 */
BYTE FASTCALL jcard_readb(WORD addr)
{
	DWORD dicrom_addr;

	/* FM77AVV[Ŷ݃T|[g */
	if (model_type < MODEL_FM77AV) {
		LOG1(Warning, "{J[h̃oCgǂݍ $2%04X", addr);
		return 0xff;
	}

	/* {J[hgȂꍇ0xffԂ */
#if XM7_VER >= 3
	if ((model_type != MODEL_FM77AV40EX) && (jcard_mode == JCARD_OFF)) {
#else
	if (jcard_mode == JCARD_OFF) {
#endif
		LOG1(Warning, "{J[h̃oCgǂݍ $2%04X", addr);
		return 0xff;
	}

	/* $28000-$29FFF : wKRAM */
	if ((addr >= 0x8000) && (addr < 0xa000)) {
		if (dicram_en) {
			return dicram[addr - 0x8000];
		}
	}

	/* $2E000-$2EFFF : ROM or gROM */
	if ((addr >= 0xe000) && (addr < 0xf000)) {
		/* ROML */
		if (dicrom_en) {
			addr &= (WORD)0x0fff;
			dicrom_addr = (dicrom_bank << 12);

#if XM7_VER >= 3
			/* gROML */
			if (extrom_sel) {
				/* oN0`31 : 1(JIS78N) */
				if (dicrom_bank < 32) {
					return kanji_romn[addr | dicrom_addr];
				}

				/* oN32`43 : gTuVXeROM(extsub.rom) */
				/* oN56`63 : F-BASIC V3.0 ROM ($8000-$FFFF) */
				return extrom[addr | (dicrom_addr - 0x20000)];
			}
#endif

			/* ROM */
			return dicrom[addr | dicrom_addr];
		}
	}

	/* gRAM */
	return extram_b[addr];
}

/*
 *	{J[h
 *	PoCg
 */
void FASTCALL jcard_writeb(WORD addr, BYTE dat)
{
	/* FM77AVV[Ŷ݃T|[g */
	if (model_type < MODEL_FM77AV) {
		LOG2(Warning, "{J[h̃oCg $2%04X <- $%02X", addr, dat);
		return;
	}

	/* {J[hgȂꍇ͉Ȃ */
#if XM7_VER >= 3
	if ((model_type != MODEL_FM77AV40EX) && (jcard_mode == JCARD_OFF)) {
#else
	if (jcard_mode == JCARD_OFF) {
#endif
		LOG2(Warning, "{J[h̃oCg $2%04X <- $%02X", addr, dat);
		return;
	}

	/* $28000-$29FFF : wKRAM */
	if ((addr >= 0x8000) && (addr < 0xa000)) {
		if (dicram_en) {
			dicram[addr - 0x8000] = dat;
			return;
		}
	}

	/* gRAM (ROM̑IԂɊւ炸݉\) */
	extram_b[addr] = dat;
}

/*
 *	{J[h
 *	Z[u
 */
BOOL FASTCALL jcard_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_byte_write(fileh, dicrom_bank)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if (!file_bool_write(fileh, extrom_sel)) {
		return FALSE;
	}
#endif
	if (!file_bool_write(fileh, dicrom_en)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, dicram_en)) {
		return FALSE;
	}
	if (!file_write(fileh, extram_b, 0x10000)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, jcard_mode)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	{J[h
 *	[h
 */
BOOL FASTCALL jcard_load(int fileh, int ver, BOOL bflag)
{
	BOOL tmp;

	LOG0(Normal, "[h");

	/* o[W`FbN */
	if ((ver < 721) && !bflag) {
		dicrom_bank = 0;
#if XM7_VER >= 3
		extrom_sel = FALSE;
#endif
		dicram_en = FALSE;
		dicrom_en = FALSE;
		return TRUE;
	}

	if (!file_byte_read(fileh, &dicrom_bank)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if (((ver >= 800) && (ver <= 999)) || (ver >= 1400)) {
		if (!file_bool_read(fileh, &extrom_sel)) {
			return FALSE;
		}
	}
	else {
		extrom_sel = FALSE;
	}
#endif
	if (!file_bool_read(fileh, &dicrom_en)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &dicram_en)) {
		return FALSE;
	}
	if (!file_read(fileh, extram_b, 0x10000)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if (((ver >= 1202) && (ver <= 1399)) || (ver >= 1402)) {
#else
	if (ver >= 1202) {
#endif
		if (!file_byte_read(fileh, &jcard_mode)) {
			return FALSE;
		}
	}
	else if (((ver >= 721) && (ver <= 799)) ||
			 ((ver >= 1200) && (ver <= 1201)) ||
#if XM7_VER >= 3
			 ((ver >= 921) && (ver <= 999)) ||
			 ((ver >= 1400) && (ver <= 1401)) ||
#endif
			 bflag) {
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
		if (tmp) {
			jcard_mode = JCARD_ON;
		}
		else {
			jcard_mode = JCARD_OFF;
		}
	}
	else {
		jcard_mode = JCARD_ON;
	}
	if (!jcard_available) {
		/* {J[hgps̏ꍇfBZ[u */
		if (jcard_mode != JCARD_OFF) {
			jcard_mode = JCARD_RAM;
		}

		/* {J[hɃJ[hgp̃Xe[g͓ǂ߂Ȃ */
		if (dicrom_bank != 0 || dicram_en || dicrom_en) {
			return FALSE;
		}
	}
#if XM7_VER >= 3
	/* AV40EXȊO */
	if (model_type != MODEL_FM77AV40EX) {
		if (extrom_sel) {
			return FALSE;
		}
	}
#endif

	return TRUE;
}

#endif	/* XM7_VER >= 2 */
