/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ {ʐMJ[h ]
 */

#ifdef JSUB

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "jsubsys.h"
#include "kanji.h"
#include "mainetc.h"
#include "device.h"
#include "event.h"
#include "jcard.h"

/*
 *	O[o [N
 */
cpu6809_t jsubcpu;					/* {TuCPUWX^ */
BYTE *jsubrom;						/* {Tuj^ROM */
BYTE *jdicrom;						/* ROM({ʐMJ[h) */
BYTE *jsub_sram;					/* [NSRAM */
BOOL jsub_available;				/* {TuVXegp\tO */
BOOL jsub_enable;					/* {TuVXegptO */
BOOL jsub_haltflag;					/* {TuVXeHALTtO */
BYTE jsub_dicbank;					/* ROMoN */
BYTE jsub_address;					/* RCBAhXJE^ */
WORD jsub_kanji_addr;				/* ROMAhXWX^ */
BOOL jsub_kanji_dipsw;				/* ROM DIPSW({ʐMJ[h) */

/*
 *	vg^Cv錾
 */
void jsub_reset(void);
void jsub_line(void);
void jsub_exec(void);
BOOL jsubcpu_event(void);

/*
 *	ID
 */
#define	ID	MAKEID('J', 'S', 'Y', 'S')
//#define JSUBSYS_LOG

/*
 *	{TuCPU
 *	
 */
BOOL FASTCALL jsubsys_init(void)
{
	/* xASăNA */
	jsubrom = NULL;
	jdicrom = NULL;
	jsub_sram = NULL;
	jsub_available = TRUE;
	jsub_enable = TRUE;
	jsub_address = 0;
	jsub_kanji_dipsw = FALSE;
#if XM7_VER >= 2
	dicrom = NULL;
	jcard_available = TRUE;
#endif

	/* m */
	jsubrom = (BYTE *)malloc(0x4000);
	if (jsubrom == NULL) {
		return FALSE;
	}
	jdicrom = (BYTE *)malloc(0x40000);
	if (jdicrom == NULL) {
		return FALSE;
	}
#if XM7_VER >= 2
	dicrom = jdicrom;
#endif
	jsub_sram = (BYTE *)malloc(0x2000);
	if (jsub_sram == NULL) {
		return FALSE;
	}

	/* ROMt@Cǂݍ */
	if (!file_load(JSUBSYS_ROM, jsubrom, 0x4000)) {
		jsub_available = FALSE;
	}
	if (!file_load(DICT_ROM, jdicrom, 0x40000)) {
		jsub_available = FALSE;
#if XM7_VER >= 2
		jcard_available = FALSE;
#endif
	}

	/* pb` */
	if (!unpatch_rom_flag) {
		/* V1L2Ƀpb` */
		if (jsubrom[0x000d] == 0x8f) {
			jsubrom[0x000d] = 0x88;
		}
	}

	/* ANZX֐ݒ */
	jsubcpu.readmem = jsubmem_readb;
	jsubcpu.writemem = jsubmem_writeb;

	return TRUE;
}

/*
 *	{TuCPU
 *	N[Abv
 */
void FASTCALL jsubsys_cleanup(void)
{
#if 0
	ASSERT(jsubrom);
	ASSERT(jdicrom);
	ASSERT(jsub_sram);
#endif

	if (jsubrom) {
		free(jsubrom);
	}
	if (jdicrom) {
		free(jdicrom);
	}
	if (jsub_sram) {
		free(jsub_sram);
	}
}

/*
 *	{TuCPU
 *	Zbg
 */
void FASTCALL jsubsys_reset(void)
{
	LOG0(Normal, "Zbg");

	jsub_address = 0;
	jsub_haltflag = FALSE;
	jsub_kanji_addr = 0;

	if (jsub_available) {
		jsub_reset();
	}
}

/*
 *	{TuCPU
 *	Pss
 */
void FASTCALL jsubcpu_execline(void)
{
	if (jsub_available && jsub_enable) {
		jsub_line();
	}
}

/*
 *	{TuCPU
 *	s
 */
void FASTCALL jsubcpu_exec(void)
{
	if (jsub_available && jsub_enable) {
		jsub_exec();
	}
}

/*
 *	{TuCPU
 *	NMI荞ݐݒ
 */
void FASTCALL jsubcpu_nmi(void)
{
	if (jsubcpu.intr & INTR_SLOAD){
		jsubcpu.intr |= INTR_NMI;
	}
}

/*
 *	{TuCPU
 *	FIRQ荞ݐݒ
 */
void FASTCALL jsubcpu_firq(void)
{
}

/*
 *	{TuCPU
 *	IRQ荞ݐݒ
 */
void FASTCALL jsubcpu_irq(void)
{
}

/*
 *	{TuCPU
 *	PoCg擾
 */
BYTE FASTCALL jsubmem_readb(WORD addr)
{
	/* {Tuj^ROM */
	if (addr >= 0xc000) {
		return jsubrom[addr - 0xc000];
	}

	/* ROM */
	if ((addr >= 0xa000) && (addr <= 0xafff)) {
		return jdicrom[(addr & 0xfff) | (jsub_dicbank << 12)];
	}

	/* SRAM */
	if ((addr >= 0x8000) && (addr <= 0x9ffe)) {
		return jsub_sram[addr - 0x8000];
	}

	/* oNEtOWX^ */
	if (addr == 0x9fff) {
		return jsub_dicbank;
	}

	LOG1(Warning, "`oCgǂݍ $%04X", addr);

	return 0xff;
}

/*
 *	{TuCPU
 *	PoCg擾(I/OȂ)
 */
BYTE FASTCALL jsubmem_readbnio(WORD addr)
{
	/* {Tuj^ROM */
	if (addr >= 0xc000) {
		return jsubrom[addr - 0xc000];
	}

	/* ROM */
	if ((addr >= 0xa000) && (addr <= 0xafff)) {
		return jdicrom[(addr & 0xfff) | (jsub_dicbank << 12)];
	}

	/* SRAM */
	if ((addr >= 0x8000) && (addr <= 0x9ffe)) {
		return jsub_sram[addr - 0x8000];
	}

	/* oNEtOWX^ */
	if (addr == 0x9fff) {
		return (BYTE)jsub_dicbank;
	}

#if 0
	ASSERT(FALSE);
#endif
	return 0xff;
}

/*
 *	{TuCPU
 *	PoCg
 */
void FASTCALL jsubmem_writeb(WORD addr, BYTE dat)
{
	/* oNEtOWX^ */
	if (addr == 0x9fff) {
		if (dat & 0x80) {
			jsub_haltflag = FALSE;
		}
		else {
			jsub_haltflag = TRUE;
		}

		jsub_dicbank = (BYTE)(dat & 0x3f);
		return;
	}

	/* SRAM */
	if ((addr >= 0x8000) && (addr <= 0x9ffe)) {
		jsub_sram[addr - 0x8000] = dat;
		return;
	}

	LOG2(Warning, "`oCg $%04X <- $%02X", addr, dat);

	return;
}

/*
 *	{TuVXe
 *	CRCBPoCgǂݍ
 */
BYTE FASTCALL jsub_readrcb(void)
{
	return jsub_sram[0x1f00 | (jsub_address++)];
}

/*
 *	{TuVXe
 *	CRCBPoCg
 */
void FASTCALL jsub_writercb(BYTE dat)
{
	jsub_sram[0x1f00 | (jsub_address++)] = dat;
}

/*
 *	{TuVXe
 *	AhXJE^NA
 */
void FASTCALL jsub_clear_address(void)
{
	jsub_address = 0;
}

/*
 *	{TuVXeC^tF[X
 *	PoCgǂݏo
 */
BOOL FASTCALL jsub_readb(WORD addr, BYTE *dat)
{
	int offset;

#if XM7_VER >= 2
	/* FM-7/FM-77/FM77AV̂ݗL */
	if ((model_type > MODEL_FM77AV) ||
		!jsub_enable) {
#else
	/* FM-8[hE{ʐMJ[h͖ */
	if ((model_type == MODEL_FM8) || !jsub_enable) {
#endif
		return FALSE;
	}

	switch (addr) {
		/* {TuVXetO */
		case 0xfd28:
			*dat = 0xff;
			if (jsub_haltflag) {
				*dat &= (BYTE)~0x80;
			}
			return TRUE;

		/* RCBf[^ǂݍ */
		case 0xfd29:
			if (jsub_haltflag) {
				*dat = jsub_readrcb();
			}
			else {
				*dat = 0xff;
			}
			return TRUE;

		/* 1f[^ */
		case 0xfd2a:		/* 1f[^LEFT */
		case 0xfd2b:		/* 1f[^RIGHT */
			if (!jsub_kanji_dipsw) {
				offset = jsub_kanji_addr << 1;
				/* ʏ̈ */
				*dat = kanji_romn[offset + (addr & 1)];
				return TRUE;
			}

			return FALSE;
	}

	return FALSE;
}

/*
 *	{TuVXeC^tF[X
 *	PoCg
 */
BOOL FASTCALL jsub_writeb(WORD addr, BYTE dat)
{
#if XM7_VER >= 2
	/* FM-7/FM-77/FM77AV̂ݗL */
	if ((model_type > MODEL_FM77AV) ||
		!jsub_enable) {
#else
	/* FM-8[hE{ʐMJ[h͖ */
	if ((model_type == MODEL_FM8) || !jsub_enable) {
#endif
		return FALSE;
	}

	switch (addr) {
		/* AhX */
		case 0xfd28:
			if (jsub_kanji_dipsw) {
				return FALSE;
			}
			jsub_kanji_addr &= 0x00ff;
			jsub_kanji_addr |= (WORD)(dat << 8);
			return TRUE;

		/* AhX */
		case 0xfd29:
			if (jsub_kanji_dipsw) {
				return FALSE;
			}
			jsub_kanji_addr &= 0xff00;
			jsub_kanji_addr |= dat;
			return TRUE;

		/* {TuVXetO */
		case 0xfd2a:
			if (dat & 0x80) {
				jsub_haltflag = FALSE;
			}
			else {
				jsub_haltflag = TRUE;
				jsub_clear_address();
			}
			return TRUE;

		/* RCBf[^ */
		case 0xfd2b:
			if (jsub_haltflag) {
				jsub_writercb(dat);
			}
			return TRUE;
	}

	return FALSE;
}

/*
 *	{TuVXe
 *	Z[u
 */
BOOL FASTCALL jsubsys_save(int fileh)
{
	LOG0(Normal, "Z[u");

	/* vbgtH[Ƃ̃pbLO邽߁A */
	if (!file_byte_write(fileh, jsubcpu.cc)) {
		return FALSE;
	}

	if (!file_byte_write(fileh, jsubcpu.dp)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.acc.d)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.x)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.y)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.u)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.s)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.pc)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.intr)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.cycle)) {
		return FALSE;
	}

	if (!file_word_write(fileh, jsubcpu.total)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, jsub_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, jsub_haltflag)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, jsub_dicbank)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, jsub_address)) {
		return FALSE;
	}
	if (!file_write(fileh, jsub_sram, 0x2000)) {
		return FALSE;
	}
	if (!file_word_write(fileh, jsub_kanji_addr)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, jsub_kanji_dipsw)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	{TuVXe
 *	[h
 */
#if XM7_VER >= 2
BOOL FASTCALL jsubsys_load(int fileh, int ver, BOOL bflag)
#else
BOOL FASTCALL jsubsys_load(int fileh, int ver)
#endif
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

#if XM7_VER >= 2
	if ((ver >= 1200) || bflag) {
#else
	if (ver >= 301) {
#endif
		/* vbgtH[Ƃ̃pbLO邽߁A */
		if (!file_byte_read(fileh, &jsubcpu.cc)) {
			return FALSE;
		}

		if (!file_byte_read(fileh, &jsubcpu.dp)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.acc.d)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.x)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.y)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.u)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.s)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.pc)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.intr)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.cycle)) {
			return FALSE;
		}

		if (!file_word_read(fileh, &jsubcpu.total)) {
			return FALSE;
		}

		if (!file_bool_read(fileh, &jsub_enable)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &jsub_haltflag)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &jsub_dicbank)) {
			return FALSE;
		}
		if (!file_byte_read(fileh, &jsub_address)) {
			return FALSE;
		}
		if (!file_read(fileh, jsub_sram, 0x2000)) {
			return FALSE;
		}
#if XM7_VER >= 2
		if (!file_word_read(fileh, &jsub_kanji_addr)) {
			return FALSE;
		}
#else
		if (((ver >= 305) && (ver <= 399)) || (ver >= 404)) {
			if (!file_word_read(fileh, &jsub_kanji_addr)) {
				return FALSE;
			}
		}
#endif

#if XM7_VER >= 2
		if (!file_bool_read(fileh, &jsub_kanji_dipsw)) {
			return FALSE;
		}
#else
		if (ver >= 404) {
			if (!file_bool_read(fileh, &jsub_kanji_dipsw)) {
				return FALSE;
			}
		}
#endif
	}

	/* {Tugps̏ꍇfBZ[u */
	if (!jsub_available) {
		jsub_enable = FALSE;
	}

	return TRUE;
}
#endif
