/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ ROM (񊿎EꐅE񐅏) ]
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "kanji.h"
#include "subctrl.h"
#include "device.h"
#include "jcard.h"
#include "jsubsys.h"

/*
 *	O[o [N
 */
WORD kanji_addr;						/* AhXWX^() */
BYTE *kanji_rom;						/* PROM */
int kanji_rom_type;						/* PROM */
#if XM7_VER >= 2
BYTE *kanji_rom2;						/* QROM */
int kanji_rom2_type;					/* QROM */
#endif
#if XM7_VER >= 3 || defined(JSUB)
BYTE *kanji_romn;						/* PROM for {ʐMJ[h/FM77AV40EXROM */
int kanji_romn_type;					/* PROM for {ʐMJ[h/FM77AV40EXROM */
#endif
BOOL kanji_enable;						/* ROMLtO for FM-7/8 */
BOOL kanji_rom_available;				/* ROMgp\tO */

/*
 *	JIS1ʔp`FbNTl
 */
#define	SUM1_JIS78	0x40ea
#define	SUM1_JIS78E	0xa36c
#define	SUM1_JIS78A	0x6a06
#define	SUM1_JIS83	0xe4bd
#define	SUM2_JIS83	0x131e
#define	SUM1_JIS78N	0xe394

/*
 *	ID
 */
#define	ID	MAKEID('K', 'A', 'N', 'J')
//#define KANJI_LOG

/*
 *	ROM
 *	ʔiȈՁj
 */
static int FASTCALL kanji_check_romtype(BYTE *rom)
{
	int i;
	BYTE high = 0, low = 0;

	/* ݃`FbN */
	if (!rom) {
		return	KANJI_ROM_NONE;
	}

	/* `FbNTvZ */
	for (i = 0x00000; i < 0x20000; i += 2) {
		high += rom[i];
		low += rom[i + 1];
	}

	/* ʔ */
	switch (((WORD)high << 8) | (WORD)low) {

	/* JIS78 Level 1 for FM-7/77 */
	case SUM1_JIS78:
		return	KANJI_ROM_JIS78L1;

	/* JIS78 Level 1 for FM-8 */
	case SUM1_JIS78E:
		return	KANJI_ROM_JIS78L1E;

	/* JIS78 Level 1 for FM77AV */
	case SUM1_JIS78A:
		return	KANJI_ROM_JIS78L1A;

	/* JIS83 Level 1 */
	case SUM1_JIS83:
		return	KANJI_ROM_JIS83L1;

	/* JIS83 Level 2 */
	case SUM2_JIS83:
		return	KANJI_ROM_JIS83L2;

	/* JIS78 Level 1 for FM77AV40EX/40SX,{ʐMJ[h */
	case SUM1_JIS78N:
		return	KANJI_ROM_JIS78L1N;

	/* siLEցEMj */
	default:
		return	KANJI_ROM_UNKNOWN;
	}
}

/*
 *	ROM
 *	
 */
BOOL FASTCALL kanji_init(void)
{
	/*  */
	kanji_rom_available = FALSE;
	kanji_enable = TRUE;
	kanji_rom_type = KANJI_ROM_UNKNOWN;
#if XM7_VER >= 2
	kanji_rom2_type = KANJI_ROM_UNKNOWN;
#endif
#if XM7_VER >= 3 || defined(JSUB)
	kanji_romn_type = KANJI_ROM_UNKNOWN;
#endif

	/* m */
#if XM7_VER >= 3 || (XM7_VER >= 2 && defined(JSUB))
	kanji_rom = (BYTE *)malloc(0x60000);
#elif XM7_VER >= 2 || defined(JSUB)
	kanji_rom = (BYTE *)malloc(0x40000);
#else
	kanji_rom = (BYTE *)malloc(0x20000);
#endif
	if (!kanji_rom) {
		return FALSE;
	}

	/* JIS1 t@Cǂݍ (KANJI.ROMKANJI1.ROMKANJIA.ROM) */
	if (file_load(KANJI_ROM, kanji_rom, 0x20000)) {
		kanji_rom_available = TRUE;
	}
	else if (file_load(KANJI_ROM1, kanji_rom, 0x20000)) {
		kanji_rom_available = TRUE;
	}
	else if (file_load(KANJI_ROMA, kanji_rom, 0x20000)) {
		kanji_rom_available = TRUE;
	}
	else if (file_load(KANJI_TMP, kanji_rom, 0x20000)) {
		kanji_rom_available = TRUE;
		kanji_rom_type = KANJI_ROM_TMP;
	}
#if XM7_VER >= 2
	else {
		/* 1ROM͕K{ */
		return FALSE;
	}
#endif

#if XM7_VER >= 2
	/* JIS2 t@Cǂݍ */
	kanji_rom2 = (BYTE *)(kanji_rom + 0x20000);
	if (!file_load(KANJI_ROM2, kanji_rom2, 0x20000)) {
		if (!file_load(KANJI_TMP2, kanji_rom2, 0x20000)) {
#if XM7_VER >= 3
			/* 2ROMFM77AV40/FM77AV40EX̂ݕK{ */
			return FALSE;
#else
			/* 2ROMȂƂ */
			kanji_rom2_type = KANJI_ROM_NONE;
			memset(kanji_rom2, 0xff, 0x20000);

#ifdef JSUB
			/* {J[hgps */
			jcard_available = FALSE;
#endif
#endif
		}
		else {
			kanji_rom2_type = KANJI_ROM_TMP;
		}
	}
#endif

#if XM7_VER >= 3 || defined(JSUB)
	/* JIS1 for {ʐMJ[h/FM77AV40EXROM t@Cǂݍ */
#if XM7_VER >= 2
	kanji_romn = (BYTE *)(kanji_rom + 0x40000);
#else
	kanji_romn = (BYTE *)(kanji_rom + 0x20000);
#endif
	if (!file_load(KANJI_ROMN, kanji_romn, 0x20000)) {
		if (!file_load(KANJI_TMPN, kanji_romn, 0x20000)) {
			/* 1 for {ʐMJ[h/FM77AV40EXROM ǂݍ߂Ȃꍇ */
			kanji_romn_type = KANJI_ROM_NONE;
			memset(kanji_romn, 0xff, 0x20000);

#ifdef JSUB
			/* {ʐMJ[hgps */
			jsub_available = FALSE;
#endif
		}
		else {
			kanji_romn_type = KANJI_ROM_TMP;
		}
	}
#endif

	/* ȈՊROMʔ */
	if (kanji_rom_type == KANJI_ROM_UNKNOWN) {
		kanji_rom_type = kanji_check_romtype(kanji_rom);
	}
#if XM7_VER >= 2
	if (kanji_rom2_type == KANJI_ROM_UNKNOWN) {
		kanji_rom2_type = kanji_check_romtype(kanji_rom2);
	}
#endif
#if XM7_VER >= 3 || defined(JSUB)
	if (kanji_romn_type == KANJI_ROM_UNKNOWN) {
		kanji_romn_type = kanji_check_romtype(kanji_romn);
	}
#endif

	return TRUE;
}

/*
 *	ROM
 *	N[Abv
 */
void FASTCALL kanji_cleanup(void)
{
#if 0
	ASSERT(kanji_rom);
#endif
	if (kanji_rom) {
		free(kanji_rom);
		kanji_rom = NULL;
#if XM7_VER >= 2
		kanji_rom2 = NULL;
#endif
#if XM7_VER >= 3 || defined(JSUB)
		kanji_romn = NULL;
#endif
	}
}

/*
 *	ROM
 *	Zbg
 */
void FASTCALL kanji_reset(void)
{
	LOG0(Normal, "Zbg");

	kanji_addr = 0;
}

/*
 *	ROM
 *	PoCgǂݏo
 */
BOOL FASTCALL kanji_readb(WORD addr, BYTE *dat)
{
	int offset;

	switch (addr) {
		/* AhX(1E2)ʁE */
		case 0xfd20:
		case 0xfd21:
#if XM7_VER >= 2
		case 0xfd2c:
		case 0xfd2d:
#endif
			*dat = 0xff;
			return TRUE;

		/* 1f[^ */
		case 0xfd22:		/* 1f[^LEFT */
		case 0xfd23:		/* 1f[^RIGHT */
#if XM7_VER == 1
			if (!kanji_rom_available) {
				*dat = 0xff;
				return TRUE;
			}
			/* LN^ZbgȂ */
			if ((model_type == MODEL_FM8) && !kanji_enable) {
				*dat = 0xff;
				return TRUE;
			}
#endif
			/* ROMJ[hȂ */
			if ((model_type == MODEL_FM7) && !kanji_enable) {
#ifdef JSUB
				/* JSUBROM$FD20ɔzuĂȂ */
				if (!(jsub_enable && jsub_kanji_dipsw)) {
					*dat = 0xff;
					return TRUE;
				}
#else
				*dat = 0xff;
				return TRUE;
#endif
			}

#if XM7_VER >= 3 || (XM7_VER == 1 && defined(L4CARD))
			/* ROMTuVXeɐڑ */
			if (subkanji_flag) {
				*dat = 0xff;
				return TRUE;
			}
#endif

			offset = kanji_addr << 1;
#ifdef JSUB
			if ((model_type == MODEL_FM7) &&
				!kanji_enable && jsub_enable && jsub_kanji_dipsw) {
				*dat = kanji_romn[offset + (addr & 1)];
			}
			else {
				*dat = kanji_rom[offset + (addr & 1)];
			}
#else
			*dat = kanji_rom[offset + (addr & 1)];
#endif
			return TRUE;

#if XM7_VER >= 2
		/* 2f[^ */
		case 0xfd2e:		/* 2f[^LEFT */
		case 0xfd2f:		/* 2f[^RIGHT */
#if XM7_VER >= 3
			if ((model_type < MODEL_FM77AV) ||
				((model_type != MODEL_FM77AV40EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((model_type < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}

#if XM7_VER >= 3
			/* ROMTuVXeɐڑ */
			if (subkanji_flag) {
				*dat = 0xff;
				return TRUE;
			}
#endif

			offset = kanji_addr << 1;
			*dat = kanji_rom2[offset + (addr & 1)];
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	ROM
 *	PoCg
 */
BOOL FASTCALL kanji_writeb(WORD addr, BYTE dat)
{
	switch (addr) {
		/* AhX(1E2) */
#if XM7_VER >= 2
		case 0xfd2c:
#if XM7_VER >= 3
			if ((model_type < MODEL_FM77AV) ||
				((model_type != MODEL_FM77AV40EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((model_type < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}
#endif
		case 0xfd20:
			kanji_addr &= 0x00ff;
			kanji_addr |= (WORD)(dat << 8);
			return TRUE;

		/* AhX(1E2) */
#if XM7_VER >= 2
		case 0xfd2d:
#if XM7_VER >= 3
			if ((model_type < MODEL_FM77AV) ||
				((model_type != MODEL_FM77AV40EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((model_type < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}
#endif
		case 0xfd21:
			kanji_addr &= 0xff00;
			kanji_addr |= dat;
			return TRUE;

		/* f[^ */
		case 0xfd22:
		case 0xfd23:
#if XM7_VER >= 2
		case 0xfd2f:
#endif
			return TRUE;

#if XM7_VER >= 2
		/* {J[h oNZNg */
		case 0xfd2e:
#if XM7_VER >= 3
			if ((model_type < MODEL_FM77AV) ||
				((model_type != MODEL_FM77AV40EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((model_type < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}

			/* bit7:wKRAM */
			if (dat & 0x80) {
				dicram_en = TRUE;
			}
			else {
				dicram_en = FALSE;
			}

			/* bit6:ROMCl[u */
			if (dat & 0x40) {
				dicrom_en = TRUE;
			}
			else {
				dicrom_en = FALSE;
			}

			/* bit0-5:ROMoN */
			dicrom_bank = (BYTE)(dat & 0x3f);
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	ROM
 *	Z[u
 */
BOOL FASTCALL kanji_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_word_write(fileh, kanji_addr)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, kanji_enable)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	ROM
 *	[h
 */
BOOL FASTCALL kanji_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_word_read(fileh, &kanji_addr)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if ((ver < 1201) || ((ver >= 1400) && (ver < 1401))) {
#elif XM7_VER >= 2
	if (ver < 1201) {
#else
	if (ver < 1001) {
#endif
		return TRUE;
	}
	if (!file_bool_read(fileh, &kanji_enable)) {
		return FALSE;
	}

	return TRUE;
}
