/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ MIDIA_v^(h!FMfMIDIJ[h?) ]
 */

#ifdef MIDI

#include <string.h>
#include "xm7.h"
#include "midi.h"
#include "device.h"
#include "rs232c.h"

/*
 *	O[o [N
 */
BOOL midi_enable;
BOOL midi_use;
BOOL midi_busy;
BOOL midi_txenable;
BOOL midi_selectmc;
BYTE midi_modereg;

/*
 *	ID
 */
#define	ID	MAKEID('M', 'I', 'D', 'I')
//#define MIDI_LOG

/*
 *	MIDIA_v^
 *	
 */
BOOL FASTCALL midi_init(void)
{
	midi_enable = TRUE;
	midi_use = FALSE;
	midi_busy = FALSE;

	return TRUE;
}

/*
 *	MIDIA_v^
 *	N[Abv
 */
void FASTCALL midi_cleanup(void)
{
}

/*
 *	MIDIA_v^
 *	Zbg
 */
void FASTCALL midi_reset(void)
{
	LOG0(Normal, "Zbg");

	midi_txenable = FALSE;
	midi_selectmc = TRUE;
	midi_modereg = 0xff;
	midi_use = FALSE;

	midi_reset_notify();
}

/*
 *	MIDIA_v^
 *	|[gON/OFF
 */
void FASTCALL midi_port_enable(BOOL flag)
{
	midi_enable = flag;
}

/*
 *	MIDIA_v^
 *	PoCgǂݏo
 */
BOOL FASTCALL midi_readb(WORD addr, BYTE *dat)
{
	if (!midi_enable) {
		return FALSE;
	}

	switch (addr) {
		case 0xfdea:	/* USART DATA */
			/* MIDI INɂ͔Ή */
			/* ǂݏoʂύX (kaikirawcbug) */
			*dat = 0xff;
			return TRUE;

		case 0xfdeb:	/* USART STATUS */
			if (midi_busy) {
				*dat = 0x02;
			}
			else {
				*dat = 0x07;
			}
			return TRUE;
	}

	return FALSE;
}

/*
 *	MIDIA_v^
 *	PoCg
 */
BOOL FASTCALL midi_writeb(WORD addr, BYTE dat)
{
	if (!midi_enable) {
		return FALSE;
	}

	switch (addr) {
		case 0xfdea:	/* USART DATA */
			/* TxE=1 [hR}hWX^ݒlȏꍇɑM */
			if (midi_txenable && (midi_modereg == 0x4e)) {
				midi_use = TRUE;
				midi_notify(dat);
			}
			return TRUE;

		case 0xfdeb:	/* USART COMMAND */
			if (midi_selectmc) {
				/* [hR}hWX^ */
				midi_modereg = dat;
				midi_selectmc = FALSE;
			}
			else {
				/* R}hWX^ */

				/* TXE */
				if (dat & RSC_TXEN) {
					midi_txenable = TRUE;
				}
				else {
					midi_txenable = FALSE;
				}

				/* Zbg */
				if (dat & RSC_IR) {
					midi_modereg = 0xff;
					midi_txenable = FALSE;
					midi_selectmc = TRUE;
				}
			}

			return TRUE;
	}

	return FALSE;
}

/*
 *	MIDIA_v^
 *	Z[u
 */
BOOL FASTCALL midi_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_bool_write(fileh, midi_enable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, midi_use)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, midi_txenable)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, midi_selectmc)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, midi_modereg)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	MIDIA_v^
 *	[h
 */
BOOL FASTCALL midi_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

#if XM7_VER >= 3
	if (((ver >= 500) && (ver < 715)) || ((ver >= 800) && (ver < 915))) {
#elif XM7_VER >= 2
	if (ver < 715) {
#else
	if (ver < 305 || (ver >= 400 && ver < 404)) {
#endif
		return TRUE;
	}

#if XM7_VER >= 3
	if ((ver >= 1403) || (ver >= 1203 && (ver < 1400))) {
#elif XM7_VER >= 2
	if (ver >= 1203) {
#else
	if (ver >= 1003) {
#endif
		if (!file_bool_read(fileh, &midi_enable)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &midi_use)) {
			return FALSE;
		}
	}
	else {
		midi_enable = TRUE;
		midi_use = FALSE;
	}
	if (!file_bool_read(fileh, &midi_txenable)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &midi_selectmc)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &midi_modereg)) {
		return FALSE;
	}

	return TRUE;
}

#endif	/* MIDI */
