/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ A^CNbN(MS58321) ]
 */

#if XM7_VER >= 2

#include <assert.h>
#include <time.h>
#include "xm7.h"
#include "event.h"
#include "rtc.h"
#include "device.h"

/*
 *	O[o [N
 */
BYTE rtc_year;							/* v N(00`99) */
BYTE rtc_month;							/* v (1`12) */
BYTE rtc_day;							/* v (0`31) */
BYTE rtc_week;							/* v j(0`6) */
BYTE rtc_hour;							/* v (0`12 or 0`24h) */
BYTE rtc_minute;						/* v (0`59) */
BYTE rtc_second;						/* v b(0`59) */
BOOL rtc_24h;							/* v 12h/24h؂ւ */
BOOL rtc_pm;							/* v AM/PMtO */
BYTE rtc_leap;							/* v [N[ */

/*
 *	X^eBbN [N
 */
static BOOL rtc_init_flag;				/* v tO */
static time_t rtc_ltime;				/* v (p) */

/*
 *	|Ήe[u
 */
static BYTE rtc_day_table[] = {			/* 228ɃZbg */
	31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

/*
 *	vg^Cv錾
 */
static BOOL FASTCALL rtc_event(void);	/* vCxg */
static BOOL FASTCALL rtc_event_adjust(void);	/* vCxg(p) */
static void FASTCALL rtc_time_adjust_sub(void);	/* AWXgTu */

/*
 *	ID
 */
#define	ID	MAKEID('R', 'T', 'C', ' ')
//#define RTC_LOG

/*
 *	v
 *	
 */
BOOL FASTCALL rtc_init(void)
{
	/* vZbgB24h, [N0 */
	rtc_24h = TRUE;
	rtc_leap = 0;

	/* tOZbg */
	rtc_init_flag = FALSE;

	return TRUE;
}

/*
 *	v
 *	N[Abv
 */
void FASTCALL rtc_cleanup(void)
{
}

/*
 *	v
 *	Zbg
 */
void FASTCALL rtc_reset(void)
{
	LOG0(Normal, "Zbg");

	/* N̂ݎݒ肷Kv */
	if (!rtc_init_flag) {
		rtc_time_adjust();
		rtc_init_flag = TRUE;
	}
}

/*
 *	v
 *	AWXg
 */
void FASTCALL rtc_time_adjust(void)
{
	/* 񎞍s */
	rtc_time_adjust_sub();

	/*  */
	rtc_ltime = 0;

	/* Cxgݒ */
	schedule_setevent(EVENT_RTC, 1000, rtc_event_adjust);
}

/*
 *	v
 *	AWXgCxg
 */
static BOOL FASTCALL rtc_event_adjust(void)
{
	if (rtc_ltime == 0) {
		/* 擾 */
		rtc_ltime = time(NULL);
	}
	else {
		/* ς܂őҋ@ */
		if (rtc_ltime != time(NULL)) {
			/* ēxAs(Cxg̍Đݒs) */
			rtc_time_adjust_sub();
		}
	}

	return TRUE;
}

/*
 *	v
 *	AWXgTu
 */
static void FASTCALL rtc_time_adjust_sub(void)
{
	time_t ltime;
	struct tm *now;

	/* ݂̎ԂǂݎAZbg */
	ltime = time(NULL);
	now = localtime(&ltime);
	rtc_year = (BYTE)(now->tm_year % 100);
	rtc_month = (BYTE)(now->tm_mon + 1);
	rtc_day = (BYTE)now->tm_mday;
	rtc_week = (BYTE)now->tm_wday;
	rtc_hour = (BYTE)now->tm_hour;
	rtc_minute = (BYTE)now->tm_min;
	rtc_second = (BYTE)now->tm_sec;

	/* AM/PMtOݒ */
	if (rtc_hour >= 12) {
		rtc_pm = TRUE;
	}
	else {
		rtc_pm = FALSE;
	}

	/* Cxgݒ */
	schedule_setevent(EVENT_RTC, 1000 * 1000, rtc_event);
}

/*
 *	v
 *	Cxg(1sec)
 */
static BOOL FASTCALL rtc_event(void)
{
	/* bAbv */
	rtc_second++;
	if (rtc_second < 60) {
		return TRUE;
	}
	rtc_second = 0;

	/* Abv */
	rtc_minute++;
	if (rtc_minute < 60) {
		return TRUE;
	}
	rtc_minute = 0;

	/* Abv */
	rtc_hour++;
	if (rtc_24h) {
		/* 24h */
		if (rtc_hour >= 12) {
			rtc_pm = TRUE;
		}
		else {
			rtc_pm = FALSE;
		}
		if (rtc_hour < 24) {
			return TRUE;
		}
	}
	else {
		/* 12h */
		if (rtc_pm) {
			/* PM */
			if (rtc_hour < 12) {
				return TRUE;
			}
		}
		else {
			/* AM */
			if (rtc_hour < 12) {
				return TRUE;
			}
			rtc_hour = 0;
			rtc_pm = TRUE;
			return TRUE;
		}
	}
	rtc_hour = 0;
	rtc_pm = FALSE;

	/* jAbv */
	rtc_week++;
	if (rtc_week > 6) {
		rtc_week = 0;
	}

	/* Abv */
	rtc_day++;
	if (rtc_day <= rtc_day_table[rtc_month - 1]) {
		return TRUE;
	}

	/* [Ñ`FbNAœ */
	if ((rtc_month == 2) && (rtc_day == 29)) {
		/* 229Ȃ̂ŁA[NȂreturn TRUE */
		if (rtc_leap == 0) {
			if ((rtc_year % 4) == 0) {
				return TRUE;
			}
		}
		else {
			if ((rtc_year % 4) == (4 - rtc_leap)) {
				return TRUE;
			}
		}
	}
	rtc_day = 1;

	/* Abv */
	rtc_month++;
	if (rtc_month <= 12) {
		return TRUE;
	}
	rtc_month = 1;

	/* NAbv */
	rtc_year++;
	if (rtc_year > 99) {
		rtc_year = 0;
	}

	return TRUE;
}

/*-[ I/O(L[{[hRg[) ]--------------------------------------*/

/*
 *	v
 *	f[^Zbg()
 */
void FASTCALL rtc_set(BYTE *packet)
{
	BYTE dat;

	ASSERT(packet);

	/* f`FbN */
	if (model_type < MODEL_FM77AV) {
		return;
	}

	/* N */
	dat = *packet++;
	rtc_year = (BYTE)((dat >> 4) * 10);
	rtc_year += (BYTE)(dat & 0x0f);

	/*  */
	dat = *packet++;
	rtc_month = (BYTE)(((dat & 0x10) >> 4) * 10);
	rtc_month += (BYTE)(dat & 0x0f);

	/*  + [N */
	dat = *packet++;
	rtc_day = (BYTE)(((dat & 0x30) >> 4) * 10);
	rtc_day += (BYTE)(dat & 0x0f);
	rtc_leap = (BYTE)((dat & 0xc0) >> 6);

	/* j,12/24,̏ */
	dat = *packet++;
	rtc_week = (BYTE)((dat >> 4) & 0x07);
	rtc_hour = (BYTE)((dat & 0x03) * 10);
	if (dat & 0x08) {
		rtc_24h = TRUE;
	}
	else {
		rtc_24h = FALSE;
	}
	if (dat & 0x04) {
		rtc_pm = TRUE;
	}
	else {
		rtc_pm = FALSE;
	}

	/* ̉ʁȀ */
	dat = *packet++;
	rtc_hour += (BYTE)(dat >> 4);
	rtc_minute = (BYTE)((dat & 0x0f) * 10);

	/* ̉ʁAb̏ */
	dat = *packet++;
	rtc_minute += (BYTE)(dat >> 4);
	rtc_second = (BYTE)((dat & 0x0f) * 10);

	/* b̉ */
	dat = *packet;
	rtc_second += (BYTE)(dat >> 4);
}

/*
 *	v
 *	f[^擾(ǂݍ)
 */
void FASTCALL rtc_get(BYTE *packet)
{
	BYTE dat;

	ASSERT(packet);

	/* f`FbN */
	if (model_type < MODEL_FM77AV) {
		return;
	}

	/* N */
	dat = (BYTE)((rtc_year / 10) << 4);
	dat |= (BYTE)(rtc_year % 10);
	*packet++ = dat;

	/*  */
	dat = (BYTE)((rtc_month / 10) << 4);
	dat |= (BYTE)(rtc_month % 10);
	*packet++ = dat;

	/*  + [N */
	dat = (BYTE)((rtc_day / 10) << 4);
	dat |= (BYTE)(rtc_day % 10);
	dat |= (BYTE)(rtc_leap * 64);
	*packet++ = dat;

	/* j,12/24,̏ */
	dat = (BYTE)(rtc_week << 4);
	dat |= (BYTE)(rtc_hour / 10);
	if (rtc_24h) {
		dat |= 0x08;
	}
	if (!rtc_24h && rtc_pm) {
		dat |= 0x04;
	}
	*packet++ = dat;

	/* ̉ʁȀ */
	dat = (BYTE)((rtc_hour % 10) << 4);
	dat |= (BYTE)(rtc_minute / 10);
	*packet++ = dat;

	/* ̉ʁAb̏ */
	dat = (BYTE)((rtc_minute % 10) << 4);
	dat |= (BYTE)(rtc_second / 10);
	*packet++ = dat;

	/* b̉ */
	dat = (BYTE)((rtc_second % 10) << 4);
	*packet = dat;
}

/*-[ t@CI/O ]----------------------------------------------------------*/

/*
 *	v
 *	Z[u
 */
BOOL FASTCALL rtc_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_bool_write(fileh, rtc_24h)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, rtc_pm)) {
		return FALSE;
	}

	if (!file_byte_write(fileh, rtc_leap)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	v
 *	[h
 */
BOOL FASTCALL rtc_load(int fileh, int ver)
{
	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_bool_read(fileh, &rtc_24h)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &rtc_pm)) {
		return FALSE;
	}

	if (!file_byte_read(fileh, &rtc_leap)) {
		return FALSE;
	}

	/* ݂̎ԂǂݎAZbg */
	rtc_time_adjust();

	return TRUE;
}

#endif	/* XM7_VER >= 2 */
