/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ TuCPU ]
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "display.h"
#include "subctrl.h"
#include "device.h"
#include "keyboard.h"
#include "multipag.h"
#include "aluline.h"
#include "l4font.h"

/*
 *	O[o [N
 */
BYTE *vram_c;						/* VRAM(^CvC) $C000(oNL) */
BYTE *subrom_c;						/* ROM (^CvC) $2800 */
BYTE *sub_ram;						/* R\[RAM $1680 */
BYTE *sub_io;						/* TuCPU I/O   $0100 */
#if XM7_VER >= 3
BYTE *subramde;						/* RAM (TypeD/E) $2000 */
BYTE *subramcg;						/* RAM (tHg)$4000 */
BYTE *subramcn;						/* R\[  $2000 */
#endif

#if XM7_VER >= 2
BYTE *vram_b;						/* VRAM(^CvB) $C000(oNL) */
BYTE *subrom_a;						/* ROM (^CvA) $2000 */
BYTE *subrom_b;						/* ROM (^CvB) $2000 */
BYTE *subromcg;						/* ROM (tHg)$2000 */

BYTE subrom_bank;					/* TuVXeROMoN */
BYTE cgrom_bank;					/* CGROMoN */
#if XM7_VER >= 3
BYTE cgram_bank;					/* CGRAMoN */
BYTE consram_bank;					/* R\[RAMoN */
#endif	/* XM7_VER >= 3 */
#endif	/* XM7_VER >= 2 */

#if XM7_VER == 1
BYTE *subrom_8;						/* ROM (FM-8)    $2800 */
BYTE *subrampcg;					/* RAM (PCG)      $800 */
BOOL pcg_flag;						/* PCGtO */
BOOL pcg_enable;					/* PCGLtO */
#ifdef L4CARD
BYTE *tvram_c;						/* Text VRAM     $1000 */
BYTE *subrom_l4;					/* ROM(400C)$4800 */
BYTE *subcg_l4;						/* CG (400C)$1000 */
BOOL enable_400linecard;			/* 400CJ[hCl[u */
BOOL detect_400linecard;			/* 400CJ[htO */
BOOL ankcg_force_internal;			/* tHggptO */
#endif
BOOL subrom_xchg;					/* TuVXeύXLtO */
#endif

/*
 *	ID
 */
#define	ID	MAKEID('S', 'M', 'E', 'M')
//#define SUBMEM_LOG

/*
 *	TuCPU
 *	
 */
BOOL FASTCALL submem_init(void)
{
	/* xASăNA */
	vram_c = NULL;
	subrom_c = NULL;
	sub_ram = NULL;
	sub_io = NULL;
#if XM7_VER == 1
	subrom_8 = NULL;
	subrampcg = NULL;
	pcg_flag = FALSE;
	pcg_enable = FALSE;
#ifdef L4CARD
	ankcg_force_internal = FALSE;
#endif
#endif

#if XM7_VER >= 2
	vram_b = NULL;
	subrom_a = NULL;
	subrom_b = NULL;
	subromcg = NULL;
#if XM7_VER >= 3
	subramde = NULL;
	subramcg = NULL;
	subramcn = NULL;
#endif
#endif

#if XM7_VER == 1 && defined(L4CARD)
	tvram_c = NULL;
	subrom_l4 = NULL;
	subcg_l4 = NULL;
#endif

	/* m(^CvC) */
#if XM7_VER >= 3
	vram_c = (BYTE *)malloc(0x30000);
#elif XM7_VER >= 2
	vram_c = (BYTE *)malloc(0x18000);
#else
	vram_c = (BYTE *)malloc(0xc000);
#endif
	if (vram_c == NULL) {
		return FALSE;
	}
	subrom_c = (BYTE *)malloc(0x2800);
	if (subrom_c == NULL) {
		return FALSE;
	}
#if XM7_VER == 1
	subrom_8 = (BYTE *)malloc(0x2800);
	if (subrom_8 == NULL) {
		return FALSE;
	}
#endif

	/* m(^CvA,B) */
#if XM7_VER >= 2
	vram_b = vram_c + 0xc000;
	subrom_a = (BYTE *)malloc(0x2000);
	if (subrom_a == NULL) {
		return FALSE;
	}
	subrom_b = (BYTE *)malloc(0x2000);
	if (subrom_b == NULL) {
		return FALSE;
	}
	subromcg = (BYTE *)malloc(0x2000);
	if (subromcg == NULL) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* m(^CvD,E) */
	subramcn = (BYTE *)malloc(0x2000);
	if (subramcn == NULL) {
		return FALSE;
	}
	subramcg = (BYTE *)malloc(0x4000);
	if (subramcg == NULL) {
		return FALSE;
	}
	subramde = (BYTE *)malloc(0x2000);
	if (subramde == NULL) {
		return FALSE;
	}
#endif
#endif

	/* m() */
	sub_ram = (BYTE *)malloc(0x1680);
	if (sub_ram == NULL) {
		return FALSE;
	}
	sub_io = (BYTE *)malloc(0x0100);
	if (sub_io == NULL) {
		return FALSE;
	}

#if XM7_VER == 1
	/* m(PCG) */
	subrampcg = (BYTE *)malloc(0x0800);
	if (subrampcg == NULL) {
		return FALSE;
	}

#ifdef L4CARD
	/* m(L4) */
	tvram_c = (BYTE *)malloc(0x1000);
	if (tvram_c == NULL) {
		return FALSE;
	}
	subrom_l4 = (BYTE *)malloc(0x4800);
	if (subrom_l4 == NULL) {
		return FALSE;
	}
	subcg_l4 = (BYTE *)malloc(0x1000);
	if (subcg_l4 == NULL) {
		return FALSE;
	}
#endif
#endif

	/* ROMt@Cǂݍ */
	if (!file_load(SUBSYSC_ROM, subrom_c, 0x2800)) {
#if XM7_VER == 1
		available_fm7roms = FALSE;
#else
		return FALSE;
#endif
	}
#if XM7_VER >= 2
	if (!file_load(SUBSYSA_ROM, subrom_a, 0x2000)) {
		return FALSE;
	}
	if (!file_load(SUBSYSB_ROM, subrom_b, 0x2000)) {
		return FALSE;
	}
	if (!file_load(SUBSYSCG_ROM, subromcg, 0x2000)) {
		return FALSE;
	}
#endif

#if XM7_VER == 1
	if (!file_load(SUBSYS8_ROM, subrom_8, 0x2800)) {
		available_fm8roms = FALSE;
	}

#ifdef L4CARD
	/* 400CΉ */
	if (!file_load(SUBSYSL4_ROM, subrom_l4, 0x4800)) {
		enable_400linecard = FALSE;
		detect_400linecard = FALSE;
	}
	else {
		enable_400linecard = TRUE;
		detect_400linecard = TRUE;

		/* ANKCG16.ROM,EXTSUB.ROM̂ǂȂꍇtHggp */
		if (ankcg_force_internal) {
			memcpy(subcg_l4, subcg_internal, 4096);
		}
		else {
			if (!file_load(ANKCG16_ROM, subcg_l4, 0x1000)) {
				if (!file_load(EXTSUB_ROM, subcg_l4, 0x1000)) {
#if 1
					memcpy(subcg_l4, subcg_internal, 4096);
#else
					int addr;
					for (addr = 0x7ff; addr >= 0; addr --) {
						subcg_l4[addr * 2 + 0] = subrom_c[addr];
						subcg_l4[addr * 2 + 1] = subrom_c[addr];
					}
#endif
				}
			}
		}
	}
#endif

	if (!available_fm8roms && !available_fm7roms) {
		return FALSE;
	}

	/* PCGGȀ */
	memset(subrampcg, 0xff, 0x0800);

	/* FM-8ROMC[WȂꍇ̃hCuԍ\ւ̑Ώ */
	/* {A擪2048oCĝ݃Rs[΂悢ÔߑŜRs[ */
	if (available_fm8roms && !available_fm7roms) {
		memcpy(subrom_c, subrom_8, 0x2800);
	}

	/* TuVXeύXLtO */
	subrom_xchg = FALSE;
#endif

	return TRUE;
}

/*
 *	TuCPU
 *	N[Abv
 */
void FASTCALL submem_cleanup(void)
{
#if 0
	ASSERT(vram_c);
	ASSERT(subrom_c);
	ASSERT(sub_ram);
	ASSERT(sub_io);
#if XM7_VER >= 2
	ASSERT(subrom_a);
	ASSERT(subrom_b);
	ASSERT(subromcg);
#if XM7_VER >= 3
	ASSERT(subramcn);
	ASSERT(subramcg);
	ASSERT(subramde);
#endif
#endif
#if XM7_VER == 1
	ASSERT(subrom_8);
	ASSERT(subrampcg);
#ifdef L4CARD
	ASSERT(tvram_c);
	ASSERT(subrom_l4);
	ASSERT(subcg_l4);
#endif
#endif
#endif

	/* rŎsꍇl */
	if (vram_c) {
		free(vram_c);
	}
	if (subrom_c) {
		free(subrom_c);
	}

#if XM7_VER >= 2
	if (subrom_a) {
		free(subrom_a);
	}
	if (subrom_b) {
		free(subrom_b);
	}
	if (subromcg) {
		free(subromcg);
	}

#if XM7_VER >= 3
	if (subramcn) {
		free(subramcn);
	}
	if (subramcg) {
		free(subramcg);
	}
	if (subramde) {
		free(subramde);
	}
#endif
#endif

	if (sub_ram) {
		free(sub_ram);
	}
	if (sub_io) {
		free(sub_io);
	}

#if XM7_VER == 1
	if (subrom_8) {
		free(subrom_8);
	}
	if (subrampcg) {
		free(subrampcg);
	}

#ifdef L4CARD
	if (tvram_c) {
		free(tvram_c);
	}
	if (subrom_l4) {
		free(subrom_l4);
	}
	if (subcg_l4) {
		free(subcg_l4);
	}
#endif
#endif
}

/*
 *	TuCPU
 *	Zbg
 */
void FASTCALL submem_reset(void)
{
	LOG0(Normal, "Zbg");

#if XM7_VER >= 2
	/* VRAMANeBuy[W */
	vram_aptr = vram_c;
	vram_active = 0;

	/* oNNA */
	subrom_bank = 0;
	cgrom_bank = 0;
#if XM7_VER >= 3
	cgram_bank = 0;
	consram_bank = 0;
#endif
#endif

	/* I/O NA */
	memset(sub_io, 0xff, 0x0100);
}

/*
 *	TuCPU
 *	PoCg擾
 */
BYTE FASTCALL submem_readb(WORD addr)
{
	BYTE dat;

#if XM7_VER == 1 && defined(L4CARD)
	/* FM-77L4 400C[h̏ */
	if (enable_400line && enable_400linecard) {
		/* GVRAM/[NRAM */
		if (addr <= 0x7fff) {
			/* [NRAM */
			if (workram_select) {
				if (multi_page & 4) {
					LOG1(Warning, "}XÑ[NRAMoCgǂݍ $%04X", addr);
					return 0xff;
				}
				return vram_c[0x8000 + ((addr + vram_offset[0]) & 0x3fff)];
			}

			/* GVRAM */
			addr += vram_offset[0];
			addr &= 0x7fff;
			if (multi_page & (1 << (addr >> 14))) {
				LOG1(Warning, "}XNVRAMoCgǂݍ $%04X", addr);
				return 0xff;
			}
			return vram_c[addr];
		}

		/* eLXgVRAM */
		if ((addr >= 0x8000) && (addr <= 0x97ff)) {
			return tvram_c[addr & 0xfff];
		}

		/* Tuj^ROM(O) */
		if ((addr >= 0x9800) && (addr <= 0xbfff)) {
			return subrom_l4[addr - 0x9800];
		}

		/* Tuj^ROM(㔼) */
		if (addr >= 0xe000) {
			return subrom_l4[addr - 0xb800];
		}
	}
#endif

	/* VRAM */
	if (addr < 0xc000) {
#if XM7_VER >= 3
		if ((model_type >= MODEL_FM77AV) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			if (mode400l) {
				/* 400C */
				if (addr >= 0x8000) {
					LOG1(Warning, "`oCgǂݍ $%04X", addr);
					return 0xff;
				}
				aluline_extrb(addr);

				if (multi_page & (1 << subram_vrambank)) {
					LOG1(Warning, "}XNVRAMoCgǂݍ $%04X", addr);
					return 0xff;
				}
				else {
					return vram_aptr[addr];
				}
			}
			else {
				/* 200C */
				aluline_extrb(addr);
				if (multi_page & (1 << (addr >> 14))) {
					LOG1(Warning, "}XNVRAMoCgǂݍ $%04X", addr);
					return 0xff;
				}
				else {
					return vram_aptr[((addr & 0xc000) << 1) | (addr & 0x3fff)];
				}
			}
		}
		else {
			LOG1(Warning, "VRAMANZXtOVRAMoCgǂݍ $%04X", addr);
			return 0xff;
		}
#elif XM7_VER >= 2
		/* 200C */
		if ((model_type >= MODEL_FM77AV) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			aluline_extrb(addr);
			if (multi_page & (1 << (addr >> 14))) {
				LOG1(Warning, "}XNVRAMoCgǂݍ $%04X", addr);
				return 0xff;
			}
			else {
				return vram_aptr[addr];
			}
		}
		else {
			LOG1(Warning, "VRAMANZXtOVRAMoCgǂݍ $%04X", addr);
			return 0xff;
		}
#else
		/* 200C */
		if ((model_type == MODEL_FM77) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			if (multi_page & (1 << (addr >> 14))) {
				LOG1(Warning, "}XNVRAMoCgǂݍ $%04X", addr);
				return 0xff;
			}
			else {
				return vram_c[addr];
			}
		}
		else {
			LOG1(Warning, "VRAMANZXtOVRAMoCgǂݍ $%04X", addr);
			return 0xff;
		}
#endif
	}

#if XM7_VER >= 3
	/* R\[RAM(RAM[hp) */
	if ((addr < 0xd000) && (subrom_bank == 4) && (consram_bank >= 1)) {
		return subramcn[(consram_bank - 1) * 0x1000 + (addr - 0xc000)];
	}
#endif

	/* [NRAM */
	if (addr < 0xd380) {
		return sub_ram[addr - 0xc000];
	}

	/* LRAM */
	if (addr < 0xd400) {
		return shared_ram[(WORD)(addr - 0xd380)];
	}

#if XM7_VER >= 2
	/* TuROM */
	if ((model_type < MODEL_FM77AV) && (addr >= 0xd800)) {
		/* ^CvC */
		return subrom_c[addr - 0xd800];
	}
	if (addr >= 0xe000) {
		switch (subrom_bank) {
			/* ^CvC */
			case 0:
				return subrom_c[addr - 0xd800];
			/* ^CvA */
			case 1:
				return subrom_a[addr - 0xe000];
			/* ^CvB */
			case 2:
				return subrom_b[addr - 0xe000];
			/* CGROM */
			case 3:
				return subromcg[addr - 0xe000];
#if XM7_VER >= 3
			/* ^CvD/E(RAM) */
			case 4:
				return subramde[addr - 0xe000];
#endif
			}
	}

	/* CGRAM,CGROM */
	if (addr >= 0xd800) {
#if XM7_VER >= 3
		if (subrom_bank == 4) {
			/* TuRAM oN */
			return subramcg[cgram_bank * 0x0800 + (addr - 0xd800)];
		}
		else if (model_type >= MODEL_FM77AV) {
			/* TuROM oN */
			return subromcg[cgrom_bank * 0x0800 + (addr - 0xd800)];
		}
		else {
			/* TuROM Type-C */
			return subrom_c[addr - 0xd800];
		}
#else
		if (model_type >= MODEL_FM77AV) {
			/* TuROM oN */
			return subromcg[cgrom_bank * 0x0800 + (addr - 0xd800)];
		}
		else {
			/* TuROM Type-C */
			return subrom_c[addr - 0xd800];
		}
#endif
	}

	/* [NRAM */
	if (model_type >= MODEL_FM77AV) {
		if (addr >= 0xd500){
			return sub_ram[(addr - 0xd500) + 0x1380];
		}
	}
#else
	if (addr >= 0xd800) {
		if ((addr < 0xe000) && pcg_flag && pcg_enable) {
			/* PCG */
			return subrampcg[addr - 0xd800];
		}
		if ((model_type == MODEL_FM8) &&
			(!subrom_xchg)) {
			/* FM-8Tuj^ */
			return subrom_8[addr - 0xd800];
		}
		else {
			/* ^CvC */
			return subrom_c[addr - 0xd800];
		}
	}
#endif

	/*
	 *	TuI/O
	 */
#if XM7_VER == 1
	/* FM-7/8/77̃TuI/O $D410`$D7FF$D400`$D40F̃~[ */
	addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x000f));
#else
	if (model_type < MODEL_FM77AV) {
		/* FM-7$D410`$D7FF$D400`$D40F̃~[ */
		addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x000f));
	}
	else if (model_type == MODEL_FM77AV) {
		/* FM77AV$D440`$D4FF$D400`$D43F̃~[ */
		addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x003f));
	}
#endif

	/* fBXvC */
	if (display_readb(addr, &dat)) {
		return dat;
	}
	/* L[{[h */
	if (keyboard_readb(addr, &dat)) {
		return dat;
	}
	/* _ZE */
#if XM7_VER >= 2
	if (aluline_readb(addr, &dat)) {
		return dat;
	}
#endif

	LOG1(Warning, "`oCgǂݍ $%04X", addr);

	return 0xff;
}

/*
 *	TuCPU
 *	PoCg擾(I/OȂ)
 */
BYTE FASTCALL submem_readbnio(WORD addr)
{
#if XM7_VER == 1 && defined(L4CARD)
	/* FM-77L4 400C[h̏ */
	if (enable_400line && enable_400linecard) {
		/* GVRAM/[NRAM */
		if (addr <= 0x7fff) {
			/* [NRAM */
			if (workram_select) {
				if (multi_page & 4) {
					return 0xff;
				}
				return vram_c[0x8000 + ((addr + vram_offset[0]) & 0x3fff)];
			}

			/* GVRAM */
			addr += vram_offset[0];
			addr &= 0x7fff;
			if (multi_page & (1 << (addr >> 14))) {
				return 0xff;
			}
			return vram_c[addr];
		}

		/* eLXgVRAM */
		if ((addr >= 0x8000) && (addr <= 0x97ff)) {
			return tvram_c[addr & 0xfff];
		}

		/* Tuj^ROM(O) */
		if ((addr >= 0x9800) && (addr <= 0xbfff)) {
			return subrom_l4[addr - 0x9800];
		}

		/* Tuj^ROM(㔼) */
		if (addr >= 0xe000) {
			return subrom_l4[addr - 0xb800];
		}
	}
#endif

	/* VRAM */
	if (addr < 0xc000) {
#if XM7_VER >= 3
		if ((model_type >= MODEL_FM77AV) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			if (mode400l) {
				/* 400C */
				if (addr < 0x8000) {
					return vram_aptr[addr];
				}
				else {
					return 0xff;
				}
			}
			else {
				/* 200C */
				return vram_aptr[((addr & 0xc000) << 1) | (addr & 0x3fff)];
			}
		}
		else {
			return 0xff;
		}
#elif XM7_VER >= 2
		/* 200C */
		if ((model_type >= MODEL_FM77AV) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			return vram_aptr[addr];
		}
		else {
			return 0xff;
		}
#else
		/* 200C */
		if ((model_type == MODEL_FM77) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			return vram_c[addr];
		}
		else {
			return 0xff;
		}
#endif
	}

#if XM7_VER >= 3
	/* R\[RAM(RAM[hp) */
	if ((addr < 0xd000) && (subrom_bank == 4) && (consram_bank >= 1)) {
		return subramcn[(consram_bank - 1) * 0x1000 + (addr - 0xc000)];
	}
#endif

	/* [NRAM */
	if (addr < 0xd380) {
		return sub_ram[addr - 0xc000];
	}

	/* LRAM */
	if (addr < 0xd400) {
		return shared_ram[(WORD)(addr - 0xd380)];
	}

#if XM7_VER >= 2
	/* TuROM */
	if ((model_type < MODEL_FM77AV) && (addr >= 0xd800)) {
		/* ^CvC */
		return subrom_c[addr - 0xd800];
	}
	if (addr >= 0xe000) {
		ASSERT(subrom_bank <= 4);
		switch (subrom_bank) {
			/* ^CvC */
			case 0:
				return subrom_c[addr - 0xd800];
			/* ^CvA */
			case 1:
				return subrom_a[addr - 0xe000];
			/* ^CvB */
			case 2:
				return subrom_b[addr - 0xe000];
			/* CGROM */
			case 3:
				return subromcg[addr - 0xe000];
#if XM7_VER >= 3
			/* ^CvD/E(RAM) */
			case 4:
				return subramde[addr - 0xe000];
#endif
			}
	}

	/* CGRAM,CGROM */
	if (addr >= 0xd800) {
#if XM7_VER >= 3
		if (subrom_bank == 4) {
			/* TuRAM oN */
			return subramcg[cgram_bank * 0x0800 + (addr - 0xd800)];
		}
		else if (model_type >= MODEL_FM77AV) {
			/* TuROM oN */
			return subromcg[cgrom_bank * 0x0800 + (addr - 0xd800)];
		}
		else {
			/* TuROM Type-C */
			return subrom_c[addr - 0xd800];
		}
#else
		if (model_type >= MODEL_FM77AV) {
			/* TuROM oN */
			return subromcg[cgrom_bank * 0x0800 + (addr - 0xd800)];
		}
		else {
			/* TuROM Type-C */
			return subrom_c[addr - 0xd800];
		}
#endif
	}

	/* [NRAM */
	if (model_type >= MODEL_FM77AV) {
		if (addr >= 0xd500) {
			return sub_ram[(addr - 0xd500) + 0x1380];
		}
	}
#else
	if (addr >= 0xd800) {
		if ((addr < 0xe000) && pcg_flag && pcg_enable) {
			/* PCG */
			return subrampcg[addr - 0xd800];
		}
		if ((model_type == MODEL_FM8) &&
			(!subrom_xchg)) {
			/* FM-8Tuj^ */
			return subrom_8[addr - 0xd800];
		}
		else {
			/* ^CvC */
			return subrom_c[addr - 0xd800];
		}
	}
#endif

	/*
	 *	TuI/O
	 */
	if (addr < 0xd800) {
#if XM7_VER == 1
		/* FM-7/8/77̃TuI/O $D410`$D7FF$D400`$D40F̃~[ */
		addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x000f));
#else
		if (model_type < MODEL_FM77AV) {
			/* FM-7̃TuI/O $D410`$D7FF$D400`$D40F̃~[ */
			addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x000f));
		}
		else if (model_type == MODEL_FM77AV) {
			/* FM77AṼTuI/O $D440`$D4FF$D400`$D43F̃~[ */
			addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x003f));
		}
#endif
		return sub_io[addr - 0xd400];
	}

	/* ɂ͗Ȃ */
	ASSERT(FALSE);
	return 0;
}

/*
 *	TuCPU
 *	PoCg
 */
void FASTCALL submem_writeb(WORD addr, BYTE dat)
{
#if XM7_VER == 1 && defined(L4CARD)
	/* FM-77L4 400C[h̏ */
	if (enable_400line && enable_400linecard) {
		/* GVRAM/[NRAM */
		if (addr <= 0x7fff) {
			/* [NRAM */
			if (workram_select) {
				if (!(multi_page & 4)) {
					vram_c[0x8000 + ((addr + vram_offset[0]) & 0x3fff)] = dat;
				}
				else {
					LOG2(Warning, "}XÑ[NRAMoCg $%04X <- $%02X", addr, dat);
				}
				return;
			}

			/* GVRAM */
			addr += vram_offset[0];
			addr &= 0x7fff;
			if (!(multi_page & (1 << (addr >> 14)))) {
				/* f[^ނȂANotifyȂ */
				if (vram_c[addr] != dat) {
					vram_c[addr] = dat;
					vram_notify(addr, dat);
				}
			}
			else {
				LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
			}
			return;
		}

		/* eLXgVRAM */
		if ((addr >= 0x8000) && (addr <= 0x97ff)) {
			tvram_c[addr & 0xfff] = dat;
			tvram_notify((WORD)(addr & 0xfff), dat);
			return;
		}

		/* Tuj^ROM */
		if (((addr >= 0x9800) && (addr <= 0xbfff)) || (addr >= 0xe000)) {
			LOG2(Warning, "Tuj^ROMւ̃oCg $%04X <- $%02X", addr, dat);
			return;
		}
	}
#endif

	/* VRAM(^CvC) */
	if (addr < 0xc000) {
#if XM7_VER >= 3
		if ((model_type >= MODEL_FM77AV) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			if (mode400l) {
				/* 400C */
				if (addr >= 0x8000) {
					LOG2(Warning, "`oCg $%04X <- $%02X", addr, dat);
					return;
				}
				if (alu_command & 0x80) {
					aluline_extrb(addr);
					return;
				}
				if (!(multi_page & (1 << subram_vrambank))) {
					if (vram_aptr[addr] != dat) {
						/* f[^ނȂANotifyȂ */
						vram_aptr[addr] = dat;
						vram_notify(addr, dat);
					}
				}
				else {
					LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
				}
			}
			else {
				/* 200C */
				if (alu_command & 0x80) {
					aluline_extrb(addr);
					return;
				}
				if (!(multi_page & (1 << (addr >> 14)))) {
					/* f[^ނȂANotifyȂ */
					if (vram_aptr[((addr & 0xc000) << 1)|(addr & 0x3fff)] != dat) {
						vram_aptr[((addr & 0xc000) << 1)|(addr & 0x3fff)] = dat;
						vram_notify(addr, dat);
					}
				}
				else {
					LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
				}
			}
		}
		else {
			LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
		}
#elif XM7_VER >= 2
		if ((model_type >= MODEL_FM77AV) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			/* 200C */
			if (alu_command & 0x80) {
				aluline_extrb(addr);
				return;
			}
			if (!(multi_page & (1 << (addr >> 14)))) {
				/* f[^ނȂANotifyȂ */
				if (vram_aptr[addr] != dat) {
					vram_aptr[addr] = dat;
					vram_notify(addr, dat);
				}
			}
			else {
				LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
			}
		}
		else {
			LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
		}
#else
		if ((model_type == MODEL_FM77) ||
			(cycle_steal || cycle_steal_mode || vrama_flag || blank_flag)) {
			if (!(multi_page & (1 << (addr >> 14)))) {
				/* f[^ނȂANotifyȂ */
				if (vram_c[addr] != dat) {
					vram_c[addr] = dat;
					vram_notify(addr, dat);
				}
			}
			else {
				LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
			}
		}
		else {
			LOG2(Warning, "}XNVRAMoCg $%04X <- $%02X", addr, dat);
		}
#endif
		return;
	}

#if XM7_VER >= 3
	/* R\[RAM(RAM[hp) */
	if ((addr < 0xd000) && (subrom_bank == 4) && (consram_bank >= 1)) {
		subramcn[(consram_bank - 1) * 0x1000 + (addr - 0xc000)] = dat;
		return;
	}
#endif

	/* [NRAM */
	if (addr < 0xd380) {
		sub_ram[addr - 0xc000] = dat;
		return;
	}

	/* LRAM */
	if (addr < 0xd400) {
		shared_ram[(WORD)(addr - 0xd380)] = dat;
		return;
	}

	/* CGROM,TuRAM ROM[hł͏݂łȂ */
	if (addr >= 0xd800) {
#if XM7_VER >= 3
		/* TuRAMłȂAveNgĂ */
		if ((subrom_bank != 4) || subram_protect) {
			if (subrom_bank != 4) {
				LOG2(Warning, "Tuj^ROMւ̃oCg $%04X <- $%02X", addr, dat);
			}
			else {
				LOG2(Warning, "TuRAMveNgԂł̃oCg $%04X <- $%02X", addr, dat);
			}
			return;
		}

		/* CGRAM */
		if (addr < 0xe000) {
			subramcg[cgram_bank * 0x0800 + (addr - 0xd800)] = dat;
			return;
		}

		/* TuRAM */
		subramde[addr - 0xe000] = dat;
		return;
#else
#if XM7_VER == 1
		/* PCG */
		if ((addr < 0xe000) && pcg_enable) {
			/* ROM/RAMǂł݉ */
			subrampcg[addr - 0xd800] = dat;
			return;
		}
#endif

		/* V1/V2ł͖ŏݕs */
		LOG2(Warning, "Tuj^ROMւ̃oCg $%04X <- $%02X", addr, dat);
		return;
#endif
	}

#if XM7_VER >= 2
	/* [NRAM */
	if (model_type >= MODEL_FM77AV) {
		if (addr >= 0xd500) {
			sub_ram[(addr - 0xd500) + 0x1380] = dat;
			return;
		}
	}
#endif

	/*
	 *	TuI/O
	 */
#if XM7_VER == 1
	/* FM-7/8/77̃TuI/O $D410`$D7FF$D400`$D40F̃~[ */
	addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x000f));
#else
	if (model_type < MODEL_FM77AV) {
		/* FM-7̃TuI/O $D410`$D7FF$D400`$D40F̃~[ */
		addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x000f));
	}
	else if (model_type == MODEL_FM77AV) {
		/* FM77AṼTuI/O $D440`$D4FF$D400`$D43F̃~[ */
		addr = (WORD)(0xd400 + ((addr - 0xd400) & 0x003f));
	}
#endif
	sub_io[addr - 0xd400] = dat;

	/* fBXvC */
	if (display_writeb(addr, dat)) {
		return;
	}
	/* L[{[h */
	if (keyboard_writeb(addr, dat)) {
		return;
	}
	/* _ZE */
#if XM7_VER >= 2
	if (aluline_writeb(addr, dat)) {
		return;
	}
#endif

	LOG2(Warning, "`oCg $%04X <- $%02X", addr, dat);
}

/*
 *	TuCPU
 *	Z[u
 */
BOOL FASTCALL submem_save(int fileh)
{
	LOG0(Normal, "Z[u");

#if XM7_VER >= 3
	/* ܂łƓŏ */
	if (mode400l) {
		if (!file_write(fileh, vram_c, 0x18000)) {
			return FALSE;
		}
	}
	else {
		if (!file_write(fileh, vram_c, 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x8000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x10000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x4000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0xc000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x14000], 0x4000)) {
			return FALSE;
		}
	}
#else
	if (!file_write(fileh, vram_c, 0x6000)) {
		return FALSE;
	}
	if (!file_write(fileh, &vram_c[0x6000], 0x6000)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_write(fileh, &vram_c[0xc000], 0x6000)) {
		return FALSE;
	}
	if (!file_write(fileh, &vram_c[0x12000], 0x6000)) {
		return FALSE;
	}
#endif
#endif

	if (!file_write(fileh, sub_ram, 0x1680)) {
		return FALSE;
	}

	if (!file_write(fileh, sub_io, 0x100)) {
		return FALSE;
	}

#if XM7_VER >= 2
	if (!file_byte_write(fileh, subrom_bank)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, cgrom_bank)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ver8g */
	if (mode400l) {
		if (!file_write(fileh, &vram_c[0x18000], 0x18000)) {
			return FALSE;
		}
	}
	else {
		if (!file_write(fileh, &vram_c[0x18000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x20000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x28000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x1c000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x24000], 0x4000)) {
			return FALSE;
		}
		if (!file_write(fileh, &vram_c[0x2c000], 0x4000)) {
			return FALSE;
		}
	}

	if (!file_write(fileh, subramde, 0x2000)) {
		return FALSE;
	}
	if (!file_write(fileh, subramcg, 0x4000)) {
		return FALSE;
	}
	if (!file_write(fileh, subramcn, 0x2000)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, cgram_bank)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, consram_bank)) {
		return FALSE;
	}
#endif
#endif

#if XM7_VER == 1 && defined(L4CARD)
	if (!file_write(fileh, tvram_c, 0x1000)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, detect_400linecard)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, enable_400linecard)) {
		return FALSE;
	}
#endif

#if XM7_VER == 1
	if (!file_write(fileh, subrampcg, 0x0800)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, pcg_flag)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, subrom_xchg)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	TuCPU
 *	[h
 */
BOOL FASTCALL submem_load(int fileh, int ver)
{
#if XM7_VER == 1
	BOOL tmp;
#endif

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

#if XM7_VER >= 3
	/* ܂łƓœǂݏo */
	if (!file_read(fileh, vram_c, 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x8000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x10000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x4000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0xc000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x14000], 0x4000)) {
		return FALSE;
	}
#else
	if (!file_read(fileh, vram_c, 0x6000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x6000], 0x6000)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_read(fileh, &vram_c[0xc000], 0x6000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x12000], 0x6000)) {
		return FALSE;
	}
#endif
#endif

	if (!file_read(fileh, sub_ram, 0x1680)) {
		return FALSE;
	}

	if (!file_read(fileh, sub_io, 0x100)) {
		return FALSE;
	}

#if XM7_VER >= 2
	if (!file_byte_read(fileh, &subrom_bank)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &cgrom_bank)) {
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ver8g */
	if ((ver < 800) || ((ver >= 1000) && (ver < 1400))) {
		consram_bank = 0;
		cgram_bank = 0;
		return TRUE;
	}

	/* ܂łƓœǂݏo */
	if (!file_read(fileh, &vram_c[0x18000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x20000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x28000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x1c000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x24000], 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, &vram_c[0x2c000], 0x4000)) {
		return FALSE;
	}

	if (!file_read(fileh, subramde, 0x2000)) {
		return FALSE;
	}
	if (!file_read(fileh, subramcg, 0x4000)) {
		return FALSE;
	}
	if (!file_read(fileh, subramcn, 0x2000)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &cgram_bank)) {
		return FALSE;
	}
	if (!file_byte_read(fileh, &consram_bank)) {
		return FALSE;
	}
	/* AV40/AV40EXȊO */
	if ((model_type != MODEL_FM77AV40) &&
		(model_type != MODEL_FM77AV40EX)) {
		if ((cgram_bank || consram_bank || (subrom_bank == 4))) {
			return FALSE;
		}
	}
#endif
#endif

#if XM7_VER == 1 && defined(L4CARD)
	if (!file_read(fileh, tvram_c, 0x1000)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &tmp)) {
		return FALSE;
	}

	/* Ver3.06g */
	if ((ver >= 407) || ((ver >= 306) && (ver <= 399))) {
		if (!file_bool_read(fileh, &enable_400linecard)) {
			return FALSE;
		}
	}
	else {
		/* LtOۑĂȂꍇtOőp */
		if (detect_400linecard && tmp) {
			enable_400linecard = tmp;
		}
		else {
			enable_400linecard = FALSE;
		}
	}
#endif

#if XM7_VER == 1
#ifdef L4CARD
	/* Ver3.10/10.09g */
	if (((ver >= 310) && (ver <= 399)) ||
		((ver >= 1009) && (ver <= 1199))) {
#else
	/* Ver3.12/10.09g */
	if (((ver >= 312) && (ver <= 399)) ||
		((ver >= 1009) && (ver <= 1199))) {
#endif
		if (!file_read(fileh, subrampcg, 0x0800)) {
			return FALSE;
		}
	}

	/* Ver3.12/10.09g */
	if (((ver >= 312) && (ver <= 399)) ||
		((ver >= 1009) && (ver <= 1199))) {
		if (!file_bool_read(fileh, &pcg_flag)) {
			return FALSE;
		}
	}

	/* Ver3.14/10.02g */
	if (((ver >= 314) && (ver <= 399)) ||
		((ver >= 1002) && (ver <= 1199))) {
		if (!file_bool_read(fileh, &tmp)) {
			return FALSE;
		}
	}
	else {
		tmp = FALSE;
	}
	if (available_fm7roms) {
		subrom_xchg = tmp;
	}
	else {
		subrom_xchg = FALSE;
	}
#endif

	return TRUE;
}
