/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *	line printer support 2010-2025 by Ben.JP
 *
 *	[ JZbge[vv^ ]
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "tapelp.h"
#include "mainetc.h"
#include "device.h"
#include "event.h"

/*
 *	O[o [N
 */
BOOL tape_in;							/* e[v ̓f[^ */
BOOL tape_out;							/* e[v o̓f[^ */
BOOL tape_motor;						/* e[v [^ */
BOOL tape_rec;							/* e[v RECtO */
BOOL tape_writep;						/* e[v ݋֎~ */
WORD tape_count;						/* e[v TCNJE^ */
DWORD tape_subcnt;						/* e[v TuJE^ */
int tape_fileh;							/* e[v t@Cnh */
DWORD tape_offset;						/* e[v t@CItZbg */
char tape_fname[256+1];					/* e[v t@Cl[ */

WORD tape_incnt;						/* e[v ǂݍ݃JE^ */
DWORD tape_fsize;						/* e[v t@CTCY */
BOOL tape_fetch;						/* e[v f[^tFb`tO */
BYTE *tape_savebuf;						/* e[v ݃obt@ */
WORD tape_saveptr;						/* e[v ݃|C^ */

BOOL tape_monitor;						/* e[v e[vj^tO */
BOOL tape_sound;						/* e[v [o̓tO */
#if XM7_VER == 1
BOOL tape_lowspeed;						/* e[v [^I1.2MHztO */
#endif

#ifdef LPRINT
BYTE lp_mode;							/* v^ ݒ胂[h */
#endif
BYTE lp_data;							/* v^ o̓f[^ */
BOOL lp_busy;							/* v^ BUSYtO */
BOOL lp_error;							/* v^ G[tO */
BOOL lp_pe;								/* v^ PEtO */
BOOL lp_ackng;							/* v^ ACKtO */
BOOL lp_online;							/* v^ IC */
BOOL lp_strobe;							/* v^ Xg[u */
int lp_fileh;							/* v^ t@Cnh */
BYTE lp_joysticktype;					/* v^ WCXeBbN^Cv */

char lp_fname[256+1];					/* v^ t@Cl[ */

/*
 *	X^eBbN [N
 */
#ifdef LPRINT
static BYTE lp_openuse;					/* v^ OPEN̎gp[h */
#endif

/*
 *	vg^Cv錾
 */
static void FASTCALL tape_flush(void);	/* e[v݃obt@tbV */

/*
 *	ID
 */
#define	ID	MAKEID('T', 'P', 'L', 'P')
//#define TAPELP_LOG

/*
 *	JZbge[vv^
 *	
 */
BOOL FASTCALL tapelp_init(void)
{
	/* [NGA */
	tape_savebuf = NULL;

	/* e[v */
	tape_fileh = -1;
	tape_fname[0] = '\0';
	tape_offset = 0;
	tape_fsize = 0;
	tape_writep = FALSE;
	tape_fetch = FALSE;
	tape_saveptr = 0;
	tape_monitor = FALSE;
	tape_motor = FALSE;
	tape_sound = FALSE;
#if XM7_VER == 1
	tape_lowspeed = TRUE;
#endif

	/* v^ */
	lp_fileh = -1;
#ifdef LPRINT
	lp_fname[0] = '\0';
	lp_init();

	lp_mode = LP_ENABLE;
	lp_openuse = LP_ENABLE;
	lp_setfile(LP_TEMPFILENAME);
#else
	lp_fname[0] = '\0';
#endif

	/* e[v݃obt@ */
	tape_savebuf = (BYTE *)malloc(TAPE_SAVEBUFSIZE);
	if (tape_savebuf == NULL) {
		return FALSE;
	}

	/* ftHg͓dgVJOY I/F */
	lp_joysticktype = LPJOY_DEMPA;

	return TRUE;
}

/*
 *	JZbge[vv^
 *	N[Abv
 */
void FASTCALL tapelp_cleanup(void)
{
#if 0
	ASSERT(tape_savebuf);
#endif

	/* t@CJĂ΁A */
	if (tape_fileh != -1) {
		tape_flush();
		file_close(tape_fileh);
		tape_fileh = -1;
	}

	/* [^OFF */
	tape_motor = FALSE;

	/* t@CJĂ΁A */
	lp_setfile(NULL);

	/* rŎsꍇl */
	if (tape_savebuf) {
		free(tape_savebuf);
	}
}

/*
 *	JZbge[vv^
 *	Zbg
 */
void FASTCALL tapelp_reset(void)
{
	LOG0(Normal, "Zbg");

	/* o̓f[^΃tbV */
	tape_flush();

	/* JZbg[^OFF */
	if (tape_motor && tape_sound) {
		wav_notify(SOUND_CMTMOTOROFF);
	}

	tape_motor = FALSE;
	tape_rec = FALSE;
	tape_count = 0;
	tape_in = TRUE;
	tape_out = FALSE;
	tape_incnt = 0;
	tape_subcnt = 0;
	tape_fetch = FALSE;

	lp_busy = FALSE;
	lp_error = FALSE;
	lp_ackng = TRUE;
	lp_pe = FALSE;
	lp_online = FALSE;
	lp_strobe = FALSE;
}

/*-[ v^ ]-------------------------------------------------------------*/

/*
 *	v^
 *	f[^o
 */
static void FASTCALL lp_output(BYTE dat)
{
	/* I[v`FbN */
	if (lp_fileh == -1) {
		return;
	}

	/* Ayh */
#ifdef LPRINT
	switch (lp_mode) {
		case LP_DISABLE:
		case LP_JASTSOUND:
			break;
		case LP_ENABLE:
			lp_writefile(&dat, 1);
			break;
		case LP_LOG:
			file_write(lp_fileh, &dat, 1);
			break;
		default:
			ASSERT(FALSE);
	}
#else
	file_write(lp_fileh, &dat, 1);
#endif
}

/*
 *	v^
 *	t@Cݒ
 */
void FASTCALL lp_setfile(char *fname)
{
	/* xJĂ΁A */
	if (lp_fileh != -1) {
#ifdef LPRINT
//		switch (lp_mode) {
		switch (lp_openuse) {
			case LP_DISABLE:
			case LP_JASTSOUND:
				break;
			case LP_ENABLE:
				lp_closefile();
				break;
			case LP_LOG:
				file_close(lp_fileh);
				break;
			default:
				break; /* ASSERT(FALSE); */
		}
#else
		file_close(lp_fileh);
#endif
		lp_fileh = -1;
	}

#ifdef LPRINT
	/* pꎞt@C폜 */
	if ((lp_openuse == LP_ENABLE) && (lp_fileh == -1)) {
		lp_removefile();
	}
#endif

	/* t@CZbg */
	if (fname == NULL) {
		lp_fname[0] = '\0';
		return;
	}
	if (strlen(fname) < sizeof(lp_fname)) {
		strcpy(lp_fname, fname);
	}
	else {
		lp_fname[0] = '\0';
	}

	/* I[vĂȂ΁CJ */
	if (lp_fileh == -1) {
		if (lp_fname[0] != '\0') {
#ifdef LPRINT
			switch (lp_mode) {
				case LP_DISABLE:
				case LP_JASTSOUND:
					break;
				case LP_ENABLE:
					lp_fileh = lp_openfile(lp_fname);
					break;
				case LP_LOG:
					lp_fileh = file_open(lp_fname, OPEN_W);
					break;
				default:
					ASSERT(FALSE);
			}

			lp_openuse = lp_mode;
#else
			lp_fileh = file_open(lp_fname, OPEN_W);
#endif
		}
	}
}

#ifdef LPRINT
/*
 *	v^
 *	
 */
void FASTCALL lp_print(void)
{
	if ((lp_mode == LP_ENABLE) && (lp_fileh != -1)) {
		lp_printfile();
	}
}
#endif

/*-[ e[v ]---------------------------------------------------------------*/

/*
 *	e[v
 *	݃obt@̃tbV
 */
static void FASTCALL tape_flush(void)
{
	if (tape_fileh != -1) {
		/* ^ԂŃobt@Ƀf[^΃t@Cɏo */
		if ((tape_rec) && (tape_saveptr > 0)) {
			file_write(tape_fileh, tape_savebuf, tape_saveptr);
		}
	}

	/* ݃|C^ */
	tape_saveptr = 0;
}

/*
 *	e[v
 *	PoCg
 */
static void FASTCALL tape_byte_write(BYTE dat)
{
	/* obt@Ƀf[^ǉ */
	tape_savebuf[tape_saveptr++] = dat;

	/* obt@ςɂȂf[^o */
	if (tape_saveptr >= TAPE_SAVEBUFSIZE) {
		tape_flush();
	}
}

/*
 *	e[v
 *	f[^
 */
static void FASTCALL tape_input(BOOL flag)
{
	BYTE high;
	BYTE low;
	WORD dat;

	/* [^Ă邩 */
	if (!tape_motor) {
		return;
	}

	/* ^ĂΓ͂łȂ */
	if (tape_rec) {
		return;
	}

	/* {ԂłȂꍇAɃf[^tFb`Ăꍇ͉Ȃ */
	if (!flag && tape_fetch) {
		return;
	}

	/* VOJE^̓JE^zĂ΁A0ɂ */
	while (tape_count >= tape_incnt) {
		tape_count -= tape_incnt;
		tape_incnt = 0;

		/* f[^tFb` */
		tape_in = FALSE;

		if (tape_fileh == -1) {
			return;
		}

		if (tape_offset >= tape_fsize) {
			return;
		}

		if (!file_seek(tape_fileh, tape_offset)) {
			return;
		}
		if (!file_read(tape_fileh, &high, 1)) {
			return;
		}
		if (!file_read(tape_fileh, &low, 1)) {
			return;
		}

		/* f[^ݒ */
		dat = (WORD)(high * 256 + low);
		if (dat > 0x7fff) {
			tape_in = TRUE;
		}

		/* f[^tFb`ς݃tOݒ */
		tape_fetch = !flag;

		/* {Ԃ̓͂łȂꍇ͂܂ */
		if (!flag) {
			return;
		}

		/* JE^ݒ */
		tape_incnt = (WORD)(dat & 0x7fff);

		/* JE^J肷 */
		if (tape_count > tape_incnt) {
			tape_count -= tape_incnt;
			tape_incnt = 0;
		}
		else {
			tape_incnt -= tape_count;
			tape_count = 0;
		}

		/* ItZbgXV */
		tape_offset += 2;
	}
}

/*
 *	e[v
 *	f[^o
 */
static void FASTCALL tape_output(BOOL flag)
{
	WORD dat;
	BYTE high, low;

	/* e[vĂ邩 */
	if (!tape_motor) {
		return;
	}

	/* ^ */
	if (!tape_rec) {
		return;
	}

	/* JE^Ă邩 */
	if (tape_count == 0) {
		return;
	}

	/* ݉\ */
	if (tape_writep) {
		return;
	}

	/* t@CI[vĂ΁Af[^ */
	dat = tape_count;
	if (dat >= 0x8000) {
		dat = 0x7fff;
	}
	if (flag) {
		dat |= 0x8000;
	}
	high = (BYTE)(dat >> 8);
	low = (BYTE)(dat & 0xff);
	if (tape_fileh != -1) {
		tape_byte_write(high);
		tape_byte_write(low);

		tape_offset += 2;
		if (tape_offset >= tape_fsize) {
			tape_fsize = tape_offset;
		}
	}

	/* JE^Zbg */
	tape_count = 0;
	tape_subcnt = 0;
}

/*
 *	e[v
 *	}[Jo
 */
static void FASTCALL tape_mark(void)
{
	/* e[vĂ邩 */
	if (!tape_motor) {
		return;
	}

	/* ^ */
	if (!tape_rec) {
		return;
	}

	/* ݉\ */
	if (tape_writep) {
		return;
	}

	/* t@CI[vĂ΁Af[^ */
	if (tape_fileh != -1) {
		tape_byte_write(0);
		tape_byte_write(0);

		tape_offset += 2;
		if (tape_offset >= tape_fsize) {
			tape_fsize = tape_offset;
		}
	}
}

/*
 *	e[v
 *	߂
 */
void FASTCALL tape_rew(void)
{
	WORD dat;

	/*  */
	if (tape_fileh == -1) {
		return;
	}

	/* assert */
	ASSERT(tape_fsize >= 16);
	ASSERT(tape_offset >= 16);
	ASSERT(!(tape_fsize & 0x01));
	ASSERT(!(tape_offset & 0x01));

	/* ^Ȃ炢tbV */
	tape_flush();

	while (tape_offset > 16) {
		/* QoCgOɖ߂Aǂݍ */
		tape_offset -= 2;
		if (!file_seek(tape_fileh, tape_offset)) {
			return;
		}
		file_read(tape_fileh, (BYTE *)&dat, 2);

		/* $0000ȂAɐݒ */
		if (dat == 0) {
			file_seek(tape_fileh, tape_offset);
			return;
		}

		/* ܓǂݍ񂾕߂ */
		if (!file_seek(tape_fileh, tape_offset)) {
			return;
		}
	}
}

/*
 *	e[v
 *	ŏ܂Ŋ߂
 */
void FASTCALL tape_rewtop(void)
{
	/*  */
	if (tape_fileh == -1) {
		return;
	}

	/* assert */
	ASSERT(tape_fsize >= 16);
	ASSERT(tape_offset >= 16);
	ASSERT(!(tape_fsize & 0x01));
	ASSERT(!(tape_offset & 0x01));

	/* ^Ȃ炢tbV */
	tape_flush();

	/* Iɍŏ܂Ŋ߂ */
	tape_offset = 16;
	file_seek(tape_fileh, tape_offset);
}

/*
 *	e[v
 *	
 */
void FASTCALL tape_ff(void)
{
	WORD dat;

	/*  */
	if (tape_fileh == -1) {
		return;
	}

	/* assert */
	ASSERT(tape_fsize >= 16);
	ASSERT(tape_offset >= 16);
	ASSERT(!(tape_fsize & 0x01));
	ASSERT(!(tape_offset & 0x01));

	/* ^Ȃ炢tbV */
	tape_flush();

	while (tape_offset < tape_fsize) {
		/* ֐i߂ */
		tape_offset += 2;
		if (tape_offset >= tape_fsize){
			return;
		}
		if (!file_seek(tape_fileh, tape_offset)) {
			return;
		}
		file_read(tape_fileh, (BYTE *)&dat, 2);

		/* $0000ȂA̎ɐݒ */
		if (dat == 0) {
			tape_offset += 2;
			if (tape_offset >= tape_fsize) {
				tape_fsize = tape_offset;
			}
			return;
		}
	}
}

#if 0
/*
 *	e[v
 *	߂()
 */
void FASTCALL tape_rew2(void)
{
	WORD dat;
	WORD flag;

	/*  */
	if (tape_fileh == -1) {
		return;
	}

	/* assert */
	ASSERT(tape_fsize >= 16);
	ASSERT(tape_offset >= 16);
	ASSERT(!(tape_fsize & 0x01));
	ASSERT(!(tape_offset & 0x01));

	/* ^Ȃ炢tbV */
	tape_flush();

	/* f[^Ԃ */
	flag = -1;

	while (tape_offset > 16) {
		/* QoCgOɖ߂Aǂݍ */
		tape_offset -= 2;
		if (!file_seek(tape_fileh, tape_offset)) {
			return;
		}
		file_read(tape_fileh, (BYTE *)&dat, 2);

		/* ΂炭f[^ωĂȂΕω܂Ŗ߂ */
		if (((WORD)(dat & 0x8000) == flag) &&
			((WORD)(dat & 0x7fff) == 0x7fff)) {
			while ((WORD)(dat & 0x8000) == flag) {
				tape_offset -= 2;
				if (!file_seek(tape_fileh, tape_offset)) {
					return;
				}
				file_read(tape_fileh, (BYTE *)&dat, 2);
			}
			tape_offset += 2;
			if (!file_seek(tape_fileh, tape_offset)) {
				return;
			}
			return;
		}

		/* f[^Ԃۑ */
		flag = (WORD)(dat & 0x8000);
	}
}

/*
 *	e[v
 *	()
 */
void FASTCALL tape_ff2(void)
{
	WORD dat;
	WORD flag;

	/*  */
	if (tape_fileh == -1) {
		return;
	}

	/* assert */
	ASSERT(tape_fsize >= 16);
	ASSERT(tape_offset >= 16);
	ASSERT(!(tape_fsize & 0x01));
	ASSERT(!(tape_offset & 0x01));

	/* ^Ȃ炢tbV */
	tape_flush();

	/* f[^Ԃ */
	flag = -1;

	while (tape_offset < tape_fsize) {
		/* ֐i߂ */
		tape_offset += 2;
		if (tape_offset >= tape_fsize){
			return;
		}
		if (!file_seek(tape_fileh, tape_offset)) {
			return;
		}
		file_read(tape_fileh, (BYTE *)&dat, 2);

		/* ΂炭f[^ωĂΕω܂Ői߂ */
		if (((WORD)(dat & 0x8000) == flag) &&
			((WORD)(dat & 0x7fff) == 0x7fff)) {
			do {
				tape_offset += 2;
				if (tape_offset >= tape_fsize){
					return;
				}
				if (!file_seek(tape_fileh, tape_offset)) {
					return;
				}
				file_read(tape_fileh, (BYTE *)&dat, 2);
			} while ((WORD)(dat & 0x8000) == flag);
			tape_offset += 2;
			if (tape_offset >= tape_fsize) {
				tape_fsize = tape_offset;
			}
			return;
		}

		/* f[^Ԃۑ */
		flag = (WORD)(dat & 0x8000);
	}
}
#endif

/*
 *	e[v
 *	t@Cݒ
 */
void FASTCALL tape_setfile(char *fname)
{
	char *header = "XM7 TAPE IMAGE 0";
	char buf[17];

	/* xJĂ΁A */
	if (tape_fileh != -1) {
		tape_flush();
		file_close(tape_fileh);
		tape_fileh = -1;
		tape_writep = FALSE;
	}

	/* t@CZbg */
	if (fname == NULL) {
		tape_fname[0] = '\0';
	}
	else {
		if (strlen(fname) < sizeof(tape_fname)) {
			strcpy(tape_fname, fname);
		}
		else {
			tape_fname[0] = '\0';
		}
	}

	/* t@CI[v݂ */
	if (tape_fname[0] != '\0') {
		tape_fileh = file_open(tape_fname, OPEN_RW);
		if (tape_fileh != -1) {
			tape_writep = FALSE;
		}
		else {
			tape_fileh = file_open(tape_fname, OPEN_R);
			tape_writep = TRUE;
		}
	}

	/* JĂ΁Awb_ǂݍ݃`FbN */
	if (tape_fileh != -1) {
		memset(buf, 0, sizeof(buf));
		file_read(tape_fileh, (BYTE*)buf, 16);
		if (strcmp(buf, header) != 0) {
			file_close(tape_fileh);
			tape_fileh = -1;
			tape_writep = FALSE;
		}
	}

	/* tȌ */
	tape_setrec(FALSE);
	tape_count = 0;
	tape_incnt = 0;
	tape_subcnt = 0;

	/* t@CJĂ΁At@CTCYAItZbg */
	if (tape_fileh != -1) {
		tape_fsize = file_getsize(tape_fileh);
		tape_offset = 16;
	}
}

/*
 *	e[v
 *	^tOݒ
 */
void FASTCALL tape_setrec(BOOL flag)
{
	/* [^Ă΁A}[J */
	if (tape_motor && !tape_rec) {
		if (flag) {
			tape_rec = TRUE;
			tape_mark();
			return;
		}
	}
	else {
		/* ^IȂA݃obt@tbV */
		if (tape_motor && tape_rec) {
			if (!flag) {
				tape_flush();
			}
		}
	}

	tape_rec = flag;
}

/*
 *	e[v
 *	TEh
 */
static BOOL FASTCALL tape_outsnd(void)
{
	if (tape_motor) {
		if (tape_rec) {
			/* ^ */
			if (!tape_writep) {
				tape_notify(tape_out);
			}
		}
		else {
			/* Đ */
			tape_input(FALSE);
			tape_notify(tape_in);
		}
	}

	return TRUE;
}

/*-[ R/W ]------------------------------------------------------------*/

/*
 *	JZbge[vv^
 *	PoCgǂݏo
 */
BOOL FASTCALL tapelp_readb(WORD addr, BYTE *dat)
{
	BYTE ret;
	BYTE joy;

	/* AhX`FbN */
	if (addr != 0xfd02) {
		return FALSE;
	}

	/* v^ Xe[^X쐬 */
	ret = 0x70;
	if (lp_busy) {
		ret |= 0x01;
	}
	if (!lp_error) {
		ret |= 0x02;
	}
	if (!lp_ackng) {
		ret |= 0x04;
	}
	if (lp_pe) {
		ret |= 0x08;
	}

	/* v^ڑȂAWCXeBbN */
#ifdef LPRINT
	if (lp_mode == LP_DISABLE) {
#else
	if ((lp_fileh == -1) || (lp_fname[0] == '\0')) {
#endif
		switch (lp_joysticktype) {
			case LPJOY_DEMPA:
				/* dgVЃWCXeBbN */
				/*  */
				ret |= 0x0f;
				/* 擾 */
				joy = joy_request(2);

				/* E */
				if (!(lp_data & 0x01) && (joy & 0x08)) {
					ret &= ~0x08;
				}
				/*  */
				if (!(lp_data & 0x02) && (joy & 0x04)) {
					ret &= ~0x08;
				}
				/*  */
				if (!(lp_data & 0x04) && (joy & 0x01)) {
					ret &= ~0x08;
				}
				/*  */
				if (!(lp_data & 0x08) && (joy & 0x02)) {
					ret &= ~0x08;
				}
				/* J2 */
				if (!(lp_data & 0x10) && (joy & 0x20)) {
					ret &= ~0x08;
				}
				/* J1 */
				if (!(lp_data & 0x20) && (joy & 0x10)) {
					ret &= ~0x08;
				}
				break;

			case LPJOY_ASCII:
				/* ASCII QuickShot I/F */
				/*  */
				ret |= 0x3f;
				/* 擾 */
				joy = joy_request(3);

				if (lp_online) {
					/* E */
					if (joy & 0x08) {
						ret &= ~0x20;
					}
					/*  */
					if (joy & 0x04) {
						ret &= ~0x10;
					}
					/*  */
					if (joy & 0x01) {
						ret &= ~0x08;
					}
					/*  */
					if (joy & 0x02) {
						ret &= ~0x04;
					}
					/* J2 */
					if (joy & 0x20) {
						ret &= ~0x02;
					}
					/* J1 */
					if (joy & 0x10) {
						ret &= ~0x01;
					}
				}
				break;

			case LPJOY_OHFM:
				/* h!FM JOY I/F */
				/*  */
				ret |= 0x0f;

				/* 1P擾 */
				joy = joy_request(4);

				/* 1P E */
				if (!(lp_data & 0x02) && (joy & 0x08)) {
					ret &= ~0x02;
				}
				/* 1P  */
				if (!(lp_data & 0x02) && (joy & 0x04)) {
					ret &= ~0x08;
				}
				/* 1P  */
				if (!(lp_data & 0x01) && (joy & 0x01)) {
					ret &= ~0x02;
				}
				/* 1P  */
				if (!(lp_data & 0x01) && (joy & 0x02)) {
					ret &= ~0x08;
				}
				/* 1P J2 */
				if (!(lp_data & 0x0c) && (joy & 0x20)) {
					ret &= ~0x02;
				}
				/* 1P J1 */
				if (!(lp_data & 0x0c) && (joy & 0x10)) {
					ret &= ~0x08;
				}

				/* 2P擾 */
				joy = joy_request(5);

				/* 2P E */
				if (!(lp_data & 0x20) && (joy & 0x08)) {
					ret &= ~0x02;
				}
				/* 2P  */
				if (!(lp_data & 0x20) && (joy & 0x04)) {
					ret &= ~0x08;
				}
				/* 2P  */
				if (!(lp_data & 0x10) && (joy & 0x01)) {
					ret &= ~0x02;
				}
				/* 2P  */
				if (!(lp_data & 0x10) && (joy & 0x02)) {
					ret &= ~0x08;
				}
				/* 2P J2 */
				if (!(lp_data & 0xc0) && (joy & 0x20)) {
					ret &= ~0x02;
				}
				/* 2P J1 */
				if (!(lp_data & 0xc0) && (joy & 0x10)) {
					ret &= ~0x08;
				}
				break;
		}
	}

	/* JZbg f[^쐬 */
	tape_input(TRUE);
	if (tape_in) {
		ret |= 0x80;
	}

	/* ok */
	*dat = ret;
	return TRUE;
}

/*
 *	JZbge[vv^
 *	PoCg
 */
BOOL FASTCALL tapelp_writeb(WORD addr, BYTE dat)
{
	switch (addr) {
		/* JZbgAv^ */
		case 0xfd00:
			/* v^ IC */
			if (dat & 0x80) {
				lp_online = FALSE;
			}
			else {
				lp_online = TRUE;
			}

			/* v^ Xg[u */
			if (dat & 0x40) {
				lp_strobe = TRUE;
			}
			else {
				if (lp_strobe && lp_online) {
					lp_output(lp_data);
					mainetc_lp();
				}
				lp_strobe = FALSE;
			}

			/* e[v o̓f[^ */
			if (dat & 0x01) {
				if (!tape_out) {
					tape_output(FALSE);
				}
				tape_out = TRUE;
			}
			else {
				if (tape_out) {
					tape_output(TRUE);
				}
				tape_out = FALSE;
			}

			/* e[v [^ */
			if (dat & 0x02) {
				if (!tape_motor) {
					/* VKX^[g */
					tape_count = 0;
					tape_subcnt = 0;
					tape_motor = TRUE;
					if (tape_rec) {
						tape_mark();
					}
					if (tape_sound) {
						wav_notify(SOUND_CMTMOTORON);
					}
					schedule_setevent(EVENT_TAPEMON, 40, tape_outsnd);
				}
			}
			else {
				if (tape_motor && tape_sound) {
					wav_notify(SOUND_CMTMOTOROFF);
				}
				schedule_delevent(EVENT_TAPEMON);

				/* [^~ */
				tape_motor = FALSE;
				tape_flush();
			}

#if XM7_VER == 1
			/* PCG */
			if (dat & 0x08) {
				pcg_flag = TRUE;
			}
			else {
				pcg_flag = FALSE;
			}
#endif

			return TRUE;

		/* v^o̓f[^ */
		case 0xfd01:
#if defined(LPRINT) && defined(JASTSOUND)
			/* WXgTEho(strobeM͍lȂ) */
			if (lp_mode == LP_JASTSOUND) {
				dac_notify(dat);
			}
#endif

			lp_data = dat;
			return TRUE;
	}

	return FALSE;
}

/*
 *	JZbge[vv^
 *	Z[u
 */
BOOL FASTCALL tapelp_save(int fileh)
{
	BOOL tmp;

	LOG0(Normal, "Z[u");

	/* Xe[gZ[uOɏ݃obt@tbV */
	tape_flush();

	if (!file_bool_write(fileh, tape_in)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, tape_out)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, tape_motor)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, tape_rec)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, tape_writep)) {
		return FALSE;
	}
	if (!file_word_write(fileh, tape_count)) {
		return FALSE;
	}
	if (!file_dword_write(fileh, tape_subcnt)) {
		return FALSE;
	}

	if (!file_dword_write(fileh, tape_offset)) {
		return FALSE;
	}
	if (!file_write(fileh, (BYTE*)tape_fname, 256 + 1)) {
		return FALSE;
	}

	tmp = (tape_fileh != -1);
	if (!file_bool_write(fileh, tmp)) {
		return FALSE;
	}

	if (!file_word_write(fileh, tape_incnt)) {
		return FALSE;
	}
	if (!file_dword_write(fileh, tape_fsize)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_bool_write(fileh, tape_lowspeed)) {
		return FALSE;
	}
#endif

	if (!file_byte_write(fileh, lp_data)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_busy)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_error)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_pe)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_ackng)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_online)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lp_strobe)) {
		return FALSE;
	}

	if (!file_write(fileh, (BYTE*)lp_fname, 256 + 1)) {
		return FALSE;
	}
	tmp = (lp_fileh != -1);
	if (!file_bool_write(fileh, tmp)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	JZbge[vv^
 *	[h
 */
BOOL FASTCALL tapelp_load(int fileh, int ver)
{
	DWORD offset;
	char fname[256 + 1];
#if XM7_VER >= 2
	BYTE tmp;
#endif
	BOOL flag;
	int pathlen;

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	/* t@C̍ő啶 */
#if XM7_VER >= 3
	if (((ver >= 715) && (ver <= 799)) || (ver >= 915)) {
#elif XM7_VER >= 2
	if (ver >= 715) {
#else
	if (((ver >= 305) && (ver <= 399)) || (ver >= 404)) {
#endif
		pathlen = 256;
	}
	else {
		pathlen = 128;
	}

	if (!file_bool_read(fileh, &tape_in)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &tape_out)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &tape_motor)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &tape_rec)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &tape_writep)) {
		return FALSE;
	}
	if (!file_word_read(fileh, &tape_count)) {
		return FALSE;
	}
#if XM7_VER >= 2
#if XM7_VER >= 3
	if ((ver >= 906) || ((ver >= 706) && (ver <= 799))) {
#else
	if (ver >= 706) {
#endif
		if (!file_dword_read(fileh, &tape_subcnt)) {
			return FALSE;
		}
	}
	else {
		if (!file_byte_read(fileh, &tmp)) {
			return FALSE;
		}
		tape_subcnt = (tmp << 4);
	}
#else
	if (!file_dword_read(fileh, &tape_subcnt)) {
		return FALSE;
	}
#endif

	if (!file_dword_read(fileh, &offset)) {
		return FALSE;
	}
	if (!file_read(fileh, (BYTE*)fname, pathlen + 1)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &flag)) {
		return FALSE;
	}

	/* }Eg */
	tape_setfile(NULL);
	if (flag) {
		tape_setfile(fname);
		if ((tape_fileh != -1) && ((tape_fsize + 1) >= offset)) {
			file_seek(tape_fileh, offset);
			tape_offset = offset;
		}
	}

	if (!file_word_read(fileh, &tape_incnt)) {
		return FALSE;
	}
	/* tape_fsize͖ */
	if (!file_dword_read(fileh, &offset)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (ver >= 1002) {
		if (!file_bool_read(fileh, &tape_lowspeed)) {
			return FALSE;
		}
	}
#endif

	if (!file_byte_read(fileh, &lp_data)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_busy)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_error)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_pe)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_ackng)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_online)) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &lp_strobe)) {
		return FALSE;
	}
	if (!file_read(fileh, (BYTE*)fname, pathlen + 1)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if (((ver >= 720) && (ver <= 799)) ||
		((ver >= 920) && (ver <= 999)) ||
		((ver >= 1209) && (ver <= 1399)) ||
		 (ver >= 1409)) {
#elif XM7_VER >= 2
	if (((ver >= 720) && (ver <= 799)) ||
		((ver >= 1209) && (ver <= 1399))) {
#else
	if (((ver >= 310) && (ver <= 399)) ||
		((ver >= 1009) && (ver <= 1199))) {
#endif
		if (!file_bool_read(fileh, &flag)) {
			return FALSE;
		}

		/* }Eg */
		lp_setfile(NULL);
		if (flag) {
			lp_setfile(fname);
		}
	}

	schedule_handle(EVENT_TAPEMON, tape_outsnd);

	/* ̑̃[NGA */
	tape_saveptr = 0;
	tape_fetch = FALSE;

	return TRUE;
}
