/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API Rg[o[ ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <winbase.h>
#include <commctrl.h>
#include <stdlib.h>
#include <assert.h>
#include "xm7.h"
#include "keyboard.h"
#include "tapelp.h"
#include "display.h"
#include "ttlpalet.h"
#include "apalet.h"
#include "subctrl.h"
#include "fdc.h"
#include "bubble.h"
#include "mouse.h"
#include "opn.h"
#include "tools.h"
#include "w32.h"
#include "w32_bar.h"
#include "w32_sch.h"
#include "w32_draw.h"
#include "w32_res.h"

/*
 *	O[o [N
 */
HWND hStatusBar;						/* Xe[^Xo[ */
BOOL bStatusMessage;					/* Xe[^XbZ[W\ */
BOOL bStatusBar;						/* Xe[^Xo[\ */
BOOL bMountFileName;					/* ^Cgo[Ƀ}Egt@C */
#if XM7_VER == 1 && defined(SFDC)
int uPaneX[9];							/* Drag&DroppyCʒu */
#else	/* XM7_VER == 1 && defined(SFDC) */
int uPaneX[5];							/* Drag&DroppyCʒu */
#endif	/* XM7_VER == 1 && defined(SFDC) */

/*
 *	X^eBbN [N
 */
static char szIdleMessage[128];			/* AChbZ[W */
static char szStatusMessage[128];		/* Xe[^XbZ[W */
static char szRunMessage[128];			/* RUNbZ[W */
static char szStopMessage[128];			/* STOPbZ[W */
static char szCaption[256];				/* LvV */
static char szCAPMessage[16];			/* CAPL[bZ[W */
static char szKANAMessage[16];			/* ȃL[bZ[W */
static char szINSMessage[16];			/* INSL[bZ[W */
static int nCAP;						/* CAPL[ */
static int nKANA;						/* ȃL[ */
static int nINS;						/* INSL[ */
#if XM7_VER == 1 && defined(SFDC)
static int nDrive[2][4];				/* tbs[hCu */
static char szDrive[2][4][128];			/* tbs[hCu */
#else	/* XM7_VER == 1 && defined(SFDC) */
static int nDrive[4];					/* tbs[hCu */
static char szDrive[4][128];			/* tbs[hCu */
#endif	/* XM7_VER == 1 && defined(SFDC) */
static UINT nDcw;						/* hCu\\ */
#if XM7_VER == 1 && defined(BUBBLE)
static int nUnit[2][4];					/* ouz_jbg */
static int nUnitP[2][4];				/* ouz_jbgCgveNg */
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
static int nTape;						/* e[v */
static UINT uResID;						/* \[XIDۑ */
static int nStatusBorder;				/* Xe[^Xo[̃{[_[` */
static BOOL (*IsAppThemed)(void);		/* e[}KpmF֐(uxtheme.dll) */
static HINSTANCE hInstTheme;			/* uxtheme.dll̃nh */
static BOOL bThemed;					/* e[}Kp */

/*-[ Xe[^Xo[ ]-------------------------------------------------------*/

/*
 *	ANZX}N
 */
#define Status_SetParts(hwnd, nParts, aWidths) \
	SendMessage((hwnd), SB_SETPARTS, (WPARAM) nParts, (LPARAM) (LPINT) aWidths)

#define Status_SetText(hwnd, iPart, uType, szText) \
	SendMessage((hwnd), SB_SETTEXT, (WPARAM) (iPart | uType), (LPARAM) (LPSTR) szText)

/*
 *	yC`
 */
enum {
		PANE_DEFAULT = 0,
#if XM7_VER == 1 && defined(SFDC)
		PANE_SDRIVE3,
		PANE_SDRIVE2,
		PANE_SDRIVE1,
		PANE_SDRIVE0,
#endif	/* XM7_VER == 1 && defined(SFDC) */
		PANE_DRIVE3,
		PANE_DRIVE2,
		PANE_DRIVE1,
		PANE_DRIVE0,
		PANE_TAPE,
#if XM7_VER == 1 && defined(BUBBLE)
		PANE_BUBBLE32,
		PANE_BUBBLE128,
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
		PANE_CAP,
		PANE_KANA,
		PANE_INS,
		PANE_MAX
};

/*
 *	Xe[^Xo[p̃bZ[W[h
 */
void FASTCALL LoadMessageStatusBar(void)
{
	/* bZ[W[h */
	if (GetMsg(hResInstance, IDS_IDLEMESSAGE,
					szIdleMessage, sizeof(szIdleMessage)) == 0) {
		szIdleMessage[0] = '\0';
	}
	if (GetMsg(hResInstance, IDS_RUNCAPTION,
					szRunMessage, sizeof(szRunMessage)) == 0) {
		szRunMessage[0] = '\0';
	}
	if (GetMsg(hResInstance, IDS_STOPCAPTION,
					szStopMessage, sizeof(szStopMessage)) == 0) {
		szStopMessage[0] = '\0';
	}
	if (GetMsg(hResInstance, IDS_BAR_CAP,
					szCAPMessage, sizeof(szCAPMessage)) == 0) {
		strncpy(szCAPMessage, "CAP", 16);
	}
	if (GetMsg(hResInstance, IDS_BAR_KANA,
					szKANAMessage, sizeof(szKANAMessage)) == 0) {
#if XM7_VER == 1
		strncpy(szKANAMessage, "Ji", 16);
#else
		strncpy(szKANAMessage, "", 16);
#endif
	}
	if (GetMsg(hResInstance, IDS_BAR_INS,
					szINSMessage, sizeof(szINSMessage)) == 0) {
		strncpy(szINSMessage, "INS", 16);
	}
}

/*
 *	Xe[^Xo[쐬
 */
HWND FASTCALL CreateStatus(HWND hParent)
{
	HWND hWnd;

	ASSERT(hParent);

	/* e[}KptO */
	bThemed = FALSE;

	/* bZ[W[h */
	LoadMessageStatusBar();

	/* Xe[^XbZ[W */
	szStatusMessage[0] = '\0';
	bStatusMessage = FALSE;
	uResID = 0;
	nDcw = 16;

	/* Xe[^Xo[쐬 */
	hWnd = CreateStatusWindow(WS_CHILD | CCS_BOTTOM | CCS_NORESIZE,
								"", hParent, ID_STATUS_BAR);
	ChangeStatusBorder(hWnd);
	ResizeStatus(hParent, hWnd);
	ShowWindow(hWnd, SW_SHOW);

	bMountFileName = TRUE;

	return hWnd;
}

/*
 *	LvV`
 */
static void FASTCALL DrawMainCaption(void)
{
	char string[256];
	char tmp[256];
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];

	/* 󋵂ɉāARs[ */
	if (run_flag) {
		strncpy(string, szRunMessage, sizeof(string));
	}
	else {
		strncpy(string, szStopMessage, sizeof(string));
	}
	strncat(string, " ", sizeof(string) - strlen(string) - 1);

	/* CPUx䗦 */
	if (bAutoSpeedAdjust) {
		_snprintf(tmp, sizeof(tmp), "(%3d%%) ", speed_ratio / 100);
		strncat(string, tmp, sizeof(string) - strlen(string) - 1);
	}

	/* Xe[g */
	if ((get_file_size(StatePath) > 0) &&
		(strncmp(StatePath, QuickPath, _MAX_PATH))) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(StatePath, NULL, NULL, fname, ext);
		_snprintf(tmp, sizeof(tmp), "- [%s%s] ", fname, ext);
		strncat(string, tmp, sizeof(string) - strlen(string) - 1);
	}

	/* LvV\ */
	if (!bMountFileName) {
		/* r */
		string[sizeof(string) - 1] = '\0';
		if (strncmp(szCaption, string, strlen(string) + 1) != 0) {
			memset(szCaption, 0, sizeof(szCaption));
			strncpy(szCaption, string, sizeof(szCaption) - 1);
			SetWindowText(hMainWnd, szCaption);
		}
		return;
	}

#if XM7_VER == 1 && defined(SFDC)
	if (fdc_enable[FDC_S] && (model_type != MODEL_FM7)) {
		/* 1MBtbs[fBXNhCu 0 */
		if (fdc_ready[FDC_S][0] != FDC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(fdc_fname[FDC_S][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* 1MBtbs[fBXNhCu 1 */
		if (fdc_ready[FDC_S][1] != FDC_TYPE_NOTREADY) {
			if ((strcmp(fdc_fname[FDC_S][0], fdc_fname[FDC_S][1]) != 0) ||
				(fdc_ready[FDC_S][0] == FDC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(fdc_fname[FDC_S][1], drive, dir, fname, ext);
				if (fdc_ready[FDC_S][0] == FDC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}

	if (fdc_enable[FDC_M] || (model_type == MODEL_FM77)) {
		/* tbs[fBXNhCu 0 */
		if (fdc_ready[FDC_M][0] != FDC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(fdc_fname[FDC_M][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* tbs[fBXNhCu 1 */
		if (fdc_ready[FDC_M][1] != FDC_TYPE_NOTREADY) {
			if ((strcmp(fdc_fname[FDC_M][0], fdc_fname[FDC_M][1]) != 0) ||
				(fdc_ready[FDC_M][0] == FDC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(fdc_fname[FDC_M][1], drive, dir, fname, ext);
				if (fdc_ready[FDC_M][0] == FDC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}
#else
#if XM7_VER >= 2
	if (fdc_enable || model_type != MODEL_FM7) {
#else
	if (fdc_enable || model_type == MODEL_FM77) {
#endif
		/* tbs[fBXNhCu 0 */
		if (fdc_ready[0] != FDC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(fdc_fname[0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* tbs[fBXNhCu 1 */
		if (fdc_ready[1] != FDC_TYPE_NOTREADY) {
			if ((strcmp(fdc_fname[0], fdc_fname[1]) != 0) ||
				(fdc_ready[0] == FDC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(fdc_fname[1], drive, dir, fname, ext);
				if (fdc_ready[0] == FDC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* e[v */
	if (tape_fileh != -1) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(tape_fname, drive, dir, fname, ext);
		_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
		strncat(string, tmp, sizeof(string) - strlen(string) - 1);
	}

#if XM7_VER == 1 && defined(BUBBLE)
	/* ou */
	if (bmc_enable[BMC_32] &&
		((model_type == MODEL_FM8) ||
		 (!opn_enable && (mos_port != 3) && (bmc_secret_flag || bmc_use[BMC_32])))) {
		/* 32KBouz_ 0 */
		if (bmc_ready[0][0] != BMC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(bmc_fname[0][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* 32KBouz_ 1 */
		if (bmc_ready[0][1] != BMC_TYPE_NOTREADY) {
			if ((strcmp(bmc_fname[0][0], bmc_fname[0][1]) != 0) ||
				(bmc_ready[0][0] == BMC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(bmc_fname[0][1], drive, dir, fname, ext);
				if (bmc_ready[0][0] == BMC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}

	if (bmc_enable[BMC_128]) {
		/* 128KBouz_ 0 */
		if (bmc_ready[1][0] != BMC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(bmc_fname[1][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* 128KBouz_ 1 */
		if (bmc_ready[1][1] != BMC_TYPE_NOTREADY) {
			if ((strcmp(bmc_fname[1][0], bmc_fname[1][1]) != 0) ||
				(bmc_ready[1][0] == BMC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(bmc_fname[1][1], drive, dir, fname, ext);
				if (bmc_ready[1][0] == BMC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}
#endif

	/* r` */
	string[255] = '\0';
	if (memcmp(szCaption, string, strlen(string) + 1) != 0) {
		strncpy(szCaption, string, sizeof(szCaption));
		SetWindowText(hMainWnd, szCaption);
	}
}

/*
 *	ACR`FbNAύX
 */
static void FASTCALL CheckIcon(void)
{
	HICON hIcon;

	/* v`FbN */
	if (model_type == nAppIcon) {
		return;
	}

	/* VACR[hAZbg */
	switch (model_type) {
#if XM7_VER == 1
		case MODEL_FM8:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_FM8ICON));
			break;
		case MODEL_FM7:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_FM7ICON));
			break;
		case MODEL_FM77:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_FM77ICON));
			break;
#else
		case MODEL_FM7:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_FM7ICON));
			break;
		case MODEL_FM77AV:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_77AVICON));
			break;
		case MODEL_FM77AV20:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_AV20ICON));
			break;
		case MODEL_FM77AV20EX:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_20EXICON));
			break;
#if XM7_VER >= 3
		case MODEL_FM77AV40:
			hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_AV40ICON));
			break;
		case MODEL_FM77AV40EX:
			if (bGravestone) {
				hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_40SXICON));
			}
			else {
//				hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_40EXICON));
				hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_40EXICON2));
			}
			break;
#endif
#endif
		default:
			hIcon = NULL;
			ASSERT(FALSE);
			break;
	}
	SendMessage(hMainWnd, WM_SETICON, ICON_BIG, (LPARAM)hIcon);
	SendMessage(hMainWnd, WM_SETICON, ICON_SMALL, (LPARAM)hIcon);

	/* ACR폜 */
	DestroyIcon(hAppIcon);

	/* L */
	hAppIcon = hIcon;
	nAppIcon = model_type;
}

/*
 *	CAPL[`
 */
static void FASTCALL DrawCAP(void)
{
	int num;

	/* ԍ */
	if (caps_flag) {
		num = 1;
	}
	else {
		num = 0;
	}

	/* Ȃ牽Ȃ */
	if (nCAP == num) {
		return;
	}

	/* `A[NXV */
	nCAP = num;
	Status_SetText(hStatusBar, PANE_CAP, SBT_OWNERDRAW, PANE_CAP);
}

/*
 *	ȃL[`
 */
static void FASTCALL DrawKANA(void)
{
	int num;

	/* ԍ */
	if (kana_flag) {
		num = 1;
	}
	else {
		num = 0;
	}

	/* Ȃ牽Ȃ */
	if (nKANA == num) {
		return;
	}

	/* `A[NXV */
	nKANA = num;
	Status_SetText(hStatusBar, PANE_KANA, SBT_OWNERDRAW, PANE_KANA);
}

/*
 *	INSL[`
 */
static void FASTCALL DrawINS(void)
{
	int num;

	/* ԍ */
	if (ins_flag) {
		num = 1;
	}
	else {
		num = 0;
	}

	/* Ȃ牽Ȃ */
	if (nINS == num) {
		return;
	}

	/* `A[NXV */
	nINS = num;
	Status_SetText(hStatusBar, PANE_INS, SBT_OWNERDRAW, PANE_INS);
}

/*
 *	hCu`
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL DrawDrive(int type, int drive, UINT nID)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL DrawDrive(int drive, UINT nID)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	char string[_MAX_FNAME + _MAX_EXT];
	char buffer[_MAX_FNAME + _MAX_EXT + 4];
	char drive_[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	int num;
	char *name;
	BOOL exdrv;

#if XM7_VER == 1 && defined(SFDC)
	ASSERT((type == FDC_M) || (type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ASSERT((drive >= 0) && (drive <= 3));

#if XM7_VER == 1 && defined(SFDC)
	/* L`FbN */
	if ((type == FDC_M) && (!fdc_enable[FDC_M] && (model_type != MODEL_FM77))) {
		return;
	}
	if ((type == FDC_S) && (!fdc_enable[FDC_S] || (model_type == MODEL_FM7))) {
		return;
	}
	exdrv = (((	fdc_ready[FDC_M][2] != FDC_TYPE_NOTREADY ||
				fdc_ready[FDC_M][3] != FDC_TYPE_NOTREADY) &&
				(fdc_enable[FDC_M] || (model_type == MODEL_FM77))) ||
			 ((	fdc_ready[FDC_S][2] != FDC_TYPE_NOTREADY ||
				fdc_ready[FDC_S][3] != FDC_TYPE_NOTREADY) &&
				(fdc_enable[FDC_S] && (model_type != MODEL_FM7))));
#else	/* XM7_VER == 1 && defined(SFDC) */
	/* L`FbN */
#if XM7_VER >= 2
	if ((!fdc_enable && (model_type == MODEL_FM7))) {
		return;
	}
#else
	if ((!fdc_enable && (model_type != MODEL_FM77))) {
		return;
	}
#endif	/* XM7_VER == 1 */
	exdrv = (	fdc_ready[2] != FDC_TYPE_NOTREADY ||
				fdc_ready[3] != FDC_TYPE_NOTREADY);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	/* ݃hCu(#2/#3)}Eg\ */
	if ((drive >= 2) && !exdrv) {
		return;
	}

	/* ԍZbg */
#if XM7_VER == 1 && defined(SFDC)
	if (fdc_ready[type][drive] == FDC_TYPE_NOTREADY) {
		num = FDC_ACCESS_NOTREADY;
	}
	else if (fdc_teject[type][drive]) {
		num = FDC_ACCESS_TEJECT;
	}
	else {
		num = fdc_access[type][drive];
		if (num == FDC_ACCESS_SEEK) {
			num = FDC_ACCESS_READY;
		}
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
		num = FDC_ACCESS_NOTREADY;
	}
	else if (fdc_teject[drive]) {
		num = FDC_ACCESS_TEJECT;
	}
	else {
		num = fdc_access[drive];
		if (num == FDC_ACCESS_SEEK) {
			num = FDC_ACCESS_READY;
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* O擾 */
	name = "";
#if XM7_VER == 1 && defined(SFDC)
	if ((fdc_ready[type][drive] == FDC_TYPE_D77) &&
		(strlen(fdc_name[type][drive][ fdc_media[type][drive] ]) > 0)) {
		name = fdc_name[type][drive][ fdc_media[type][drive] ];
	}
	else if (fdc_ready[type][drive] != FDC_TYPE_NOTREADY) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(fdc_fname[type][drive], NULL, NULL, fname, ext);
		strncpy(string, fname, sizeof(string));
		strncat(string, ext, sizeof(string) - strlen(string) - 1);
		if ((fdc_ready[type][drive] == FDC_TYPE_D77) && (fdc_medias[type][drive] > 1)) {
			_snprintf(buffer, sizeof(buffer), "%s #%02d", string, fdc_media[type][drive] + 1);
			name = buffer;
		}
		else {
			name = string;
		}
	}
#else
	if ((fdc_ready[drive] == FDC_TYPE_D77) &&
		(strlen(fdc_name[drive][ fdc_media[drive] ]) > 0)) {
		name = fdc_name[drive][ fdc_media[drive] ];
	}
	else if (fdc_ready[drive] != FDC_TYPE_NOTREADY) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(fdc_fname[drive], drive_, dir, fname, ext);
		strncpy(string, fname, sizeof(string));
		strncat(string, ext, sizeof(string) - strlen(string) - 1);
		if ((fdc_ready[drive] == FDC_TYPE_D77) && (fdc_medias[drive] > 1)) {
			_snprintf(buffer, sizeof(buffer), "%s #%02d", string,
					  fdc_media[drive] + 1);
			name = buffer;
		}
		else {
			name = string;
		}
	}
#endif

	/* fBAI[o[͊gqŕ\ */
	if (strlen(name) > nDcw) {
		if (nDcw < 16) {
			/* gq̂ݎo */
#if XM7_VER == 1 && defined(SFDC)
			_splitpath(fdc_fname[type][drive], NULL, NULL, NULL, ext);
#else
			_splitpath(fdc_fname[drive], NULL, NULL, NULL, ext);
#endif
			strncpy(string, ext, sizeof(string));
			for (name = string; *name != '\0'; name ++) {
				*name = (char)toupper(*name);
			}
			if (strlen(ext) > 1) {
				name = string + 1;
			}
			else {
				name = string;
			}
			if (nDcw >= 7) {
				/* fBAԍݒ */
#if XM7_VER == 1 && defined(SFDC)
				if ((fdc_ready[type][drive] == FDC_TYPE_D77) && (fdc_medias[type][drive] > 1)) {
					_snprintf(buffer, sizeof(buffer), "%s #%02d", name, fdc_media[type][drive] + 1);
					name = buffer;
				}
#else
				if ((fdc_ready[drive] == FDC_TYPE_D77) && (fdc_medias[drive] > 1)) {
					_snprintf(buffer, sizeof(buffer), "%s #%02d", name, fdc_media[drive] + 1);
					name = buffer;
				}
#endif
				else {
					_snprintf(buffer, sizeof(buffer), "%s File", name);
					name = buffer;
				}
			}
		}
	}

	/* ԍr */
#if XM7_VER == 1 && defined(SFDC)
	if (nDrive[type][drive] == num) {
		if (strcmp(szDrive[type][drive], name) == 0) {
			return;
		}
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (nDrive[drive] == num) {
		if (strcmp(szDrive[drive], name) == 0) {
			return;
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* ` */
#if XM7_VER == 1 && defined(SFDC)
	nDrive[type][drive] = num;
	strncpy(szDrive[type][drive], name, sizeof(szDrive[type][drive]));
#else	/* XM7_VER == 1 && defined(SFDC) */
	nDrive[drive] = num;
	strncpy(szDrive[drive], name, sizeof(szDrive[drive]));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	Status_SetText(hStatusBar, nID, SBT_OWNERDRAW, nID);
}

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	ouz_jbg`
 */
static void FASTCALL DrawUnit(int type, UINT nID)
{
	int i, units, num, writep, update;

	ASSERT((type == BMC_32) || (type == BMC_128));

	/* L`FbN */
	if (!bmc_enable[type] ||
		((type == BMC_32) &&
		 (model_type != MODEL_FM8) &&
		 (opn_enable || (mos_port == 3) || !(bmc_secret_flag || bmc_use[BMC_32])))) {
		return;
	}

	/* jbg */
	if (type) {
		units = BMC_UNITS_128;
	}
	else {
		units = BMC_UNITS_32;
	}

	/* jbg[v */
	update = FALSE;
	for (i = 0; i < units; i ++) {
		/* ԍZbg */
		if (bmc_ready[type][i] == BMC_TYPE_NOTREADY) {
			num = BMC_ACCESS_NOTREADY;
		}
		else if (bmc_teject[type][i]) {
			num = BMC_ACCESS_TEJECT;
		}
		else {
			num = bmc_access[type][i];
		}

		/* CgveNg */
		writep = bmc_writep[type][i];

		/* r */
		if ((nUnit[type][i] == num) &&
			(nUnitP[type][i] == writep)) {
			continue;
		}

		nUnit[type][i] = num;
		nUnitP[type][i] = writep;
		update = TRUE;
	}

	/* update */
	if (!update) {
		return;
	}

	/* ` */
	Status_SetText(hStatusBar, nID, SBT_OWNERDRAW, nID);
}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

/*
 *	e[v`
 */
static void FASTCALL DrawTape(void)
{
	int num;

	/* io[vZ */
	num = 30000;
	if (tape_fileh != -1) {
		num = (int)((tape_offset >> 8) % 10000);
		if (tape_motor) {
			if (tape_rec) {
				num += 20000;
			}
			else {
				num += 10000;
			}
		}
	}

	/* ԍr */
	if (nTape == num) {
		return;
	}

	/* ` */
	nTape = num;
	Status_SetText(hStatusBar, PANE_TAPE, SBT_OWNERDRAW, PANE_TAPE);
}

/*
 *	`
 */
void FASTCALL DrawStatus(void)
{
	/* EChE`FbN */
	if (!hMainWnd) {
		return;
	}

	DrawMainCaption();
	CheckIcon();

	/* SʁAXe[^Xo[`FbN */
	if (bFullScreen || !hStatusBar) {
		return;
	}

	DrawCAP();
	DrawKANA();
	DrawINS();
#if XM7_VER == 1 && defined(SFDC)
	DrawDrive(FDC_M, 0, PANE_DRIVE0);
	DrawDrive(FDC_M, 1, PANE_DRIVE1);
	DrawDrive(FDC_M, 2, PANE_DRIVE2);
	DrawDrive(FDC_M, 3, PANE_DRIVE3);
	DrawDrive(FDC_S, 0, PANE_SDRIVE0);
	DrawDrive(FDC_S, 1, PANE_SDRIVE1);
	DrawDrive(FDC_S, 2, PANE_SDRIVE2);
	DrawDrive(FDC_S, 3, PANE_SDRIVE3);
#else	/* XM7_VER == 1 && defined(SFDC) */
	DrawDrive(0, PANE_DRIVE0);
	DrawDrive(1, PANE_DRIVE1);
	DrawDrive(2, PANE_DRIVE2);
	DrawDrive(3, PANE_DRIVE3);
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER == 1 && defined(BUBBLE)
	DrawUnit(BMC_32, PANE_BUBBLE32);
	DrawUnit(BMC_128, PANE_BUBBLE128);
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
	DrawTape();
}

/*
 *	ĕ`
 */
void FASTCALL PaintStatus(void)
{
	int i;
	/* L[NׂăNA */
	szCaption[0] = '\0';
	nCAP = -1;
	nKANA = -1;
	nINS = -1;
	for (i = 0; i < 4; i ++) {
 #if XM7_VER == 1 && defined(SFDC)
		nDrive[FDC_M][i] = -1;
		nDrive[FDC_S][i] = -1;
		szDrive[FDC_M][i][0] = '\0';
		szDrive[FDC_S][i][0] = '\0';
 #else	/* XM7_VER == 1 && defined(SFDC) */
		nDrive[i] = -1;
		szDrive[i][0] = '\0';
 #endif	/* XM7_VER == 1 && defined(SFDC) */
 #if XM7_VER == 1 && defined(BUBBLE)
		nUnit[BMC_32][i] = -1;
		nUnitP[BMC_32][i] = -1;
		nUnit[BMC_128][i] = -1;
		nUnitP[BMC_128][i] = -1;
 #endif	/* XM7_VER == 1 && defined(SFDC) */
	}
	nTape = -1;

	/* ` */
	DrawStatus();
}

/*
 *	I[i[h[
 */
void FASTCALL OwnerDrawStatus(DRAWITEMSTRUCT *pDI)
{
	HBRUSH hBrush;
	COLORREF fColor;
	COLORREF bColor;
	RECT srect;
	char string[_MAX_FNAME + _MAX_EXT + 4];
	int drive;
#if XM7_VER == 1
	int type;
#endif	/* XM7_VER == 1 */
	int ndrive;
	char *szdrive;
#if XM7_VER == 1 && defined(BUBBLE)
	int i, units, nunit;
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
#if defined(ENGLISH_SUPPORT) || defined(RES_DLL)
	HFONT hFont;
	HFONT hFont0;
#endif

	ASSERT(pDI);

	/* AF */
	switch (pDI->itemData) {
		/* tbs[fBXN */
		case PANE_DRIVE3:
		case PANE_DRIVE2:
		case PANE_DRIVE1:
		case PANE_DRIVE0:
#if XM7_VER == 1 && defined(SFDC)
		case PANE_SDRIVE3:
		case PANE_SDRIVE2:
		case PANE_SDRIVE1:
		case PANE_SDRIVE0:
#endif	/* XM7_VER == 1 && defined(SFDC) */
			switch (pDI->itemData) {
			case PANE_DRIVE3:
				drive = 3;
#if XM7_VER == 1 && defined(SFDC)
				type = FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
				break;
			case PANE_DRIVE2:
				drive = 2;
#if XM7_VER == 1 && defined(SFDC)
				type = FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
				break;
			case PANE_DRIVE1:
				drive = 1;
#if XM7_VER == 1 && defined(SFDC)
				type = FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
				break;
			case PANE_DRIVE0:
				drive = 0;
#if XM7_VER == 1 && defined(SFDC)
				type = FDC_M;
#endif	/* XM7_VER == 1 && defined(SFDC) */
				break;
#if XM7_VER == 1 && defined(SFDC)
			case PANE_SDRIVE3:
				drive = 3;
				type = FDC_S;
				break;
			case PANE_SDRIVE2:
				drive = 2;
				type = FDC_S;
				break;
			case PANE_SDRIVE1:
				drive = 1;
				type = FDC_S;
				break;
			case PANE_SDRIVE0:
				drive = 0;
				type = FDC_S;
				break;
#endif	/* XM7_VER == 1 && defined(SFDC) */
			}
#if XM7_VER == 1 && defined(SFDC)
			ndrive  = nDrive[type][drive];
			szdrive = szDrive[type][drive];
#else	/* XM7_VER == 1 && defined(SFDC) */
			ndrive  = nDrive[drive];
			szdrive = szDrive[drive];
#endif	/* XM7_VER == 1 && defined(SFDC) */
			if (ndrive == FDC_ACCESS_NOTREADY) {
				strncpy(string, "", sizeof(string));
			}
			else {
				strncpy(string, szdrive, sizeof(string));
			}
			fColor = RGB(255, 255, 255);
			if ((ndrive == FDC_ACCESS_NOTREADY) ||
				(ndrive == FDC_ACCESS_TEJECT)) {
				bColor = RGB(0, 0, 0);
			}
			else if (ndrive == FDC_ACCESS_READ) {
				bColor = RGB(191, 0, 0);
			}
			else if (ndrive == FDC_ACCESS_WRITE) {
				bColor = RGB(0, 0, 191);
			}
			else {
				bColor = RGB(63, 63, 63);
			}
			break;

		/* e[v */
		case PANE_TAPE:
			if (nTape >= 30000) {
				string[0] = '\0';
			}
			else {
				_snprintf(string, sizeof(string), "%04d", nTape % 10000);
			}
			fColor = RGB(255, 255, 255);
			bColor = RGB(63, 63, 63);
			if (nTape >= 10000) {
				if (nTape >= 30000) {
					bColor = RGB(0, 0, 0);
				}
				else if (nTape >= 20000) {
					bColor = RGB(0, 0, 191);
				}
				else {
					bColor = RGB(191, 0, 0);
				}
			}
			break;

#if XM7_VER == 1 && defined(BUBBLE)
		/* ouz_jbg */
		case PANE_BUBBLE32:
		case PANE_BUBBLE128:
			if (pDI->itemData == PANE_BUBBLE32) {
				type = BMC_32;
				units = BMC_UNITS_32;
			}
			else {
				type = BMC_128;
				units = BMC_UNITS_128;
			}
			string[0] = '\0';

			/* uVœh */
			hBrush = CreateSolidBrush(RGB(0, 0, 0));
			if (hBrush) {
				FillRect(pDI->hDC, &(pDI->rcItem), hBrush);
				DeleteObject(hBrush);
			}

			for (i = 0; i < units; i ++) {
				nunit  = nUnit[type][i];
				if ((nunit == BMC_ACCESS_NOTREADY) ||
					(nunit == BMC_ACCESS_TEJECT)) {
					fColor = RGB(0, 0, 0);
					bColor = RGB(0, 0, 0);
				}
				else {
					if (nunit == BMC_ACCESS_READ) {
						fColor = RGB(191, 0, 0);
					}
					else if (nunit == BMC_ACCESS_WRITE) {
						fColor = RGB(0, 0, 191);
					}
					else {
						fColor = RGB(95, 95, 95);
					}
					if (nUnitP[type][i]) {
						bColor = RGB(0, 191, 0);
					}
					else {
						bColor = RGB(95, 95, 95);
					}
				}

				hBrush = CreateSolidBrush(fColor);
				if (hBrush) {
					srect.left = pDI->rcItem.left + i * 6 + 1;
					srect.right = pDI->rcItem.left + i * 6 + 6;
					srect.top = pDI->rcItem.top + 1;
					srect.bottom = pDI->rcItem.top + 4;
					FillRect(pDI->hDC, &srect, hBrush);
					DeleteObject(hBrush);
				}

				hBrush = CreateSolidBrush(bColor);
				if (hBrush) {
					srect.left = pDI->rcItem.left + i * 6 + 1;
					srect.right = pDI->rcItem.left + i * 6 + 6;
					srect.top = pDI->rcItem.top + 5;
					srect.bottom = pDI->rcItem.top + 8;
					FillRect(pDI->hDC, &srect, hBrush);
					DeleteObject(hBrush);
				}

				/* jbgԍ` */
				SetDIBitsToDevice(pDI->hDC, pDI->rcItem.left + i * 6 + 1, pDI->rcItem.bottom - 6, 6, 6, 0, 0, 0, 6,
									&bmpUnit[i][sizeof(BITMAPINFOHEADER)],
									(BITMAPINFO*)&bmpUnit[i][0],
									DIB_RGB_COLORS);
			}
			return;
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

		/* CAP */
		case PANE_CAP:
			fColor = RGB(255, 255, 255);
			strncpy(string, szCAPMessage, sizeof(string));
			if (nCAP) {
				bColor = RGB(255, 0, 0);
			}
			else {
				bColor = RGB(0, 0, 0);
			}
			break;

		/*  */
		case PANE_KANA:
			fColor = RGB(255, 255, 255);
			strncpy(string, szKANAMessage, sizeof(string));
			if (nKANA) {
				bColor = RGB(255, 0, 0);
			}
			else {
				bColor = RGB(0, 0, 0);
			}
			break;

		/* INS */
		case PANE_INS:
			fColor = RGB(255, 255, 255);
			strncpy(string, szINSMessage, sizeof(string));
			if (nINS) {
				bColor = RGB(255, 0, 0);
			}
			else {
				bColor = RGB(0, 0, 0);
			}
			break;

		/* ȊO */
		default:
			ASSERT(FALSE);
			return;
	}

	/* uVœh */
	hBrush = CreateSolidBrush(bColor);
	if (hBrush) {
		FillRect(pDI->hDC, &(pDI->rcItem), hBrush);
		DeleteObject(hBrush);
	}

#if defined(ENGLISH_SUPPORT) || defined(RES_DLL)
	if (!IsJapanese(FALSE) && (pDI->itemData == PANE_KANA)) {
		/* "KANA"̂݉kڂ̏tHg쐬 */
		hFont = CreateFont(
			-11,					/* tHg̍(傫) */
			-4,						/* tHg̕(0) */
			0,						/* px(0) */
			0,						/* px(0) */
			FW_DONTCARE,			/* ̑ */
			FALSE,					/* tHgC^bNȂTRUE */
			FALSE,					/* ȂTRUE */
			FALSE,					/* ȂTRUE */
			ANSI_CHARSET,			/* tHg̕Zbg */
			OUT_DEFAULT_PRECIS,		/* o͐x̐ݒ */
			CLIP_DEFAULT_PRECIS,	/* NbsOx */
			DEFAULT_QUALITY,		/* tHg̏o͕i */
			FIXED_PITCH,			/* tHg̃sb`ƃt@~ */
			"Tahoma"				/* tHg̃^CvtFCX̎w */
		);
		hFont0 = (HFONT)SelectObject(pDI->hDC, hFont);		/* tHgݒ */
	}
#endif

	/* eLXg̉e` */
	srect.left = pDI->rcItem.left + 1;
	srect.right = pDI->rcItem.right + 1;
	srect.top = pDI->rcItem.top + 1;
	srect.bottom = pDI->rcItem.bottom + 1;
	SetTextColor(pDI->hDC, RGB(0, 0, 0));
	SetBkMode(pDI->hDC, TRANSPARENT);
	DrawText(pDI->hDC, string, strlen(string), &srect,
				DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);

	/* eLXg` */
	SetTextColor(pDI->hDC, fColor);
	SetBkMode(pDI->hDC, TRANSPARENT);	/* for WinXP Luna */
	DrawText(pDI->hDC, string, strlen(string), &(pDI->rcItem),
				DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);

#if defined(ENGLISH_SUPPORT) || defined(RES_DLL)
	if (!IsJapanese(FALSE) && (pDI->itemData == PANE_KANA)) {
		SelectObject(pDI->hDC, hFont0);						/* tHgɖ߂ */
		DeleteObject(hFont);
	}
#endif

	/* hCuԍ` */
	if ((pDI->itemData == PANE_DRIVE0) || (pDI->itemData == PANE_DRIVE1) ||
#if XM7_VER == 1 && defined(SFDC)
		(pDI->itemData == PANE_SDRIVE0) || (pDI->itemData == PANE_SDRIVE1) ||
		(pDI->itemData == PANE_SDRIVE2) || (pDI->itemData == PANE_SDRIVE3) ||
#endif	/* XM7_VER == 1 && defined(SFDC) */
		(pDI->itemData == PANE_DRIVE2) || (pDI->itemData == PANE_DRIVE3)) {
		SetDIBitsToDevice(pDI->hDC, pDI->rcItem.right - 8, pDI->rcItem.bottom - 8, 8, 8, 0, 0, 0, 8,
							&bmpDrive[drive][sizeof(BITMAPINFOHEADER)],
							(BITMAPINFO*)&bmpDrive[drive][0],
							DIB_RGB_COLORS);
#if XM7_VER == 1 && defined(SFDC)
		if (type == FDC_S) {
#if 1
			SetDIBitsToDevice(pDI->hDC, pDI->rcItem.left, pDI->rcItem.top, 15, 6, 0, 0, 0, 6,
								&bmp2HD[sizeof(BITMAPINFOHEADER)],
								(BITMAPINFO*)&bmp2HD[0],
								DIB_RGB_COLORS);
#else
			const int nums[] = {2, 4, 5};
			for (drive = 0; drive <= 2; drive ++) {
				SetDIBitsToDevice(pDI->hDC, pDI->rcItem.left + drive * 8, pDI->rcItem.top, 8, 8, 0, 0, 0, 8,
									&bmpDrive[nums[drive]][sizeof(BITMAPINFOHEADER)],
									(BITMAPINFO*)&bmpDrive[nums[drive]][0],
									DIB_RGB_COLORS);
			}
#endif
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	}
}

/*
 *	TCYύX
 */
void FASTCALL SizeStatus(LONG cx)
{
	HDC hDC;
	TEXTMETRIC tm;
	LONG cw;
	UINT uPane[PANE_MAX];
	BOOL bsize;

	ASSERT(cx > 0);
	ASSERT(hStatusBar);

	/* eLXggbN擾 */
	hDC = GetDC(hStatusBar);
	GetTextMetrics(hDC, &tm);
	ReleaseDC(hStatusBar, hDC);
	cw = tm.tmAveCharWidth;

	/* yCTCY(yCE[̈ʒuݒ) */
	uPane[PANE_INS] = cx;
	uPane[PANE_KANA] = uPane[PANE_INS] - cw * 4;
	uPane[PANE_CAP] = uPane[PANE_KANA] - cw * 4;
#if XM7_VER == 1 && defined(BUBBLE)
	uPane[PANE_BUBBLE128] = uPane[PANE_CAP] - cw * 4;
	if (bmc_enable[BMC_128]) {
		uPane[PANE_BUBBLE32] = uPane[PANE_BUBBLE128] - 29;
	}
	else {
		uPane[PANE_BUBBLE32] = uPane[PANE_BUBBLE128];
	}
	if (bmc_enable[BMC_32] &&
		((model_type == MODEL_FM8) ||
		 (!opn_enable && (mos_port != 3) && (bmc_secret_flag || bmc_use[BMC_32])))) {
		uPane[PANE_TAPE] = uPane[PANE_BUBBLE32] - 17;
	}
	else {
		uPane[PANE_TAPE] = uPane[PANE_BUBBLE32];
	}
	uPane[PANE_DRIVE0] = uPane[PANE_TAPE] - cw * 4;
#else
	uPane[PANE_TAPE] = uPane[PANE_CAP] - cw * 4;
	uPane[PANE_DRIVE0] = uPane[PANE_TAPE] - cw * 5;
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
#ifdef FILTERDLL
	bsize = ((nScalingSize > DISPLAYSIZE_100) &&
			 (nScalingSize != DISPLAYSIZE_050) &&
			 (nScalingSize != DISPLAYSIZE_120));
#else
	bsize = (nScalingSize != DISPLAYSIZE_100);
#endif
#if XM7_VER == 1 && defined(SFDC)
	if ((fdc_enable[FDC_S] && (model_type != MODEL_FM7)) &&
		(fdc_enable[FDC_M] || (model_type == MODEL_FM77))) {
		/* 2HD + 2D */
		if (fdc_ready[FDC_M][2] != FDC_TYPE_NOTREADY ||
			fdc_ready[FDC_M][3] != FDC_TYPE_NOTREADY ||
			fdc_ready[FDC_S][2] != FDC_TYPE_NOTREADY ||
			fdc_ready[FDC_S][3] != FDC_TYPE_NOTREADY) {
#ifdef FILTERDLL
			if (((nScalingSize == DISPLAYSIZE_100)) || (nScalingSize == DISPLAYSIZE_050)) {
				nDcw = 4;
			}
			else if (nScalingSize == DISPLAYSIZE_180) {
				nDcw = 12;
			}
			else if (nScalingSize == DISPLAYSIZE_150) {
				nDcw = 9;
			}
			else if (nScalingSize == DISPLAYSIZE_120) {
				nDcw = 6;
			}
			else {
				nDcw = 14;
			}
#else
			if (nScalingSize == DISPLAYSIZE_100) {
				nDcw = 4;
			}
			else {
				nDcw = 14;
			}
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0] - cw * nDcw;
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1] - cw * nDcw;
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2] - cw * nDcw;
			uPane[PANE_SDRIVE0] = uPane[PANE_DRIVE3] - cw * nDcw;
			uPane[PANE_SDRIVE1] = uPane[PANE_SDRIVE0] - cw * nDcw;
			uPane[PANE_SDRIVE2] = uPane[PANE_SDRIVE1] - cw * nDcw;
			uPane[PANE_SDRIVE3] = uPane[PANE_SDRIVE2] - cw * nDcw;
			uPane[PANE_DEFAULT] = uPane[PANE_SDRIVE3] - cw * nDcw;
		}
		else {
#ifdef FILTERDLL
			if (nScalingSize == DISPLAYSIZE_050) {
				nDcw = 4;
			}
			else if (nScalingSize == DISPLAYSIZE_120) {
				nDcw = 12;
			}
			else if (bsize) {
				nDcw = 16;
			}
			else {
				nDcw = 8;
			}
#else
			if (bsize) {
				nDcw = 16;
			}
			else {
				nDcw = 8;
			}
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0] - cw * nDcw;
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1] - cw * nDcw;
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2];
			uPane[PANE_SDRIVE0] = uPane[PANE_DRIVE3];
			uPane[PANE_SDRIVE1] = uPane[PANE_SDRIVE0] - cw * nDcw;
			uPane[PANE_SDRIVE2] = uPane[PANE_SDRIVE1] - cw * nDcw;
			uPane[PANE_SDRIVE3] = uPane[PANE_SDRIVE2];
			uPane[PANE_DEFAULT] = uPane[PANE_SDRIVE3];
		}
	}
	else if (fdc_enable[FDC_S] && (model_type == MODEL_FM8)) {
		/* 2HD ONLY */
		if (fdc_ready[FDC_S][2] != FDC_TYPE_NOTREADY ||
			fdc_ready[FDC_S][3] != FDC_TYPE_NOTREADY) {
#ifdef FILTERDLL
			if (nScalingSize == DISPLAYSIZE_050) {
				nDcw = 4;
			}
			else if (nScalingSize == DISPLAYSIZE_120) {
				nDcw = 12;
			}
			else if (bsize) {
				nDcw = 16;
			}
			else {
				nDcw = 8;
			}
#else
			if (bsize) {
				nDcw = 16;
			}
			else {
				nDcw = 8;
			}
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0];
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1];
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2];
			uPane[PANE_SDRIVE0] = uPane[PANE_DRIVE3];
			uPane[PANE_SDRIVE1] = uPane[PANE_SDRIVE0] - cw * nDcw;
			uPane[PANE_SDRIVE2] = uPane[PANE_SDRIVE1] - cw * nDcw;
			uPane[PANE_SDRIVE3] = uPane[PANE_SDRIVE2] - cw * nDcw;
			uPane[PANE_DEFAULT] = uPane[PANE_SDRIVE3] - cw * nDcw;
		}
		else {
#ifdef FILTERDLL
			if (nScalingSize == DISPLAYSIZE_050) {
				nDcw = 4;
			}
			else {
				nDcw = 16;
			}
#else
			nDcw = 16;
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0];
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1];
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2];
			uPane[PANE_SDRIVE0] = uPane[PANE_DRIVE3];
			uPane[PANE_SDRIVE1] = uPane[PANE_SDRIVE0] - cw * nDcw;
			uPane[PANE_SDRIVE2] = uPane[PANE_SDRIVE1] - cw * nDcw;
			uPane[PANE_SDRIVE3] = uPane[PANE_SDRIVE2];
			uPane[PANE_DEFAULT] = uPane[PANE_SDRIVE3];
		}
	}
	else if (fdc_enable[FDC_M] || (model_type == MODEL_FM77)) {
		/* 2D ONLY */
		if (fdc_ready[FDC_M][2] != FDC_TYPE_NOTREADY ||
			fdc_ready[FDC_M][3] != FDC_TYPE_NOTREADY) {
#ifdef FILTERDLL
			if (nScalingSize == DISPLAYSIZE_050) {
				nDcw = 4;
			}
			else if (nScalingSize == DISPLAYSIZE_120) {
				nDcw = 12;
			}
			else if (bsize) {
				nDcw = 16;
			}
			else {
				nDcw = 8;
			}
#else
			if (bsize) {
				nDcw = 16;
			}
			else {
				nDcw = 8;
			}
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0] - cw * nDcw;
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1] - cw * nDcw;
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2] - cw * nDcw;
			uPane[PANE_SDRIVE0] = uPane[PANE_DRIVE3] - cw * nDcw;
			uPane[PANE_SDRIVE1] = uPane[PANE_SDRIVE0];
			uPane[PANE_SDRIVE2] = uPane[PANE_SDRIVE1];
			uPane[PANE_SDRIVE3] = uPane[PANE_SDRIVE2];
			uPane[PANE_DEFAULT] = uPane[PANE_SDRIVE3];
		}
		else {
#ifdef FILTERDLL
			if (nScalingSize == DISPLAYSIZE_050) {
				nDcw = 4;
			}
			else {
				nDcw = 16;
			}
#else
			nDcw = 16;
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0] - cw * nDcw;
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1] - cw * nDcw;
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2];
			uPane[PANE_SDRIVE0] = uPane[PANE_DRIVE3];
			uPane[PANE_SDRIVE1] = uPane[PANE_SDRIVE0];
			uPane[PANE_SDRIVE2] = uPane[PANE_SDRIVE1];
			uPane[PANE_SDRIVE3] = uPane[PANE_SDRIVE2];
			uPane[PANE_DEFAULT] = uPane[PANE_SDRIVE3];
		}
	}
	else {
		/* NONE */
		nDcw = 16;
		uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0];
		uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1];
		uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2];
		uPane[PANE_SDRIVE0] = uPane[PANE_DRIVE3];
		uPane[PANE_SDRIVE1] = uPane[PANE_SDRIVE0];
		uPane[PANE_SDRIVE2] = uPane[PANE_SDRIVE1];
		uPane[PANE_SDRIVE3] = uPane[PANE_SDRIVE2];
		uPane[PANE_DEFAULT] = uPane[PANE_SDRIVE3];
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
  #if XM7_VER >= 2
	if (fdc_enable || (model_type != MODEL_FM7)) {
  #else
	if (fdc_enable || (model_type == MODEL_FM77)) {
  #endif
		/* 2D ONLY */
		if (fdc_ready[2] != FDC_TYPE_NOTREADY ||
			fdc_ready[3] != FDC_TYPE_NOTREADY) {
#ifdef FILTERDLL
				if (nScalingSize == DISPLAYSIZE_050) {
					nDcw = 4;
				}
				else if (nScalingSize == DISPLAYSIZE_120) {
					nDcw = 12;
				}
				else if (bsize) {
					nDcw = 16;
				}
				else {
					nDcw = 8;
				}
#else
				if (bsize) {
					nDcw = 16;
				}
				else {
					nDcw = 8;
				}
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0] - cw * nDcw;
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1] - cw * nDcw;
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2] - cw * nDcw;
			uPane[PANE_DEFAULT] = uPane[PANE_DRIVE3] - cw * nDcw;
		}
		else {
#ifdef FILTERDLL
			if (nScalingSize == DISPLAYSIZE_050) {
				nDcw = 4;
			}
			else {
				nDcw = 16;
			}
#else
			nDcw = 16;
#endif
			uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0] - cw * nDcw;
			uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1] - cw * nDcw;
			uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2];
			uPane[PANE_DEFAULT] = uPane[PANE_DRIVE3];
		}
	}
	else {
		/* NONE */
		nDcw = 16;
		uPane[PANE_DRIVE1] = uPane[PANE_DRIVE0];
		uPane[PANE_DRIVE2] = uPane[PANE_DRIVE1];
		uPane[PANE_DRIVE3] = uPane[PANE_DRIVE2];
		uPane[PANE_DEFAULT] = uPane[PANE_DRIVE3];
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* Drag&DroppyCWݒ */
#if XM7_VER == 1 && defined(SFDC)
	uPaneX[0] = uPane[PANE_DEFAULT];
	uPaneX[1] = uPane[PANE_SDRIVE3];
	uPaneX[2] = uPane[PANE_SDRIVE2];
	uPaneX[3] = uPane[PANE_SDRIVE1];
	uPaneX[4] = uPane[PANE_SDRIVE0];
	uPaneX[5] = uPane[PANE_DRIVE3];
	uPaneX[6] = uPane[PANE_DRIVE2];
	uPaneX[7] = uPane[PANE_DRIVE1];
	uPaneX[8] = uPane[PANE_DRIVE0];
#else	/* XM7_VER == 1 && defined(SFDC) */
	uPaneX[0] = uPane[PANE_DEFAULT];
	uPaneX[1] = uPane[PANE_DRIVE3];
	uPaneX[2] = uPane[PANE_DRIVE2];
	uPaneX[3] = uPane[PANE_DRIVE1];
	uPaneX[4] = uPane[PANE_DRIVE0];
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* yCTCYݒ */
	Status_SetParts(hStatusBar, countof(uPane), uPane);

	/* ĕ` */
	PaintStatus();
}

/*
 *	Xe[^Xo[TCY
 */
void FASTCALL ResizeStatus(HWND hwnd, HWND hstatus)
{
	RECT rect;
	int height;

	/* wiFύX */
	SendMessage(hstatus, SB_SETBKCOLOR, 0, GetSysColor(COLOR_3DFACE));

	/* ␳ */
	GetWindowRect(hMainWnd, &rect);
	switch (nScalingSize) {
		case DISPLAYSIZE_100:		/* x1.0(640x400) */
		default:
			height = 400;
			break;

#ifdef FILTERDLL
		case DISPLAYSIZE_150:		/* x1.5(960x600) */
			height = 600;
			break;
#endif

		case DISPLAYSIZE_200:		/* x2.0(1280x800) */
			height = 800;
			break;

#ifdef FILTERDLL
		case DISPLAYSIZE_050:		/* x0.5(320x200) */
			height = 200;
			break;

		case DISPLAYSIZE_180:		/* x1.8(1152x720) */
			height = 720;
			break;

		case DISPLAYSIZE_120:		/* x1.2(768x480) */
			height = 480;
			break;
#endif

		case DISPLAYSIZE_300:		/* x3.0(1920x1200) */
			height = 1200;
			break;

		case DISPLAYSIZE_400:		/* x4.0(2560x1600) */
			height = 1600;
			break;
	}
	MoveWindow(hstatus, 0, height, (rect.right - rect.left),
			   GetSystemMetrics(SM_CYMENU), TRUE);

	/* t[EChẼTCY␳ */
	GetClientRect(hwnd, &rect);
	PostMessage(hwnd, WM_SIZE, 0, MAKELPARAM(rect.right, rect.bottom));
}

/*
 *	Xe[^XbZ[W\
 */
void FASTCALL SetStatusMessage(UINT ID)
{
	ASSERT(hStatusBar);

	/* Xe[^Xo[\ĂȂ΁AȂ */
	if (!IsWindowVisible(hStatusBar)) {
		return;
	}

	/* wID̕񃊃\[X[h݂ */
	if (GetMsg(hResInstance, ID, szStatusMessage, sizeof(szStatusMessage)) != 0) {
		/* Xe[^XbZ[WZbg */
		Status_SetText(hStatusBar, 0, nStatusBorder, szStatusMessage);

		/* Xe[^XbZ[Wp^C}ݒ肷 */
		if (bStatusMessage) {
			KillTimer(hMainWnd, ID_STATUS_BAR);
		}
		SetTimer(hMainWnd, ID_STATUS_BAR, 1500, NULL);

		/* Xe[^XbZ[W\tO𗧂Ă */
		bStatusMessage = TRUE;
	}
	else {
		/* \[X[hɎsꍇA\tO~낷 */
		bStatusMessage = FALSE;
	}
}

/*
 *	Xe[^XbZ[WI
 */
void FASTCALL EndStatusMessage(void)
{
	char buffer[128];

	ASSERT(hStatusBar);

	/* ƂƃXe[^XbZ[W\ĂȂ΁AȂ */
	if (!bStatusMessage) {
		return;
	}

	/* Xe[^XbZ[W\tO~낷 */
	bStatusMessage = FALSE;

	/* Xe[^Xo[\ĂȂ΁AȂ */
	if (!IsWindowVisible(hStatusBar)) {
		return;
	}

	if (uResID != 0) {
		/* 񃊃\[X[h݂ */
		if (GetMsg(hResInstance, uResID, buffer, sizeof(buffer)) == 0) {
			buffer[0] = '\0';
		}
	}
	else {
		/* AChbZ[WRs[ */
		strncpy(buffer, szIdleMessage, sizeof(buffer));
	}

	/* Zbg */
	Status_SetText(hStatusBar, 0, nStatusBorder, buffer);
}

/*
 *	j[ZNg
 */
void FASTCALL OnMenuSelect(WPARAM wParam)
{
	char buffer[128];
	UINT uID;

	ASSERT(hStatusBar);

	/* Xe[^Xo[\ĂȂ΁AȂ */
	if (!IsWindowVisible(hStatusBar)) {
		return;
	}

	/* [h郊\[XID */
	uID = (UINT)LOWORD(wParam);
	if (((uID >= IDM_D0MEDIA00) && (uID <= IDM_D0MEDIA15)) ||
		((uID >= IDM_D1MEDIA00) && (uID <= IDM_D1MEDIA15)) ||
#if XM7_VER == 1 && defined(SFDC)
		((uID >= IDM_D4MEDIA00) && (uID <= IDM_D4MEDIA15)) ||
		((uID >= IDM_D5MEDIA00) && (uID <= IDM_D5MEDIA15)) ||
		((uID >= IDM_D6MEDIA00) && (uID <= IDM_D6MEDIA15)) ||
		((uID >= IDM_D7MEDIA00) && (uID <= IDM_D7MEDIA15)) ||
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER == 1 && defined(BUBBLE)
		((uID >= IDM_B0MEDIA00) && (uID <= IDM_B0MEDIA15)) ||
		((uID >= IDM_B1MEDIA00) && (uID <= IDM_B1MEDIA15)) ||
		((uID >= IDM_B2MEDIA00) && (uID <= IDM_B2MEDIA15)) ||
		((uID >= IDM_B3MEDIA00) && (uID <= IDM_B3MEDIA15)) ||
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
		((uID >= IDM_D2MEDIA00) && (uID <= IDM_D2MEDIA15)) ||
		((uID >= IDM_D3MEDIA00) && (uID <= IDM_D3MEDIA15))) {
		/* fBA */
		uID = IDS_MEDIA_CHANGE;
	}
	if ((uID >= IDM_STATE_FILE0) && (uID <= IDM_STATE_FILE9)) {
		/* z}VXe[gt@CJ */
		uID = IDM_OPEN;
	}
	if ((uID >= IDM_D0_FILE0) && (uID <= IDM_D0_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D0OPEN;
	}
	if ((uID >= IDM_D1_FILE0) && (uID <= IDM_D1_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D1OPEN;
	}
	if ((uID >= IDM_D2_FILE0) && (uID <= IDM_D2_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D2OPEN;
	}
	if ((uID >= IDM_D3_FILE0) && (uID <= IDM_D3_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D3OPEN;
	}
#if XM7_VER == 1 && defined(SFDC)
	if ((uID >= IDM_D4_FILE0) && (uID <= IDM_D4_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D4OPEN;
	}
	if ((uID >= IDM_D5_FILE0) && (uID <= IDM_D5_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D5OPEN;
	}
	if ((uID >= IDM_D6_FILE0) && (uID <= IDM_D6_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D6OPEN;
	}
	if ((uID >= IDM_D7_FILE0) && (uID <= IDM_D7_FILE9)) {
		/* fBXNC[W}Eg */
		uID = IDM_D7OPEN;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if ((uID >= IDM_T_FILE0) && (uID <= IDM_T_FILE9)) {
		/* e[vC[W}Eg */
		uID = IDM_TOPEN;
	}
#if XM7_VER == 1 && defined(BUBBLE)
	if ((uID >= IDM_B0_FILE0) && (uID <= IDM_B0_FILE9)) {
		/* ouJZbgC[W}Eg */
		uID = IDM_B0OPEN;
	}
	if ((uID >= IDM_B1_FILE0) && (uID <= IDM_B1_FILE9)) {
		/* ouJZbgC[W}Eg */
		uID = IDM_B1OPEN;
	}
	if ((uID >= IDM_B2_FILE0) && (uID <= IDM_B2_FILE9)) {
		/* ouJZbgC[W}Eg */
		uID = IDM_B2OPEN;
	}
	if ((uID >= IDM_B3_FILE0) && (uID <= IDM_B3_FILE9)) {
		/* ouJZbgC[W}Eg */
		uID = IDM_B3OPEN;
	}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
#ifdef FMDLL
	if ((uID >= IDM_EXTSND1) && (uID <= IDM_EXTSND9)) {
		/* ODLLύX */
		uID = IDM_EXT_SOUND_DLL;
	}
#endif
#ifdef FILTERDLL
	if ((uID >= IDM_EXTFLT1) && (uID <= IDM_EXTFLT9)) {
		/* O摜tB^DLLύX */
		uID = IDM_EXT_FILTER_DLL;
	}
#endif

	/* 񃊃\[X[h݂ */
	if (GetMsg(hResInstance, uID, buffer, sizeof(buffer)) == 0) {
		buffer[0] = '\0';
	}

	/* Zbg */
	Status_SetText(hStatusBar, 0, nStatusBorder, buffer);

	/* \[XID([htO)ۑ */
	uResID = uID;
}

/*
 *	j[I
 */
void FASTCALL OnExitMenuLoop(void)
{
	ASSERT(hStatusBar);

	/* Xe[^Xo[\ĂȂ΁AȂ */
	if (!IsWindowVisible(hStatusBar)) {
		return;
	}

	/* AChbZ[W\ */
	if (bStatusMessage) {
		Status_SetText(hStatusBar, 0, nStatusBorder, szStatusMessage);
	}
	else {
		Status_SetText(hStatusBar, 0, nStatusBorder, szIdleMessage);
	}

	/* \[XID([htO)ۑ */
	uResID = 0;
}

/*-[ WindowsXPΉ ]----------------------------------------------------*/

/*
 *	WindowsXP rWAX^CΉp 08/14
 *	UXTHEME.DLL
 */
void FASTCALL InitThemeDLL(void)
{
	char path[_MAX_PATH];

	/* uxtheme.dll[h */
	GetSystemDirectory(path, sizeof(path));
	strncat(path, "\\uxtheme.dll", sizeof(path) - strlen(path) - 1);
	hInstTheme = LoadLibrary(path);

	/* ֐̐擪AhXݒ */
	if (hInstTheme) {
		IsAppThemed = (BOOL(*)(void))GetProcAddress(hInstTheme, "IsAppThemed");

		if (!IsAppThemed) {
			/* s */
			CleanThemeDLL();
		}
	}
}

/*
 *	WindowsXP rWAX^CΉp
 *	UXTHEME.DLLN[Abv
 */
void FASTCALL CleanThemeDLL(void)
{
	/* DLLǂݍ܂ĂJ */
	if (hInstTheme) {
		FreeLibrary(hInstTheme);
		IsAppThemed = NULL;
		hInstTheme = NULL;
	}
}

/*
 *	WindowsXP rWAX^CΉp
 *	Xe[^XbZ[W{[_[`ԕύX
 */
void FASTCALL ChangeStatusBorder(HWND hwnd)
{
	/* e[}KpԂł͘g`悷 */
	if (hInstTheme && IsAppThemed) {
		if (!IsAppThemed()) {
			nStatusBorder = SBT_NOBORDERS;
			bThemed = FALSE;
		}
		else {
			nStatusBorder = 0;
			bThemed = TRUE;
		}
	}
	else {
		nStatusBorder = SBT_NOBORDERS;
		bThemed = FALSE;
	}

	/* AChbZ[W\ */
	if (bStatusMessage) {
		Status_SetText(hwnd, 0, nStatusBorder, szStatusMessage);
	}
	else {
		Status_SetText(hwnd, 0, nStatusBorder, szIdleMessage);
	}
}

#endif	/* _WIN32 */
