/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API DirectDraw ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#ifdef DINPUT8
#define DIRECTDRAW_VERSION		0x700	/* DirectX7w */
#else
#define DIRECTDRAW_VERSION		0x300	/* DirectX3w */
#endif
#include <ddraw.h>
#include <objbase.h>
#include <assert.h>
#include <shellapi.h>
#include "xm7.h"
#include "subctrl.h"
#include "display.h"
#include "multipag.h"
#include "ttlpalet.h"
#include "apalet.h"
#include "fdc.h"
#include "tapelp.h"
#include "bubble.h"
#include "keyboard.h"
#include "mouse.h"
#include "opn.h"
#include "tools.h"
#include "w32.h"
#include "w32_bar.h"
#include "w32_sch.h"
#include "w32_res.h"
#include "w32_draw.h"
#include "w32_dd.h"
#include "w32_kbd.h"
#include "w32_sub.h"

/*
 *	O[o [N
 */
#ifdef __cplusplus
extern "C" {
#endif
#if XM7_VER == 1
DWORD rgbTTLDD[16];						/* 640x200 pbg */
#else
DWORD rgbTTLDD[8];						/* 640x200 pbg */
DWORD rgbAnalogDD[4096];				/* 320x200 pbg */
#endif
BYTE nDDResolutionMode;					/* tXN[𑜓x */
BOOL bDD480Status;						/* 640x480Xe[^XtO */
BOOL bDDtruecolor;						/* TrueColorDtO */
BYTE DDDrawFlag[4000];					/* 8x8 ĕ`̈tO */
#if XM7_VER == 1 && defined(L4CARD)
DWORD rgbTTLL4DD[32];					/* 640x400 PFpbg */
#endif
#ifdef __cplusplus
}
#endif

/*
 *	X^eBbN [N
 */
#if XM7_VER >= 3
static BYTE nMode;						/* ʃ[h */
#elif XM7_VER >= 2
static BOOL bAnalog;					/* AiO[htO */
#elif XM7_VER == 1 && defined(L4CARD)
static BOOL b400Line;					/* 400C[htO */
#endif
static RECT BkRect;						/* EChE̋` */
static BOOL bMouseCursor;				/* }EXJ[\tO */
static HMENU hMainMenu;					/* Cj[nh */
static HBRUSH hDrawBrush;				/* h[EBhẼuV */
static HBRUSH hMainBrush;				/* CEBhẼuV */
static LPDIRECTDRAW2 lpdd2;				/* DirectDraw2 */
static LPDIRECTDRAWSURFACE lpdds[2];	/* DirectDrawSurface3 */
static LPDIRECTDRAWCLIPPER lpddc;		/* DirectDrawClipper */
static UINT nPixelFormat;				/* 320x200 sNZtH[}bg */
static WORD nDrawTop;					/* `͈͏ */
static WORD nDrawBottom;				/* `͈͉ */
static WORD nDrawLeft;					/* `͈͍ */
static WORD nDrawRight;					/* `͈͉E */
static BOOL bPaletFlag;					/* pbgύXtO */
static BYTE nDDResolution;				/* tXN[ ۂ̉𑜓x */
static BOOL bClearFlag;					/* ㉺NAtO */
#if XM7_VER >= 3
static BOOL bWindowOpen;				/* n[hEFAEBhE */
static WORD nWindowDx1;					/* EBhEXW */
static WORD nWindowDy1;					/* EBhEYW */
static WORD nWindowDx2;					/* EBhEEXW */
static WORD nWindowDy2;					/* EBhEEYW */
#endif

static char szRunMessage[128];			/* RUNbZ[W */
static char szStopMessage[128];			/* STOPbZ[W */
static char szCaption[256];				/* LvV */
static char szCAPMessage[16];			/* CAPL[bZ[W */
static char szKANAMessage[16];			/* ȃL[bZ[W */
static char szINSMessage[16];			/* INSL[bZ[W */
static int nCAP;						/* CAPL[ */
static int nKANA;						/* ȃL[ */
static int nINS;						/* INSL[ */
#if XM7_VER == 1 && defined(SFDC)
static int nDrive[2][4];				/* tbs[hCu */
static char szDrive[2][4][16 + 1];		/* tbs[hCu */
#else	/* XM7_VER == 1 && defined(SFDC) */
static int nDrive[4];					/* tbs[hCu */
static char szDrive[4][16 + 1];			/* tbs[hCu */
#endif	/* XM7_VER == 1 && defined(SFDC) */
static int nTape;						/* e[v */
static BOOL bCaption;					/* LvV\ */
static int nScrX;						/* 𑜓x */
static int nScrY;						/* c𑜓x */

/*
 * ddraw.lib̃NsvƂ邽߂̒ǉ
 */
typedef HRESULT(WINAPI *DIRECTDRAWCREATEFUNC)(GUID*, LPDIRECTDRAW*, IUnknown*);
static DIRECTDRAWCREATEFUNC pCreateFunc;/* DirectDrawCreate */
static HMODULE hDDrawDll;				/* ddraw.dll */

/*
 *	}`j^Ή\
 *	Windows95łRpCʂ߈Ӑ}IɖOύXĂ
 */
typedef struct tagMONITOR_INFO {
    DWORD  cbSize;
    RECT   rcMonitor;
    RECT   rcWork;
    DWORD  dwFlags;
} MONITOR_INFO, *LPMONITOR_INFO;

/*
 *	}`j^Ή֘A
 */
static HMODULE hModUser32;				/* user32.dllW[ */
static HMONITOR (WINAPI* monitorfromwindow)(HWND, DWORD);
static BOOL (WINAPI* getmonitorinfo)(HMONITOR, LPMONITOR_INFO);
static HRESULT (WINAPI* ddenumerateex)(LPDDENUMCALLBACKEX, LPVOID, DWORD);
static HMONITOR (WINAPI* monitorfromrect)(LPCRECT, DWORD);
static HMONITOR hmonitor;				/* T hmonitor */
static GUID gmonitor;					/* hmonitor ɑΉ GUID */
static POINT ptMouseBak;				/* EBhEÃ}EXʒu */
static POINT ptFullScr;					/* tXN[}EXʒu */

static POINT ptFullRes[2];				/* tXN[𑜓x */
static int nMlt;						/* { */

/*
 *	AZu֐̂߂̃vg^Cv錾
 */
#ifdef __cplusplus
extern "C" {
#endif
#if XM7_VER == 1
/* 15/16rbgJ[ */
void Render640DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void RenderL4DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render640mDD(LPVOID lpSurface, LONG lPitch, int first, int last);

/* 32rbgJ[ */
void Render640Tc32DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void RenderL4Tc32DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render640mTc32DD(LPVOID lpSurface, LONG lPitch, int first, int last);

/* Ăяop */
static void (*Render640)(LPVOID lpSurface, LONG lPitch, int first, int last);
static void (*Render640m)(LPVOID lpSurface, LONG lPitch, int first, int last);
#ifdef L4CARD
static void (*RenderL4)(LPVOID lpSurface, LONG lPitch, int first, int last);
#endif
#elif XM7_VER == 2
/* 15/16rbgJ[ */
void Render640DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render320DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render640cDD(LPVOID lpSurface, LONG lPitch, int first, int last);

/* 32rbgJ[ */
void Render640Tc32DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render320Tc32DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render640cTc32DD(LPVOID lpSurface, LONG lPitch, int first, int last);

/* Ăяop */
static void (*Render640)(LPVOID lpSurface, LONG lPitch, int first, int last);
static void (*Render320)(LPVOID lpSurface, LONG lPitch, int first, int last);
static void (*Render640c)(LPVOID lpSurface, LONG lPitch, int first, int last);
#elif XM7_VER >= 3
/* 15/16rbgJ[ */
void Render640DD2(LPVOID lpSurface, LONG lPitch,
					int first, int last, int scale);
void Render640wDD2(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx, int scale);
void Render320DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render320wDD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx);
void Render256k555DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int multipage);
void Render256k565DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int multipage);
void Render640cDD(LPVOID lpSurface, LONG lPitch,
					int first, int last);
void Render640cwDD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx);

/* 24rbgJ[ */
void Render640Tc24DD2(LPVOID lpSurface, LONG lPitch,
					int first, int last, int scale);
void Render640wTc24DD2(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx, int scale);
void Render320Tc24DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render320wTc24DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx);
void Render256kTc24DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int multipage);
void Render640cTc24DD(LPVOID lpSurface, LONG lPitch,
					int first, int last);
void Render640cwTc24DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx);

/* 32rbgJ[ */
void Render640Tc32DD2(LPVOID lpSurface, LONG lPitch,
					int first, int last, int scale);
void Render640wTc32DD2(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx, int scale);
void Render320Tc32DD(LPVOID lpSurface, LONG lPitch, int first, int last);
void Render320wTc32DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx);
void Render256kTc32DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int multipage);
void Render640cTc32DD(LPVOID lpSurface, LONG lPitch,
					int first, int last);
void Render640cwTc32DD(LPVOID lpSurface, LONG lPitch,
					int first, int last, int firstx, int lastx);

/* Ăяop */
static void (*Render640)(LPVOID lpSurface, LONG lPitch,
				int first, int last, int scale);
static void (*Render640w)(LPVOID lpSurface, LONG lPitch,
				int first, int last, int firstx, int lastx, int scale);
static void (*Render320)(LPVOID lpSurface, LONG lPitch, int first, int last);
static void (*Render320w)(LPVOID lpSurface, LONG lPitch,
				int first, int last, int firstx, int lastx);
static void (*Render256k)(LPVOID lpSurface, LONG lPitch,
				int first, int last, int multipage);
static void (*Render640c)(LPVOID lpSurface, LONG lPitch,
				int first, int last);
static void (*Render640cw)(LPVOID lpSurface, LONG lPitch,
				int first, int last, int firstx, int lastx);
#endif
#ifdef __cplusplus
}
#endif

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	16F[hppbgR[h
 */
static const WORD Palet16_L4[] = {
	0x0000,	0x0017,	0xb800,	0xb817,	0x05e0,	0x05f7,	0xbde0,	0xbdf7,
	0x4208,	0x421f,	0xfa08,	0xfa1f,	0x47e8,	0x47ff,	0xffe8,	0xffff,
	0x0000,	0x0017,	0x5c00,	0x5c17,	0x02e0,	0x02f7,	0x5ee0,	0x5ef7,
	0x2108,	0x211f,	0x7d08,	0x7d1f,	0x23e8,	0x23ff,	0x7fe8,	0x7fff,
	0x0000,	0x1984,	0x32e9,	0x334a,	0x4cef,	0x5510,	0x5571,	0x5db1,
	0x2227,	0x3309,	0x446d,	0x4cae,	0x6694,	0x6ed5,	0x7736,	0x7777,
	0x0000,	0x0cc4,	0x1969,	0x19aa,	0x266f, 0x2a90,	0x2ab1,	0x2ed1,
	0x1107,	0x1989,	0x222d,	0x264e,	0x3354,	0x3755,	0x3b96,	0x3bb7,
	0x1000,	0x2861,	0x4922,	0x5143,	0x8265,	0x8a86,	0x92c6,	0x9ae6,
	0x38a1,	0x4922,	0x69e4,	0x7a25,	0xbba8,	0xc3c9,	0xc42a,	0xc44a,
	0x0800,	0x1421,	0x2482,	0x28a3,	0x4125,	0x4546,	0x4966,	0x4d66,
	0x1c41,	0x2482,	0x34e4,	0x3d05,	0x5dc8,	0x61e9,	0x620a,	0x622a,
};
static const DWORD Palet16Tc_L4[] = {
	/* J[j^ */
	0x00000000,	0x000000bb,	0x00bb0000,	0x00bb00bb,
	0x0000bb00,	0x0000bbbb,	0x00bbbb00,	0x00bbbbbb,
	0x00444444,	0x004444ff,	0x00ff4444,	0x00ff44ff,
	0x0044ff44,	0x0044ffff,	0x00ffff44,	0x00ffffff,
	/* O[j^ */
	0x00000000,	0x00173025,	0x002e5e45,	0x00336953,
	0x004c9e7c,	0x004fa380,	0x0055b08b,	0x0057b58e,
	0x00214536,	0x002e6046,	0x00448c6e,	0x00499777,
	0x0066d3a6,	0x0069d9ab,	0x0070e8b7,	0x0073eebb,
	/* Ao[CG[j^ */
	0x00170000,	0x002a0e08,	0x00492415,	0x00512a19,
	0x00814c2d,	0x008b5030,	0x00945935,	0x00985d37,
	0x0038170e,	0x004b2516,	0x006f3f26,	0x007a472b,
	0x00b97445,	0x00c27b49,	0x00c28750,	0x00c28b53,
};
#endif

/*
 *	vg^Cv錾
 */
static void FASTCALL AllClear(BOOL clear_flag);
static void FASTCALL StatusLine(BOOL force_redraw);

/*
 *	tXN[[hp̃bZ[W[h
 */
void FASTCALL LoadMessageInitDD(void)
{
	/* bZ[W[h */
	if (GetMsg(hResInstance, IDS_RUNCAPTION,
					szRunMessage, sizeof(szRunMessage)) == 0) {
		szRunMessage[0] = '\0';
	}
	if (GetMsg(hResInstance, IDS_STOPCAPTION,
					szStopMessage, sizeof(szStopMessage)) == 0) {
		szStopMessage[0] = '\0';
	}
	if (GetMsg(hResInstance, IDS_BAR_CAP,
					szCAPMessage, sizeof(szCAPMessage)) == 0) {
		strncpy(szCAPMessage, "CAP", 16);
	}
	if (GetMsg(hResInstance, IDS_BAR_KANA,
					szKANAMessage, sizeof(szKANAMessage)) == 0) {
#if XM7_VER == 1
		strncpy(szKANAMessage, "Ji", 16);
#else
		strncpy(szKANAMessage, "", 16);
#endif
	}
	if (GetMsg(hResInstance, IDS_BAR_INS,
					szINSMessage, sizeof(szINSMessage)) == 0) {
		strncpy(szINSMessage, "INS", 16);
	}
}

/*
 *	
 */
void FASTCALL InitDD(void)
{
	int i;
	char path[_MAX_PATH];
	int xsize, ysize;

	/* [NGA(ݒ胏[N͕ύXȂ) */
#if XM7_VER >= 3
	nMode = SCR_200LINE;
#elif XM7_VER >= 2
	bAnalog = FALSE;
#elif XM7_VER == 1 && defined(L4CARD)
	b400Line = FALSE;
#endif
	bMouseCursor = TRUE;
	lpdd2 = NULL;
	memset(lpdds, 0, sizeof(lpdds));
	lpddc = NULL;
	bClearFlag = TRUE;
	ptFullScr.x = 0;
	ptFullScr.y = 0;

	/* Xe[^XC */
	szCaption[0] = '\0';
	nCAP = -1;
	nKANA = -1;
	nINS = -1;
	for (i = 0; i < 4; i ++) {
#if XM7_VER == 1 && defined(SFDC)
		nDrive[FDC_M][i] = -1;
		nDrive[FDC_S][i] = -1;
		szDrive[FDC_M][i][0] = '\0';
		szDrive[FDC_S][i][0] = '\0';
#else	/* XM7_VER == 1 && defined(SFDC) */
		nDrive[i] = -1;
		szDrive[i][0] = '\0';
#endif	/* XM7_VER == 1 && defined(SFDC) */
	}
	nTape = -1;
	bCaption = TRUE;

	/* }`j^ΉpMonitorFromWindow[h */
	GetSystemDirectory(path, sizeof(path));
	strncat(path, "\\user32.dll", sizeof(path) - strlen(path) - 1);
	hModUser32 = LoadLibrary(path);
	if (hModUser32) {
		monitorfromwindow = (HMONITOR (WINAPI*)(HWND, DWORD))GetProcAddress(hModUser32, "MonitorFromWindow");
		getmonitorinfo = (BOOL (WINAPI*)(HMONITOR, LPMONITOR_INFO))GetProcAddress(hModUser32, "GetMonitorInfoA");
		monitorfromrect = (HMONITOR (WINAPI*)(LPCRECT, DWORD))GetProcAddress(hModUser32, "MonitorFromRect");
	}
	else {
		monitorfromwindow = NULL;
		getmonitorinfo = NULL;
		monitorfromrect = NULL;
	}

	/* 𑜓x */
	ptFullRes[0].x = GetSystemMetrics(SM_CXSCREEN);
	ptFullRes[0].y = GetSystemMetrics(SM_CYSCREEN);
	if (monitorfromrect && getmonitorinfo) {
		/* }`j^ */
		RECT srect;
		MONITOR_INFO minfo;

		/* EBhÊ郂j^nh擾 */
		GetWindowRect(hMainWnd, &srect);
		hmonitor = monitorfromrect(&srect, MONITOR_DEFAULTTONEAREST);
		memset(&minfo, 0, sizeof(minfo));
		minfo.cbSize = sizeof(minfo);
		getmonitorinfo(hmonitor, &minfo);

		/* ^XNo[̂郂j^nh擾 */
		APPBARDATA ad;
		ad.cbSize = sizeof(APPBARDATA);
		ad.hWnd = NULL;
		SHAppBarMessage(ABM_GETTASKBARPOS, &ad);
		HMONITOR hmonitor2 = monitorfromrect(&(ad.rc), MONITOR_DEFAULTTONEAREST);

		if (hmonitor != hmonitor2) {
			srect = minfo.rcWork;
			ptFullRes[0].x = srect.right - srect.left;
			ptFullRes[0].y = srect.bottom - srect.top;
		}
	}
	/* { */
	if (bDD480Status) {
		xsize = 640;
		ysize = 480;
	}
	else {
		xsize = 640;
		ysize = 400;
	}
	if ((ptFullRes[0].x / xsize) <= (ptFullRes[0].y / ysize)) {
		nMlt = ptFullRes[0].x / xsize;
	}
	else {
		nMlt = ptFullRes[0].y / ysize;
	}
	/* ] */
	ptFullRes[1].x = (ptFullRes[0].x - xsize * nMlt) / 2;
	ptFullRes[1].y = (ptFullRes[0].y - ysize * nMlt) / 2;

	/* bZ[W[h */
	LoadMessageInitDD();

	/* DirectDrawCreatẽ|C^𓾂 */
	hDDrawDll = LoadLibrary("ddraw.dll");
	if (hDDrawDll) {
		pCreateFunc = (DIRECTDRAWCREATEFUNC)GetProcAddress(hDDrawDll, "DirectDrawCreate");
	}
}

/*
 *	N[Abv
 */
void FASTCALL CleanDD(void)
{
	DWORD dwStyle;
	RECT brect;
	int i;
	int width;
	int height;
#ifdef FILTERDLL
	double d;
#endif

	/* DirectDrawClipper */
	if (lpddc) {
		lpddc->Release();
		lpddc = NULL;
	}

	/* DirectDrawSurface3 */
	for (i=0; i<2; i++) {
		if (lpdds[i]) {
			lpdds[i]->Release();
			lpdds[i] = NULL;
		}
	}

	/* DirectDraw2 */
	if (lpdd2) {
		lpdd2->Release();
		lpdd2 = NULL;
	}

	/* EChEX^C߂ */
	dwStyle = GetWindowLongPtr(hMainWnd, GWL_STYLE);
	dwStyle &= ~WS_POPUP;
	dwStyle |= (WS_CAPTION | WS_BORDER | WS_SYSMENU);
	SetWindowLongPtr(hMainWnd, GWL_STYLE, dwStyle);
	dwStyle = GetWindowLongPtr(hMainWnd, GWL_EXSTYLE);
	dwStyle |= WS_EX_WINDOWEDGE;
	SetWindowLongPtr(hMainWnd, GWL_EXSTYLE, dwStyle);

	/* Windows Vista/Windows 7ł̃ACR\A */
	SetClassLongPtr(hMainWnd, GCLP_HICONSM,
		(LONG_PTR)LoadIcon(hAppInstance, IDI_APPLICATION));

	/* Windows 8ł̃j[EuV */
	if (bWin8flag) {
		SetMenu(hMainWnd, hMainMenu);
#ifdef _M_AMD64
		SetClassLongPtr(hMainWnd, GCLP_HBRBACKGROUND, (LONG)hMainBrush);
		SetClassLongPtr(hDrawWnd, GCLP_HBRBACKGROUND, (LONG)hDrawBrush);
#else
		SetClassLong(hMainWnd, GCL_HBRBACKGROUND, (LONG)hMainBrush);
		SetClassLong(hDrawWnd, GCL_HBRBACKGROUND, (LONG)hDrawBrush);
#endif
	}

	/* EChEʒu߂ */
#ifdef FILTERDLL
	if (bAspectRatio && bFilterDll && bUseFilterDll) {
		d = dAspectRatio;			/* 0.899(0.4495) */
	}
	else {
		d = 1.000;					/* 1.000(0.5000) */
	}
	switch (nScalingSize) {
		case DISPLAYSIZE_100:		/* x1.0(640x400) */
		default:
			width  = 640 * 1.0;
			height = 400 * 1.0 / d;
			break;

		case DISPLAYSIZE_150:		/* x1.5(960x600) */
			width  = 640 * 1.5;
			height = 400 * 1.5 / d;
			break;

		case DISPLAYSIZE_200:		/* x2.0(1280x800) */
			width  = 640 * 2.0;
			height = 400 * 2.0 / d;
			break;

		case DISPLAYSIZE_050:		/* x0.5(320x200) */
			width  = 640 * 0.5;
			height = 400 * 0.5 / d;
			break;

		case DISPLAYSIZE_180:		/* x1.8(1152x720) */
			width  = 640 * 1.8;
			height = 400 * 1.8 / d;
			break;

		case DISPLAYSIZE_120:		/* x1.2(768x480) */
			width  = 640 * 1.2;
			height = 400 * 1.2 / d;
			break;

		case DISPLAYSIZE_300:		/* x3.0(1920x1200) */
			width  = 640 * 3.0;
			height = 400 * 3.0 / d;
			break;

		case DISPLAYSIZE_400:		/* x4.0(2560x1600) */
			width  = 640 * 4.0;
			height = 400 * 4.0 / d;
			break;
	}
#else
	switch (nScalingSize) {
		case DISPLAYSIZE_100:		/* x1.0(640x400) */
		default:
			width  = 640 * 1.0;
			height = 400 * 1.0;
			break;

		case DISPLAYSIZE_200:		/* x2.0(1280x800) */
			width  = 640 * 2.0;
			height = 400 * 2.0;
			break;

		case DISPLAYSIZE_300:		/* x3.0(1920x1200) */
			width  = 640 * 3.0;
			height = 400 * 3.0;
			break;

		case DISPLAYSIZE_400:		/* x4.0(2560x1600) */
			width  = 640 * 4.0;
			height = 400 * 4.0;
			break;
	}
#endif
	SetWindowPos(hMainWnd, HWND_NOTOPMOST, BkRect.left, BkRect.top,
		(BkRect.right - BkRect.left), (BkRect.bottom - BkRect.top),
		SWP_DRAWFRAME);

	MoveWindow(hDrawWnd, 0, 0, width, height, TRUE);
	if (hStatusBar) {
		GetWindowRect(hStatusBar, &brect);
		MoveWindow(hStatusBar, 0, height,
					(brect.right - brect.left),
					(brect.bottom - brect.top),
					TRUE);
	}

	/* user32.dll */
	if (hModUser32) {
		FreeLibrary(hModUser32);
		hModUser32 = NULL;
	}

	/* }EX|C^ʒu𕜋A */
	SetCursorPos(ptMouseBak.x, ptMouseBak.y);

	/* ddraw.dll */
	if (hDDrawDll) {
		FreeLibrary(hDDrawDll);
		hDDrawDll = NULL;
	}
}


/*
 *	}`j^pDirectDraw񋓃R[obN֐
 */
static BOOL WINAPI DDEnumCallback(GUID FAR* guid, LPSTR desc, LPSTR name, LPVOID context, HMONITOR hm)
{
	UNUSED(desc);
	UNUSED(name);
	UNUSED(context);

	if (hm == hmonitor) {
		gmonitor = *guid;
		return 0;
	}
	return 1;
}

/*
 *	ʐݒTu
 */
static BOOL FASTCALL SetDisplayMode(int x, int y)
{
#if XM7_VER >= 3
	if (bDDtruecolor) {
		nPixelFormat = 3;
		if (FAILED(lpdd2->SetDisplayMode(x, y, 32, 0, 0))) {
			nPixelFormat = 4;
			if (FAILED(lpdd2->SetDisplayMode(x, y, 24, 0, 0))) {
				nPixelFormat = 0;
				if (FAILED(lpdd2->SetDisplayMode(x, y, 16, 0, 0))) {
					/* s */
					return FALSE;
				}
			}
		}
	}
	else {
		if (bWin8flag) {
			nPixelFormat = 3;
			if (FAILED(lpdd2->SetDisplayMode(x, y, 32, 0, 0))) {
				nPixelFormat = 0;
				if (FAILED(lpdd2->SetDisplayMode(x, y, 16, 0, 0))) {
					/* s */
					return FALSE;
				}
			}
		}
		else {
			nPixelFormat = 0;
			if (FAILED(lpdd2->SetDisplayMode(x, y, 16, 0, 0))) {
				nPixelFormat = 3;
				if (FAILED(lpdd2->SetDisplayMode(x, y, 32, 0, 0))) {
					/* s */
					return FALSE;
				}
			}
		}
	}
#else
	if (bWin8flag || bDDtruecolor) {
		nPixelFormat = 3;
		if (FAILED(lpdd2->SetDisplayMode(x, y, 32, 0, 0))) {
			nPixelFormat = 0;
			if (FAILED(lpdd2->SetDisplayMode(x, y, 16, 0, 0))) {
				/* s */
				return FALSE;
			}
		}
	}
	else {
		nPixelFormat = 0;
		if (FAILED(lpdd2->SetDisplayMode(x, y, 16, 0, 0))) {
			nPixelFormat = 3;
			if (FAILED(lpdd2->SetDisplayMode(x, y, 32, 0, 0))) {
				/* s */
				return FALSE;
			}
		}
	}
#endif

	/*  */
	nScrX = x;
	nScrY = y;
	return TRUE;
}

/*
 *	ZNg
 */
BOOL FASTCALL SelectDD(void)
{
	char path[_MAX_PATH];
	DWORD dwStyle;
	LPDIRECTDRAW lpdd;
	DDSURFACEDESC ddsd;
	DDPIXELFORMAT ddpf;
	RECT brect;
	LPDIRECTDRAW lpdd1;
	HMODULE hModDDraw;
	RECT srect;
	MONITOR_INFO minfo;

	/* assert */
	ASSERT(hMainWnd);

	/* EChE`E}EX|C^ʒuL */
	GetWindowRect(hMainWnd, &BkRect);
	GetCursorPos(&ptMouseBak);

	/* j[nhL */
	if (bWin8flag) {
		hMainMenu = GetMenu(hMainWnd);
	}

	/* EChEX^CύX */
	dwStyle = GetWindowLongPtr(hMainWnd, GWL_STYLE);
	dwStyle &= ~(WS_CAPTION | WS_BORDER | WS_SYSMENU);
	dwStyle |= WS_POPUP;
	SetWindowLongPtr(hMainWnd, GWL_STYLE, dwStyle);
	dwStyle = GetWindowLongPtr(hMainWnd, GWL_EXSTYLE);
	dwStyle &= ~WS_EX_WINDOWEDGE;
	SetWindowLongPtr(hMainWnd, GWL_EXSTYLE, dwStyle);

	/* DirectDrawIuWFNg쐬(}`j^Ή) */
	lpdd = NULL;
	hmonitor = NULL;

	/* }`j^ΉpDirectDraw */
	memset(&gmonitor, 0, sizeof(gmonitor));
	if (monitorfromwindow) {
		GetSystemDirectory(path, sizeof(path));
		strncat(path, "\\ddraw.dll", sizeof(path) - strlen(path) - 1);
		hModDDraw = LoadLibrary(path);
		if (hModDDraw) {
			ddenumerateex = (HRESULT (WINAPI*)(LPDDENUMCALLBACKEX, LPVOID,
				DWORD))GetProcAddress(hModDDraw,"DirectDrawEnumerateExA");
			if (ddenumerateex) {
				hmonitor = monitorfromwindow(hMainWnd, 1);
				ddenumerateex(DDEnumCallback, NULL,
							  DDENUM_ATTACHEDSECONDARYDEVICES);
			}
			FreeLibrary(hModDDraw);
		}
	}

	if (!FAILED(CoCreateInstance(CLSID_DirectDraw, 0, CLSCTX_ALL, IID_IDirectDraw, (void**)&lpdd1))) {
		if (!FAILED(IDirectDraw_Initialize(lpdd1, &gmonitor))) {
			lpdd = lpdd1;
		}
	}

	if (!lpdd) {
		/* DirectDrawIuWFNg쐬 */
		if (pCreateFunc) {
			if (FAILED(pCreateFunc(NULL, &lpdd, NULL))) {
				return FALSE;
			}
		}
		else {
			return FALSE;
		}
	}

	/* [hݒ */
	if (FAILED(lpdd->SetCooperativeLevel(hMainWnd, DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN))) {
		lpdd->Release();
		return FALSE;
	}

	/* DirectDraw2C^tF[X擾 */
	if (FAILED(lpdd->QueryInterface(IID_IDirectDraw2, (LPVOID*)&lpdd2))) {
		lpdd->Release();
		return FALSE;
	}

	/* ܂ŗ΁ADirectDraw͂͂KvȂ */
	lpdd->Release();

	/* ʃ[hݒ */
	switch (nDDResolutionMode) {
		case 0:	/* 640 x 400 */
			nDDResolution = DDRES_400LINE;
			if (!SetDisplayMode(640, 400)) {
				nDDResolution = DDRES_480LINE;
				if (!SetDisplayMode(640, 480)) {
					return FALSE;
				}
			}
			break;
		case 1:	/* 640 x 480 */
			nDDResolution = DDRES_480LINE;
			if (!SetDisplayMode(640, 480)) {
				nDDResolution = DDRES_400LINE;
				if (!SetDisplayMode(640, 400)) {
					return FALSE;
				}
			}
			break;
		case 2:	/* 1920 x 1200 */
			nDDResolution = DDRES_WUXGA;
			if (!SetDisplayMode(1920, 1200)) {
				nDDResolution = DDRES_400LINE;
				if (!SetDisplayMode(640, 400)) {
					nDDResolution = DDRES_480LINE;
					if (!SetDisplayMode(640, 480)) {
						return FALSE;
					}
				}
			}
			break;
		case 3:	/* 1280 x 1024 */
			nDDResolution = DDRES_SXGA;
			if (!SetDisplayMode(1280, 1024)) {
				nDDResolution = DDRES_480LINE;
				if (!SetDisplayMode(640, 480)) {
					nDDResolution = DDRES_400LINE;
					if (!SetDisplayMode(640, 400)) {
						return FALSE;
					}
				}
			}
			break;
		case 4:	/* 1280 x 800 */
			nDDResolution = DDRES_WXGA800;
			if (!SetDisplayMode(1280, 800)) {
				nDDResolution = DDRES_400LINE;
				if (!SetDisplayMode(640, 400)) {
					nDDResolution = DDRES_480LINE;
					if (!SetDisplayMode(640, 480)) {
						return FALSE;
					}
				}
			}
			break;
		case 5:
			nDDResolution = DDRES_USER;
			if (!SetDisplayMode(ptFullRes[0].x,ptFullRes[0].y)) {
				nDDResolution = DDRES_400LINE;
				if (!SetDisplayMode(640, 400)) {
					nDDResolution = DDRES_480LINE;
					if (!SetDisplayMode(640, 480)) {
						return FALSE;
					}
				}
			}
			break;
		default:ASSERT(FALSE);
				return FALSE;
	}

	/* vC}T[tFCX쐬 */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS;
	if (!bWin8flag &&
		((nDDResolution == DDRES_400LINE) || (nDDResolution == DDRES_480LINE))){
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	}
	else {
		ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_SYSTEMMEMORY;
	}
	if (FAILED(lpdd2->CreateSurface(&ddsd, &lpdds[0], NULL))) {
		return FALSE;
	}

	/* [NT[tFCX쐬(DDSCAPS_SYSTEMMEMORYw肷) */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
	ddsd.dwWidth = 640;
	if (nDDResolution == DDRES_480LINE) {
		ddsd.dwHeight = 480;
	}
	else if (nDDResolution == DDRES_SXGA) {
		ddsd.dwHeight = 512;
	}
	else if (nDDResolution == DDRES_USER) {
		ddsd.dwWidth = ptFullRes[0].x / nMlt;
		ddsd.dwHeight = ptFullRes[0].y / nMlt;
	}
	else {
		ddsd.dwHeight = 400;
	}
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
	if (FAILED(lpdd2->CreateSurface(&ddsd, &lpdds[1], NULL))) {
		return FALSE;
	}

	if (nPixelFormat == 0) {
		/* sNZtH[}bg𓾂 */
		memset(&ddpf, 0, sizeof(ddpf));
		ddpf.dwSize = sizeof(ddpf);
		if (FAILED(lpdds[1]->GetPixelFormat(&ddpf))) {
			return FALSE;
		}

		/* sNZtH[}bg`FbN HELŋK肳Ă2^Cv̂ݑΉ */
		if (!(ddpf.dwFlags & DDPF_RGB)) {
			return FALSE;
		}
		if ((ddpf.dwRBitMask == 0xf800) &&
			(ddpf.dwGBitMask == 0x07e0) &&
			(ddpf.dwBBitMask == 0x001f)) {
			nPixelFormat = 1;
		}
		if ((ddpf.dwRBitMask == 0x7c00) &&
			(ddpf.dwGBitMask == 0x03e0) &&
			(ddpf.dwBBitMask == 0x001f)) {
			nPixelFormat = 2;
		}

		if (nPixelFormat == 0) {
			return FALSE;
		}
	}

#if XM7_VER == 1
	/* _O֐ݒ */
	if (nPixelFormat == 3) {
		Render640 = Render640Tc32DD;
		Render640m = Render640mTc32DD;
#ifdef L4CARD
		RenderL4 = RenderL4Tc32DD;
#endif
	}
	else {
		Render640 = Render640DD;
		Render640m = Render640mDD;
#ifdef L4CARD
		RenderL4 = RenderL4DD;
#endif
	}
#elif XM7_VER == 2
	/* _O֐ݒ */
	if (nPixelFormat == 3) {
		Render640 = Render640Tc32DD;
		Render320 = Render320Tc32DD;
		Render640c = Render640cTc32DD;
	}
	else {
		Render640 = Render640DD;
		Render320 = Render320DD;
		Render640c = Render640cDD;
	}
#elif XM7_VER >= 3
	/* _O֐ݒ */
	if (nPixelFormat == 1) {
		Render640 = Render640DD2;
		Render320 = Render320DD;
		Render256k = Render256k565DD;
		Render640w = Render640wDD2;
		Render320w = Render320wDD;
		Render640c = Render640cDD;
		Render640cw = Render640cwDD;
	}
	else if (nPixelFormat == 2) {
		Render640 = Render640DD2;
		Render320 = Render320DD;
		Render256k = Render256k555DD;
		Render640w = Render640wDD2;
		Render320w = Render320wDD;
		Render640c = Render640cDD;
		Render640cw = Render640cwDD;
	}
	else if (nPixelFormat == 3) {
		Render640 = Render640Tc32DD2;
		Render320 = Render320Tc32DD;
		Render256k = Render256kTc32DD;
		Render640w = Render640wTc32DD2;
		Render320w = Render320wTc32DD;
		Render640c = Render640cTc32DD;
		Render640cw = Render640cwTc32DD;
	}
	else {
		Render640 = Render640Tc24DD2;
		Render320 = Render320Tc24DD;
		Render256k = Render256kTc24DD;
		Render640w = Render640wTc24DD2;
		Render320w = Render320wTc24DD;
		Render640c = Render640cTc24DD;
		Render640cw = Render640cwTc24DD;
	}
#endif

	/* Nbp[쐬A蓖 */
	if (FAILED(lpdd2->CreateClipper(NULL, &lpddc, NULL))) {
		return FALSE;
	}
	if (FAILED(lpddc->SetHWnd(NULL, hDrawWnd))) {
		return FALSE;
	}

	/* EChETCYύX */
	if (nDDResolution != DDRES_400LINE) {
		if (hmonitor && getmonitorinfo) {
			memset(&minfo, 0, sizeof(minfo));
			minfo.cbSize = sizeof(minfo);
			getmonitorinfo(hmonitor, &minfo);
			srect = minfo.rcWork;
		}
		else {
			srect.left = 0;
			srect.top = 0;
			srect.right = nScrX;
			srect.bottom = nScrY;
		}
		if (hStatusBar) {
			GetWindowRect(hStatusBar, &brect);
		}
		else {
			brect.top = 0;
			brect.bottom = 0;
		}
		MoveWindow(hMainWnd, srect.left, srect.top, nScrX, nScrY + (brect.bottom - brect.top), TRUE);
		MoveWindow(hDrawWnd, 0, 0, nScrX, nScrY, TRUE);
		if (hStatusBar) {
			MoveWindow(hStatusBar, 0, nScrY, 0, (brect.bottom - brect.top),
						TRUE);
		}
	}

	/* j[o[EuVL(Windows 8̂) */
	if (bWin8flag) {
#ifdef _M_AMD64
		hDrawBrush = (HBRUSH)GetClassLongPtr(hDrawWnd, GCLP_HBRBACKGROUND);
		hMainBrush = (HBRUSH)GetClassLongPtr(hMainWnd, GCLP_HBRBACKGROUND);
		SetClassLongPtr(hDrawWnd, GCLP_HBRBACKGROUND, NULL);
		SetClassLongPtr(hMainWnd, GCLP_HBRBACKGROUND, NULL);
#else
		hDrawBrush = (HBRUSH)GetClassLong(hDrawWnd, GCL_HBRBACKGROUND);
		hMainBrush = (HBRUSH)GetClassLong(hMainWnd, GCL_HBRBACKGROUND);
		SetClassLong(hDrawWnd, GCL_HBRBACKGROUND, NULL);
		SetClassLong(hMainWnd, GCL_HBRBACKGROUND, NULL);
#endif

		SetMenu(hMainWnd, NULL);
		DrawMenuBar(hMainWnd);
	}
	else {
		DrawMenuBar(NULL);
	}

	/* LvV\tOݒ */
	if ((nDDResolution == DDRES_480LINE) || (nDDResolution == DDRES_SXGA)) {
		bCaption = bDD480Status;
	}
	else if (nDDResolution == DDRES_USER) {
		bCaption = bDD480Status;
	}
	else {
		bCaption = TRUE;
	}

	/* }EXʒuʒɈړ */
	GetWindowRect(hMainWnd, &srect);
	ptFullScr.x = (srect.right + srect.left) / 2;
	ptFullScr.y = (srect.bottom + srect.top) / 2;
	SetCursorPos(ptFullScr.x, ptFullScr.y);

	/* [NZbg */
#if XM7_VER >= 3
	nMode = SCR_200LINE;
#elif XM7_VER >= 2
	bAnalog = FALSE;
#elif XM7_VER == 1 && defined(L4CARD)
	b400Line = FALSE;
#endif
	ReDrawDD();
	AllClear(TRUE);
	InvalidateRect(hMainWnd, NULL, TRUE);
	InvalidateRect(hDrawWnd, NULL, TRUE);

	/*  */
	return TRUE;
}

/*-[ ` ]-----------------------------------------------------------------*/

#if XM7_VER == 2 && defined(FMTV151)
/*
 *	V2
 */
static void FASTCALL DrawV2_DD(BYTE *pSurface, LONG lPitch)
{
	BYTE	x, y;
	int		xx, yy;
	WORD	col;
	WORD	*pbits;
	DWORD	*pbitsTC;

	/* pbge[u */
	static const WORD V2rgbTable[] = {
		0x0000 | 0x0000 | 0x0000,
		0x0000 | 0x07e0 | 0x0000,
		0x0000 | 0x0000 | 0x0000,
		0x0000 | 0x03e0 | 0x0000,
		0x0000 | 0x0000 | 0x0000
	};
	static const WORD V2rgbTableTc[] = {
		0x00000000,	0x0000ff00,	0x00000000
	};

	if (!bFMTV151) {
		return;
	}

	if (!bRasterRendering) {
		if ((nDrawBottom < (V2YPoint * 2)) ||
			(nDrawTop > ((V2YPoint + V2YSize * V2YPxSz) * 2 - 1)) ||
			(nDrawRight < V2XPoint) ||
			(nDrawLeft > (V2XPoint + V2XSize * V2XPxSz - 1))) {
			return;
		}
	}


	if (nPixelFormat == 3) {
		pbitsTC = (DWORD *)(pSurface + V2YPoint * lPitch * 2 + V2XPoint * 4);

		for (y = 0; y < V2YSize; y ++) {
			for (x = 0; x < V2XSize; x ++) {
				if (nV2data[y * V2XSize + x]) {
					col = V2rgbTableTc[nV2data[y * V2XSize + x] + 0];
					for (yy = y * V2YPxSz; yy < (y + 1) * V2YPxSz; yy ++) {
						for (xx = x * V2XPxSz; xx < (x + 1) * V2XPxSz; xx ++) {
							pbitsTC[yy * (lPitch / 2) + xx] = col;
							if (!bAnalog && bPseudo400Line) {
								pbitsTC[yy * (lPitch/2) +(lPitch/4)+xx] = col;
							}
						}
					}
				}
			}
		}
	}
	else {
		pbits = (WORD *)(pSurface + V2YPoint * lPitch * 2 + V2XPoint * 2);

		for (y = 0; y < V2YSize; y ++) {
			for (x = 0; x < V2XSize; x ++) {
				if (nV2data[y * V2XSize + x]) {
					if (nPixelFormat == 1) {
						col = V2rgbTable[nV2data[y * V2XSize + x] + 0];
					}
					else {
						col = V2rgbTable[nV2data[y * V2XSize + x] + 2];
					}
					for (yy = y * V2YPxSz; yy < (y + 1) * V2YPxSz; yy ++) {
						for (xx = x * V2XPxSz; xx < (x + 1) * V2XPxSz; xx ++) {
							pbits[yy * lPitch + xx] = col;
							if (!bAnalog && bPseudo400Line) {
								pbits[yy * lPitch + (lPitch / 2) + xx] = col;
							}
						}
					}
				}
			}
		}
	}
}
#endif

/*
 *	SĂ̍ĕ`tOݒ
 */
static void FASTCALL SetDrawFlag(BOOL flag)
{
	memset(DDDrawFlag, (BYTE)flag, sizeof(DDDrawFlag));
}

/*
 *	S̈NA
 */
static void FASTCALL AllClear(BOOL clear_flag)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	BYTE *p;
	int i;
	int lines;

	/* tO`FbN */
	if (!bClearFlag) {
		return;
	}

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* I[NA */
	if (clear_flag) {
		p = (BYTE *)ddsd.lpSurface;
		if (nDDResolution == DDRES_480LINE) {
			lines = 480;
		}
		else if (nDDResolution == DDRES_SXGA) {
			lines = 512;
		}
		else if (nDDResolution == DDRES_USER) {
			lines = (ptFullRes[0].y / nMlt);
		}
		else {
			lines = 400;
		}
		if (nDDResolution == DDRES_USER) {
			if (ptFullRes[1].x > 0) {
				for (i=0; i<lines; i++) {
					if (nPixelFormat == 3) {
						memset(p, 0, (ptFullRes[0].x / nMlt) * 4);
					}
#if XM7_VER >=3
					else if (nPixelFormat == 4) {
						memset(p, 0, (ptFullRes[0].x / nMlt) * 3);
					}
#endif
					else {
						memset(p, 0, (ptFullRes[0].x / nMlt) * 2);
					}
					p += ddsd.lPitch;
				}
				lines = 0;
			}
		}
		for (i=0; i<lines; i++) {
			if (nPixelFormat == 3) {
				memset(p, 0, 640 * 4);
			}
#if XM7_VER >= 3
			else if (nPixelFormat == 4) {
				memset(p, 0, 640 * 3);
			}
#endif
			else {
				memset(p, 0, 640 * 2);
			}
			p += ddsd.lPitch;
		}
	}

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);

	/* [NZbg */
	bClearFlag = FALSE;
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	SetDrawFlag(TRUE);
#if XM7_VER >= 3
	bWindowOpen = FALSE;
	nWindowDx1 = 640;
	nWindowDy1 = 400;
	nWindowDx2 = 0;
	nWindowDy2 = 0;
#endif
	SetDirtyFlag(0, 400, TRUE);

	/* Xe[^XCNA */
	szCaption[0] = '\0';
	nCAP = -1;
	nKANA = -1;
	nINS = -1;
	for (i = 0; i < 4; i ++) {
 #if XM7_VER == 1 && defined(SFDC)
		nDrive[FDC_M][i] = -1;
		nDrive[FDC_S][i] = -1;
		szDrive[FDC_M][i][0] = '\0';
		szDrive[FDC_S][i][0] = '\0';
 #else	/* XM7_VER == 1 && defined(SFDC) */
		nDrive[i] = -1;
		szDrive[i][0] = '\0';
 #endif	/* XM7_VER == 1 && defined(SFDC) */
	}
	nTape = -1;

	/* 640x480̂Blt */
	if ((nDDResolution == DDRES_400LINE) ||
		(nDDResolution == DDRES_WUXGA) ||
		(nDDResolution == DDRES_WXGA800)) {
		return;
	}

	if (nDDResolution == DDRES_USER) {
		if (ptFullRes[1].x > 0) {
			/* ] */
			rect.left = 0;
			rect.right = ptFullRes[1].x / nMlt;
			rect.top = 0;
			rect.bottom = ptFullRes[0].y / nMlt;
			drect.left = rect.left * nMlt;
			drect.right = rect.right * nMlt;
			drect.top = rect.top * nMlt;
			drect.bottom = rect.bottom * nMlt;
			hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
			if (hResult != DD_OK) {
				/* T[tFCXXgĂ΁AXgA */
				if (hResult == DDERR_SURFACELOST) {
					lpdds[0]->Restore();
					lpdds[1]->Restore();
				}
				/* ͑S̈XV */
				ReDrawDD();
				return;
			}

			/* E] */
			rect.left = (ptFullRes[0].x - ptFullRes[1].x) / nMlt;
			rect.right = ptFullRes[0].x / nMlt;
			rect.top = 0;
			rect.bottom = ptFullRes[0].y / nMlt;
			drect.left = rect.left * nMlt;
			drect.right = rect.right * nMlt;
			drect.top = rect.top * nMlt;
			drect.bottom = rect.bottom * nMlt;
			hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
			if (hResult != DD_OK) {
				/* T[tFCXXgĂ΁AXgA */
				if (hResult == DDERR_SURFACELOST) {
					lpdds[0]->Restore();
					lpdds[1]->Restore();
				}
				/* ͑S̈XV */
				ReDrawDD();
				return;
			}
		}

		if (ptFullRes[1].y > 0) {
			/* 㕔` */
			rect.left = ptFullRes[1].x / nMlt;
			rect.right = (ptFullRes[0].x - ptFullRes[1].x) / nMlt;
			rect.top = 0;
			rect.bottom = ptFullRes[1].y / nMlt;
			if (bCaption) {
				rect.bottom += 40;
			}
			drect.left = rect.left * nMlt;
			drect.right = rect.right * nMlt;
			drect.top = rect.top * nMlt;
			drect.bottom = rect.bottom * nMlt;
			hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
			if (hResult != DD_OK) {
				/* T[tFCXXgĂ΁AXgA */
				if (hResult == DDERR_SURFACELOST) {
					lpdds[0]->Restore();
					lpdds[1]->Restore();
				}
				/* ͑S̈XV */
				ReDrawDD();
				return;
			}

			/* ` */
			rect.left = ptFullRes[1].x / nMlt;
			rect.right = (ptFullRes[0].x - ptFullRes[1].x) / nMlt;
			rect.top = ptFullRes[1].y / nMlt + 400;
			rect.bottom = ptFullRes[0].y / nMlt;
			if (bCaption) {
				rect.top += 80;
			}
			drect.left = rect.left * nMlt;
			drect.right = rect.right * nMlt;
			drect.top = ptFullRes[1].y + 400 * nMlt;
			drect.bottom = ptFullRes[0].y;
			if (bCaption) {
				drect.top += 40 * nMlt;
			}
			hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
			if (hResult != DD_OK) {
				/* T[tFCXXgĂ΁AXgA */
				if (hResult == DDERR_SURFACELOST) {
					lpdds[0]->Restore();
					lpdds[1]->Restore();
				}
				/* ͑S̈XV */
				ReDrawDD();
				return;
			}
		}
		return;
	}

	/* ΁ABlt */
	if (nDDResolution == DDRES_480LINE) {
		rect.top = 0;
		rect.left = 0;
		rect.right = 640;
		rect.bottom = 40;
		drect = rect;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top = 0;
		rect.left = 0;
		rect.right = 640;
		rect.bottom = 56;
		drect.top = 0;
		drect.left = 0;
		drect.right = 1280;
		drect.bottom = 112;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* CWP[^ĕ` */
	StatusLine(TRUE);
}

/*
 *	XLC`
 */
static void FASTCALL RenderFullScan(BYTE *pSurface, LONG lPitch)
{
	WORD u;
	WORD top, bottom;

	/* X^Pʃ_OɉʑŜKv̂ */
	if (nDrawTop >= nDrawBottom) {
		top = 0;
		bottom = 400;
	}
	else {
		top = nDrawTop;
		bottom = nDrawBottom;
	}

	/* ݒ */
	pSurface += top * lPitch;

	/* tO`FbN */
	if (bFullScan) {
		/* [v */
		for (u=top; u<bottom; u += (WORD)2) {
			if (nPixelFormat == 3) {
				memcpy(&pSurface[lPitch], pSurface, 640 * 4);
			}
#if XM7_VER >= 3
			else if (nPixelFormat == 4) {
				memcpy(&pSurface[lPitch], pSurface, 640 * 3);
			}
#endif
			else {
				memcpy(&pSurface[lPitch], pSurface, 640 * 2);
			}
			pSurface += (lPitch * 2);
		}
	}
	else {
		/* X^Pʃ_O͂ŃNAȂƂȂ */
		if (bRasterRendering) {
			/* [v */
			for (u=top; u<bottom; u += (WORD)2) {
				if (nPixelFormat == 3) {
					memset(&pSurface[lPitch], 0x00, 640 * 4);
				}
#if XM7_VER >= 3
				else if (nPixelFormat == 4) {
					memset(&pSurface[lPitch], 0x00, 640 * 3);
				}
#endif
				else {
					memset(&pSurface[lPitch], 0x00, 640 * 2);
				}
				pSurface += (lPitch * 2);
			}
		}
	}
}

/*
 *	Xe[^XC(LvV)`
 */
static void FASTCALL DrawCaption(void)
{
	HDC hDC;
	HRESULT hResult;
	RECT rect;
	RECT drect;

	ASSERT((nDDResolution == DDRES_480LINE) || (nDDResolution == DDRES_SXGA) || (nDDResolution == DDRES_USER));
	ASSERT(bCaption);

	/* DC擾 */
	hResult = lpdds[1]->GetDC(&hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* ExtTextOutgAxŕ` */
	SetBkColor(hDC, RGB(0, 0, 0));
	SetTextColor(hDC, RGB(255, 255, 255));
	if (nDDResolution == DDRES_480LINE) {
		rect.top = 0;
		rect.left = 0;
		rect.right = 640;
		rect.bottom = 40;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top = 0;
		rect.left = 0;
		rect.right = 640;
		rect.bottom = 56;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.left = ptFullRes[1].x / nMlt;
		rect.right = ptFullRes[1].x / nMlt + 640;
		rect.top = ptFullRes[1].y / nMlt;
		rect.bottom = ptFullRes[1].y / nMlt + 40;
	}
	if (nDDResolution == DDRES_USER) {
		ExtTextOut(hDC, rect.left, rect.top, ETO_OPAQUE, &rect,
						 szCaption, strlen(szCaption), NULL);
	}
	else {
		ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &rect,
						 szCaption, strlen(szCaption), NULL);
	}

	/* DC */
	hResult = lpdds[1]->ReleaseDC(hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* Blt */
	if (nDDResolution == DDRES_480LINE) {
		drect.top = 0;
		drect.left = 0;
		drect.right = 640;
		drect.bottom = 40;
	}
	else if (nDDResolution == DDRES_SXGA) {
		drect.top = 0;
		drect.left = 0;
		drect.right = 1280;
		drect.bottom = 112;
	}
	else if (nDDResolution == DDRES_USER) {
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}
}

/*
 *	Xe[^XC(LvV)
 */
static BOOL FASTCALL StatusCaption(void)
{
	char string[256];
	char tmp[256];
	char drive[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];

	/* 󋵂ɉāARs[ */
	if (run_flag) {
		strncpy(string, szRunMessage, sizeof(string));
	}
	else {
		strncpy(string, szStopMessage, sizeof(string));
	}
	strncat(string, " ", sizeof(string) - strlen(string) - 1);

	/* CPUx䗦 */
	if (bAutoSpeedAdjust) {
		_snprintf(tmp, sizeof(tmp), "(%3d%%) ", speed_ratio / 100);
		strncat(string, tmp, sizeof(string) - strlen(string) - 1);
	}

	/* Xe[g */
	if ((get_file_size(StatePath) > 0) &&
		(strncmp(StatePath, QuickPath, _MAX_PATH))) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(StatePath, NULL, NULL, fname, ext);
		_snprintf(tmp, sizeof(tmp), "- [%s%s] ", fname, ext);
		strncat(string, tmp, sizeof(string) - strlen(string) - 1);
	}

	/* LvV\ */
	if (!bMountFileName) {
		/* r */
		string[sizeof(string) - 1] = '\0';
		if (strncmp(szCaption, string, strlen(string) + 1) != 0) {
			memset(szCaption, 0, sizeof(szCaption));
			strncpy(szCaption, string, sizeof(szCaption) - 1);
			return TRUE;
		}
		return FALSE;
	}

#if XM7_VER == 1 && defined(SFDC)
	if (fdc_enable[FDC_S] && (model_type != MODEL_FM7)) {
		/* 1MBtbs[fBXNhCu 0 */
		if (fdc_ready[FDC_S][0] != FDC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(fdc_fname[FDC_S][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* 1MBtbs[fBXNhCu 1 */
		if (fdc_ready[FDC_S][1] != FDC_TYPE_NOTREADY) {
			if ((strcmp(fdc_fname[FDC_S][0], fdc_fname[FDC_S][1]) != 0) ||
				(fdc_ready[FDC_S][0] == FDC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(fdc_fname[FDC_S][1], drive, dir, fname, ext);
				if (fdc_ready[FDC_S][0] == FDC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}

	if (fdc_enable[FDC_M] || (model_type == MODEL_FM77)) {
		/* tbs[fBXNhCu 0 */
		if (fdc_ready[FDC_M][0] != FDC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(fdc_fname[FDC_M][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* tbs[fBXNhCu 1 */
		if (fdc_ready[FDC_M][1] != FDC_TYPE_NOTREADY) {
			if ((strcmp(fdc_fname[FDC_M][0], fdc_fname[FDC_M][1]) != 0) ||
				(fdc_ready[FDC_M][0] == FDC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(fdc_fname[FDC_M][1], drive, dir, fname, ext);
				if (fdc_ready[FDC_M][0] == FDC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}
#else
 #if XM7_VER >= 2
	if (fdc_enable || (model_type != MODEL_FM7)) {
 #else
	if (fdc_enable || (model_type == MODEL_FM77)) {
 #endif
		/* tbs[fBXNhCu 0 */
		if (fdc_ready[0] != FDC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(fdc_fname[0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* tbs[fBXNhCu 1 */
		if (fdc_ready[1] != FDC_TYPE_NOTREADY) {
			if ((strcmp(fdc_fname[0], fdc_fname[1]) != 0) ||
				(fdc_ready[0] == FDC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(fdc_fname[1], drive, dir, fname, ext);
				if (fdc_ready[0] == FDC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* e[v */
	if (tape_fileh != -1) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(tape_fname, drive, dir, fname, ext);
		_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
		strncat(string, tmp, sizeof(string) - strlen(string) - 1);
	}

#if XM7_VER == 1 && defined(BUBBLE)
	/* ou */
	if (bmc_enable[BMC_32] &&
		((model_type == MODEL_FM8) ||
		 (!opn_enable && (mos_port != 3) && (bmc_secret_flag || bmc_use[BMC_32])))) {
		/* 32KBouz_ 0 */
		if (bmc_ready[0][0] != BMC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(bmc_fname[0][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* 32KBouz_ 1 */
		if (bmc_ready[0][1] != BMC_TYPE_NOTREADY) {
			if ((strcmp(bmc_fname[0][0], bmc_fname[0][1]) != 0) ||
				(bmc_ready[0][0] == BMC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(bmc_fname[0][1], drive, dir, fname, ext);
				if (bmc_ready[0][0] == BMC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}

	if (bmc_enable[BMC_128]) {
		/* 128KBouz_ 0 */
		if (bmc_ready[1][0] != BMC_TYPE_NOTREADY) {
			/* t@Cl[{gq̂ݎo */
			_splitpath(bmc_fname[1][0], drive, dir, fname, ext);
			_snprintf(tmp, sizeof(tmp), "- %s%s ", fname, ext);
			strncat(string, tmp, sizeof(string) - strlen(string) - 1);
		}

		/* 128KBouz_ 1 */
		if (bmc_ready[1][1] != BMC_TYPE_NOTREADY) {
			if ((strcmp(bmc_fname[1][0], bmc_fname[1][1]) != 0) ||
				(bmc_ready[1][0] == BMC_TYPE_NOTREADY)) {
				/* t@Cl[{gq̂ݎo */
				_splitpath(bmc_fname[1][1], drive, dir, fname, ext);
				if (bmc_ready[1][0] == BMC_TYPE_NOTREADY) {
					_snprintf(tmp, sizeof(tmp), "- (%s%s) ", fname, ext);
				}
				else {
					_snprintf(tmp, sizeof(tmp), "(%s%s) ", fname, ext);
				}
				strncat(string, tmp, sizeof(string) - strlen(string) - 1);
			}
		}
	}
#endif

	/* r` */
	string[255] = '\0';
	if (memcmp(szCaption, string, strlen(string) + 1) != 0) {
		strncpy(szCaption, string, sizeof(szCaption));
		return TRUE;
	}

	/* OƓȂ̂ŁA`悵ȂĂ悢 */
	return FALSE;
}

/*
 *	Xe[^XC(CAP)
 */
static BOOL FASTCALL StatusCAP(void)
{
	HDC hDC;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	int num;

	/* l擾Ar */
	if (caps_flag) {
		num = 1;
	}
	else {
		num = 0;
	}
	if (num == nCAP) {
		return FALSE;
	}

	/* DC擾 */
	hResult = lpdds[1]->GetDC(&hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	/* -1ȂAS̈NA */
	if (nDDResolution == DDRES_480LINE) {
		rect.left = 0;
		rect.right = 640;
		rect.top = 440;
		rect.bottom = 480;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.left = ptFullRes[1].x / nMlt;
		rect.right = ptFullRes[1].x / nMlt + 640;
		rect.top = ptFullRes[1].y / nMlt + 440;
		rect.bottom = ptFullRes[1].y / nMlt + 480;
	}
	else {
		rect.left = 0;
		rect.right = 640;
		rect.top = 472;
		rect.bottom = 512;
	}
	drect = rect;
	SetBkColor(hDC, RGB(0, 0, 0));
	SetTextColor(hDC, RGB(255, 255, 255));
	if (nCAP == -1) {
		/* NA */
		ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &rect, NULL, 0, NULL);

		if (nDDResolution == DDRES_USER) {
			/* "CAP"̕` */
			TextOut(hDC, 500 + ptFullRes[1].x / nMlt, drect.top + 4, szCAPMessage, strlen(szCAPMessage));

			/* N` */
			rect.left = 500 + ptFullRes[1].x / nMlt;
		}
		else
		{
			/* "CAP"̕` */
			TextOut(hDC, 500, drect.top + 4, szCAPMessage, strlen(szCAPMessage));

			/* N` */
			rect.left = 500;
		}
		rect.right = rect.left + 30;
		rect.top = drect.top + 25;
		rect.bottom = rect.top + 10;
		FrameRect(hDC, &rect, (HBRUSH)GetStockObject(WHITE_BRUSH));
	}

	/* ŃRs[ */
	nCAP = num;

	/* CF` */
	if (nDDResolution == DDRES_USER) {
		rect.left = 501 + ptFullRes[1].x / nMlt;
	}
	else
	{
		rect.left = 501;
	}
	rect.right = rect.left + 28;
	rect.top = drect.top + 26;
	rect.bottom = rect.top + 8;
	if (nCAP == 1) {
		SetBkColor(hDC, RGB(255, 0, 0));
	}
	else {
		SetBkColor(hDC, RGB(0, 0, 0));
	}
	ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &rect, NULL, 0, NULL);

	/* DC */
	hResult = lpdds[1]->ReleaseDC(hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	return TRUE;
}

/*
 *	Xe[^XC()
 */
static BOOL FASTCALL StatusKANA(void)
{
	HDC hDC;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	int num;

	/* l擾Ar */
	if (kana_flag) {
		num = 1;
	}
	else {
		num = 0;
	}
	if (num == nKANA) {
		return FALSE;
	}

	/* DC擾 */
	hResult = lpdds[1]->GetDC(&hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	if (nDDResolution == DDRES_480LINE) {
		rect.left = 0;
		rect.right = 640;
		rect.top = 440;
		rect.bottom = 480;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.left = ptFullRes[1].x / nMlt;
		rect.right = ptFullRes[1].x / nMlt + 640;
		rect.top = ptFullRes[1].y / nMlt + 440;
		rect.bottom = ptFullRes[1].y / nMlt + 480;
	}
	else {
		rect.left = 0;
		rect.right = 640;
		rect.top = 472;
		rect.bottom = 512;
	}
	drect = rect;
	SetBkColor(hDC, RGB(0, 0, 0));
	SetTextColor(hDC, RGB(255, 255, 255));
	if (nKANA == -1) {
		if (nDDResolution == DDRES_USER) {
			/* ""̕` */
			TextOut(hDC, 546 + ptFullRes[1].x / nMlt, drect.top + 4, szKANAMessage, strlen(szKANAMessage));

			/* N` */
			rect.left = 546 + ptFullRes[1].x / nMlt;
		}
		else
		{
			/* ""̕` */
			TextOut(hDC, 546, drect.top + 4, szKANAMessage, strlen(szKANAMessage));

			/* N` */
			rect.left = 546;
		}
		rect.right = rect.left + 30;
		rect.top = drect.top + 25;
		rect.bottom = rect.top + 10;
		FrameRect(hDC, &rect, (HBRUSH)GetStockObject(WHITE_BRUSH));
	}

	/* ŃRs[ */
	nKANA = num;

	/* CF` */
	if (nDDResolution == DDRES_USER) {
		rect.left = 547 + ptFullRes[1].x / nMlt;
	}
	else
	{
		rect.left = 547;
	}
	rect.right = rect.left + 28;
	rect.top = drect.top + 26;
	rect.bottom = rect.top + 8;
	if (nKANA == 1) {
		SetBkColor(hDC, RGB(255, 0, 0));
	}
	else {
		SetBkColor(hDC, RGB(0, 0, 0));
	}
	ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &rect, NULL, 0, NULL);

	/* DC */
	hResult = lpdds[1]->ReleaseDC(hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	return TRUE;
}

/*
 *	Xe[^XC(INS)
 */
static BOOL FASTCALL StatusINS(void)
{
	HDC hDC;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	int num;

	/* l擾Ar */
	if (ins_flag) {
		num = 1;
	}
	else {
		num = 0;
	}
	if (num == nINS) {
		return FALSE;
	}

	/* DC擾 */
	hResult = lpdds[1]->GetDC(&hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	if (nDDResolution == DDRES_480LINE) {
		rect.left = 0;
		rect.right = 640;
		rect.top = 440;
		rect.bottom = 480;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.left = ptFullRes[1].x / nMlt;
		rect.right = ptFullRes[1].x / nMlt + 640;
		rect.top = ptFullRes[1].y / nMlt + 440;
		rect.bottom = ptFullRes[1].y / nMlt + 480;
	}
	else {
		rect.left = 0;
		rect.right = 640;
		rect.top = 472;
		rect.bottom = 512;
	}
	drect = rect;
	SetBkColor(hDC, RGB(0, 0, 0));
	SetTextColor(hDC, RGB(255, 255, 255));
	if (nINS == -1) {
		if (nDDResolution == DDRES_USER) {
			/* "INS"̕` */
			TextOut(hDC, 595 + ptFullRes[1].x / nMlt, drect.top + 4, szINSMessage, strlen(szINSMessage));

			/* N` */
			rect.left = 593 + ptFullRes[1].x / nMlt;
		}
		else
		{
			/* "INS"̕` */
			TextOut(hDC, 595, drect.top + 4, szINSMessage, strlen(szINSMessage));

			/* N` */
			rect.left = 593;
		}
		rect.right = rect.left + 30;
		rect.top = drect.top + 25;
		rect.bottom = rect.top + 10;
		FrameRect(hDC, &rect, (HBRUSH)GetStockObject(WHITE_BRUSH));
	}

	/* ŃRs[ */
	nINS = num;

	/* CF` */
	if (nDDResolution == DDRES_USER) {
		rect.left = 594 + ptFullRes[1].x / nMlt;
	}
	else
	{
		rect.left = 594;
	}
	rect.right = rect.left + 28;
	rect.top = drect.top + 26;
	rect.bottom = rect.top + 8;
	if (nINS == 1) {
		SetBkColor(hDC, RGB(255, 0, 0));
	}
	else {
		SetBkColor(hDC, RGB(0, 0, 0));
	}
	ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &rect, NULL, 0, NULL);

	/* DC */
	hResult = lpdds[1]->ReleaseDC(hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	return TRUE;
}

/*
 *	Xe[^XC(tbs[hCu)
 */
#if XM7_VER == 1 && defined(SFDC)
static BOOL FASTCALL StatusDrive(int type, int drive)
#else	/* XM7_VER == 1 && defined(SFDC) */
static BOOL FASTCALL StatusDrive(int drive)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	char string[_MAX_FNAME + _MAX_EXT];
	char buffer[_MAX_FNAME + _MAX_EXT];
	char drive_[_MAX_DRIVE];
	char dir[_MAX_DIR];
	char fname[_MAX_FNAME];
	char ext[_MAX_EXT];
	HDC hDC;
	HRESULT hResult;
	RECT rect, srect;
	int num;
	char *name;
	BOOL exdrv;
	UINT dcw;

#if XM7_VER == 1 && defined(SFDC)
	ASSERT((type == FDC_M) || (type == FDC_S));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	ASSERT((drive >= 0) && (drive <= 3));

#if XM7_VER == 1 && defined(SFDC)
	/* L`FbN */
	if ((type == FDC_M) && (!fdc_enable[FDC_M] && (model_type != MODEL_FM77))) {
		return TRUE;
	}
	if ((type == FDC_S) && (!fdc_enable[FDC_S] || (model_type == MODEL_FM7))) {
		return TRUE;
	}
	exdrv = (((	fdc_ready[FDC_M][2] != FDC_TYPE_NOTREADY ||
				fdc_ready[FDC_M][3] != FDC_TYPE_NOTREADY) &&
				(fdc_enable[FDC_M] || (model_type == MODEL_FM77))) ||
			 ((	fdc_ready[FDC_S][2] != FDC_TYPE_NOTREADY ||
				fdc_ready[FDC_S][3] != FDC_TYPE_NOTREADY) &&
				(fdc_enable[FDC_S] && (model_type != MODEL_FM7))));
#else	/* XM7_VER == 1 && defined(SFDC) */
	/* L`FbN */
#if XM7_VER >= 2
	if ((!fdc_enable && (model_type == MODEL_FM7))) {
		return TRUE;
	}
#else
	if ((!fdc_enable && (model_type != MODEL_FM77))) {
		return TRUE;
	}
#endif	/* XM7_VER == 1 */
	exdrv = (	fdc_ready[2] != FDC_TYPE_NOTREADY ||
				fdc_ready[3] != FDC_TYPE_NOTREADY);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	/* ݃hCu(#2/#3)}Eg\ */
	if ((drive >= 2) && !exdrv) {
		return TRUE;
	}

	/* ԍZbg */
#if XM7_VER == 1 && defined(SFDC)
	if (fdc_ready[type][drive] == FDC_TYPE_NOTREADY) {
		num = FDC_ACCESS_NOTREADY;
	}
	else if (fdc_teject[type][drive]) {
		num = FDC_ACCESS_TEJECT;
	}
	else {
		num = fdc_access[type][drive];
		if (num == FDC_ACCESS_SEEK) {
			num = FDC_ACCESS_READY;
		}
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (fdc_ready[drive] == FDC_TYPE_NOTREADY) {
		num = FDC_ACCESS_NOTREADY;
	}
	else if (fdc_teject[drive]) {
		num = FDC_ACCESS_TEJECT;
	}
	else {
		num = fdc_access[drive];
		if (num == FDC_ACCESS_SEEK) {
			num = FDC_ACCESS_READY;
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* O擾 */
	name = "";
#if XM7_VER == 1 && defined(SFDC)
	if ((fdc_ready[type][drive] == FDC_TYPE_D77) &&
		(strlen(fdc_name[type][drive][ fdc_media[type][drive] ]) > 0)) {
		name = fdc_name[type][drive][ fdc_media[type][drive] ];
	}
	else if (fdc_ready[type][drive] != FDC_TYPE_NOTREADY) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(fdc_fname[type][drive], NULL, NULL, fname, ext);
		strncpy(string, fname, sizeof(string));
		strncat(string, ext, sizeof(string) - strlen(string) - 1);
		if ((fdc_ready[type][drive] == FDC_TYPE_D77) && (fdc_medias[type][drive] > 1)) {
			_snprintf(buffer, sizeof(buffer), "%s #%02d", string, fdc_media[type][drive] + 1);
			name = buffer;
		}
		else {
			name = string;
		}
	}
#else
	if ((fdc_ready[drive] == FDC_TYPE_D77) &&
		(strlen(fdc_name[drive][ fdc_media[drive] ]) > 0)) {
		name = fdc_name[drive][ fdc_media[drive] ];
	}
	else if (fdc_ready[drive] != FDC_TYPE_NOTREADY) {
		/* t@Cl[{gq̂ݎo */
		_splitpath(fdc_fname[drive], drive_, dir, fname, ext);
		strncpy(string, fname, sizeof(string));
		strncat(string, ext, sizeof(string) - strlen(string) - 1);
		if ((fdc_ready[drive] == FDC_TYPE_D77) && (fdc_medias[drive] > 1)) {
			_snprintf(buffer, sizeof(buffer), "%s #%02d", string,
					  fdc_media[drive] + 1);
			name = buffer;
		}
		else {
			name = string;
		}
	}
#endif

	/* 炩߃fBA\\߂Ă */
#if XM7_VER == 1 && defined(SFDC)
	if ((fdc_enable[FDC_S] && (model_type != MODEL_FM7)) &&
		(fdc_enable[FDC_M] || (model_type == MODEL_FM77))) {
		if (exdrv) {
			dcw = 6;
		}
		else {
			dcw = 12;
		}
	}
	else if (fdc_enable[FDC_S] && (model_type == MODEL_FM8)) {
		if (exdrv) {
			dcw = 12;
		}
		else {
			dcw = 16;
		}
	}
	else if (fdc_enable[FDC_M] || (model_type == MODEL_FM77)) {
		if (exdrv) {
			dcw = 12;
		}
		else {
			dcw = 16;
		}
	}
	else {
		dcw = 16;
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
 #if XM7_VER >= 2
	if (fdc_enable || (model_type != MODEL_FM7)) {
 #else
	if (fdc_enable || (model_type == MODEL_FM77)) {
 #endif
		if (exdrv) {
			dcw = 12;
		}
		else {
			dcw = 16;
		}
	}
	else {
		dcw = 16;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* fBAI[o[͊gqŕ\ */
	if (strlen(name) > dcw) {
		if (dcw < 16) {
			/* gq̂ݎo */
#if XM7_VER == 1 && defined(SFDC)
			_splitpath(fdc_fname[type][drive], NULL, NULL, NULL, ext);
#else
			_splitpath(fdc_fname[drive], NULL, NULL, NULL, ext);
#endif
			strncpy(string, ext, sizeof(string));
			for (name = string; *name != '\0'; name ++) {
				*name = toupper(*name);
			}
			if (strlen(ext) > 1) {
				name = string + 1;
			}
			else {
				name = string;
			}
			if (dcw >= 7) {
				/* fBAԍݒ */
#if XM7_VER == 1 && defined(SFDC)
				if ((fdc_ready[type][drive] == FDC_TYPE_D77) && (fdc_medias[type][drive] > 1)) {
					_snprintf(buffer, sizeof(buffer), "%s #%02d", name, fdc_media[type][drive] + 1);
					name = buffer;
				}
#else
				if ((fdc_ready[drive] == FDC_TYPE_D77) && (fdc_medias[drive] > 1)) {
					_snprintf(buffer, sizeof(buffer), "%s #%02d", name, fdc_media[drive] + 1);
					name = buffer;
				}
#endif
				else {
					_snprintf(buffer, sizeof(buffer), "%s File", name);
					name = buffer;
				}
			}
		}
	}

	/* ԍr */
#if XM7_VER == 1 && defined(SFDC)
	if (nDrive[type][drive] == num) {
		if (strcmp(szDrive[type][drive], name) == 0) {
			return FALSE;
		}
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (nDrive[drive] == num) {
		if (strcmp(szDrive[drive], name) == 0) {
			return FALSE;
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* DC擾 */
	hResult = lpdds[1]->GetDC(&hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	/* ŃRs[ */
#if XM7_VER == 1 && defined(SFDC)
	nDrive[type][drive] = num;
	strncpy(szDrive[type][drive], name, sizeof(szDrive[type][drive]));
#else	/* XM7_VER == 1 && defined(SFDC) */
	nDrive[drive] = num;
	strncpy(szDrive[drive], name, sizeof(szDrive[drive]));
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* Wݒ */
	SetBkColor(hDC, RGB(0, 0, 0));
#if XM7_VER == 1 && defined(SFDC)
	if ((fdc_enable[FDC_S] && (model_type != MODEL_FM7)) &&
		(fdc_enable[FDC_M] || (model_type == MODEL_FM77))) {
		if (exdrv) {
			if (type == FDC_S) {
				rect.left = (3 - drive) * 55;
				rect.right = ((3 - drive) + 1) * 55 - 4;
			}
			else {
				rect.left = (3 - drive) * 55 + 225;
				rect.right = ((3 - drive) + 1) * 55 + 225 - 4;
			}
		}
		else {
			if (type == FDC_S) {
				rect.left = (drive ^ 1) * 110;
				rect.right = ((drive ^ 1) + 1) * 110 - 4;
			}
			else {
				rect.left = (drive ^ 1) * 110 + 225;
				rect.right = ((drive ^ 1) + 1) * 110 + 225 - 4;
			}
		}
	}
	else if (fdc_enable[FDC_S] && (model_type == MODEL_FM8)) {
		if (exdrv) {
			rect.left = (3 - drive) * 110;
			rect.right = ((3 - drive) + 1) * 110 - 4;
		}
		else {
			rect.left = (drive ^ 1) * 160;
			rect.right = ((drive ^ 1) + 1) * 160 - 4;
		}
	}
	else if (fdc_enable[FDC_M] || (model_type == MODEL_FM77)) {
		if (exdrv) {
			rect.left = (3 - drive) * 110;
			rect.right = ((3 - drive) + 1) * 110 - 4;
		}
		else {
			rect.left = (drive ^ 1) * 160;
			rect.right = ((drive ^ 1) + 1) * 160 - 4;
		}
	}
	else {
		rect.left = (drive ^ 1) * 160;
		rect.right = ((drive ^ 1) + 1) * 160 - 4;
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
  #if XM7_VER >= 2
	if (fdc_enable || (model_type != MODEL_FM7)) {
  #else
	if (fdc_enable || (model_type == MODEL_FM77)) {
  #endif
		if (exdrv) {
			rect.left = (3 - drive) * 110;
			rect.right = ((3 - drive) + 1) * 110 - 4;
		}
		else {
			rect.left = (drive ^ 1) * 160;
			rect.right = ((drive ^ 1) + 1) * 160 - 4;
		}
	}
	else {
		rect.left = (drive ^ 1) * 160;
		rect.right = ((drive ^ 1) + 1) * 160 - 4;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if (nDDResolution == DDRES_480LINE) {
		rect.top = 444;
		rect.bottom = 474;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		rect.top = ptFullRes[1].y / nMlt + 444;
		rect.bottom = ptFullRes[1].y / nMlt + 474;
	}
	else {
		rect.top = 476;
		rect.bottom = 506;
	}
	srect.left = rect.left + 1;
	srect.right = rect.right + 1;
	srect.top = rect.top + 1;
	srect.bottom = rect.bottom + 1;

	/* F */
	if ((num != FDC_ACCESS_NOTREADY) && (num != FDC_ACCESS_TEJECT)) {
		SetBkColor(hDC, RGB(63, 63, 63));
	}
	if (num == FDC_ACCESS_READ) {
		SetBkColor(hDC, RGB(191, 0, 0));
	}
	if (num == FDC_ACCESS_WRITE) {
		SetBkColor(hDC, RGB(0, 0, 191));
	}

	/* wihԂ */
	ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &rect, NULL, 0, NULL);

	/* DrawText */
	SetTextColor(hDC, RGB(0, 0, 0));
	SetBkMode(hDC, TRANSPARENT);
	DrawText(hDC, name, strlen(name), &srect,
		DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);
	SetTextColor(hDC, RGB(255, 255, 255));
	SetBkMode(hDC, TRANSPARENT);
	DrawText(hDC, name, strlen(name), &rect,
		DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);

	/* hCuԍ` */
	if (num != FDC_ACCESS_NOTREADY) {
		SetDIBitsToDevice(hDC, rect.right - 8, rect.bottom - 8, 8, 8, 0, 0, 0, 8,
							&bmpDrive[drive][sizeof(BITMAPINFOHEADER)],
							(BITMAPINFO*)&bmpDrive[drive][0],
							DIB_RGB_COLORS);
#if XM7_VER == 1 && defined(SFDC)
		if (type == FDC_S) {
#if 1
			SetDIBitsToDevice(hDC, rect.left, rect.top, 15, 6, 0, 0, 0, 6,
								&bmp2HD[sizeof(BITMAPINFOHEADER)],
								(BITMAPINFO*)&bmp2HD[0],
								DIB_RGB_COLORS);
#else
			const int nums[] = {2, 4, 5};
			for (num = 0; num <= 2; num ++) {
				SetDIBitsToDevice(hDC, rect.left + num * 8, rect.top, 8, 8, 0, 0, 0, 8,
									&bmpDrive[nums[num]][sizeof(BITMAPINFOHEADER)],
									(BITMAPINFO*)&bmpDrive[nums[num]][0],
									DIB_RGB_COLORS);
			}
#endif
		}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	}

	/* DC */
	hResult = lpdds[1]->ReleaseDC(hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	return TRUE;
}

/*
 *	Xe[^XC(e[v)
 */
static BOOL FASTCALL StatusTape(void)
{
	HDC hDC;
	HRESULT hResult;
	RECT rect, srect;
	int num;
	char string[64];

	/* ԍZbg */
	num = 30000;
	if (tape_fileh != -1) {
		num = (int)((tape_offset >> 8) % 10000);
		if (tape_motor) {
			if (tape_rec) {
				num += 20000;
			}
			else {
				num += 10000;
			}
		}
	}

	/* ԍr */
	if (nTape == num) {
		return FALSE;
	}

	/* DC擾 */
	hResult = lpdds[1]->GetDC(&hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	/* ŃRs[ */
	nTape = num;

	/* Wݒ */
	SetBkColor(hDC, RGB(0, 0, 0));
	rect.left = 450;
	rect.right = rect.left + 40;
	if (nDDResolution == DDRES_480LINE) {
		rect.top = 444;
		rect.bottom = 474;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		rect.top = ptFullRes[1].y / nMlt + 444;
		rect.bottom = ptFullRes[1].y / nMlt + 474;
	}
	else {
		rect.top = 476;
		rect.bottom = 506;
	}
	srect.left = rect.left + 1;
	srect.right = rect.right + 1;
	srect.top = rect.top + 1;
	srect.bottom = rect.bottom + 1;

	/* FA񌈒 */
	if (nTape >= 30000) {
		string[0] = '\0';
	}
	else {
		_snprintf(string, sizeof(string), "%04d", nTape % 10000);
		if (nTape >= 10000) {
			if (nTape >= 20000) {
				SetBkColor(hDC, RGB(0, 0, 191));
			}
			else {
				SetBkColor(hDC, RGB(191, 0, 0));
			}
		}
		else {
			SetBkColor(hDC, RGB(63, 63, 63));
		}
	}

	/* wihԂ */
	ExtTextOut(hDC, 0, 0, ETO_OPAQUE, &rect, NULL, 0, NULL);

	/* DrawText */
	SetTextColor(hDC, RGB(0, 0, 0));
	SetBkMode(hDC, TRANSPARENT);
	DrawText(hDC, string, strlen(string), &srect,
		DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);
	SetTextColor(hDC, RGB(255, 255, 255));
	SetBkMode(hDC, TRANSPARENT);
	DrawText(hDC, string, strlen(string), &rect,
		DT_CENTER | DT_VCENTER | DT_SINGLELINE | DT_NOPREFIX);

	/* DC */
	hResult = lpdds[1]->ReleaseDC(hDC);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return FALSE;
	}

	return TRUE;
}

/*
 *	Xe[^XC`
 */
static void FASTCALL StatusLine(BOOL force_redraw)
{
	BOOL flag;
	HRESULT hResult;
	RECT rect;
	RECT drect;

	/* 640x480/1280x1024AXe[^X̏ꍇ̂ */
	if ((nDDResolution == DDRES_400LINE) ||
		(nDDResolution == DDRES_WUXGA) ||
		(nDDResolution == DDRES_WXGA800) ||
		!bCaption) {
		return;
	}

	/* LvV */
	if (StatusCaption()) {
		DrawCaption();
	}

	flag = FALSE;

	/* L[{[hXe[^X */
	if (StatusCAP()) {
		flag = TRUE;
	}
	if (StatusKANA()) {
		flag = TRUE;
	}
	if (StatusINS()) {
		flag = TRUE;
	}
#if XM7_VER == 1 && defined(SFDC)
	if (StatusDrive(FDC_M, 0)) {
		flag = TRUE;
	}
	if (StatusDrive(FDC_M, 1)) {
		flag = TRUE;
	}
	if (StatusDrive(FDC_M, 2)) {
		flag = TRUE;
	}
	if (StatusDrive(FDC_M, 3)) {
		flag = TRUE;
	}
	if (StatusDrive(FDC_S, 0)) {
		flag = TRUE;
	}
	if (StatusDrive(FDC_S, 1)) {
		flag = TRUE;
	}
	if (StatusDrive(FDC_S, 2)) {
		flag = TRUE;
	}
	if (StatusDrive(FDC_S, 3)) {
		flag = TRUE;
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (StatusDrive(0)) {
		flag = TRUE;
	}
	if (StatusDrive(1)) {
		flag = TRUE;
	}
	if (StatusDrive(2)) {
		flag = TRUE;
	}
	if (StatusDrive(3)) {
		flag = TRUE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if (StatusTape()) {
		flag = TRUE;
	}

	/* tO~Ă΁A`悷KvȂ */
	if (!flag && !force_redraw) {
		return;
	}

	/* Blt */
	if (nDDResolution == DDRES_480LINE) {
		rect.left = 0;
		rect.right = 640;
		rect.top = 440;
		rect.bottom = 480;
		drect = rect;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.left = ptFullRes[1].x / nMlt;
		rect.right = ptFullRes[1].x / nMlt + 640;
		rect.top = ptFullRes[1].y / nMlt + 440;
		rect.bottom = ptFullRes[1].y / nMlt + 480;
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
	}
	else {
		rect.left = 0;
		rect.right = 640;
		rect.top = 456;
		rect.bottom = 512;
		drect.left = rect.left * 2;
		drect.right = rect.right * 2;
		drect.top = rect.top * 2;
		drect.bottom = rect.bottom * 2;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}
}

/*
 *	640x200AfW^[h
 *	pbgݒ
 */
static void FASTCALL Palet640(void)
{
	int i;
	int vpage;
	int base;
	BYTE col;

	/* HighColorppbge[u */
	static const WORD rgbTable[] = {
		/* nPixelFormat = 1 */
		0x0000 | 0x0000 | 0x0000,
		0x0000 | 0x0000 | 0x001f,
		0xf800 | 0x0000 | 0x0000,
		0xf800 | 0x0000 | 0x001f,
		0x0000 | 0x07e0 | 0x0000,
		0x0000 | 0x07e0 | 0x001f,
		0xf800 | 0x07e0 | 0x0000,
		0xf800 | 0x07e0 | 0x001f,

		/* nPixelFormat = 2 */
		0x0000 | 0x0000 | 0x0000,
		0x0000 | 0x0000 | 0x001f,
		0x7c00 | 0x0000 | 0x0000,
		0x7c00 | 0x0000 | 0x001f,
		0x0000 | 0x03e0 | 0x0000,
		0x0000 | 0x03e0 | 0x001f,
		0x7c00 | 0x03e0 | 0x0000,
		0x7c00 | 0x03e0 | 0x001f,

		/* nPixelFormat = 1, Green Monitor */
		0x0000 | 0x0000 | 0x0000,
		0x2000 | 0x0200 | 0x0006,
		0x3800 | 0x03c0 | 0x000c,
		0x4000 | 0x0420 | 0x000d,
		0x6000 | 0x0640 | 0x0014,
		0x6800 | 0x06a0 | 0x0014,
		0x7000 | 0x0720 | 0x0016,
		0x7000 | 0x0760 | 0x0017,

		/* nPixelFormat = 2, Green Monitor */
		0x0000 | 0x0000 | 0x0000,
		0x1000 | 0x0100 | 0x0006,
		0x1c00 | 0x01e0 | 0x000c,
		0x2000 | 0x0200 | 0x000d,
		0x3000 | 0x0320 | 0x0014,
		0x3400 | 0x0340 | 0x0014,
		0x3800 | 0x0380 | 0x0016,
		0x3800 | 0x03a0 | 0x0017,

		/* nPixelFormat = 1, Amber Yellow Monitor */
		0x1800 | 0x0000 | 0x0000,
		0x5000 | 0x0100 | 0x0002,
		0x8800 | 0x0200 | 0x0004,
		0x9000 | 0x0240 | 0x0005,
		0xd000 | 0x0380 | 0x0008,
		0xd800 | 0x03a0 | 0x0008,
		0xe800 | 0x03e0 | 0x0009,
		0xe800 | 0x0400 | 0x0009,

		/* nPixelFormat = 2, Amber Yellow Monitor */
		0x0c00 | 0x0000 | 0x0000,
		0x2800 | 0x0080 | 0x0002,
		0x4400 | 0x0100 | 0x0004,
		0x4800 | 0x0120 | 0x0005,
		0x6800 | 0x01c0 | 0x0008,
		0x6c00 | 0x01c0 | 0x0008,
		0x7400 | 0x01e0 | 0x0009,
		0x7400 | 0x0200 | 0x0009,
	};

	/* TrueColorppbge[u */
	static const DWORD rgbTableTc[] = {
		/* nPixelFormat = 3 (TrueColor), Color Monitor */
		0x00000000,	0x000000ff,	0x00ff0000,	0x00ff00ff,
		0x0000ff00,	0x0000ffff,	0x00ffff00,	0x00ffffff,
		/* nPixelFormat = 3 (TrueColor), Green Monitor */
		0x00000000,	0x001f4033,	0x003a785f,	0x0041876a,
		0x0063cca1,	0x0067d5a7,	0x0070e7b6,	0x0073eebb,
		/* nPixelFormat = 3 (TrueColor), Amber Yellow Monitor */
		0x001e0000,	0x00562314,	0x00894327,	0x00944a2b,
		0x00d17141,	0x00d97643,	0x00e87f49,	0x00ee834b,
	};

	/* tOZbgĂȂ΁AȂ */
	if (!bPaletFlag) {
		return;
	}

	/* }`y[WA\v[𓾂 */
	vpage = (~(multi_page >> 4)) & 0x07;

	/* CfbNXx[Xݒ */
	switch (nMonitorType) {
		case MONITOR_COLOR:
			if (nPixelFormat == 1) {
				base = 0;
			}
			else if (nPixelFormat == 2) {
				base = 8;
			}
			else {
				base = 0;
			}
			break;
		case MONITOR_GREEN:
			if (nPixelFormat == 1) {
				base = 16;
			}
			else if (nPixelFormat == 2) {
				base = 24;
			}
			else {
				base = 8;
			}
			break;
		case MONITOR_AMBERYELLOW:
			if (nPixelFormat == 1) {
				base = 32;
			}
			else if (nPixelFormat == 2) {
				base = 40;
			}
			else {
				base = 16;
			}
			break;
		default:
			ASSERT(FALSE);
			return;
	}

	/* 640x200AfW^pbg */
	for (i=0; i<8; i++) {
		if (crt_flag) {
			/* CRT ON */
#if XM7_VER >= 2
			col = (BYTE)(ttl_palet[i & vpage] & 0x07);
			if ((nPixelFormat == 1) ||
				(nPixelFormat == 2)) {
				rgbTTLDD[i] = rgbTable[base + col];
			}
			else {
				rgbTTLDD[i] = rgbTableTc[base + col];
			}
#else
			/* FM-8[hł͖ */
			if ((model_type == MODEL_FM8) && !ttl_palet_en) {
				col = (BYTE)i;
			}
			else {
				col = (BYTE)(ttl_palet[i & vpage] & 0x07);
			}
			if (bPseudo400Line) {
				if ((nPixelFormat == 1) ||
					(nPixelFormat == 2)) {
					if (col & 2) {
						rgbTTLDD[i + 8] = rgbTable[7 + base];
					}
					else {
						rgbTTLDD[i + 8] = rgbTable[0 + base];
					}
					if (col & 4) {
						rgbTTLDD[i] = rgbTable[7 + base];
					}
					else {
						rgbTTLDD[i] = rgbTable[0 + base];
					}
				}
				else {
					if (col & 2) {
						rgbTTLDD[i + 8] = rgbTableTc[7 + base];
					}
					else {
						rgbTTLDD[i + 8] = rgbTableTc[0 + base];
					}
					if (col & 4) {
						rgbTTLDD[i] = rgbTableTc[7 + base];
					}
					else {
						rgbTTLDD[i] = rgbTableTc[0 + base];
					}
				}
			}
			else {
				if ((nPixelFormat == 1) ||
					(nPixelFormat == 2)) {
					rgbTTLDD[i] = rgbTable[base + col];
				}
				else {
					rgbTTLDD[i] = rgbTableTc[base + col];
				}
			}
#endif
		}
		else {
			/* CRT OFF */
			if ((nPixelFormat == 1) ||
				(nPixelFormat == 2)) {
				rgbTTLDD[i] = rgbTable[base + 0];
			}
			else {
				rgbTTLDD[i] = rgbTableTc[base + 0];
			}
		}
	}

	/* tO~낷 */
	bPaletFlag = FALSE;
}

/*
 *	640x200AfW^[h
 *	`
 */
static void FASTCALL Draw640(void)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	BYTE *p;
#if XM7_VER >= 3
	WORD wdtop,wdbtm;
#endif

	/* I[NA */
	AllClear(TRUE);

	/* pbgݒ */
	Palet640();

	/* Xe[^XC */
	StatusLine(FALSE);

	/* _O`FbN */
	if ((nDrawTop >= nDrawBottom) || (nDrawLeft >= nDrawRight)) {
		return;
	}

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

#if XM7_VER >= 3
	/* EBhEI[v */
	if (window_open) {
		/* EBhEO 㑤̕` */
		if ((nDrawTop >> 1) < window_dy1) {
			if (bPseudo400Line) {
				Render640c(p, ddsd.lPitch, nDrawTop >> 1, window_dy1);
			}
			else {
				Render640(p, ddsd.lPitch, nDrawTop >> 1, window_dy1, 2);
			}
		}

		/* EBhE̕` */
		if ((nDrawTop >> 1) > window_dy1) {
			wdtop = (WORD)(nDrawTop >> 1);
		}
		else {
			wdtop = window_dy1;
		}

		if ((nDrawBottom >> 1) < window_dy2) {
			wdbtm = (WORD)(nDrawBottom >> 1);
		}
		else {
			wdbtm = window_dy2;
		}

		if (wdbtm > wdtop) {
			if (bPseudo400Line) {
				Render640cw(p, ddsd.lPitch,
					wdtop, wdbtm, window_dx1, window_dx2);
			}
			else {
				Render640w(p, ddsd.lPitch,
					wdtop, wdbtm, window_dx1, window_dx2, 2);
			}
		}

		/* EBhEO ̕` */
		if ((nDrawBottom >> 1) > window_dy2) {
			if (bPseudo400Line) {
				Render640c(p, ddsd.lPitch, window_dy2, nDrawBottom >> 1);
			}
			else {
				Render640(p, ddsd.lPitch, window_dy2, nDrawBottom >> 1, 2);
			}
		}
	}
	else {
		if (bPseudo400Line) {
			Render640c(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1);
		}
		else {
			Render640(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1, 2);
		}
	}
	if (!bPseudo400Line) {
		RenderFullScan(p, ddsd.lPitch);
	}
#elif XM7_VER == 2
	if (bPseudo400Line) {
		Render640c(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1);
	}
	else {
		Render640(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1);
	}
#ifdef FMTV151
	DrawV2_DD(p, ddsd.lPitch);
#endif
	if (!bPseudo400Line) {
		RenderFullScan(p, ddsd.lPitch);
	}
#else
	if (bPseudo400Line) {
		Render640m(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1);
	}
	else {
		Render640(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1);
		RenderFullScan(p, ddsd.lPitch);
	}
#endif

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);

	/* Blt */
	rect.top = nDrawTop;
	rect.left = nDrawLeft;
	rect.right = nDrawRight;
	rect.bottom = nDrawBottom;
	drect = rect;
	if (nDDResolution == DDRES_480LINE) {
		rect.top += 40;
		rect.bottom += 40;
		drect.top += 40;
		drect.bottom += 40;
	}
	else if (nDDResolution == DDRES_WUXGA) {
		drect.top *= 3;
		drect.bottom *= 3;
		drect.left *= 3;
		drect.right *= 3;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top += 56;
		rect.bottom += 56;
		drect.left *= 2;
		drect.right *= 2;
		drect.top = rect.top * 2;
		drect.bottom = rect.bottom * 2;
	}
	else if (nDDResolution == DDRES_WXGA800) {
		drect.top *= 2;
		drect.bottom *= 2;
		drect.left *= 2;
		drect.right *= 2;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.top += ptFullRes[1].y / nMlt;
		rect.bottom += ptFullRes[1].y / nMlt;
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		if (bCaption) {
			rect.top += 40;
			rect.bottom += 40;
		}
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* ɔA[NZbg */
	nDrawTop = 400;
	nDrawBottom = 0;
	nDrawLeft = 640;
	nDrawRight = 0;
	SetDrawFlag(FALSE);
}

#if XM7_VER >= 3
/*
 *	640x400AfW^[h
 *	`
 */
static void FASTCALL Draw400l(void)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	BYTE *p;
	WORD wdtop,wdbtm;

	/* I[NA */
	AllClear(TRUE);

	/* pbgݒ */
	Palet640();

	/* Xe[^XC */
	StatusLine(FALSE);

	/* _O`FbN */
	if ((nDrawTop >= nDrawBottom) || (nDrawLeft >= nDrawRight)) {
		return;
	}

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

	/* EBhEI[v */
	if (window_open) {
		if (nDrawTop < window_dy1) {
			Render640(p, ddsd.lPitch, nDrawTop, window_dy1, 1);
		}

		if (nDrawTop > window_dy1) {
			wdtop = nDrawTop;
		}
		else {
			wdtop = window_dy1;
		}
		if (nDrawBottom < window_dy2) {
			wdbtm = nDrawBottom;
		}
		else {
			wdbtm = window_dy2;
		}
		if (wdbtm > wdtop) {
			Render640w(p, ddsd.lPitch, wdtop, wdbtm, window_dx1, window_dx2, 1);
		}
		if (nDrawBottom > window_dy2) {
			Render640(p, ddsd.lPitch, window_dy2, nDrawBottom, 1);
		}
	}
	else {
		Render640(p, ddsd.lPitch, nDrawTop, nDrawBottom, 1);
	}

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);

	/* Blt */
	rect.top = nDrawTop;
	rect.left = nDrawLeft;
	rect.right = nDrawRight;
	rect.bottom = nDrawBottom;
	drect = rect;
	if (nDDResolution == DDRES_480LINE) {
		rect.top += 40;
		rect.bottom += 40;
		drect.top += 40;
		drect.bottom += 40;
	}
	else if (nDDResolution == DDRES_WUXGA) {
		drect.top *= 3;
		drect.bottom *= 3;
		drect.left *= 3;
		drect.right *= 3;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top += 56;
		rect.bottom += 56;
		drect.left *= 2;
		drect.right *= 2;
		drect.top = rect.top * 2;
		drect.bottom = rect.bottom * 2;
	}
	else if (nDDResolution == DDRES_WXGA800) {
		drect.top *= 2;
		drect.bottom *= 2;
		drect.left *= 2;
		drect.right *= 2;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.top += ptFullRes[1].y / nMlt;
		rect.bottom += ptFullRes[1].y / nMlt;
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		if (bCaption) {
			rect.top += 40;
			rect.bottom += 40;
		}
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* ɔA[NZbg */
	nDrawTop = 400;
	nDrawBottom = 0;
	nDrawLeft = 640;
	nDrawRight = 0;
	SetDrawFlag(FALSE);
}
#endif

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	640x400APF[h
 *	pbgݒ
 */
static void FASTCALL PaletL4(void)
{
	int i;
	int base;

	/* tOZbgĂȂ΁AȂ */
	if (!bPaletFlag) {
		return;
	}

	/* pbgJnʒuݒ */
	switch (nMonitorType) {
		case MONITOR_COLOR:
			if (nPixelFormat == 1) {
				base = 0;
			}
			else if (nPixelFormat == 2) {
				base = 16;
			}
			else {
				base = 0;
			}
			break;
		case MONITOR_GREEN:
			if (nPixelFormat == 1) {
				base = 32;
			}
			else if (nPixelFormat == 2) {
				base = 48;
			}
			else {
				base = 16;
			}
			break;
		case MONITOR_AMBERYELLOW:
			if (nPixelFormat == 1) {
				base = 64;
			}
			else if (nPixelFormat == 2) {
				base = 96;
			}
			else {
				base = 32;
			}
			break;
		default:
			ASSERT(FALSE);
			return;
	}

	if (nPixelFormat == 3) {
		/* 640x400AOtBbNpbg */
		for (i=0; i<3; i++) {
			if (crt_flag) {
				if ((i == 0) || (multi_page & (0x08 << i))) {
					rgbTTLL4DD[i] = (DWORD)Palet16Tc_L4[(ttl_palet[0] & 0x0f) + base];
				}
				else {
					rgbTTLL4DD[i] = (DWORD)Palet16Tc_L4[(ttl_palet[1] & 0x0f) + base];
				}
			}
			else {
				rgbTTLL4DD[i] = (DWORD)Palet16Tc_L4[base];
			}
		}
	}
	else {
		/* 640x400AOtBbNpbg */
		for (i=0; i<3; i++) {
			if (crt_flag) {
				if ((i == 0) || (multi_page & (0x08 << i))) {
					rgbTTLL4DD[i] = (DWORD)Palet16_L4[(ttl_palet[0] & 0x0f) + base];
				}
				else {
					rgbTTLL4DD[i] = (DWORD)Palet16_L4[(ttl_palet[1] & 0x0f) + base];
				}
			}
			else {
				rgbTTLL4DD[i] = (DWORD)Palet16_L4[base];
			}
		}
	}

	/* tO~낷 */
	bPaletFlag = FALSE;
}

/*
 *	eLXg
 *	_O
 */
static void FASTCALL RenderTextDD(BYTE *p, DWORD pitch,
						int first, int last, int left, int right)
{
	DWORD	tmp;
	int		x, y, x2;
	WORD	addr;
	DWORD	col;
	BYTE	cursor_start, cursor_end, cursor_type;
	BYTE	raster, lines, maxx;
	BYTE	chr, atr, chr_dat;
	BYTE	line, line_old;
	int		xsize;
	int		base;

	/* \OFFȂ牽Ȃ */
	if (!crt_flag) {
		return;
	}

	/* cursor */
	cursor_start	= (BYTE)(crtc_register[10] & 0x1f);
	cursor_end		= (BYTE)(crtc_register[11] & 0x1f);
	cursor_type		= (BYTE)((crtc_register[10] & 0x60) >> 5);

	/* ʕ\ */
	maxx			= (BYTE)(crtc_register[1] << 1);
	lines			= (BYTE)((crtc_register[9] & 0x1f) + 1);

	/* pbgJnʒuݒ */
	switch (nMonitorType) {
		case MONITOR_COLOR:
			if (nPixelFormat == 1) {
				base = 0;
			}
			else if (nPixelFormat == 2) {
				base = 16;
			}
			else {
				base = 0;
			}
			break;
		case MONITOR_GREEN:
			if (nPixelFormat == 1) {
				base = 32;
			}
			else if (nPixelFormat == 2) {
				base = 48;
			}
			else {
				base = 16;
			}
			break;
		case MONITOR_AMBERYELLOW:
			if (nPixelFormat == 1) {
				base = 64;
			}
			else if (nPixelFormat == 2) {
				base = 96;
			}
			else {
				base = 32;
			}
			break;
		default:
			ASSERT(FALSE);
			return;
	}

	if (width40_flag) {
		xsize = 16;
	}
	else {
		xsize = 8;
	}

	for (x=left / xsize; x<right / xsize; x++) {
		line_old = 255;
		for (y=first; y<last; y++) {
			raster = (BYTE)(y % lines);
			line = (BYTE)(y / lines);
			if (line != line_old) {
				addr = (WORD)((text_start_addr + (line * maxx + x) * 2) & 0x0ffe);
				chr = tvram_c[addr + 0];
				atr = tvram_c[addr + 1];

				if (atr & 0x20) {
					/* Intensity attribute ON */
					if (nPixelFormat == 3) {
						col = Palet16Tc_L4[(atr & 0x07) + base + 8];
					}
					else {
						col = Palet16_L4[(atr & 0x07) + base + 8];
					}
				}
				else {
					/* Intensity attribute OFF */
					if (nPixelFormat == 3) {
						col = Palet16Tc_L4[(atr & 0x07) + base + 0];
					}
					else {
						col = Palet16_L4[(atr & 0x07) + base + 0];
					}
				}
			}

			/* Blink attribute */
			if ((!(atr & 0x10) || text_blink) && (raster < 16)) {
				chr_dat = subcg_l4[(WORD)(chr * 16) + raster];
			}
			else {
				chr_dat = 0x00;
			}

			/* Reverse attribute */
			if (atr & 0x08) {
				chr_dat = (BYTE)~chr_dat;
			}

			/* قłāA[ */
			if ((addr == cursor_addr) && (cursor_type != 1) &&
				(cursor_blink || (cursor_type == 0))) {
				if ((raster >= cursor_start) && (raster <= cursor_end)) {
					chr_dat = (BYTE)~chr_dat;
				}
			}

			if (chr_dat) {
				for (x2=0; x2<8; x2++) {
					if (chr_dat & (1 << (7-x2))) {
						if (nPixelFormat == 3) {
							if (width40_flag) {
								tmp = y * pitch + (((x << 3) + x2) << 3);
								*(DWORD *)(&p[tmp + 0]) = col;
								*(DWORD *)(&p[tmp + 4]) = col;
							}
							else {
								tmp = y * pitch + (((x << 3) + x2) << 2);
								*(DWORD *)(&p[tmp + 0]) = col;
							}
						}
						else {
							if (width40_flag) {
								tmp = y * pitch + (((x << 3) + x2) << 2);
								*(WORD *)(&p[tmp + 0]) = (WORD)col;
								*(WORD *)(&p[tmp + 2]) = (WORD)col;
							}
							else {
								tmp = y * pitch + (((x << 3) + x2) << 1);
								*(WORD *)(&p[tmp + 0]) = (WORD)col;
							}
						}
					}
				}
			}

			line_old = line;
		}
	}
}

/*
 *	640x400APF[h
 *	`
 */
static void FASTCALL DrawL4(void)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	BYTE *p;

	/* I[NA */
	AllClear(TRUE);

	/* pbgݒ */
	PaletL4();

	/* Xe[^XC */
	StatusLine(FALSE);

	/* _O`FbN */
	if ((nDrawTop >= nDrawBottom) || (nDrawLeft >= nDrawRight)) {
		return;
	}

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

	RenderL4(p, ddsd.lPitch, nDrawTop, nDrawBottom);
	RenderTextDD(p, ddsd.lPitch, nDrawTop, nDrawBottom, nDrawLeft, nDrawRight);

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);

	/* Blt */
	rect.top = nDrawTop;
	rect.left = nDrawLeft;
	rect.right = nDrawRight;
	rect.bottom = nDrawBottom;
	drect = rect;
	if (nDDResolution == DDRES_480LINE) {
		rect.top += 40;
		rect.bottom += 40;
		drect.top += 40;
		drect.bottom += 40;
	}
	else if (nDDResolution == DDRES_WUXGA) {
		drect.top *= 3;
		drect.bottom *= 3;
		drect.left *= 3;
		drect.right *= 3;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top += 56;
		rect.bottom += 56;
		drect.left *= 2;
		drect.right *= 2;
		drect.top = rect.top * 2;
		drect.bottom = rect.bottom * 2;
	}
	else if (nDDResolution == DDRES_WXGA800) {
		drect.top *= 2;
		drect.bottom *= 2;
		drect.left *= 2;
		drect.right *= 2;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.top += ptFullRes[1].y / nMlt;
		rect.bottom += ptFullRes[1].y / nMlt;
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		if (bCaption) {
			rect.top += 40;
			rect.bottom += 40;
		}
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* ɔA[NZbg */
	nDrawTop = 400;
	nDrawBottom = 0;
	nDrawLeft = 640;
	nDrawRight = 0;
	SetDrawFlag(FALSE);
}
#endif

#if XM7_VER >= 2
/*
 *	320x200AAiO[h
 *	pbgݒ
 */
static void FASTCALL Palet320(void)
{
	int i, j;
	DWORD r, g, b;
	int amask;
	WORD color;

	/* tOZbgĂȂ΁AȂ */
	if (!bPaletFlag) {
		return;
	}

	/* AiO}XN쐬 */
	amask = 0;
	if (!(multi_page & 0x10)) {
		amask |= 0x000f;
	}
	if (!(multi_page & 0x20)) {
		amask |= 0x00f0;
	}
	if (!(multi_page & 0x40)) {
		amask |= 0x0f00;
	}

	for (i=0; i<4096; i++) {
		/* ŉʂ5bitÂB,G,R */
#if XM7_VER >= 2
		/* TTL RGB */
		if (crt_flag && (nMonitorType != MONITOR_COLOR)) {
			r = 0;
			g = 0;
			b = 0;
			j = 0;
			/* R */
			if ((i & amask) & 0x0080) {
				j |= 2;
			}

			/* G */
			if ((i & amask) & 0x0800) {
				j |= 4;
			}

			/* B */
			if ((i & amask) & 0x0008) {
				j |= 1;
			}

			/* RGB */
			if (ttl_palet[j & 7] & 0x01) {
				b = 0x0f;
			}
			if (ttl_palet[j & 7] & 0x02) {
				r = 0x0f;
			}
			if (ttl_palet[j & 7] & 0x04) {
				g = 0x0f;
			}
		}
		else
#endif
		if (crt_flag) {
			j = i & amask;
			r = (WORD)apalet_r[j];
			g = (WORD)apalet_g[j];
			b = (WORD)apalet_b[j];
		}
		else {
			r = 0;
			g = 0;
			b = 0;
		}

		/* sNZ^CvɉADWORDf[^쐬 */
		if (nPixelFormat >= 3) {
			/* R8bit, G8bit, B8bit (TrueColor) */
			r = (DWORD)((r << 20) | (r << 16));
			g = (DWORD)((g << 12) | (g <<  8));
			b = (DWORD)((b <<  4) | (b      ));

			/* Zbg */
			rgbAnalogDD[i] = (DWORD)(r | g | b);
		}
		else {
			if (nPixelFormat == 1) {
				/* R5bit, G6bit, B5bit^Cv */
				r <<= 12;
				if (r > 0) {
					r |= 0x0800;
				}

				g <<= 7;
				if (g > 0) {
					g |= 0x0060;
				}

				b <<= 1;
				if (b > 0) {
					b |= 0x0001;
				}
			}
			else if (nPixelFormat == 2) {
				/* R5bit, G5bit, B5bit^Cv */
				r <<= 11;
				if (r > 0) {
					r |= 0x0400;
				}

				g <<= 6;
				if (g > 0) {
					g |= 0x0020;
				}

				b <<= 1;
				if (b > 0) {
					b |= 0x0001;
				}
			}

			/* Zbg */
			color = (WORD)(r | g | b);
			rgbAnalogDD[i] = (DWORD)((color << 16) | color);
		}
	}

	/* tO~낷 */
	bPaletFlag = FALSE;
}

/*
 *	320x200AAiO[h
 *	`
 */
static void FASTCALL Draw320(void)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	BYTE *p;
#if XM7_VER >= 3
	WORD wdtop,wdbtm;
#endif

	/* I[NA */
	AllClear(TRUE);

	/* pbgݒ */
	Palet320();

	/* Xe[^XC */
	StatusLine(FALSE);

	/* _O`FbN */
	if ((nDrawTop >= nDrawBottom) || (nDrawLeft >= nDrawRight)) {
		return;
	}

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

#if XM7_VER >= 3
	/* EBhEI[v */
	if (window_open) {
		/* EBhEO 㑤̕` */
		if ((nDrawTop >> 1) < window_dy1) {
			Render320(p, ddsd.lPitch, nDrawTop >> 1, window_dy1);
		}

		/* EBhE̕` */
		if ((nDrawTop >> 1) > window_dy1) {
			wdtop = (WORD)(nDrawTop >> 1);
		}
		else {
			wdtop = window_dy1;
		}

		if ((nDrawBottom >> 1) < window_dy2) {
			wdbtm = (WORD)(nDrawBottom >> 1);
		}
		else {
			wdbtm = window_dy2;
		}

		if (wdbtm > wdtop) {
			Render320w(p, ddsd.lPitch, wdtop, wdbtm, window_dx1, window_dx2);
		}

		/* EBhEO ̕` */
		if ((nDrawBottom >> 1) > window_dy2) {
			Render320(p, ddsd.lPitch, window_dy2, nDrawBottom >> 1);
		}
	}
	else {
		Render320(p, ddsd.lPitch, nDrawTop>>1, nDrawBottom>>1);
	}
	RenderFullScan(p, ddsd.lPitch);
#else
	Render320(p, ddsd.lPitch, nDrawTop>>1, nDrawBottom>>1);
#ifdef FMTV151
	DrawV2_DD(p, ddsd.lPitch);
#endif
	RenderFullScan(p, ddsd.lPitch);
#endif

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);

	/* Blt */
	rect.top = nDrawTop;
	rect.left = nDrawLeft;
	rect.right = nDrawRight;
	rect.bottom = nDrawBottom;
	drect = rect;
	if (nDDResolution == DDRES_480LINE) {
		rect.top += 40;
		rect.bottom += 40;
		drect.top += 40;
		drect.bottom += 40;
	}
	else if (nDDResolution == DDRES_WUXGA) {
		drect.top *= 3;
		drect.bottom *= 3;
		drect.left *= 3;
		drect.right *= 3;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top += 56;
		rect.bottom += 56;
		drect.left *= 2;
		drect.right *= 2;
		drect.top = rect.top * 2;
		drect.bottom = rect.bottom * 2;
	}
	else if (nDDResolution == DDRES_WXGA800) {
		drect.top *= 2;
		drect.bottom *= 2;
		drect.left *= 2;
		drect.right *= 2;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.top += ptFullRes[1].y / nMlt;
		rect.bottom += ptFullRes[1].y / nMlt;
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		if (bCaption) {
			rect.top += 40;
			rect.bottom += 40;
		}
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* ɔA[NZbg */
	nDrawTop = 400;
	nDrawBottom = 0;
	nDrawLeft = 640;
	nDrawRight = 0;
	SetDrawFlag(FALSE);
}
#endif

#if XM7_VER >= 3
/*
 *	320x200A26F[h
 *	`
 */
static void FASTCALL Draw256k(void)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	BYTE *p;

	/* I[NA */
	AllClear(TRUE);

	/* Xe[^XC */
	StatusLine(FALSE);

	/* _O`FbN */
	if ((nDrawTop >= nDrawBottom) || (nDrawLeft >= nDrawRight)) {
		return;
	}

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

	if (crt_flag) {
		Render256k(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1, multi_page);
	}
	else {
		Render256k(p, ddsd.lPitch, nDrawTop >> 1, nDrawBottom >> 1, 0xff);
	}
	RenderFullScan(p, ddsd.lPitch);

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);

	/* Blt */
	rect.top = nDrawTop;
	rect.left = nDrawLeft;
	rect.right = nDrawRight;
	rect.bottom = nDrawBottom;
	drect = rect;
	if (nDDResolution == DDRES_480LINE) {
		rect.top += 40;
		rect.bottom += 40;
		drect.top += 40;
		drect.bottom += 40;
	}
	else if (nDDResolution == DDRES_WUXGA) {
		drect.top *= 3;
		drect.bottom *= 3;
		drect.left *= 3;
		drect.right *= 3;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top += 56;
		rect.bottom += 56;
		drect.left *= 2;
		drect.right *= 2;
		drect.top = rect.top * 2;
		drect.bottom = rect.bottom * 2;
	}
	else if (nDDResolution == DDRES_WXGA800) {
		drect.top *= 2;
		drect.bottom *= 2;
		drect.left *= 2;
		drect.right *= 2;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.top += ptFullRes[1].y / nMlt;
		rect.bottom += ptFullRes[1].y / nMlt;
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		if (bCaption) {
			rect.top += 40;
			rect.bottom += 40;
		}
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* ɔA[NZbg */
	nDrawTop = 400;
	nDrawBottom = 0;
	nDrawLeft = 640;
	nDrawRight = 0;
	SetDrawFlag(FALSE);
}
#endif

/*
 *	``FbN
 */
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(L4CARD))
static void FASTCALL DrawDDCheck(void)
{
#if XM7_VER >= 3
	/* Ȃ蔲(H */
	if (nMode != screen_mode) {
		ReDrawDD();
		nMode = screen_mode;
	}
#elif XM7_VER >= 2
	/* 320x200 */
	if (mode320) {
		if (!bAnalog) {
			ReDrawDD();
			bAnalog = TRUE;
		}
		return;
	}

	/* 640x200 */
	if (bAnalog) {
		ReDrawDD();
		bAnalog = FALSE;
	}
#elif XM7_VER == 1 && defined(L4CARD)
	/* 640x400 */
	if (enable_400line) {
		if (!b400Line) {
			ReDrawDD();
			b400Line = TRUE;
		}
		return;
	}

	/* 640x200 */
	if (b400Line) {
		ReDrawDD();
		b400Line = FALSE;
	}
#endif
}
#endif

/*
 *	`
 */
void FASTCALL DrawDD(void)
{
	/*  */
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(L4CARD))
	DrawDDCheck();
#endif

	/* ` */
#if XM7_VER >= 3
	switch (nMode) {
		case SCR_400LINE	:	Draw400l();
								return;
		case SCR_262144		:	Draw256k();
								return;
		case SCR_4096		:	Draw320();
								return;
		case SCR_200LINE	:	Draw640();
								return;
	}
#elif XM7_VER >= 2
	if (bAnalog) {
		Draw320();
		return;
	}
#elif XM7_VER == 1 && defined(L4CARD)
	if (b400Line) {
		DrawL4();
		return;
	}
#endif
	Draw640();
}

/*
 *	640x200AfW^[h
 *	X^`
 */
static void FASTCALL DrawRaster640(int nRaster)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	BYTE *p;

	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 200) {
		return;
	}

	/* pbgݒ */
	Palet640();

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

#if XM7_VER >= 3
	/* EBhEI[v */
	if (window_open) {
		/* EBhEO̕` */
		if ((nRaster < window_dy1) || (nRaster > window_dy2)) {
			if (bPseudo400Line) {
				Render640c(p, ddsd.lPitch, nRaster, nRaster + 1);
			}
			else {
				Render640(p, ddsd.lPitch, nRaster, nRaster + 1, 2);
			}
		}
		else {
			if (bPseudo400Line) {
				Render640cw(p, ddsd.lPitch,
					nRaster, nRaster + 1, window_dx1, window_dx2);
			}
			else {
				Render640w(p, ddsd.lPitch,
					nRaster, nRaster + 1, window_dx1, window_dx2, 2);
			}
		}
	}
	else {
		if (bPseudo400Line) {
			Render640c(p, ddsd.lPitch, nRaster, nRaster + 1);
		}
		else {
			Render640(p, ddsd.lPitch, nRaster, nRaster + 1, 2);
		}
	}
#elif XM7_VER == 2
	if (bPseudo400Line) {
		Render640c(p, ddsd.lPitch, nRaster, nRaster + 1);
	}
	else {
		Render640(p, ddsd.lPitch, nRaster, nRaster + 1);
	}
#else
	if (bPseudo400Line) {
		Render640m(p, ddsd.lPitch, nRaster, nRaster + 1);
	}
	else {
		Render640(p, ddsd.lPitch, nRaster, nRaster + 1);
	}
#endif

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);
}

#if XM7_VER >= 3
/*
 *	640x400AfW^[h
 *	X^`
 */
static void FASTCALL DrawRaster400l(int nRaster)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	BYTE *p;

	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 400) {
		return;
	}

	/* pbgݒ */
	Palet640();

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

	/* EBhEI[v */
	if (window_open) {
		/* EBhEO̕` */
		if ((nRaster < window_dy1) || (nRaster > window_dy2)) {
			Render640(p, ddsd.lPitch, nRaster, nRaster + 1, 1);
		}
		else {
			Render640w(p, ddsd.lPitch, nRaster, nRaster + 1, window_dx1, window_dx2, 1);
		}
	}
	else {
		Render640(p, ddsd.lPitch, nRaster, nRaster + 1, 1);
	}

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);
}
#endif

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	eLXg
 *	X^`
 */
static void FASTCALL RenderTextRasterDD(BYTE *p, DWORD pitch,
						int nRaster, int left, int right)
{
	DWORD	tmp;
	int		x, x2;
	WORD	addr;
	DWORD	col;
	BYTE	cursor_start, cursor_end, cursor_type;
	BYTE	raster, lines, maxx;
	BYTE	chr, atr, chr_dat;
	BYTE	line;
	int		xsize;
	int		base;

	/* pbgJnʒuݒ */
	switch (nMonitorType) {
		case MONITOR_COLOR:
			if (nPixelFormat == 1) {
				base = 0;
			}
			else if (nPixelFormat == 2) {
				base = 16;
			}
			else {
				base = 0;
			}
			break;
		case MONITOR_GREEN:
			if (nPixelFormat == 1) {
				base = 32;
			}
			else if (nPixelFormat == 2) {
				base = 48;
			}
			else {
				base = 16;
			}
			break;
		case MONITOR_AMBERYELLOW:
			if (nPixelFormat == 1) {
				base = 64;
			}
			else if (nPixelFormat == 2) {
				base = 96;
			}
			else {
				base = 32;
			}
			break;
		default:
			ASSERT(FALSE);
			return;
	}

	/* \OFFȂ牽Ȃ */
	if (!crt_flag) {
		return;
	}

	/* cursor */
	cursor_start	= (BYTE)(crtc_register[10] & 0x1f);
	cursor_end		= (BYTE)(crtc_register[11] & 0x1f);
	cursor_type		= (BYTE)((crtc_register[10] & 0x60) >> 5);

	/* ʕ\ */
	maxx			= (BYTE)(crtc_register[1] << 1);
	lines			= (BYTE)((crtc_register[9] & 0x1f) + 1);

	if (width40_flag) {
		xsize = 16;
	}
	else {
		xsize = 8;
	}

	for (x=left / xsize; x<right / xsize; x++) {
		raster = (BYTE)(nRaster % lines);
		line = (BYTE)(nRaster / lines);
		addr = (WORD)((text_start_addr + (line * maxx + x) * 2) & 0x0ffe);
		chr = tvram_c[addr + 0];
		atr = tvram_c[addr + 1];

		if (atr & 0x20) {
			/* Intensity attribute ON */
			if (nPixelFormat == 3) {
				col = Palet16Tc_L4[(atr & 0x07) + base + 8];
			}
			else {
				col = Palet16_L4[(atr & 0x07) + base + 8];
			}
		}
		else {
			/* Intensity attribute OFF */
			if (nPixelFormat == 3) {
				col = Palet16Tc_L4[(atr & 0x07) + base + 0];
			}
			else {
				col = Palet16_L4[(atr & 0x07) + base + 0];
			}
		}

		/* Blink attribute */
		if ((!(atr & 0x10) || text_blink) && (raster < 16)) {
			chr_dat = subcg_l4[(WORD)(chr * 16) + raster];
		}
		else {
			chr_dat = 0x00;
		}

		/* Reverse attribute */
		if (atr & 0x08) {
			chr_dat = (BYTE)~chr_dat;
		}

		/* قłāA[ */
		if ((addr == cursor_addr) && (cursor_type != 1) &&
			(cursor_blink || (cursor_type == 0))) {
			if ((raster >= cursor_start) && (raster <= cursor_end)) {
				chr_dat = (BYTE)~chr_dat;
			}
		}

		if (chr_dat) {
			for (x2=0; x2<8; x2++) {
				if (chr_dat & (1 << (7-x2))) {
					if (nPixelFormat == 3) {
						if (width40_flag) {
							tmp = nRaster * pitch + (((x << 3) + x2) << 3);
							*(DWORD *)(&p[tmp + 0]) = col;
							*(DWORD *)(&p[tmp + 4]) = col;
						}
						else {
							tmp = nRaster * pitch + (((x << 3) + x2) << 2);
							*(DWORD *)(&p[tmp + 0]) = col;
						}
					}
					else {
						if (width40_flag) {
							tmp = nRaster * pitch + (((x << 3) + x2) << 2);
							*(WORD *)(&p[tmp + 0]) = (WORD)col;
							*(WORD *)(&p[tmp + 2]) = (WORD)col;
						}
						else {
							tmp = nRaster * pitch + (((x << 3) + x2) << 1);
							*(WORD *)(&p[tmp + 0]) = (WORD)col;
						}
					}
				}
			}
		}
	}
}

/*
 *	640x400APF[h
 *	X^`
 */
static void FASTCALL DrawRasterL4(int nRaster)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	BYTE *p;

	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 400) {
		return;
	}

	/* pbgݒ */
	PaletL4();

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

	RenderL4(p, ddsd.lPitch, nRaster, nRaster + 1);
	RenderTextRasterDD(p, ddsd.lPitch, nRaster, 0, 640);

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);
}
#endif

#if XM7_VER >= 2
/*
 *	320x200AAiO[h
 *	X^`
 */
static void FASTCALL DrawRaster320(int nRaster)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	BYTE *p;

	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 200) {
		return;
	}

	/* pbgݒ */
	Palet320();

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

#if XM7_VER >= 3
	/* EBhEI[v */
	if (window_open) {
		/* EBhEO̕` */
		if ((nRaster < window_dy1) || (nRaster > window_dy2)) {
			Render320(p, ddsd.lPitch, nRaster, nRaster + 1);
		}
		else {
			Render320w(p, ddsd.lPitch, nRaster, nRaster + 1, window_dx1, window_dx2);
		}
	}
	else {
		Render320(p, ddsd.lPitch, nRaster, nRaster + 1);
	}
#else
	Render320(p, ddsd.lPitch, nRaster, nRaster + 1);
#endif

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);
}
#endif

#if XM7_VER >= 3
/*
 *	320x200A26F[h
 *	X^`
 */
static void FASTCALL DrawRaster256k(int nRaster)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	BYTE *p;

	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 200) {
		return;
	}

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

	if (crt_flag) {
		Render256k(p, ddsd.lPitch, nRaster, nRaster + 1, multi_page);
	}
	else {
		Render256k(p, ddsd.lPitch, nRaster, nRaster + 1, 0xff);
	}

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);
}
#endif

/*
 *	X^Pʃ_O
 *	`
 */
void FASTCALL DrawPostRenderDD(void)
{
	DDSURFACEDESC ddsd;
	HRESULT hResult;
	RECT rect;
	RECT drect;
	BYTE *p;

	/* I[NA */
	AllClear(FALSE);

	/* Xe[^XC */
	StatusLine(FALSE);

	/* T[tFCXbN */
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	hResult = lpdds[1]->Lock(NULL, &ddsd,
							 DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* _O㏈ */
	p = (BYTE *)ddsd.lpSurface;
	if (nDDResolution == DDRES_480LINE) {
		p += (ddsd.lPitch * 40);
	}
	else if (nDDResolution == DDRES_SXGA) {
		p += (ddsd.lPitch * 56);
	}
	else if (nDDResolution == DDRES_USER) {
		p += (ddsd.lPitch * (ptFullRes[1].y / nMlt) + (ddsd.lPitch * ptFullRes[1].x / ptFullRes[0].x));
		if (bCaption) {
			p += (ddsd.lPitch * 40);
		}
	}

#if XM7_VER >= 3
	if ((nMode != SCR_400LINE) && !bPseudo400Line) {
		RenderFullScan(p, ddsd.lPitch);
	}
#elif XM7_VER == 2 && defined(FMTV151)
	DrawV2_DD(p, ddsd.lPitch);
	if (!bPseudo400Line) {
		RenderFullScan(p, ddsd.lPitch);
	}
#elif XM7_VER == 1
#ifdef L4CARD
	if (!b400Line && !bPseudo400Line) {
#else
	if (!bPseudo400Line) {
#endif
		RenderFullScan(p, ddsd.lPitch);
	}
#else
	if (!bPseudo400Line) {
		RenderFullScan(p, ddsd.lPitch);
	}
#endif

	/* T[tFCXAbN */
	lpdds[1]->Unlock(ddsd.lpSurface);

	/* Blt */
	rect.left = 0;
	rect.right = 640;

	/* X^Pʃ_OɂЂԂ邱Ƃ̂ */
	if (nDrawTop >= nDrawBottom) {
		rect.top = 0;
		rect.bottom = 400;
	}
	else {
		rect.top = nDrawTop;
		rect.bottom = nDrawBottom;
	}

	drect = rect;

	if (nDDResolution == DDRES_480LINE) {
		rect.top += 40;
		rect.bottom += 40;
		drect.top += 40;
		drect.bottom += 40;
	}
	else if (nDDResolution == DDRES_WUXGA) {
		drect.top *= 3;
		drect.bottom *= 3;
		drect.left *= 3;
		drect.right *= 3;
	}
	else if (nDDResolution == DDRES_SXGA) {
		rect.top += 56;
		rect.bottom += 56;
		drect.left *= 2;
		drect.right *= 2;
		drect.top = rect.top * 2;
		drect.bottom = rect.bottom * 2;
	}
	else if (nDDResolution == DDRES_WXGA800) {
		drect.top *= 2;
		drect.bottom *= 2;
		drect.left *= 2;
		drect.right *= 2;
	}
	else if (nDDResolution == DDRES_USER) {
		rect.top += ptFullRes[1].y / nMlt;
		rect.bottom += ptFullRes[1].y / nMlt;
		rect.left += ptFullRes[1].x / nMlt;
		rect.right += ptFullRes[1].x / nMlt;
		if (bCaption) {
			rect.top += 40;
			rect.bottom += 40;
		}
		drect.top = rect.top * nMlt;
		drect.bottom = rect.bottom * nMlt;
		drect.left = rect.left * nMlt;
		drect.right = rect.right * nMlt;
	}
	hResult = lpdds[0]->Blt(&drect, lpdds[1], &rect, DDBLT_WAIT, NULL);
	if (hResult != DD_OK) {
		/* T[tFCXXgĂ΁AXgA */
		if (hResult == DDERR_SURFACELOST) {
			lpdds[0]->Restore();
			lpdds[1]->Restore();
		}
		/* ͑S̈XV */
		ReDrawDD();
		return;
	}

	/* ɔA[NZbg */
	nDrawTop = 400;
	nDrawBottom = 0;
	nDrawLeft = 640;
	nDrawRight = 0;
	SetDrawFlag(TRUE);
}

/*
 *	X^Pʃ_O
 *	eX^_O
 */
void FASTCALL DrawRasterDD(int nRaster)
{
	/*  */
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(L4CARD))
	DrawDDCheck();
#endif

#if XM7_VER >= 3
	/* ꂩgĕ` */
	switch (nMode) {
		case SCR_400LINE	:	DrawRaster400l(nRaster);
								return;
		case SCR_262144		:	DrawRaster256k(nRaster);
								return;
		case SCR_4096		:	DrawRaster320(nRaster);
								return;
		case SCR_200LINE	:	DrawRaster640(nRaster);
								return;
	}
#elif XM7_VER >= 2
	/* ǂ炩gĕ` */
	if (bAnalog) {
		DrawRaster320(nRaster);
	}
	else {
		DrawRaster640(nRaster);
	}
#elif defined(L4CARD)
	/* ǂ炩gĕ` */
	if (b400Line) {
		DrawRasterL4(nRaster);
	}
	else {
		DrawRaster640(nRaster);
	}
#else
	DrawRaster640(nRaster);
#endif
}

/*
 *	j[Jn
 */
void FASTCALL EnterMenuDD(HWND hWnd, BOOL flag)
{
	ASSERT(hWnd);

	/* Nbp[̗LŔł */
	if (!lpddc) {
		return;
	}

	/* Nbp[Zbg */
	LockVM();
	lpdds[0]->SetClipper(lpddc);
	UnlockVM();

	/* }EXJ[\on */
	if (!bMouseCursor) {
		ShowCursor(TRUE);
		bMouseCursor = TRUE;
	}

#ifdef MOUSE
	/* }EXLv`~ */
	SetMouseCapture(FALSE);
#endif

	/* tO~Ăꍇ̓j[o[̏ɋA */
	if (!flag) {
		return;
	}

	/* j[o[` */
	if (bWin8flag) {
		/* Windows 8ł͌̃Cj[𕜌 */
		if (hWnd == hMainWnd) {
			SetMenu(hMainWnd, hMainMenu);
		}
	}
	DrawMenuBar(hMainWnd);
}

/*
 *	j[I
 */
void FASTCALL ExitMenuDD(BOOL flag)
{
	/* Nbp[̗LŔł */
	if (!lpddc) {
		return;
	}

	/* Nbp[ */
	LockVM();
	lpdds[0]->SetClipper(NULL);
	UnlockVM();

#ifdef MOUSE
	/* }EXLv`ݒ */
	SetMouseCapture(TRUE);
#endif

	/* }EXJ[\OFF */
	if (bMouseCursor) {
		ShowCursor(FALSE);
		bMouseCursor = FALSE;
	}

	/* tO~Ăꍇ̓j[o[̏ɋA */
	if (!flag) {
		return;
	}

	/* j[o[ */
	if (bWin8flag) {
		/* Windows 8ł̓Cj[ꎞIɉj[o[` */
		SetMenu(hMainWnd, NULL);
		DrawMenuBar(hMainWnd);
	}
	else {
		DrawMenuBar(NULL);
	}

	/* ĕ\ */
	ReDrawDD();
}

/*-[ VMƂ̐ڑ ]-----------------------------------------------------------*/

/*
 *	VRAMZbg
 */
void FASTCALL VramDD(WORD addr)
{
	WORD x;
	WORD y;

	/* yWZo */
#if XM7_VER >= 3
	switch (nMode) {
		case SCR_400LINE	:	addr &= 0x7fff;
								x = (WORD)((addr % 80) << 3);
								y = (WORD)(addr / 80);
								break;
		case SCR_262144		:
		case SCR_4096		:	addr &= 0x1fff;
								x = (WORD)((addr % 40) << 4);
								y = (WORD)((addr / 40) << 1);
								break;
		case SCR_200LINE	:	addr &= 0x3fff;
								x = (WORD)((addr % 80) << 3);
								y = (WORD)((addr / 80) << 1);
								break;
	}
#elif XM7_VER >= 2
	if (bAnalog) {
		addr &= 0x1fff;
		x = (WORD)((addr % 40) << 4);
		y = (WORD)((addr / 40) << 1);
	}
	else {
		addr &= 0x3fff;
		x = (WORD)((addr % 80) << 3);
		y = (WORD)((addr / 80) << 1);
	}
#elif defined(L4CARD)
	if (b400Line) {
		addr -= vram_offset[0];
		addr &= 0x7fff;
		x = (WORD)((addr % 80) << 3);
		y = (WORD)(addr / 80);
	}
	else {
		addr &= 0x3fff;
		x = (WORD)((addr % 80) << 3);
		y = (WORD)((addr / 80) << 1);
	}
#else
	addr &= 0x3fff;
	x = (WORD)((addr % 80) << 3);
	y = (WORD)((addr / 80) << 1);
#endif

	/* I[o[`FbN */
	if ((x >= 640) || (y >= 400)) {
		return;
	}

	/* ĕ`tOݒ */
	DDDrawFlag[(y >> 3) * 80 + (x >> 3)] = 1;

	/* XV */
	if (nDrawTop > y) {
		nDrawTop = y;
	}
	if (nDrawBottom <= y) {
#if XM7_VER >= 3
		if (nMode == SCR_400LINE) {
			nDrawBottom = (WORD)(y + 1);
		}
		else {
			nDrawBottom = (WORD)(y + 2);
		}
#else
		nDrawBottom = (WORD)(y + 2);
#endif
	}

	/* XV */
	if (nDrawLeft > x) {
		nDrawLeft = x;
	}
	if (nDrawRight <= x) {
#if XM7_VER >= 2
#if XM7_VER >= 3
		if (nMode & SCR_ANALOG) {
#else
		if (bAnalog) {
#endif
			nDrawRight = (WORD)(x + 16);
		}
		else {
			nDrawRight = (WORD)(x + 8);
		}
#else
		nDrawRight = (WORD)(x + 8);
#endif
	}
}

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	eLXgVRAMZbg
 */
void FASTCALL TvramDD(WORD addr)
{
	WORD ysize;
	WORD x;
	WORD y;
	BYTE maxy;
	WORD yy;

	maxy = (BYTE)(crtc_register[6] & 0x7f);

	/* WZo */
	addr = (WORD)((addr - text_start_addr) & 0x0ffe);
	ysize = (WORD)((crtc_register[9] & 0x1f) + 1);
	if (width40_flag) {
		x = (WORD)((addr % 80) * 8);
		y = (WORD)(addr / 80);
	}
	else {
		x = (WORD)((addr % 160) * 4);
		y = (WORD)(addr / 160);
	}

	/* I[o[`FbN */
	if (y >= maxy) {
		return;
	}
	y = (WORD)(y * ysize);
	if (y > 400 - ysize) {
		y = (WORD)(400 - ysize);
	}

	/* ĕ`tOݒ */
	for (yy = y; yy < (WORD)(y + ysize); yy += (WORD)8) {
		DDDrawFlag[(yy >> 3) * 80 + (x >> 3)] = 1;
		if (width40_flag) {
			DDDrawFlag[(yy >> 3) * 80 + (x >> 3) + 1] = 1;
		}
	}

	/* XV */
	if (nDrawTop > y) {
		nDrawTop = y;
	}
	if (nDrawBottom <= (y + ysize - 1)) {
		nDrawBottom = (WORD)(y + ysize);
	}

	/* XV */
	if (nDrawLeft > x) {
		nDrawLeft = x;
	}
	if (nDrawRight <= x) {
		if (width40_flag) {
			nDrawRight = (WORD)(x + 16);
		}
		else {
			nDrawRight = (WORD)(x + 8);
		}
	}
}

/*
 *	eLXgVRAMĕ`v
 */
void FASTCALL ReDrawTVRamDD(void)
{
	WORD maxy;

	maxy = (WORD)((crtc_register[6] & 0x7f) * ((crtc_register[9] & 0x1f) + 1));

	nDrawTop = 0;
	if (nDrawBottom < maxy) {
		nDrawBottom = maxy;
	}
	nDrawLeft = 0;
	nDrawRight = 640;
	SetDrawFlag(TRUE);
}
#endif

/*
 *	TTLpbgZbg
 */
void FASTCALL DigitalDD(void)
{
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	SetDrawFlag(TRUE);
}

/*
 *	AiOpbgZbg
 */
void FASTCALL AnalogDD(void)
{
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	SetDrawFlag(TRUE);
}

/*
 *	ĕ`v
 */
void FASTCALL ReDrawDD(void)
{
	/* S̈惌_O */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	bClearFlag = TRUE;
	SetDrawFlag(TRUE);
	SetDirtyFlag(0, 400, TRUE);
}

#if XM7_VER >= 3
/*
 *	n[hEFAEBhEʒm
 */
void FASTCALL WindowDD(void)
{
	WORD tmpLeft, tmpRight;
	WORD tmpTop, tmpBottom;
	WORD tmpDx1, tmpDx2;
	WORD tmpDy1, tmpDy2;
	BYTE *p;
	int i;

	/* 26F[h͉Ȃ */
	if (nMode == SCR_262144) {
		return;
	}

	/* OăNbsO */
	window_clip(nMode);

	/* EBhETCY␳ */
	tmpDx1 = window_dx1;
	tmpDy1 = window_dy1;
	tmpDx2 = window_dx2;
	tmpDy2 = window_dy2;
	if (nMode != SCR_400LINE) {
		tmpDy1 <<= 1;
		tmpDy2 <<= 1;
	}
	if (nMode == SCR_4096) {
		tmpDx1 <<= 1;
		tmpDx2 <<= 1;
	}

	if (bWindowOpen != window_open) {
		if (window_open) {
			/* EBhEJꍇ */
			tmpLeft = tmpDx1;
			tmpRight = tmpDx2;
			tmpTop = tmpDy1;
			tmpBottom = tmpDy2;
		}
		else {
			/* EBhEꍇ */
			tmpLeft = nWindowDx1;
			tmpRight = nWindowDx2;
			tmpTop = nWindowDy1;
			tmpBottom = nWindowDy2;
		}
	}
	else {
		if (window_open) {
			/* XV̈TCŶ݂̂ɐݒ */
			tmpTop = nDrawTop;
			tmpBottom = nDrawBottom;
			tmpLeft = nDrawLeft;
			tmpRight = nDrawRight;

			/* WύX`FbN */
			if (!((nWindowDx1 == tmpDx1) &&
				  (nWindowDy1 == tmpDy1) &&
				  (nWindowDx2 == tmpDx2) &&
				  (nWindowDy2 == tmpDy2))) {
				/* X */
				if (nWindowDx1 < tmpDx1) {
					tmpLeft = nWindowDx1;
				}
				else {
					tmpLeft = tmpDx1;
				}

				/* EX */
				if (nWindowDx2 > tmpDx2) {
					tmpRight = nWindowDx2;
				}
				else {
					tmpRight = tmpDx2;
				}

				/* Y */
				if (nWindowDy1 < tmpDy1) {
					tmpTop = nWindowDy1;
				}
				else {
					tmpTop = tmpDy1;
				}

				/* EY */
				if (nWindowDy2 > tmpDy2) {
					tmpBottom = nWindowDy2;
				}
				else {
					tmpBottom = tmpDy2;
				}
			}
		}
		else {
			/* EBhEJĂȂ̂ŉȂ */
			return;
		}
	}

	/* O̍ĕ`̈ƔrčLΗ̈XV */
	if (tmpLeft < nDrawLeft) {
		nDrawLeft = tmpLeft;
	}
	if (tmpRight > nDrawRight) {
		nDrawRight = tmpRight;
	}
	if (tmpTop < nDrawTop) {
		nDrawTop = tmpTop;
	}
	if (tmpBottom > nDrawBottom) {
		nDrawBottom = tmpBottom;
	}

	/* ĕ`tOXV */
	if ((nDrawLeft < nDrawRight) && (nDrawTop < nDrawBottom)) {
		if (nMode == SCR_400LINE) {
			SetDirtyFlag(nDrawTop, nDrawBottom + 1, TRUE);
		}
		else {
			SetDirtyFlag(nDrawTop >> 1, (nDrawBottom >> 1) + 1, TRUE);
		}
		p = &DDDrawFlag[(nDrawTop >> 3) * 80 + (nDrawLeft >> 3)];
		for (i = (nDrawTop >> 3); i < ((nDrawBottom + 7) >> 3) ; i++) {
			memset(p, 1, (nDrawRight - nDrawLeft) >> 3);
			p += 80;
		}
	}

	/* EBhEI[vԂۑ */
	bWindowOpen = window_open;
	nWindowDx1 = tmpDx1;
	nWindowDy1 = tmpDy1;
	nWindowDx2 = tmpDx2;
	nWindowDy2 = tmpDy2;
}
#endif

#endif	/* _WIN32 */
