/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API \ ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>
#include "xm7.h"
#include "device.h"
#include "subctrl.h"
#include "display.h"
#include "multipag.h"
#include "bubble.h"
#include "w32.h"
#include "w32_gdi.h"
#include "w32_dd.h"
#include "w32_draw.h"
#include "w32_bar.h"
#include "w32_sub.h"
#include "w32_res.h"

/*
 *	O[o [N
 */
BOOL bFullScreen;							/* tXN[tO */
BOOL bFullScan;								/* tXL(Window) */
int nScalingSize;							/* n{\TCY */
#ifdef FILTERDLL
BOOL bAspectRatio;							/* AXyNg䒲tO */
double dAspectRatio;						/* AXyNg */
#endif
BOOL bPseudo400Line;						/* ^400CtO */
MONITORTYPE nMonitorType;					/* j^^Cv */
BOOL bRasterRendering;						/* X^_OtO */
BYTE bmpDrive[6][0x100];					/* Xe[^XhCuBMP */
BOOL bPaletteRefresh;						/* pbgXVtO */
BOOL bMultiPageRefresh;						/* }`y[WXVtO */
#if XM7_VER == 1
BYTE nBubbleCommand[2];						/* R}hWX^ */
WORD wBubblePageReg[2];						/* y[WWX^ */
WORD wBubbleCountReg[2];					/* y[WJEgWX^ */
#endif

/*
 *	X^eBbN [N
 */
static BOOL bNextFrameRender;				/* t[`tO */
static BOOL bDirtyLine[400];				/* vtO */

/*
 *	
 */
void FASTCALL InitDraw(void)
{
	/* [NGA */
	bFullScreen = FALSE;
	bFullScan = FALSE;
	nScalingSize = DISPLAYSIZE_100;			/* x1.0(640x400) */
#ifdef FILTERDLL
	bAspectRatio = FALSE;					/* AXyNg䒲tO */
	dAspectRatio = ASPECT_RATIO;			/* AXyNg */
#endif
	bPseudo400Line = FALSE;
	nMonitorType = MONITOR_COLOR;
	bRasterRendering = FALSE;
	bNextFrameRender = FALSE;
	memset(bDirtyLine, 0, sizeof(bDirtyLine));
	bPaletteRefresh = TRUE;
	bMultiPageRefresh = TRUE;
#if XM7_VER == 1
	memset(nBubbleCommand, 0, sizeof(nBubbleCommand));
	memset(wBubblePageReg, 0, sizeof(wBubblePageReg));
	memset(wBubbleCountReg, 0, sizeof(wBubbleCountReg));
#endif

	/* NGDIȂ̂ŁAGDI */
	InitGDI();
}

/*
 *	N[Abv
 */
void FASTCALL CleanDraw(void)
{
	/* tXN[tOɉāAN[Abv */
	if (bFullScreen) {
		CleanDD();
	}
	else {
		CleanGDI();
	}
}

/*
 *  ZNg
 */
BOOL FASTCALL SelectDraw(HWND hWnd)
{
	int i, j, k;
	BITMAPINFOHEADER *pbmi;
	BYTE *p;
	BYTE dat;
	const BYTE * const string = "0123HD";

	ASSERT(hWnd);

	/* NA */
	memset(bmpDrive, 0, sizeof(bmpDrive));

	/* hCuprbg}bv쐬 */
	for (i=0; i<(int)strlen(string); i++) {

		/* rbg}bvwb_ݒ */
		pbmi = (BITMAPINFOHEADER*)&bmpDrive[i][0];
		pbmi->biSize = sizeof(BITMAPINFOHEADER);
		pbmi->biWidth = 8;
		pbmi->biHeight = -8;
		pbmi->biPlanes = 1;
		pbmi->biBitCount = 24;
		pbmi->biCompression = BI_RGB;
		pbmi->biSizeImage = 8 * 8 * 3;

		/* AhXݒ */
		p = &bmpDrive[i][sizeof(BITMAPINFOHEADER)];

		for (j=0; j<8; j++) {
			/* CGROMf[^擾 */
			dat = subrom_c[string[i] * 8 + j];

			/* 1rbgƂɌ */
			for (k=0; k<8; k++) {
				if (dat & 0x80) {
					p[0] = 208;
					p[1] = 208;
					p[2] = 208;
				}
				dat <<= 1;
				p += 3;
			}
		}
	}

	/* NGDII */
	return SelectGDI(hWnd);
}

/*
 *	[h؂ւ
 */
void FASTCALL ModeDraw(HWND hWnd, BOOL bFullFlag)
{
	ASSERT(hWnd);

	/* ƈvĂ΁AςKvȂ */
	if (bFullFlag == bFullScreen) {
		return;
	}

	if (bFullFlag) {
		/* tXN[ */
		CleanGDI();
		InitDD();
		bFullScreen = TRUE;
		if (!SelectDD()) {
			/* tXN[s */
			bFullScreen = FALSE;
			CleanDD();
			InitGDI();
			SelectGDI(hWnd);
		}
	}
	else {
		/* EChE */
		bFullScreen = FALSE;
		CleanDD();
		InitGDI();
		if (!SelectGDI(hWnd)) {
			/* EChEs */
			CleanGDI();
			bFullScreen = TRUE;
			InitDD();
			SelectDD();
		}
	}

	/* ĕ`̂ߏ׍H */
	display_notify();
}

/*
 *	[hċN
 */
void FASTCALL ReDraw(HWND hWnd)
{
	ASSERT(hWnd);

	/* tXN[[hȂȂɂȂ */
	if (bFullScreen) {
		return;
	}

	CleanGDI();
	InitGDI();
	SelectGDI(hWnd);

	/* ĕ`̂ߏ׍H */
	display_notify();
}

/*
 *	`(ʏ)
 */
void FASTCALL OnDraw(HWND hWnd, HDC hDC)
{
	ASSERT(hWnd);
	ASSERT(hDC);

	/* G[Ȃ牽Ȃ */
	if (nErrorCode == 0) {
		if (bFullScreen) {
			if (bRasterRendering) {
				DrawPostRenderDD();
			}
			else {
				DrawDD();
			}
		}
		else {
			if (bRasterRendering) {
				DrawPostRenderGDI(hWnd, hDC);
			}
			else {
				DrawGDI(hWnd, hDC);
			}
		}
	}
}

/*
 *	`(WM_PAINT)
 */
void FASTCALL OnPaint(HWND hWnd)
{
	PAINTSTRUCT ps;
	HDC hDC;

	ASSERT(hWnd);

	hDC = BeginPaint(hWnd, &ps);

	/* G[Ȃ牽Ȃ */
	if (nErrorCode == 0) {
		/* ĕ`w */
		if (bFullScreen) {
			ReDrawDD();
		}
		else {
			ReDrawGDI();
		}

		/* ` */
		OnDraw(hWnd, hDC);
	}

	EndPaint(hWnd, &ps);
}

/*
 *	j[Jn
 */
void FASTCALL EnterMenu(HWND hWnd, BOOL flag)
{
	if (bFullScreen) {
		EnterMenuDD(hWnd, flag);
	}
	else {
		EnterMenuGDI(hWnd);
	}
}

/*
 *	j[I
 */
void FASTCALL ExitMenu(BOOL flag)
{
	if (bFullScreen) {
		ExitMenuDD(flag);
	}
	else {
		ExitMenuGDI();
	}
}

#ifdef FILTERDLL
/*
 *	O摜tB^p[^ݒ
 */
BOOL FASTCALL SetFilterParameter(WORD wParam)
{
	if (bFullScreen) {
		return FALSE;
	}
	else {
		return SetFilterParameterGDI(wParam);
	}
}
#endif

#if defined(ENGLISH_SUPPORT) || defined(RES_DLL)
/*
 *	ʃbZ[W[h
 */
void FASTCALL LoadMessageScreen(void)
{
	if (bFullScreen) {
		LoadMessageInitDD();
	}
	else {
		LoadMessageStatusBar();
	}
}
#endif

/*-[ VMƂ̐ڑ ]-----------------------------------------------------------*/

/*
 *	ĕ`惉X^ꊇݒ
 */
void FASTCALL SetDirtyFlag(int top, int bottom, BOOL flag)
{
	int y;

	if (bRasterRendering) {
		for (y = top; y < bottom; y++) {
			bDirtyLine[y] = flag;
		}
	}
}

/*
 *	VRAMݒʒm
 */
void FASTCALL vram_notify(WORD addr, BYTE dat)
{
	WORD y;

	UNUSED(dat);

#if XM7_VER >= 2
#if XM7_VER >= 3
	if (screen_mode == SCR_200LINE) {
#else
	if (!mode320) {
#endif
		if (vram_active != vram_display) {
			return;
		}
	}
#endif

	/* YWZo */
	if (bRasterRendering) {
#if XM7_VER >= 3
		switch (screen_mode) {
			case SCR_400LINE	:	addr &= 0x7fff;
									y = (WORD)(addr / 80);
									break;
			case SCR_262144		:
			case SCR_4096		:	addr &= 0x1fff;
									y = (WORD)(addr / 40);
									break;
			case SCR_200LINE	:	addr &= 0x3fff;
									y = (WORD)(addr / 80);
									break;
		}
#elif XM7_VER >= 2
		if (mode320) {
			addr &= 0x1fff;
			y = (WORD)(addr / 40);
		}
		else {
			addr &= 0x3fff;
			y = (WORD)(addr / 80);
		}
#elif defined(L4CARD)
		if (enable_400line) {
			addr -= vram_offset[0];
			addr &= 0x7fff;
			y = (WORD)(addr / 80);
		}
		else {
			addr &= 0x3fff;
			y = (WORD)(addr / 80);
		}
#else
		addr &= 0x3fff;
		y = (WORD)(addr / 80);
#endif
		if (y < 400) {
			bDirtyLine[y] = TRUE;
		}
	}

	if (bFullScreen) {
		VramDD(addr);
	}
	else {
		VramGDI(addr);
	}
}


/*
 *	TTLpbgʒm
 */
void FASTCALL ttlpalet_notify(void)
{
	/* pbgEChEXV */
	bPaletteRefresh = TRUE;

#if XM7_VER >= 2
#if XM7_VER >= 3
	if (screen_mode & SCR_ANALOG) {
#else
	if (mode320) {
#endif
		return;
	}
#endif

	if (bRasterRendering) {
		bNextFrameRender = TRUE;
		SetDirtyFlag(now_raster, 400, TRUE);
	}

	if (bFullScreen) {
		DigitalDD();
	}
	else {
		DigitalGDI();
	}
}

/*
 *	AiOpbgʒm
 */
#if XM7_VER >= 2
void FASTCALL apalet_notify(void)
{
	/* pbgEChEXV */
	bPaletteRefresh = TRUE;

#if XM7_VER >= 3
	if (!screen_mode == SCR_4096) {
#else
	if (!mode320) {
#endif
		return;
	}

	if (bRasterRendering) {
		bNextFrameRender = TRUE;
		SetDirtyFlag(now_raster, 400, TRUE);
	}

	if (bFullScreen) {
		AnalogDD();
	}
	else {
		AnalogGDI();
	}
}
#endif

/*
 *	}`y[WANeBuC[ύXʒm
 */
void FASTCALL multipage_notify(void)
{
	/* pbgEChE(}`y[W)XV */
	bMultiPageRefresh = TRUE;
}

/*
 *  ĕ`vʒm
 */
void FASTCALL display_notify(void)
{
	int i;
	int raster;

	if (bRasterRendering) {
		bNextFrameRender = TRUE;
		SetDirtyFlag(0, 400, TRUE);
		if (!run_flag) {
			raster = now_raster;
			for (i = 0; i < 400; i++) {
				now_raster = i;
				hblank_notify();
			}
			now_raster = raster;
		}
	}

	if (bFullScreen) {
		/* ReDraw͖ʂȃNA܂ނ̂ŁA׍H */
		AnalogDD();
	}
	else {
		AnalogGDI();
	}
}

/*
 *	VBLANKIvʒm
 */
void FASTCALL vblankperiod_notify(void)
{
	BOOL flag;
	int y;
	int ymax;

	if (bRasterRendering) {
		/* ̃t[Iɏ邩 */
		if (bNextFrameRender) {
			bNextFrameRender = FALSE;
			SetDirtyFlag(0, 400, TRUE);
		}
		else {
			/* Kv`FbN */
#if XM7_VER >= 3
			if (screen_mode == SCR_400LINE) {
				ymax = 400;
			}
			else {
				ymax = 200;
			}
#elif XM7_VER == 1 && defined(L4CARD)
			if (enable_400line) {
				ymax = 400;
			}
			else {
				ymax = 200;
			}
#else
			ymax = 200;
#endif
			flag = FALSE;
			for (y = 0; y < ymax; y++) {
				flag |= bDirtyLine[y];
			}
			if (!flag) {
				return;
			}
		}

		/* S̈_O */
		if (bFullScreen) {
			/* ReDraw͖ʂȃNA܂ނ̂ŁA׍H */
			AnalogDD();
		}
		else {
			AnalogGDI();
		}
	}
}

/*
 *	HBLANKvʒm
 */
void FASTCALL hblank_notify(void)
{
	/* VMĂȂꍇ͉Ȃ */
	if (!run_flag) {
		return;
	}

	if (bRasterRendering) {
		/* \͈͓(now_rasterbDirtyLine̔z͈͓)̏ꍇ̂ݕ` */
		if (now_raster < countof(bDirtyLine)) {
			if (bDirtyLine[now_raster]) {
				bDirtyLine[now_raster] = FALSE;
				if (bFullScreen) {
					DrawRasterDD(now_raster);
				}
				else {
					DrawRasterGDI(now_raster);
				}
			}
		}
	}
}

/*
 *	fBW^CYvʒm
 */
void FASTCALL digitize_notify(void)
{
}

#if XM7_VER >= 3
/*
 *	n[hEFAEBhEʒm
 */
void FASTCALL window_notify(void)
{
	if (bFullScreen) {
		WindowDD();
	}
	else {
		WindowGDI();
	}
}
#endif

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	eLXgVRAMݒʒm
 */
void FASTCALL tvram_notify(WORD addr, BYTE dat)
{
	WORD ysize;
	WORD y;

	UNUSED(dat);

	if (bRasterRendering) {
		/* YWZo */
		addr = (WORD)((addr - text_start_addr) & 0x0ffe);
		ysize = (WORD)((crtc_register[9] & 0x1f) + 1);
		if (width40_flag) {
			y = (WORD)((addr / 80) * ysize);
		}
		else {
			y = (WORD)((addr / 160) * ysize);
		}

		/* I[o[`FbN */
		if (y >= 400) {
			return;
		}

		/* ĕ`tOݒ */
		SetDirtyFlag(y, y + ysize, TRUE);
	}

	if (bFullScreen) {
		TvramDD(addr);
	}
	else {
		TvramGDI(addr);
	}
}

/*
 *  eLXgĕ`vʒm
 */
void FASTCALL tvram_redraw_notify(void)
{
	bNextFrameRender = TRUE;
	SetDirtyFlag(now_raster, 400, TRUE);

	if (bFullScreen) {
		ReDrawTVRamDD();
	}
	else {
		ReDrawTVRamGDI();
	}
}
#endif

#if XM7_VER == 1
/*
 *	ouRg[vʒm
 */
void FASTCALL bubble_notify(int type)
{
	nBubbleCommand[type] = bmc_command[type];
	wBubblePageReg[type] = bmc_pagereg[type];
	wBubbleCountReg[type] = bmc_countreg[type];
}
#endif

/*-[ V2 ]-------------------------------------------------------------------*/

#if XM7_VER == 2 && defined(FMTV151)
/*
 *	V2
 *	0=߁A1=΁A2=
 */
const BYTE nV2data[] = {
/*						  1 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 */
/*		1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 */
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,1,1,1,1,1,1,1,2,0,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,1,1,1,1,1,1,1,1,1,2,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,1,1,1,1,2,0,0,1,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,1,1,1,2,0,0,0,0,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,0,0,0,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,0,0,1,1,1,1,2,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,0,1,1,1,1,2,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,0,1,1,1,1,2,0,0,
		1,1,1,2,0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,1,1,1,1,2,0,0,0,
		1,1,1,1,2,0,0,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,2,0,0,0,0,
		0,1,1,1,1,2,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,2,0,0,0,0,0,
		0,0,1,1,1,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,2,0,0,0,0,0,0,
		0,0,0,1,1,1,1,1,2,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,2,
		0,0,0,0,1,1,1,2,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,1,1,2,
};

#endif

/*-[ 2HD ]------------------------------------------------------------------*/

#if XM7_VER == 1 && defined(SFDC)
/*
 *	2HD
 */
const BYTE bmp2HD[] = {
		0x28, 0x00, 0x00, 0x00, 0x0F, 0x00, 0x00, 0x00,
		0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x20, 0x01, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
};
#endif	/* XM7_VER == 1 && defined(SFDC) */

/*-[ BBL ]------------------------------------------------------------------*/

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	Unit Number
 */
const BYTE bmpUnit[][0xa0] = {
	{	0x28, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00,
		0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	},
	{	0x28, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00,
		0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	},
	{	0x28, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00,
		0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	},
	{	0x28, 0x00, 0x00, 0x00, 0x06, 0x00, 0x00, 0x00,
		0x06, 0x00, 0x00, 0x00, 0x01, 0x00, 0x18, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x78, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
		0xFF, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00,
		0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0xFF, 0x00,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,
		0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
		0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	},
};
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

#endif	/* _WIN32 */
