/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API t@CI/O ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>
#include <io.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include "xm7.h"
#include "device.h"

/*
 *	t@C[h(ROMp)
 */
BOOL FASTCALL file_load(char *fname, BYTE *buf, int size)
{
	char path[_MAX_PATH];
	char drvname[_MAX_DRIVE];
	char dirname[_MAX_DIR];
	char filename[_MAX_FNAME];
	char extname[_MAX_EXT];

	int handle;

	/* assert */
	ASSERT(fname);
	ASSERT(buf);
	ASSERT(size > 0);

	/* t@CƊgqo */
	_splitpath(fname, NULL, NULL, filename, extname);

	/* XM7̃pX𓾂 */
	memset(path, 0, sizeof(path));
	GetModuleFileName(NULL, path, sizeof(path));
	_splitpath(path, drvname, dirname, NULL, NULL);

	/* pX */
	_makepath(path, drvname, dirname, filename, extname);

	handle = open(path, O_BINARY | O_RDONLY);
	if (handle == -1) {
		return FALSE;
	}

	if (read(handle, buf, size) != size) {
		close(handle);
		return FALSE;
	}

	close(handle);
	return TRUE;
}

/*
 *	t@CZ[u(wKRAMp)
 */
BOOL FASTCALL file_save(char *fname, BYTE *buf, int size)
{
	char path[_MAX_PATH];
	char drvname[_MAX_DRIVE];
	char dirname[_MAX_DIR];
	char filename[_MAX_FNAME];
	char extname[_MAX_EXT];

	int handle;

	/* assert */
	ASSERT(fname);
	ASSERT(buf);
	ASSERT(size > 0);

	/* t@CƊgqo */
	_splitpath(fname, NULL, NULL, filename, extname);

	/* XM7̃pX𓾂 */
	memset(path, 0, sizeof(path));
	GetModuleFileName(NULL, path, sizeof(path));
	_splitpath(path, drvname, dirname, NULL, NULL);

	/* pX */
	_makepath(path, drvname, dirname, filename, extname);

	handle = open(path, O_BINARY | O_CREAT | O_WRONLY, S_IWRITE);
	if (handle == -1) {
		return FALSE;
	}

	if (write(handle, buf, size) != size) {
		close(handle);
		return FALSE;
	}

	close(handle);
	return TRUE;
}

/*
 *	t@Cl[(Xe[gp)
 */
BOOL FASTCALL file_rename_state(char *fname)
{
	int i, handle;
	char path[_MAX_PATH];
	char drvname[_MAX_DRIVE];
	char dirname[_MAX_DIR];
	char filename[_MAX_FNAME];
	char extname[_MAX_EXT];
	char number[16];

	/* assert */
	ASSERT(fname);

	/* t@C݃`FbN */
	handle = open(fname, O_BINARY | O_RDONLY);
	if (handle == -1) {
		return FALSE;
	}
	close(handle);

#if 1
	/* t@C̑݊mF */
	for (i = 0; i < 99; i ++) {
		/* [pX𕪉 */
		_splitpath(fname, drvname, dirname, filename, extname);

		/* t@Cɔԍt */
		_snprintf(number, sizeof(number), "_%02d", i);
		strncat(filename, number, sizeof(filename)-strlen(filename)-1);

		/* pX */
		_makepath(path, drvname, dirname, filename, extname);

		/* t@C݃`FbN */
		handle = open(path, O_BINARY | O_RDONLY);
		if (handle == -1) {
			break;
		}
		close(handle);
	}
#else
	/* [pX𕪉 */
	_splitpath(fname, drvname, dirname, filename, NULL);

	/* pX */
	_makepath(path, drvname, dirname, filename, ".BAK");
#endif

	/* t@C폜 */
	remove(path);

	/* t@Cl[ */
	rename(fname, path);

	return TRUE;
}

/*
 *	t@CI[v
 */
int FASTCALL file_open(char *fname, int mode)
{
	/* assert */
	ASSERT(fname);

	switch (mode) {
		case OPEN_R:
			return open(fname, O_BINARY | O_RDONLY);
		case OPEN_W:
			return open(fname, O_BINARY | O_CREAT | O_TRUNC | O_WRONLY,
						S_IWRITE);
		case OPEN_RW:
			/* CD-ROM̓ǂݍ݂RWĂ܂ */
			if (access(fname, 0x06) != 0) {
				return -1;
			}
			return open(fname, O_BINARY | O_RDWR);

	}

	ASSERT(FALSE);
	return -1;
}

/*
 *	t@CN[Y
 */
void FASTCALL file_close(int handle)
{
	/* assert */
	ASSERT(handle >= 0);

	close(handle);
}

/*
 *	t@CTCY擾
 */
DWORD FASTCALL file_getsize(int handle)
{
	long now;
	long end;

	/* assert */
	ASSERT(handle >= 0);

	now = tell(handle);
	if (now == -1) {
		return 0;
	}

	end = lseek(handle, 0L, SEEK_END);
	lseek(handle, now, SEEK_SET);

	return end;
}

/*
 *	t@CV[N
 */
BOOL FASTCALL file_seek(int handle, DWORD offset)
{
	long now;

	/* assert */
	ASSERT(handle >= 0);

	now = lseek(handle, offset, SEEK_SET);
	if (now != (long)offset) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	t@Cǂݏo
 */
BOOL FASTCALL file_read(int handle, BYTE *ptr, DWORD size)
{
	unsigned int cnt;

	/* assert */
	ASSERT(handle >= 0);
	ASSERT(ptr);
	ASSERT(size > 0);

	cnt = read(handle, ptr, size);
	if (cnt != size) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	t@C
 */
BOOL FASTCALL file_write(int handle, BYTE *ptr, DWORD size)
{
	unsigned int cnt;

	/* assert */
	ASSERT(handle >= 0);
	ASSERT(ptr);
	ASSERT(size > 0);

	cnt = write(handle, ptr, size);
	if (cnt != size) {
		return FALSE;
	}

	return TRUE;
}

#ifdef DEBUG
/*
 *	t@C|C^擾
 */
long FASTCALL file_getptr(int handle)
{
	/* assert */
	ASSERT(handle >= 0);

	return tell(handle);
}
#endif

#endif	/* _WIN32 */
