/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API EChE\ ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>
#include <stdlib.h>
#include "xm7.h"
#include "subctrl.h"
#include "display.h"
#include "multipag.h"
#include "ttlpalet.h"
#include "apalet.h"
#include "mouse.h"
#include "w32.h"
#include "w32_draw.h"
#include "w32_gdi.h"
#include "w32_kbd.h"
#include "w32_sub.h"
#ifdef FILTERDLL
#include "filter.h"
#endif

/*
 *	O[o [N
 */
#ifdef __cplusplus
extern "C" {
#endif
DWORD rgbTTLGDI[16];						/* fW^pbg */
DWORD rgbAnalogGDI[4096];					/* AiOpbg */
BYTE *pBitsGDI;								/* rbgf[^ */
#ifdef FILTERDLL
BYTE *pBitsGDI32;							/* rbgf[^ */
#endif
BYTE GDIDrawFlag[4000];						/* 8x8 ĕ`̈tO */
#if XM7_VER == 1 && defined(L4CARD)
DWORD rgbTTLL4GDI[32];						/* fW^pbg (L4) */
#endif
#ifdef FILTERDLL
BOOL bFilterDll;							/* O摜tB^DLLFtO */
BOOL bUseFilterDll;							/* O摜tB^DLLgptO */
char szFilterDll[_MAX_FNAME];				/* O摜tB^DLLt@C */
DWORD dwFilterDllFlg;						/* O摜tB^DLLOtO */
DWORD dwFilterDllDepFlg;					/* O摜tB^DLLtO */
#endif
#ifdef __cplusplus
}
#endif

/*
 *	X^eBbN [N
 */
#if XM7_VER >= 3
static BYTE nMode;							/* ʃ[h */
#elif XM7_VER >= 2
static BOOL bAnalog;						/* AiO[htO */
#elif XM7_VER == 1 && defined(L4CARD)
static BOOL b400Line;						/* 400C[htO */
#endif
static BYTE nNowBpp;						/* ݂̃rbg[x */
static BOOL bMouseCursor;					/* }EXJ[\tO */
static HBITMAP hBitmap;						/* rbg}bv nh */
#ifdef FILTERDLL
static HBITMAP hBitmap32;					/* rbg}bv nh */
#endif
static WORD nDrawTop;						/* `͈͏ */
static WORD nDrawBottom;					/* `͈͉ */
static WORD nDrawLeft;						/* `͈͍ */
static WORD nDrawRight;						/* `͈͉E */
static BOOL bPaletFlag;						/* pbgύXtO */
static BOOL bClearFlag;						/* NAtO */
static BOOL bSelectCancelFlag;				/* ZNgLZtO */
#if XM7_VER >= 3
static BOOL bWindowOpen;					/* n[hEFAEBhE */
static WORD nWindowDx1;						/* EBhEXW */
static WORD nWindowDy1;						/* EBhEYW */
static WORD nWindowDx2;						/* EBhEEXW */
static WORD nWindowDy2;						/* EBhEEYW */
#endif
#ifdef FILTERDLL
static CFILTER	*pFilterDll = NULL;
static LPVOID	pFilter = NULL;
#endif

/*
 *	AZu֐̂߂̃vg^Cv錾
 */
#ifdef __cplusplus
extern "C" {
#endif
#if XM7_VER >= 3
extern void Render640GDI2(int first, int last, int pitch, int scale);
extern void Render640wGDI2(int first, int last, int firstx, int lastx, int pitch, int scale);
extern void Render320GDI(int first, int last);
extern void Render320wGDI(int first, int last, int firstx, int lastx);
extern void Render320GDI32bpp(int first, int last);
extern void Render320wGDI32bpp(int first, int last, int firstx, int lastx);
extern void Render256kGDI(int first, int last, int multipage);
extern void Render640cGDI(int first, int last);
extern void Render640cwGDI(int first, int last, int firstx, int lastx);
#else
extern void Render640GDI(int first, int last);
extern void Render320GDI(int first, int last);
extern void RenderL4GDI(int first, int last);
extern void Render640mGDI(int first, int last);
extern void Render640cGDI(int first, int last);
#endif
#ifdef __cplusplus
}
#endif

/*
 *	vg^Cv錾
 */
static void FASTCALL SetDrawFlag(BOOL flag);

/*
 *	
 */
void FASTCALL InitGDI(void)
{
	/* [NGA */
#if XM7_VER >= 3
	nMode = SCR_200LINE;
#elif XM7_VER >= 2
	bAnalog = FALSE;
#elif XM7_VER == 1 && defined(L4CARD)
	b400Line = FALSE;
#endif
	nNowBpp = 4;
	bMouseCursor = TRUE;
	hBitmap = NULL;
	pBitsGDI = NULL;
#ifdef FILTERDLL
	hBitmap32 = NULL;
	pBitsGDI32 = NULL;
#endif
	memset(rgbTTLGDI, 0, sizeof(rgbTTLGDI));
#if XM7_VER >= 2
	memset(rgbAnalogGDI, 0, sizeof(rgbAnalogGDI));
#elif XM7_VER == 1 && defined(L4CARD)
	memset(rgbTTLL4GDI, 0, sizeof(rgbTTLL4GDI));
#endif

	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = FALSE;
	bSelectCancelFlag = FALSE;
	SetDrawFlag(FALSE);

#if XM7_VER >= 3
	bWindowOpen = FALSE;
	nWindowDx1 = 640;
	nWindowDy1 = 400;
	nWindowDx2 = 0;
	nWindowDy2 = 0;
#endif

#ifdef FILTERDLL
	if (pFilterDll) {
		if (pFilter) {
			pFilterDll->Release(pFilter);
		}
		delete pFilterDll;
	}
	pFilter = NULL;

	pFilterDll = new CFILTER(szFilterDll);
	bFilterDll = pFilterDll->IsDll();

	if (bFilterDll && bUseFilterDll) {
		DWORD	dwNeed;
		DWORD	dwCaps = pFilterDll->GetCaps(pFilter);

		if (dwCaps & FILTER_FLAG_DRAW_HDC) {
			dwNeed = (DWORD)GetDC(hDrawWnd);
		}
		else {
			dwNeed = (DWORD)hDrawWnd;
		}

		/* C^[XC̖邳x/255ʃ[h */
		if (!pFilterDll->Create(&pFilter, dwNeed, 640, 400,
							dwFilterDllDepFlg << 16 | dwFilterDllFlg)) {
			if (pFilter) {
				pFilterDll->Release(pFilter);
				pFilter = NULL;
			}
			delete pFilterDll;
			pFilterDll = NULL;
		}

		if (dwCaps & FILTER_FLAG_DRAW_HDC) {
			ReleaseDC(hDrawWnd, (HDC)dwNeed);
		}
	}
	else {
		delete pFilterDll;
		pFilterDll = NULL;
	}
#endif
}

/*
 *	N[Abv
 */
void FASTCALL CleanGDI(void)
{
	if (hBitmap) {
		DeleteObject(hBitmap);
		hBitmap = NULL;
		pBitsGDI = NULL;
	}
#ifdef FILTERDLL
	if (hBitmap32) {
		DeleteObject(hBitmap32);
		hBitmap32= NULL;
		pBitsGDI32 = NULL;
	}

	if (pFilterDll) {
		if (pFilter) {
			pFilterDll->Release(pFilter);
		}
		delete pFilterDll;
		pFilterDll = NULL;
	}
	pFilter = NULL;
#endif
}

/*
 *	GDIZNg
 */
static BOOL FASTCALL SelectSub(HWND hWnd)
{
	BITMAPINFOHEADER *pbmi;
	HDC hDC;
#ifdef FILTERDLL
	BOOL	bFlag = FALSE;
	LPDWORD	lpdw;
#endif

	ASSERT(hWnd);

	/* DIBZNVɑ݂ꍇAj */
	if (hBitmap) {
		DeleteObject(hBitmap);
		hBitmap = NULL;
		pBitsGDI = NULL;
	}

#ifdef FILTERDLL
	if (hBitmap32) {
		DeleteObject(hBitmap32);
		hBitmap32 = NULL;
		pBitsGDI32 = NULL;
	}
#endif

	/* m */
	pbmi = (BITMAPINFOHEADER*)malloc(sizeof(BITMAPINFOHEADER) +
									 sizeof(RGBQUAD) * 256);
	if (!pbmi) {
		return FALSE;
	}
	memset(pbmi, 0, sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * 256);

	/* rbg}bv쐬 */
	pbmi->biSize = sizeof(BITMAPINFOHEADER);
	pbmi->biWidth = 640;
	pbmi->biHeight = -400;
	pbmi->biPlanes = 1;
	pbmi->biBitCount = nNowBpp;
	pbmi->biCompression = BI_RGB;

	/* DC擾ADIBZNV쐬 */
	hDC = GetDC(hWnd);
	hBitmap = CreateDIBSection(hDC, (BITMAPINFO*)pbmi, DIB_RGB_COLORS,
								(void**)&pBitsGDI, NULL, 0);
#ifdef FILTERDLL
	if (pFilter) {
		/* GdiFlush();ꂽقƂb */
		/* dll֏ރtH[}bgƓDIB쐬 */
		pbmi->biBitCount = (WORD)(pFilterDll->GetCaps(pFilter) &
						FILTER_FLAG_PIXEL_MASK) >> FILTER_FLAG_PIXEL_SHIFT;

		/* Tvł8̓T|[gO */
		switch (pbmi->biBitCount) {
			case 15:
				pbmi->biCompression   = BI_BITFIELDS;
				lpdw = (LPDWORD)(pbmi + 1);
				*(lpdw + 0) = 0x7c00;	/* R5 */
				*(lpdw + 1) = 0x03e0;	/* G5 */
				*(lpdw + 2) = 0x001f;	/* B5 */
				bFlag = TRUE;
				break;
			case 16:
				pbmi->biCompression   = BI_BITFIELDS;
				lpdw = (LPDWORD)(pbmi + 1);
				*(lpdw + 0) = 0xf800;	/* R5 */
				*(lpdw + 1) = 0x07e0;	/* G6 */
				*(lpdw + 2) = 0x001f;	/* B5 */
				bFlag = TRUE;
				break;
			case 24:
			case 32:
				bFlag = TRUE;
				break;
			default:
				break;
		}

		/* FILTER_FLAG_RENDER_HDCȂKv */
		if (bFlag) {
			/* OŃtH[}bgϊ̖ʓ|Ȃ̂APIɔC */
			hBitmap32 = CreateDIBSection(hDC, (BITMAPINFO*)pbmi, DIB_RGB_COLORS,
										(void**)&pBitsGDI32, NULL, 0);
		}
	}
#endif
	ReleaseDC(hWnd, hDC);
	free(pbmi);
	if (!hBitmap) {
		return FALSE;
	}
#ifdef FILTERDLL
	if (bFlag && !hBitmap32) {
		return FALSE;
	}
#endif

	/* SGAAxNA */
	if (!bRasterRendering) {
		memset(pBitsGDI, 0, (640 * 400 * nNowBpp) >> 3);
		bClearFlag = TRUE;
	}

	return TRUE;
}

/*
 *	SĂ̍ĕ`tOݒ
 */
static void FASTCALL SetDrawFlag(BOOL flag)
{
	memset(GDIDrawFlag, (BYTE)flag, sizeof(GDIDrawFlag));
}

/*
 *  640x200AfW^[h
 *	ZNg
 */
static BOOL FASTCALL Select640(HWND hWnd)
{
	if ((!pBitsGDI) || (nNowBpp != 4)) {
		nNowBpp = 4;
		if (!SelectSub(hWnd)) {
			return FALSE;
		}
	}

	/* S̈斳 */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	SetDrawFlag(TRUE);

#if XM7_VER >= 3
	/* fW^/200C[h */
	nMode = SCR_200LINE;
#elif XM7_VER >= 2
	/* fW^[h */
	bAnalog = FALSE;
#elif XM7_VER == 1 && defined(L4CARD)
	/* 200C[h */
	b400Line = FALSE;
#endif

	return TRUE;
}

#if XM7_VER == 1 && defined(L4CARD)
/*
 *  640x400APF[h
 *	ZNg
 */
static BOOL FASTCALL SelectL4(HWND hWnd)
{
	if ((!pBitsGDI) || (nNowBpp != 8)) {
		nNowBpp = 8;
		if (!SelectSub(hWnd)) {
			return FALSE;
		}
	}

	/* S̈斳 */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	SetDrawFlag(TRUE);

	/* 400C[h */
	b400Line = TRUE;

	return TRUE;
}
#endif

#if XM7_VER >= 3
/*
 *  640x400AfW^[h
 *	ZNg
 */
static BOOL FASTCALL Select400l(HWND hWnd)
{
	if ((!pBitsGDI) || (nNowBpp != 4)) {
		nNowBpp = 4;
		if (!SelectSub(hWnd)) {
			return FALSE;
		}
	}

	/* S̈斳 */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	SetDrawFlag(TRUE);

	/* fW^/400C[h */
	nMode = SCR_400LINE;

	return TRUE;
}
#endif

#if XM7_VER >= 2
/*
 *  320x200AAiO[h
 *	ZNg
 */
static BOOL FASTCALL Select320(HWND hWnd)
{
#if XM7_VER == 2
	if ((!pBitsGDI) || (nNowBpp != 16)) {
		nNowBpp = 16;
		if (!SelectSub(hWnd)) {
			return FALSE;
		}
	}
#else
	if (bRasterRendering) {
		if ((!pBitsGDI) || (nNowBpp != 32)) {
			nNowBpp = 32;
			if (!SelectSub(hWnd)) {
				return FALSE;
			}
		}
	}
	else {
		if ((!pBitsGDI) || (nNowBpp != 16)) {
			nNowBpp = 16;
			if (!SelectSub(hWnd)) {
				return FALSE;
			}
		}
	}
#endif

	/* S̈斳 */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	SetDrawFlag(TRUE);

#if XM7_VER >= 3
	/* AiO/200C[h */
	nMode = SCR_4096;
#else
	/* AiO[h */
	bAnalog = TRUE;
#endif

	return TRUE;
}
#endif

#if XM7_VER >= 3
/*
 *  320x200A26F[h
 *	ZNg
 */
static BOOL FASTCALL Select256k(HWND hWnd)
{
	if ((!pBitsGDI) || (nNowBpp != 32)) {
		nNowBpp = 32;
		if (!SelectSub(hWnd)) {
			return FALSE;
		}
	}

	/* S̈斳 */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	SetDrawFlag(TRUE);

	/* AiO(26F)/200C[h */
	nMode = SCR_262144;

	return TRUE;
}
#endif

/*
 *	ZNg`FbN
 */
static BOOL FASTCALL SelectCheck(void)
{
	/* ZNgLZ */
	if (bSelectCancelFlag) {
		bSelectCancelFlag = FALSE;
		return FALSE;
	}

#if XM7_VER >= 3
	/* Ȃ蔲(H */
	if (nMode == screen_mode) {
		return TRUE;
	}
	else {
		return FALSE;
	}
#elif XM7_VER >= 2
	/* 320x200 */
	if (mode320) {
		if (bAnalog) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}

	/* 640x200 */
	if (!bAnalog) {
		return TRUE;
	}
	else {
		return FALSE;
	}
#elif defined(L4CARD)
	/* 640x400 */
	if (enable_400line && enable_400linecard) {
		if (b400Line) {
			return TRUE;
		}
		else {
			return FALSE;
		}
	}

	/* 640x200 */
	if (!b400Line) {
		return TRUE;
	}
	else {
		return FALSE;
	}
#else
	return TRUE;
#endif
}

/*
 *	ZNg
 */
BOOL FASTCALL SelectGDI(HWND hWnd)
{
	ASSERT(hWnd);

	/* Ȃ */
	if (!pBitsGDI) {
#if XM7_VER >= 3
		switch (screen_mode) {
			case SCR_400LINE	:	return Select400l(hWnd);
			case SCR_262144		:	return Select256k(hWnd);
			case SCR_4096		:	return Select320(hWnd);
			case SCR_200LINE	:	return Select640(hWnd);
		}
#elif XM7_VER >= 2
		if (mode320) {
			return Select320(hWnd);
		}
		return Select640(hWnd);
#elif defined(L4CARD)
		if (enable_400line && enable_400linecard) {
			return SelectL4(hWnd);
		}
		return Select640(hWnd);
#else
		return Select640(hWnd);
#endif
	}

	/* vĂ邩`FbN */
	if (SelectCheck()) {
		return TRUE;
	}

	/* ZNg */
#if XM7_VER >= 3
	switch (screen_mode) {
		case SCR_400LINE	:	return Select400l(hWnd);
		case SCR_262144		:	return Select256k(hWnd);
		case SCR_4096		:	return Select320(hWnd);
		default				:	return Select640(hWnd);
	}
#else
#if XM7_VER >= 2
	if (mode320) {
		return Select320(hWnd);
	}
#elif defined(L4CARD)
	if (enable_400line && enable_400linecard) {
		return SelectL4(hWnd);
	}
#endif
	return Select640(hWnd);
#endif
}

/*-[ ` ]-----------------------------------------------------------------*/

#if ((XM7_VER == 2) && defined(FMTV151))
/*
 *	V2 (4bpp)
 */
static void FASTCALL DrawV2_4bpp(void)
{
	BYTE	x, y;
	int		xx, yy;
	WORD	col;
	BYTE	*pbits;

	if (!bFMTV151) {
		return;
	}

	if (!bRasterRendering) {
		if ((nDrawBottom < (V2YPoint * 2)) ||
			(nDrawTop > ((V2YPoint + V2YSize * V2YPxSz) * 2 - 1)) ||
			(nDrawRight < V2XPoint) ||
			(nDrawLeft > (V2XPoint + V2XSize * V2XPxSz - 1))) {
			return;
		}
	}

	pbits = (BYTE *)(pBitsGDI + V2YPoint * 640 + (V2XPoint >> 1));

	for (y = 0; y < V2YSize; y ++) {
		for (x = 0; x < V2XSize; x ++) {
			if (nV2data[y * V2XSize + x]) {
				col = (WORD)(10 - nV2data[y * V2XSize + x]);
				for (yy = y * V2YPxSz; yy < (y + 1) * V2YPxSz; yy ++) {
					for (xx = x * V2XPxSz; xx < (x + 1) * V2XPxSz; xx ++) {
						if (xx & 1) {
							pbits[yy * 640 + (xx >> 1)] &= (BYTE)0xf0;
							pbits[yy * 640 + (xx >> 1)] |= col;
							if (bPseudo400Line) {
								pbits[yy * 640+320+(xx >> 1)] &= (BYTE)0xf0;
								pbits[yy * 640+320+(xx >> 1)] |= col;
							}
						}
						else {
							pbits[yy * 640 + (xx >> 1)] &= (BYTE)0x0f;
							pbits[yy * 640 + (xx >> 1)] |= (BYTE)(col << 4);
							if (bPseudo400Line) {
								pbits[yy*640+320+(xx>>1)] &= (BYTE)0x0f;
								pbits[yy*640+320+(xx>>1)] |= (BYTE)(col << 4);
							}
						}
					}
				}
			}
		}
	}
}

/*
 *	V2 (16bpp)
 */
static void FASTCALL DrawV2_16bpp(void)
{
	BYTE	x, y;
	int		xx, yy;
	WORD	col;
	WORD	*pbits;

	/* pbge[u */
	static const WORD V2rgbTable[] = {
		0x0000 | 0x0000 | 0x0000,
		0x0000 | 0x03e0 | 0x0000,
	};

	if (!bFMTV151) {
		return;
	}

	if (!bRasterRendering) {
		if ((nDrawBottom < (V2YPoint * 2)) ||
			(nDrawTop > ((V2YPoint + V2YSize * V2YPxSz) * 2 - 1)) ||
			(nDrawRight < V2XPoint) ||
			(nDrawLeft > (V2XPoint + V2XSize * V2XPxSz - 1))) {
			return;
		}
	}

	pbits = (WORD *)(pBitsGDI + V2YPoint * 2560 + V2XPoint * 2);

	for (y = 0; y < V2YSize; y ++) {
		for (x = 0; x < V2XSize; x ++) {
			if (nV2data[y * V2XSize + x]) {
				col = V2rgbTable[nV2data[y * V2XSize + x]];
				for (yy = y * V2YPxSz; yy < (y + 1) * V2YPxSz; yy ++) {
					for (xx = x * V2XPxSz; xx < (x + 1) * V2XPxSz; xx ++) {
						pbits[yy * 1280 + xx] = col;
					}
				}
			}
		}
	}
}
#endif

/*
 *	I[NA
 */
static void FASTCALL AllClear(void)
{
	/* ׂăNA */
	memset(pBitsGDI, 0, (640 * 400 * nNowBpp) >> 3);

	/* S̈_OΏۂƂ */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	SetDrawFlag(TRUE);

	bClearFlag = FALSE;
}

/*
 *	tXL
 */
static void FASTCALL RenderFullScan(void)
{
	BYTE *p;
	BYTE *q;
	WORD u;
	WORD top, bottom;

	/* X^Pʃ_OɉʑŜKv̂ */
	if (nDrawTop >= nDrawBottom) {
		top = 0;
		bottom = 400;
	}
	else {
		top = nDrawTop;
		bottom = nDrawBottom;
	}

	/* |C^ */
	p = &pBitsGDI[(top * 640 * nNowBpp) >> 3];
	q = p + ((640 * nNowBpp) >> 3);

	/* [v */
	for (u=top; u<bottom; u += (WORD)2) {
		memcpy(q, p, (640 * nNowBpp) >> 3);
		p += ((640 * 2 * nNowBpp) >> 3);
		q += ((640 * 2 * nNowBpp) >> 3);
	}
}

/*
 *	Cݒ (fW^[h)
 */
static void FASTCALL RenderSetOddLineDigital(void)
{
	BYTE *p;
	WORD u;
	WORD top, bottom;

	/* X^Pʃ_OɉʑŜKv̂ */
	if (nDrawTop >= nDrawBottom) {
		top = 0;
		bottom = 400;
	}
	else {
		top = nDrawTop;
		bottom = nDrawBottom;
	}

	/* |C^ */
	p = &pBitsGDI[((top + 1) * 640 * nNowBpp) >> 3];

	/* [v */
	for (u=top; u<bottom; u += (WORD)2) {
		memset(p, 0x88, (640 * nNowBpp) >> 3);
		p += ((640 * 2 * nNowBpp) >> 3);
	}
}

/*
 *	Cݒ (AiO[h)
 */
#if XM7_VER >= 2
static void FASTCALL RenderSetOddLineAnalog(void)
{
	BYTE *p;
	WORD u;
	WORD top, bottom;

	/* X^Pʃ_OɉʑŜKv̂ */
	if (nDrawTop >= nDrawBottom) {
		top = 0;
		bottom = 400;
	}
	else {
		top = nDrawTop;
		bottom = nDrawBottom;
	}

	/* |C^ */
	p = &pBitsGDI[((top + 1) * 640 * nNowBpp) >> 3];

	/* [v */
	for (u=top; u<bottom; u += (WORD)2) {
		memset(p, 0x00, (640 * nNowBpp) >> 3);
		p += ((640 * 2 * nNowBpp) >> 3);
	}
}
#endif

/*
 *	640x200/400AfW^[h
 *	pbgݒ
 */
static void FASTCALL Palet640(void)
{
	int i;
	int vpage;
	BYTE col;
	int base;

	/* pbge[u */
	static DWORD rgbTable[] = {
		/* J[j^ */
		0x00000000,
		0x000000ff,
		0x00ff0000,
		0x00ff00ff,
		0x0000ff00,
		0x0000ffff,
		0x00ffff00,
		0x00ffffff,
		/* O[j^ */
		0x00000000,
		0x001f4033,
		0x003a785f,
		0x0041876a,
		0x0063cca1,
		0x0067d5a7,
		0x0070e7b6,
		0x0073eebb,
		/* Ao[CG[j^ */
		0x001e0000,
		0x00562314,
		0x00894327,
		0x00944a2b,
		0x00d17141,
		0x00d97643,
		0x00e87f49,
		0x00ee834b,
	};

	/* CfbNXx[Xݒ */
#if XM7_VER >= 2
	if ((nMonitorType == MONITOR_GREEN) &&
		(model_type < MODEL_FM77AV)) {
#else
	if (nMonitorType == MONITOR_GREEN) {
#endif
		base = 8;
	}
#if XM7_VER >= 2
	else if ((nMonitorType == MONITOR_AMBERYELLOW) &&
			 (model_type < MODEL_FM77AV)) {
#else
	else if (nMonitorType == MONITOR_AMBERYELLOW) {
#endif
		base = 16;
	}
	else {
		base = 0;
	}

	/* }`y[WA\v[𓾂 */
	vpage = (~(multi_page >> 4)) & 0x07;

	/* 640x200/400AfW^pbg */
	for (i=0; i<8; i++) {
		if (crt_flag) {
			/* CRT ON */
#if XM7_VER == 1
			if ((model_type == MODEL_FM8) && !ttl_palet_en) {
				col = (BYTE)i;
			}
			else {
				col = (BYTE)(ttl_palet[i & vpage] & 0x07);
			}
#else
			col = (BYTE)(ttl_palet[i & vpage] & 0x07);
#endif

			if (bPseudo400Line) {
				if (col & 2) {
					rgbTTLGDI[i + 8] = rgbTable[base + 7];
				}
				else {
					rgbTTLGDI[i + 8] = rgbTable[base + 0];
				}
				if (col & 4) {
					rgbTTLGDI[i] = rgbTable[base + 7];
				}
				else {
					rgbTTLGDI[i] = rgbTable[base + 0];
				}
			}
			else {
				rgbTTLGDI[i] = rgbTable[base + col];
			}
		}
		else {
			/* CRT OFF */
			rgbTTLGDI[i] = rgbTable[base];
		}
	}

	/* FMTV-151 V2`lR[p */
#if XM7_VER == 2 && defined(FMTV151)
	rgbTTLGDI[8] = rgbTable[0];
	rgbTTLGDI[9] = rgbTable[4];
#endif
}

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	640x400APF[h
 *	pbgݒ
 */
static void FASTCALL PaletL4(void)
{
	int i;
	int base;

	/* pbge[u(16F) */
	static const DWORD rgbTable16[] = {
		/* J[j^ */
		0x00000000,
		0x000000bb,
		0x00bb0000,
		0x00bb00bb,
		0x0000bb00,
		0x0000bbbb,
		0x00bbbb00,
		0x00bbbbbb,
		0x00444444,
		0x004444ff,
		0x00ff4444,
		0x00ff44ff,
		0x0044ff44,
		0x0044ffff,
		0x00ffff44,
		0x00ffffff,
		/* O[j^ */
		0x00000000,
		0x00173025,
		0x002e5e45,
		0x00336953,
		0x004c9e7c,
		0x004fa380,
		0x0055b08b,
		0x0057b58e,
		0x00214536,
		0x002e6046,
		0x00448c6e,
		0x00499777,
		0x0066d3a6,
		0x0069d9ab,
		0x0070e8b7,
		0x0073eebb,
		/* Ao[CG[j^ */
		0x00170000,
		0x002a0e08,
		0x00492415,
		0x00512a19,
		0x00814c2d,
		0x008b5030,
		0x00945935,
		0x00985d37,
		0x0038170e,
		0x004b2516,
		0x006f3f26,
		0x007a472b,
		0x00b97445,
		0x00c27b49,
		0x00c28750,
		0x00c28b53,
	};

	/* CfbNXx[Xݒ */
	if (nMonitorType == MONITOR_GREEN) {
		base = 16;
	}
	else if (nMonitorType == MONITOR_AMBERYELLOW) {
		base = 32;
	}
	else {
		base = 0;
	}

	/* 640x400AeLXgpbg */
	for (i=0; i<16; i++) {
		if (crt_flag) {
			/* CRT ON */
			rgbTTLL4GDI[i + 16] = rgbTable16[base + i];
		}
		else {
			/* CRT OFF */
			rgbTTLL4GDI[i + 16] = rgbTable16[base];
		}
	}

	/* 640x400AOtBbNpbg */
	for (i=0; i<3; i++) {
		if (crt_flag) {
			if ((i == 0) || (multi_page & (0x08 << i))) {
				rgbTTLL4GDI[i] = rgbTable16[(ttl_palet[0] & 0x0f) + base];
			}
			else {
				rgbTTLL4GDI[i] = rgbTable16[(ttl_palet[1] & 0x0f) + base];
			}
		}
		else {
			rgbTTLL4GDI[i] = rgbTable16[base];
		}
	}
}
#endif

#ifdef FILTERDLL
/*
 *	ňU32rbg}bvɕϊĂFilterփRs[
 *	{Ȃ璼Filterփ_O
 */
static BOOL FASTCALL FilterDraw(HDC hdcDraw, HDC hdcDst, HDC hdcSrc)
{
#define	USE_INNER_RENDER	1			/* 0:ORender 1:DLLRender */
	DWORD		dwCaps;
#if !USE_INNER_RENDER
	FILTERDESC	sdesc;
	LPBYTE		lpSrc;
	LPBYTE		lpDst;
	DWORD		dwSize;
	int			y;
#endif
	DWORD		dwNeed;
	double d;

	if (!pFilter) {
		return FALSE;
	}

	dwCaps = pFilterDll->GetCaps(pFilter);

#if USE_INNER_RENDER
	if (dwCaps & FILTER_FLAG_RENDER_HDC) {
		/* dwDstɂ͕`pHDC */
		/* dwPitchdummy */
		pFilterDll->Render(pFilter, (DWORD)hdcDraw, (DWORD)hdcSrc, NULL);
	}
	else {
		/* U32ɕϊă_[ */
		BitBlt(hdcDst, 0, 0, 640, 400, hdcSrc, 0, 0, SRCCOPY);
		/* dwDstNULL̏ꍇGetDescsdesc.lpSurfaceƓ */
		/* dwPitch̓\[X */
		pFilterDll->Render(pFilter, NULL, (DWORD)pBitsGDI32, 640 * 4);
	}
#else
	/* U32ɕϊă_[ */
	BitBlt(hdcDst, 0, 0, 640, 400, hdcSrc, 0, 0, SRCCOPY);
	sdesc.dwSize = sizeof(FILTERDESC);

	if (dwCaps & FILTER_FLAG_LOCK) {
		pFilterDll->Lock(pFilter, &sdesc);	/* LockĎ擾 */
	}
	else {
		pFilterDll->GetDesc(pFilter, &sdesc);
	}

	lpSrc = pBitsGDI32;
	lpDst = (LPBYTE)sdesc.lpSurface;
	dwSize = 640 * (((sdesc.dwFlags & FILTER_FLAG_PIXEL_MASK) >>
										FILTER_FLAG_PIXEL_SHIFT) / 8);

	for (y=0; y<400; y++) {
		CopyMemory(lpDst, lpSrc, dwSize);
		lpSrc += dwSize;
		lpDst += sdesc.dwPitch;
	}

	if (dwCaps & FILTER_FLAG_LOCK) {
		pFilterDll->Unlock(pFilter);
	}
#endif

	dwNeed = (dwCaps & FILTER_FLAG_DRAW_HDC) ? (DWORD)hdcDraw : (DWORD)hDrawWnd;
	if (bAspectRatio) {
		d = dAspectRatio;			/* 0.899(0.4495) */
	}
	else {
		d = 1.000;					/* 1.000(0.5000) */
	}
	switch (nScalingSize) {
		case DISPLAYSIZE_100:		/* x1.0(640x400) */
		default:
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 1.0, 400 * 1.0 / d);
			break;

		case DISPLAYSIZE_150:		/* x1.5(960x600) */
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 1.5, 400 * 1.5 / d);
			break;

		case DISPLAYSIZE_200:		/* x2.0(1280x800) */
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 2.0, 400 * 2.0 / d);
			break;

		case DISPLAYSIZE_050:		/* x0.5(320x200) */
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 0.5, 400 * 0.5 / d);
			break;

		case DISPLAYSIZE_180:		/* x1.8(1152x720) */
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 1.8, 400 * 1.8 / d);
			break;

		case DISPLAYSIZE_120:		/* x1.2(768x480) */
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 1.2, 400 * 1.2 / d);
			break;

		case DISPLAYSIZE_300:		/* x3.0(1920x1200) */
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 3.0, 400 * 3.0 / d);
			break;

		case DISPLAYSIZE_400:		/* x4.0(2560x1600) */
			pFilterDll->Draw(pFilter, dwNeed, 0, 0, 640 * 4.0, 400 * 4.0 / d);
			break;
	}

	return TRUE;
}
#endif

/*
 *	gk`
 *	`
 */
#ifdef FILTERDLL
static BOOL FASTCALL DrawBlt(HDC hDC, HDC hMemDC32, HDC hMemDC, WORD wdleft, WORD wdtop, WORD wdright, WORD wdbtm)
#else
static BOOL FASTCALL DrawBlt(HDC hDC, HDC hMemDC, WORD wdleft, WORD wdtop, WORD wdright, WORD wdbtm)
#endif
{
	int nsize;

#ifdef FILTERDLL
	if (pFilter) {
		FilterDraw(hDC, hMemDC32, hMemDC);
		return TRUE;
	}
#endif

	if (nScalingSize != DISPLAYSIZE_100) {
		if (nScalingSize == DISPLAYSIZE_200) {
			nsize = 2;
		}
		else  if (nScalingSize == DISPLAYSIZE_300) {
			nsize = 3;
		}
		else  if (nScalingSize == DISPLAYSIZE_400) {
			nsize = 4;
		}
		else {
			nsize = 1;
		}
		StretchBlt(hDC, wdleft * nsize, wdtop * nsize, (wdright - wdleft) * nsize, (wdbtm - wdtop) * nsize,
			hMemDC, wdleft, wdtop, (wdright - wdleft), (wdbtm - wdtop), SRCCOPY);
	}
	else {
		BitBlt(hDC, wdleft, wdtop,
			(wdright - wdleft), (wdbtm - wdtop),
			hMemDC, wdleft, wdtop, SRCCOPY);
	}

	return TRUE;
}

/*
 *	640x200AfW^[h
 *	`
 */
static void FASTCALL Draw640(HWND hWnd, HDC hDC)
{
	HDC hMemDC;
	HBITMAP hDefBitmap;
#ifdef FILTERDLL
	HDC hMemDC32;
	HBITMAP hDefBitmap32;
#endif
#if XM7_VER >= 3
	WORD wdtop, wdbtm;
#endif

	ASSERT(hWnd);
	ASSERT(hDC);

	/* pbgݒ */
	if (bPaletFlag) {
		Palet640();
	}

	/* NA */
	if (bClearFlag) {
		AllClear();
	}

	/* _O */
	if ((nDrawTop < nDrawBottom) && (nDrawLeft < nDrawRight)) {
#if XM7_VER >= 3
		/* EBhEI[v */
		if (window_open) {
			/* EBhEO 㑤̕` */
			if ((nDrawTop >> 1) < window_dy1) {
				if (bPseudo400Line) {
					Render640cGDI(nDrawTop >> 1, window_dy1);
				}
				else {
					Render640GDI2(nDrawTop >> 1, window_dy1, 320, 4);
				}
			}

			/* EBhE̕` */
			if ((nDrawTop >> 1) > window_dy1) {
				wdtop = (WORD)(nDrawTop >> 1);
			}
			else {
				wdtop = window_dy1;
			}

			if ((nDrawBottom >> 1) < window_dy2) {
				wdbtm = (WORD)(nDrawBottom >> 1);
			}
			else {
				wdbtm = window_dy2;
			}

			if (wdbtm > wdtop) {
				if (bPseudo400Line) {
					Render640cwGDI(wdtop, wdbtm, window_dx1, window_dx2);
				}
				else {
					Render640wGDI2(wdtop, wdbtm, window_dx1, window_dx2,320,4);
				}
			}

			/* EBhEO ̕` */
			if ((nDrawBottom >> 1) > window_dy2) {
				if (bPseudo400Line) {
					Render640cGDI(window_dy2, nDrawBottom >> 1);
				}
				else {
					Render640GDI2(window_dy2, nDrawBottom >> 1, 320, 4);
				}
			}
		}
		else {
			if (bPseudo400Line) {
				Render640cGDI(nDrawTop >> 1, nDrawBottom >> 1);
			}
			else {
				Render640GDI2(nDrawTop >> 1, nDrawBottom >> 1, 320, 4);
			}
		}
#elif XM7_VER == 2
		if (bPseudo400Line) {
			Render640cGDI(nDrawTop >> 1, nDrawBottom >> 1);
		}
		else {
			Render640GDI(nDrawTop >> 1, nDrawBottom >> 1);
		}
#ifdef FMTV151
		DrawV2_4bpp();
#endif	/* FMTV-151 */
#else
		if (bPseudo400Line) {
			Render640mGDI(nDrawTop >> 1, nDrawBottom >> 1);
		}
		else {
			Render640GDI(nDrawTop >> 1, nDrawBottom >> 1);
		}
#endif	/* XM7_VER >= 3 */

		if (!bPseudo400Line) {
			if (bFullScan) {
				RenderFullScan();
			}
			else {
				RenderSetOddLineDigital();
			}
		}
	}
	else {
		/* pbgύXĂȂꍇ͉Ȃ */
		if (!bPaletFlag) {
			return;
		}
	}

	/* pbgύXĂꍇ͑S̈ĕ` */
	if (bPaletFlag) {
		nDrawTop = 0;
		nDrawBottom = 400;
		nDrawLeft = 0;
		nDrawRight = 640;
		SetDrawFlag(TRUE);
	}

	/* DC쐬AIuWFNgI */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hDefBitmap = (HBITMAP)SelectObject(hMemDC, hBitmap);
#ifdef FILTERDLL
	if (pFilter) {
		hMemDC32 = CreateCompatibleDC(hDC);
		ASSERT(hMemDC32);
		hDefBitmap32 = (HBITMAP)SelectObject(hMemDC32, hBitmap32);
	}
#endif

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hDefBitmap) {
		SetDIBColorTable(hMemDC, 0, 16, (RGBQUAD *)rgbTTLGDI);
#ifdef FILTERDLL
		DrawBlt(hDC, hMemDC32, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#else
		DrawBlt(hDC, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#endif
		SelectObject(hMemDC, hDefBitmap);
#ifdef FILTERDLL
		if (pFilter) {
			SelectObject(hMemDC32, hDefBitmap32);
		}
#endif

		/* ɔA[NZbg */
		nDrawTop = 400;
		nDrawBottom = 0;
		nDrawLeft = 640;
		nDrawRight = 0;
		bPaletFlag = FALSE;
		SetDrawFlag(FALSE);
	}
	else {
		/* ĕ`N */
		InvalidateRect(hWnd, NULL, FALSE);
	}

	/* DC폜 */
	DeleteDC(hMemDC);
#ifdef FILTERDLL
	if (pFilter) {
		DeleteDC(hMemDC32);
	}
#endif
}

#if XM7_VER >= 3
/*
 *	640x400AfW^[h
 *	`
 */
static void FASTCALL Draw400l(HWND hWnd, HDC hDC)
{
	HDC hMemDC;
	HBITMAP hDefBitmap;
#ifdef FILTERDLL
	HDC hMemDC32;
	HBITMAP hDefBitmap32;
#endif
	WORD wdtop, wdbtm;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* pbgݒ */
	if (bPaletFlag) {
		Palet640();
	}

	/* NA */
	if (bClearFlag) {
		AllClear();
	}

	/* _O */
	if ((nDrawTop < nDrawBottom) && (nDrawLeft < nDrawRight)) {
		/* EBhEI[v */
		if (window_open) {
			/* EBhEO 㑤̕` */
			if (nDrawTop < window_dy1) {
				Render640GDI2(nDrawTop, window_dy1, 0, 2);
			}

			/* EBhE̕` */
			if (nDrawTop > window_dy1) {
				wdtop = nDrawTop;
			}
			else {
				wdtop = window_dy1;
			}

			if (nDrawBottom < window_dy2) {
				wdbtm = nDrawBottom;
			}
			else {
				wdbtm = window_dy2;
			}

			if (wdbtm > wdtop) {
				Render640wGDI2(wdtop, wdbtm, window_dx1, window_dx2, 0, 2);
			}

			/* EBhEO ̕` */
			if (nDrawBottom > window_dy2) {
				Render640GDI2(window_dy2, nDrawBottom, 0, 2);
			}
		}
		else {
			Render640GDI2(nDrawTop, nDrawBottom, 0, 2);
		}
	}
	else {
		/* pbgύXĂȂꍇ͉Ȃ */
		if (!bPaletFlag) {
			return;
		}
	}

	/* pbgύXĂꍇ͑S̈ĕ` */
	if (bPaletFlag) {
		nDrawTop = 0;
		nDrawBottom = 400;
		nDrawLeft = 0;
		nDrawRight = 640;
		SetDrawFlag(TRUE);
	}

	/* DC쐬AIuWFNgI */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hDefBitmap = (HBITMAP)SelectObject(hMemDC, hBitmap);
#ifdef FILTERDLL
	if (pFilter) {
		hMemDC32 = CreateCompatibleDC(hDC);
		ASSERT(hMemDC32);
		hDefBitmap32 = (HBITMAP)SelectObject(hMemDC32, hBitmap32);
	}
#endif

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hDefBitmap) {
		SetDIBColorTable(hMemDC, 0, 16, (RGBQUAD *)rgbTTLGDI);
#ifdef FILTERDLL
		DrawBlt(hDC, hMemDC32, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#else
		DrawBlt(hDC, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#endif
		SelectObject(hMemDC, hDefBitmap);
#ifdef FILTERDLL
		if (pFilter) {
			SelectObject(hMemDC32, hDefBitmap32);
		}
#endif

		/* ɔA[NZbg */
		nDrawTop = 400;
		nDrawBottom = 0;
		nDrawLeft = 640;
		nDrawRight = 0;
		bPaletFlag = FALSE;
		SetDrawFlag(FALSE);
	}
	else {
		/* ĕ`N */
		InvalidateRect(hWnd, NULL, FALSE);
	}

	/* DC폜 */
	DeleteDC(hMemDC);
#ifdef FILTERDLL
	if (pFilter) {
		DeleteDC(hMemDC32);
	}
#endif
}
#endif

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	eLXg
 *	_O
 */
static void FASTCALL RenderTextGDI(int first, int last, int left, int right)
{
	DWORD	tmp;
	int		x, y, x2;
	WORD	addr;
	BYTE	cursor_start, cursor_end, cursor_type;
	BYTE	raster, lines, maxx;
	BYTE	chr, atr, chr_dat;
	BYTE	line, line_old;
	BYTE	col;
	BYTE	*p;
	int		xsize;

	/* \OFF̏ꍇAȂ */
	if (!crt_flag) {
		return;
	}

	/* cursor */
	cursor_start	= (BYTE)(crtc_register[10] & 0x1f);
	cursor_end		= (BYTE)(crtc_register[11] & 0x1f);
	cursor_type		= (BYTE)((crtc_register[10] & 0x60) >> 5);

	/* ʕ\ */
	maxx			= (BYTE)(crtc_register[1] << 1);
	lines			= (BYTE)((crtc_register[9] & 0x1f) + 1);

	/* 1XTCY */
	if (width40_flag) {
		xsize = 16;
	}
	else {
		xsize = 8;
	}

	for (x = (left / xsize); x < (right / xsize); x++) {
		line_old = 255;
		for (y = first; y < last; y++) {
			p = (BYTE *)pBitsGDI + y * 640;
			raster = (BYTE)(y % lines);
			line = (BYTE)(y / lines);
			if (line != line_old) {
				addr = (WORD)((text_start_addr + (line * maxx + x) * 2) & 0x0ffe);
				chr = tvram_c[addr + 0];
				atr = tvram_c[addr + 1];

				/* Agr[g3rbg`Fݒ */
				col = (BYTE)((atr & 0x07) | 0x10);

				/* CeVeBAgr[g */
				if (atr & 0x20) {
					col |= (BYTE)8;
				}
			}

			/* tHgf[^擾(uNAgr[g܂) */
			if ((!(atr & 0x10) || text_blink) && (raster < 16)) {
				chr_dat = subcg_l4[(WORD)(chr * 16) + raster];
			}
			else {
				chr_dat = 0x00;
			}

			/* o[XAgr[g */
			if (atr & 0x08) {
				chr_dat ^= (BYTE)0xFF;
			}

			/* J[\ */
			if ((addr == cursor_addr) && (cursor_type != 1) &&
				(cursor_blink || (cursor_type == 0))) {
				if ((raster >= cursor_start) && (raster <= cursor_end)) {
					chr_dat ^= (BYTE)0xFF;
				}
			}

			/* `揈 */
			if (chr_dat) {
				for (x2 = 0; x2 < 8; x2++) {
					tmp = (DWORD)((x << 3) + x2);
					if (chr_dat & (1 << (7 ^ x2))) {
						if (width40_flag) {
							p[tmp * 2 + 0] = col;
							p[tmp * 2 + 1] = col;
						}
						else {
							p[tmp] = col;
						}
					}
				}
			}

			/* `悵sʒuۑ */
			line_old = line;
		}
	}
}

/*
 *	640x400APF[h
 *	`
 */
static void FASTCALL DrawL4(HWND hWnd, HDC hDC)
{
	HDC hMemDC;
	HBITMAP hDefBitmap;
#ifdef FILTERDLL
	HDC hMemDC32;
	HBITMAP hDefBitmap32;
#endif

	ASSERT(hWnd);
	ASSERT(hDC);

	/* pbgݒ */
	if (bPaletFlag) {
		PaletL4();
	}

	/* NA */
	if (bClearFlag) {
		AllClear();
	}

	/* _O */
	if ((nDrawTop < nDrawBottom) && (nDrawLeft < nDrawRight)) {
		RenderL4GDI(nDrawTop, nDrawBottom);
		RenderTextGDI(nDrawTop, nDrawBottom, nDrawLeft, nDrawRight);
	}
	else {
		/* pbgύXĂȂꍇ͉Ȃ */
		if (!bPaletFlag) {
			return;
		}
	}

	/* pbgύXĂꍇ͑S̈ĕ` */
	if (bPaletFlag) {
		nDrawTop = 0;
		nDrawBottom = 400;
		nDrawLeft = 0;
		nDrawRight = 640;
		SetDrawFlag(TRUE);
	}

	/* DC쐬AIuWFNgI */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hDefBitmap = (HBITMAP)SelectObject(hMemDC, hBitmap);
#ifdef FILTERDLL
	if (pFilter) {
		hMemDC32 = CreateCompatibleDC(hDC);
		ASSERT(hMemDC32);
		hDefBitmap32 = (HBITMAP)SelectObject(hMemDC32, hBitmap32);
	}
#endif

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hDefBitmap) {
		SetDIBColorTable(hMemDC, 0, 32, (RGBQUAD *)rgbTTLL4GDI);
#ifdef FILTERDLL
		DrawBlt(hDC, hMemDC32, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#else
		DrawBlt(hDC, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#endif
		SelectObject(hMemDC, hDefBitmap);
#ifdef FILTERDLL
		if (pFilter) {
			SelectObject(hMemDC32, hDefBitmap32);
		}
#endif

		/* ɔA[NZbg */
		nDrawTop = 400;
		nDrawBottom = 0;
		nDrawLeft = 640;
		nDrawRight = 0;
		bPaletFlag = FALSE;
		SetDrawFlag(FALSE);
	}
	else {
		/* ĕ`N */
		InvalidateRect(hWnd, NULL, FALSE);
	}

	/* DC폜 */
	DeleteDC(hMemDC);
#ifdef FILTERDLL
	if (pFilter) {
		DeleteDC(hMemDC32);
	}
#endif
}
#endif

#if XM7_VER >= 2
/*
 *	320x200AAiO[h
 *	pbgݒ
 */
static void FASTCALL Palet320(void)
{
	int i, j;
	DWORD color;
	DWORD r, g, b;
	int amask;

	/* AiO}XN쐬 */
	amask = 0;
	if (!(multi_page & 0x10)) {
		amask |= 0x000f;
	}
	if (!(multi_page & 0x20)) {
		amask |= 0x00f0;
	}
	if (!(multi_page & 0x40)) {
		amask |= 0x0f00;
	}

	for (i=0; i<4096; i++) {
		/* ŉʂ5bitÂB,G,R */
		color = 0;
		/* TTL RGB */
		if (crt_flag &&
			(nMonitorType != MONITOR_COLOR) &&
			(model_type == MODEL_FM77AV)) {
			r = 0;
			g = 0;
			b = 0;
			j = 0;
			/* R */
			if ((i & amask) & 0x0080) {
				j |= 2;
			}

			/* G */
			if ((i & amask) & 0x0800) {
				j |= 4;
			}

			/* B */
			if ((i & amask) & 0x0008) {
				j |= 1;
			}

			/* RGB */
			if (ttl_palet[j & 7] & 0x01) {
				b = 0x0f;
			}
			if (ttl_palet[j & 7] & 0x02) {
				r = 0x0f;
			}
			if (ttl_palet[j & 7] & 0x04) {
				g = 0x0f;
			}
		}
		else
		if (crt_flag) {
			j = i & amask;
			r = apalet_r[j];
			g = apalet_g[j];
			b = apalet_b[j];
		}
		else {
			r = 0;
			g = 0;
			b = 0;
		}

#if XM7_VER >= 3
		if (bRasterRendering) {
			/* R */
			color |= (WORD)r * 0x11;
			color <<= 8;

			/* G */
			color |= (WORD)g * 0x11;
			color <<= 8;

			/* B */
			color |= (WORD)b * 0x11;

			/* Zbg */
			rgbAnalogGDI[i] = color;

			continue;
		}
#endif

		/* R */
		r <<= 1;
		if (r > 0) {
			r |= 0x01;
		}
		color |= (WORD)r;
		color <<= 5;

		/* G */
		g <<= 1;
		if (g > 0) {
			g |= 0x01;
		}
		color |= (WORD)g;
		color <<= 5;

		/* B */
		b <<= 1;
		if (b > 0) {
			b |= 0x01;
		}
		color |= (WORD)b;

		/* Zbg */
		rgbAnalogGDI[i] = (DWORD)((color << 16) | color);
	}
}

/*
 *	320x200AAiO[h
 *	`
 */
static void FASTCALL Draw320(HWND hWnd, HDC hDC)
{
	HDC hMemDC;
	HBITMAP hDefBitmap;
#ifdef FILTERDLL
	HDC hMemDC32;
	HBITMAP hDefBitmap32;
#endif
#if XM7_VER >= 3
	WORD wdtop, wdbtm;
#endif

	ASSERT(hWnd);
	ASSERT(hDC);

	/* pbgݒ */
	if (bPaletFlag) {
		Palet320();
	}

	/* NA */
	if (bClearFlag) {
		AllClear();
	}

	/* _O */
	if (nDrawTop >= nDrawBottom) {
		return;
	}

#if XM7_VER >= 3
	/* EBhEI[v */
	if (window_open) {
		/* EBhEO 㑤̕` */
		if ((nDrawTop >> 1) < window_dy1) {
			Render320GDI(nDrawTop >> 1, window_dy1);
		}

		/* EBhE̕` */
		if ((nDrawTop >> 1) > window_dy1) {
			wdtop = (WORD)(nDrawTop >> 1);
		}
		else {
			wdtop = window_dy1;
		}

		if ((nDrawBottom >> 1) < window_dy2) {
			wdbtm = (WORD)(nDrawBottom >> 1);
		}
		else {
			wdbtm = window_dy2;
		}

		if (wdbtm > wdtop) {
			Render320wGDI(wdtop, wdbtm, window_dx1, window_dx2);
		}

		/* EBhEO ̕` */
		if ((nDrawBottom >> 1) > window_dy2) {
			Render320GDI(window_dy2, nDrawBottom >> 1);
		}
	}
	else {
		Render320GDI(nDrawTop >> 1, nDrawBottom >> 1);
	}
#else
	Render320GDI(nDrawTop >> 1, nDrawBottom >> 1);
#ifdef FMTV151
	DrawV2_16bpp();
#endif
#endif

	if (bFullScan) {
		RenderFullScan();
	}

	/* DC쐬AIuWFNgI */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hDefBitmap = (HBITMAP)SelectObject(hMemDC, hBitmap);
#ifdef FILTERDLL
	if (pFilter) {
		hMemDC32 = CreateCompatibleDC(hDC);
		ASSERT(hMemDC32);
		hDefBitmap32 = (HBITMAP)SelectObject(hMemDC32, hBitmap32);
	}
#endif

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hDefBitmap) {
#ifdef FILTERDLL
		DrawBlt(hDC, hMemDC32, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#else
		DrawBlt(hDC, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#endif
		SelectObject(hMemDC, hDefBitmap);
#ifdef FILTERDLL
		if (pFilter) {
			SelectObject(hMemDC32, hDefBitmap32);
		}
#endif

		/* ɔA[NZbg */
		nDrawTop = 400;
		nDrawBottom = 0;
		nDrawLeft = 640;
		nDrawRight = 0;
		bPaletFlag = FALSE;
		SetDrawFlag(FALSE);
	}
	else {
		/* ĕ`N */
		InvalidateRect(hWnd, NULL, FALSE);
	}

	/* DC */
	DeleteDC(hMemDC);
#ifdef FILTERDLL
	if (pFilter) {
		DeleteDC(hMemDC32);
	}
#endif
}
#endif

#if XM7_VER >= 3
/*
 *	320x200A26F[h
 *	`
 */
static void FASTCALL Draw256k(HWND hWnd, HDC hDC)
{
	HDC hMemDC;
	HBITMAP hDefBitmap;
#ifdef FILTERDLL
	HDC hMemDC32;
	HBITMAP hDefBitmap32;
#endif

	ASSERT(hWnd);
	ASSERT(hDC);

	/* NA */
	if (bClearFlag) {
		AllClear();
	}

	/* _O */
	if (nDrawTop >= nDrawBottom) {
		return;
	}
	if (crt_flag) {
		Render256kGDI(nDrawTop >> 1, nDrawBottom >> 1, multi_page);
	}
	else {
		Render256kGDI(nDrawTop >> 1, nDrawBottom >> 1, 0xff);
	}
	if (bFullScan) {
		RenderFullScan();
	}

	/* DC쐬AIuWFNgI */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hDefBitmap = (HBITMAP)SelectObject(hMemDC, hBitmap);
#ifdef FILTERDLL
	if (pFilter) {
		hMemDC32 = CreateCompatibleDC(hDC);
		ASSERT(hMemDC32);
		hDefBitmap32 = (HBITMAP)SelectObject(hMemDC32, hBitmap32);
	}
#endif

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hDefBitmap) {
#ifdef FILTERDLL
		DrawBlt(hDC, hMemDC32, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#else
		DrawBlt(hDC, hMemDC, nDrawLeft, nDrawTop, nDrawRight, nDrawBottom);
#endif
		SelectObject(hMemDC, hDefBitmap);
#ifdef FILTERDLL
		if (pFilter) {
			SelectObject(hMemDC32, hDefBitmap32);
		}
#endif

		/* ɔA[NZbg */
		nDrawTop = 400;
		nDrawBottom = 0;
		nDrawLeft = 640;
		nDrawRight = 0;
		bPaletFlag = FALSE;
		SetDrawFlag(FALSE);
	}
	else {
		/* ĕ`N */
		InvalidateRect(hWnd, NULL, FALSE);
	}

	/* DC */
	DeleteDC(hMemDC);
#ifdef FILTERDLL
	if (pFilter) {
		DeleteDC(hMemDC32);
	}
#endif
}
#endif

/*
 *	`
 */
void FASTCALL DrawGDI(HWND hWnd, HDC hDC)
{
	ASSERT(hWnd);
	ASSERT(hDC);

	/* 640-320 ؂ւ */
	SelectGDI(hWnd);

#if XM7_VER >= 3
	/* ꂩgĕ` */
	switch (nMode) {
		case SCR_400LINE	:	Draw400l(hWnd, hDC);
								return;
		case SCR_262144		:	Draw256k(hWnd, hDC);
								return;
		case SCR_4096		:	Draw320(hWnd, hDC);
								return;
		case SCR_200LINE	:	Draw640(hWnd, hDC);
								return;
	}
#elif XM7_VER >= 2
	/* ǂ炩gĕ` */
	if (bAnalog) {
		Draw320(hWnd, hDC);
	}
	else {
		Draw640(hWnd, hDC);
	}
#elif defined(L4CARD)
	/* ǂ炩gĕ` */
	if (b400Line) {
		DrawL4(hWnd, hDC);
	}
	else {
		Draw640(hWnd, hDC);
	}
#else
	Draw640(hWnd, hDC);
#endif
}

/*
 *	640x200AfW^[h
 *	X^`
 */
static void FASTCALL DrawRaster640(int nRaster)
{
	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 200) {
		return;
	}

	/* _O */
#if XM7_VER >= 3
	/* EBhEI[v */
	if (window_open) {
		/* EBhEO̕` */
		if ((nRaster < window_dy1) || (nRaster > window_dy2)) {
			if (bPseudo400Line) {
				Render640cGDI(nRaster, nRaster + 1);
			}
			else {
				Render640GDI2(nRaster, nRaster + 1, 320, 4);
			}
		}
		else {
			if (bPseudo400Line) {
				Render640cwGDI(nRaster, nRaster + 1, window_dx1, window_dx2);
			}
			else {
				Render640wGDI2(nRaster,nRaster+1,window_dx1,window_dx2,320,4);
			}
		}
	}
	else {
		if (bPseudo400Line) {
			Render640cGDI(nRaster, nRaster + 1);
		}
		else {
			Render640GDI2(nRaster, nRaster + 1, 320, 4);
		}
	}
#elif XM7_VER == 2
	if (bPseudo400Line) {
		Render640cGDI(nRaster, nRaster + 1);
	}
	else {
		Render640GDI(nRaster, nRaster + 1);
	}
#else
	if (bPseudo400Line) {
		Render640mGDI(nRaster, nRaster + 1);
	}
	else {
		Render640GDI(nRaster, nRaster + 1);
	}
#endif	/* XM7_VER >= 3 */
}

#if XM7_VER >= 3
/*
 *	640x400AfW^[h
 *	X^`
 */
static void FASTCALL DrawRaster400l(int nRaster)
{
	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 400) {
		return;
	}

	/* _O */
	/* EBhEI[v */
	if (window_open) {
		/* EBhEO̕` */
		if ((nRaster < window_dy1) || (nRaster > window_dy2)) {
			Render640GDI2(nRaster, nRaster + 1, 0, 2);
		}
		else {
			Render640wGDI2(nRaster, nRaster + 1, window_dx1, window_dx2, 0, 2);
		}
	}
	else {
		Render640GDI2(nRaster, nRaster + 1, 0, 2);
	}
}
#endif

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	eLXg
 *	X^`
 */
static void FASTCALL RenderTextRasterGDI(int nRaster, int left, int right)
{
	DWORD	tmp;
	int		x, x2;
	WORD	addr;
	BYTE	cursor_start, cursor_end, cursor_type;
	BYTE	raster, lines, maxx;
	BYTE	chr, atr, chr_dat;
	BYTE	line;
	BYTE	col;
	BYTE	*p;
	int		xsize;

	/* \OFF̏ꍇAȂ */
	if (!crt_flag) {
		return;
	}

	/* cursor */
	cursor_start	= (BYTE)(crtc_register[10] & 0x1f);
	cursor_end		= (BYTE)(crtc_register[11] & 0x1f);
	cursor_type		= (BYTE)((crtc_register[10] & 0x60) >> 5);

	/* ʕ\ */
	maxx			= (BYTE)(crtc_register[1] << 1);
	lines			= (BYTE)((crtc_register[9] & 0x1f) + 1);

	/* 1XTCY */
	if (width40_flag) {
		xsize = 16;
	}
	else {
		xsize = 8;
	}

	for (x = (left / xsize); x < (right / xsize); x++) {
		p = (BYTE *)pBitsGDI + nRaster * 640;
		raster = (BYTE)(nRaster % lines);
		line = (BYTE)(nRaster / lines);
		addr = (WORD)((text_start_addr + (line * maxx + x) * 2) & 0x0ffe);
		chr = tvram_c[addr + 0];
		atr = tvram_c[addr + 1];

		/* Agr[g3rbg`Fݒ */
		col = (BYTE)((atr & 0x07) | 0x10);

		/* CeVeBAgr[g */
		if (atr & 0x20) {
			col |= (BYTE)8;
		}

		/* tHgf[^擾(uNAgr[g܂) */
		if ((!(atr & 0x10) || text_blink) && (raster < 16)) {
			chr_dat = subcg_l4[(WORD)(chr * 16) + (nRaster % lines)];
		}
		else {
			chr_dat = 0x00;
		}

		/* o[XAgr[g */
		if (atr & 0x08) {
			chr_dat ^= (BYTE)0xFF;
		}

		/* J[\ */
		if ((addr == cursor_addr) && (cursor_type != 1) &&
			(cursor_blink || (cursor_type == 0))) {
			if ((raster >= cursor_start) && (raster <= cursor_end)) {
				chr_dat ^= (BYTE)0xFF;
			}
		}

		/* `揈 */
		if (chr_dat) {
			for (x2 = 0; x2 < 8; x2++) {
				tmp = (DWORD)((x << 3) + x2);
				if (chr_dat & (1 << (7 ^ x2))) {
					if (width40_flag) {
						p[tmp * 2 + 0] = col;
						p[tmp * 2 + 1] = col;
					}
					else {
						p[tmp] = col;
					}
				}
			}
		}
	}
}

/*
 *	640x400APF[h
 *	X^`
 */
static void FASTCALL DrawRasterL4(int nRaster)
{
	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 400) {
		return;
	}

	/* _O */
	RenderL4GDI(nRaster, nRaster + 1);
	RenderTextRasterGDI(nRaster, 0, 640);
}
#endif

#if XM7_VER >= 2
/*
 *	320x200AAiO[h
 *	X^`
 */
static void FASTCALL DrawRaster320(int nRaster)
{
	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 200) {
		return;
	}

	/* pbgݒ */
	Palet320();

	/* _O */
	/* EBhEI[v */
#if XM7_VER >= 3
	if (window_open) {
		/* EBhEO̕` */
		if ((nRaster < window_dy1) || (nRaster > window_dy2)) {
			Render320GDI32bpp(nRaster, nRaster + 1);
		}
		else {
			Render320wGDI32bpp(nRaster, nRaster + 1, window_dx1, window_dx2);
		}
	}
	else {
		Render320GDI32bpp(nRaster, nRaster + 1);
	}
#else
	Render320GDI(nRaster, nRaster + 1);
#endif
}
#endif

#if XM7_VER >= 3
/*
 *	320x200A26F[h
 *	X^`
 */
static void FASTCALL DrawRaster256k(int nRaster)
{
	/* \͈͊Ȍꍇ͉Ȃ */
	if (nRaster >= 200) {
		return;
	}

	/* _O */
	if (crt_flag) {
		Render256kGDI(nRaster, nRaster + 1, multi_page);
	}
	else {
		Render256kGDI(nRaster, nRaster + 1, 0xff);
	}
}
#endif

/*
 *	X^Pʃ_O
 *	`
 */
void FASTCALL DrawPostRenderGDI(HWND hWnd, HDC hDC)
{
	HDC hMemDC;
	HBITMAP hDefBitmap;
	WORD wdtop, wdbtm;
#ifdef FILTERDLL
	HDC hMemDC32;
	HBITMAP hDefBitmap32;
#endif

	ASSERT(hWnd);
	ASSERT(hDC);

	/* pbgݒ */
#if XM7_VER >= 3
	if (nMode == SCR_400LINE) {
		if (bPaletFlag) {
			Palet640();
		}
	}
	else if (nMode == SCR_200LINE) {
		if (bPaletFlag) {
			Palet640();
		}
		if (!bPseudo400Line) {
			if (bFullScan) {
				RenderFullScan();
			}
			else {
				RenderSetOddLineDigital();
			}
		}
	}
	else {
		if (bFullScan) {
			RenderFullScan();
		}
		else {
			RenderSetOddLineAnalog();
		}
	}
#elif XM7_VER >= 2
	if (bAnalog) {
#ifdef FMTV151
		DrawV2_16bpp();
#endif
		if (bFullScan) {
			RenderFullScan();
		}
		else {
			RenderSetOddLineAnalog();
		}
	}
	else {
#ifdef FMTV151
		DrawV2_4bpp();
#endif
		if (bPaletFlag) {
			Palet640();
		}
		if (!bPseudo400Line) {
			if (bFullScan) {
				RenderFullScan();
			}
			else {
				RenderSetOddLineDigital();
			}
		}
	}
#elif defined(L4CARD)
	if (b400Line) {
		if (bPaletFlag) {
			PaletL4();
		}
	}
	else {
		if (bPaletFlag) {
			Palet640();
		}
		if (!bPseudo400Line) {
			if (bFullScan) {
				RenderFullScan();
			}
			else {
				RenderSetOddLineDigital();
			}
		}
	}
#else
	if (bPaletFlag) {
		Palet640();
	}
	if (!bPseudo400Line) {
		if (bFullScan) {
			RenderFullScan();
		}
		else {
			RenderSetOddLineDigital();
		}
	}
#endif

	/* DC쐬AIuWFNgI */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hDefBitmap = (HBITMAP)SelectObject(hMemDC, hBitmap);
#ifdef FILTERDLL
	if (pFilter) {
		hMemDC32 = CreateCompatibleDC(hDC);
		ASSERT(hMemDC32);
		hDefBitmap32 = (HBITMAP)SelectObject(hMemDC32, hBitmap32);
	}
#endif

	/* X^Pʃ_OɂЂԂ邱Ƃ̂ */
	if (nDrawTop >= nDrawBottom) {
		wdtop = 0;
		wdbtm = 400;
	}
	else {
		wdtop = nDrawTop;
		wdbtm = nDrawBottom;
	}

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hDefBitmap) {
	/* pbgݒ */
#if XM7_VER >= 3
		switch (nMode) {
			case SCR_400LINE	:
			case SCR_200LINE	:
				SetDIBColorTable(hMemDC, 0, 16, (RGBQUAD *)rgbTTLGDI);
				break;
		}
#elif XM7_VER >= 2
		if (!bAnalog) {
			SetDIBColorTable(hMemDC, 0, 16, (RGBQUAD *)rgbTTLGDI);
		}
#elif defined(L4CARD)
		if (b400Line) {
			SetDIBColorTable(hMemDC, 0, 32, (RGBQUAD *)rgbTTLL4GDI);
		}
		else {
			SetDIBColorTable(hMemDC, 0, 16, (RGBQUAD *)rgbTTLGDI);
		}
#else
		SetDIBColorTable(hMemDC, 0, 16, (RGBQUAD *)rgbTTLGDI);
#endif

#ifdef FILTERDLL
		DrawBlt(hDC, hMemDC32, hMemDC, 0, wdtop, 640, wdbtm);
#else
		DrawBlt(hDC, hMemDC, 0, wdtop, 640, wdbtm);
#endif
		SelectObject(hMemDC, hDefBitmap);
#ifdef FILTERDLL
		if (pFilter) {
			SelectObject(hMemDC32, hDefBitmap32);
		}
#endif

		/* ɔA[NZbg */
		nDrawTop = 400;
		nDrawBottom = 0;
		nDrawLeft = 640;
		nDrawRight = 0;
		SetDrawFlag(TRUE);
	}
	else {
		/* ĕ`N */
		InvalidateRect(hWnd, NULL, FALSE);
	}

	/* DC폜 */
	DeleteDC(hMemDC);
#ifdef FILTERDLL
	if (pFilter) {
		DeleteDC(hMemDC32);
	}
#endif
}

/*
 *	X^Pʃ_O
 *	eX^_O
 */
void FASTCALL DrawRasterGDI(int nRaster)
{
	/* 640-320 ؂ւ */
	SelectGDI(hDrawWnd);

#if XM7_VER >= 3
	/* ꂩgĕ` */
	switch (nMode) {
		case SCR_400LINE	:	DrawRaster400l(nRaster);
								return;
		case SCR_262144		:	DrawRaster256k(nRaster);
								return;
		case SCR_4096		:	DrawRaster320(nRaster);
								return;
		case SCR_200LINE	:	DrawRaster640(nRaster);
								return;
	}
#elif XM7_VER >= 2
	/* ǂ炩gĕ` */
	if (bAnalog) {
		DrawRaster320(nRaster);
	}
	else {
		DrawRaster640(nRaster);
	}
#elif defined(L4CARD)
	/* ǂ炩gĕ` */
	if (b400Line) {
		DrawRasterL4(nRaster);
	}
	else {
		DrawRaster640(nRaster);
	}
#else
	DrawRaster640(nRaster);
#endif
}

/*
 *	j[Jn
 */
void FASTCALL EnterMenuGDI(HWND hWnd)
{
	ASSERT(hWnd);
	UNUSED(hWnd);

	/* rbg}bvnh̗LŔł */
	if (!hBitmap) {
		return;
	}

	/* }EXJ[\on */
	if (!bMouseCursor) {
		ShowCursor(TRUE);
		bMouseCursor = TRUE;
	}

#ifdef MOUSE
	/* }EXLv`~ */
	SetMouseCapture(FALSE);
#endif
}

/*
 *	j[I
 */
void FASTCALL ExitMenuGDI(void)
{
	/* rbg}bvnh̗LŔł */
	if (!hBitmap) {
		return;
	}

	/* }EXJ[\on */
	if (!bMouseCursor) {
		ShowCursor(TRUE);
		bMouseCursor = TRUE;
	}

#ifdef MOUSE
	/* }EXLv`Jn */
	SetMouseCapture(TRUE);
#endif
}

#ifdef FILTERDLL
/*
 *	O摜tB^p[^ݒ
 */
BOOL FASTCALL SetFilterParameterGDI(WORD wParam)
{
	if (!pFilter) {
		return FALSE;
	}

	dwFilterDllDepFlg = (DWORD)(wParam & 0xffff);
	pFilterDll->SetFilter(pFilter, dwFilterDllDepFlg << 16 |
												dwFilterDllFlg);

	return TRUE;
}
#endif

/*-[ VMƂ̐ڑ ]-----------------------------------------------------------*/

/*
 *	VRAMZbg
 */
void FASTCALL VramGDI(WORD addr)
{
	WORD x;
	WORD y;

	/* yWZo */
#if XM7_VER >= 3
	switch (nMode) {
		case SCR_400LINE	:	addr &= 0x7fff;
								x = (WORD)((addr % 80) << 3);
								y = (WORD)(addr / 80);
								break;
		case SCR_262144		:
		case SCR_4096		:	addr &= 0x1fff;
								x = (WORD)((addr % 40) << 4);
								y = (WORD)((addr / 40) << 1);
								break;
		case SCR_200LINE	:	addr &= 0x3fff;
								x = (WORD)((addr % 80) << 3);
								y = (WORD)((addr / 80) << 1);
								break;
	}
#elif XM7_VER >= 2
	if (bAnalog) {
		addr &= 0x1fff;
		x = (WORD)((addr % 40) << 4);
		y = (WORD)((addr / 40) << 1);
	}
	else {
		addr &= 0x3fff;
		x = (WORD)((addr % 80) << 3);
		y = (WORD)((addr / 80) << 1);
	}
#elif defined(L4CARD)
	if (b400Line) {
		addr -= vram_offset[0];
		addr &= 0x7fff;
		x = (WORD)((addr % 80) << 3);
		y = (WORD)(addr / 80);
	}
	else {
		addr &= 0x3fff;
		x = (WORD)((addr % 80) << 3);
		y = (WORD)((addr / 80) << 1);
	}
#else
	addr &= 0x3fff;
	x = (WORD)((addr % 80) << 3);
	y = (WORD)((addr / 80) << 1);
#endif

	/* I[o[`FbN */
	if ((x >= 640) || (y >= 400)) {
		return;
	}

	/* ĕ`tOݒ */
	GDIDrawFlag[(y >> 3) * 80 + (x >> 3)] = 1;

	/* XV */
	if (nDrawTop > y) {
		nDrawTop = y;
	}
	if (nDrawBottom <= y) {
#if XM7_VER >= 3
		if (nMode == SCR_400LINE) {
			nDrawBottom = (WORD)(y + 1);
		}
		else {
			nDrawBottom = (WORD)(y + 2);
		}
#else
		nDrawBottom = (WORD)(y + 2);
#endif
	}

	/* XV */
	if (nDrawLeft > x) {
		nDrawLeft = x;
	}
	if (nDrawRight <= x) {
#if XM7_VER >= 2
#if XM7_VER >= 3
		if (nMode & SCR_ANALOG) {
#else
		if (bAnalog) {
#endif
			nDrawRight = (WORD)(x + 16);
		}
		else {
			nDrawRight = (WORD)(x + 8);
		}
#else
		nDrawRight = (WORD)(x + 8);
#endif
	}
}

#if XM7_VER == 1 && defined(L4CARD)
/*
 *	eLXgVRAMZbg
 */
void FASTCALL TvramGDI(WORD addr)
{
	WORD ysize;
	WORD x;
	WORD y;
	BYTE maxy;
	WORD yy;

	maxy = (BYTE)(crtc_register[6] & 0x7f);

	/* WZo */
	addr = (WORD)((addr - text_start_addr) & 0x0ffe);
	ysize = (WORD)((crtc_register[9] & 0x1f) + 1);
	if (width40_flag) {
		x = (WORD)((addr % 80) * 8);
		y = (WORD)(addr / 80);
	}
	else {
		x = (WORD)((addr % 160) * 4);
		y = (WORD)(addr / 160);
	}

	/* I[o[`FbN */
	if (y >= maxy) {
		return;
	}
	y = (WORD)(y * ysize);
	if (y > 400 - ysize) {
		y = (WORD)(400 - ysize);
	}

	/* ĕ`tOݒ */
	for (yy = y; yy < (WORD)(y + ysize); yy += (WORD)8) {
		GDIDrawFlag[(yy >> 3) * 80 + (x >> 3)] = 1;
		if (width40_flag) {
			GDIDrawFlag[(yy >> 3) * 80 + (x >> 3) + 1] = 1;
		}
	}

	/* XV */
	if (nDrawTop > y) {
		nDrawTop = y;
	}
	if (nDrawBottom <= (y + ysize - 1)) {
		nDrawBottom = (WORD)(y + ysize);
	}

	/* XV */
	if (nDrawLeft > x) {
		nDrawLeft = x;
	}
	if (nDrawRight <= x) {
		if (width40_flag) {
			nDrawRight = (WORD)(x + 16);
		}
		else {
			nDrawRight = (WORD)(x + 8);
		}
	}
}

/*
 *	eLXgVRAMĕ`v
 */
void FASTCALL ReDrawTVRamGDI(void)
{
	WORD maxy;

	maxy = (WORD)((crtc_register[6] & 0x7f) * ((crtc_register[9] & 0x1f) + 1));

	nDrawTop = 0;
	if (nDrawBottom < maxy) {
		nDrawBottom = maxy;
	}
	nDrawLeft = 0;
	nDrawRight = 640;
	SetDrawFlag(TRUE);
}
#endif

/*
 *	TTLpbgZbg
 */
void FASTCALL DigitalGDI(void)
{
	/* svȃ_O}邽߁Äݒ͕`掞ɍs */
	bPaletFlag = TRUE;
}

/*
 *	AiOpbgZbg
 */
void FASTCALL AnalogGDI(void)
{
	bPaletFlag = TRUE;
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	SetDrawFlag(TRUE);
}

/*
 *	ĕ`v
 */
void FASTCALL ReDrawGDI(void)
{
	/* ĕ` */
	nDrawTop = 0;
	nDrawBottom = 400;
	nDrawLeft = 0;
	nDrawRight = 640;
	bPaletFlag = TRUE;
	bClearFlag = TRUE;
	SetDrawFlag(TRUE);
}

#if XM7_VER >= 3
/*
 *	n[hEFAEBhEʒm
 */
void FASTCALL WindowGDI(void)
{
	WORD tmpLeft, tmpRight;
	WORD tmpTop, tmpBottom;
	WORD tmpDx1, tmpDx2;
	WORD tmpDy1, tmpDy2;
	BYTE *p;
	int i;

	/* 26F[h͉Ȃ */
	if (nMode == SCR_262144) {
		return;
	}

	/* OăNbsO */
	window_clip(nMode);

	/* EBhETCY␳ */
	tmpDx1 = window_dx1;
	tmpDy1 = window_dy1;
	tmpDx2 = window_dx2;
	tmpDy2 = window_dy2;
	if (nMode != SCR_400LINE) {
		tmpDy1 <<= 1;
		tmpDy2 <<= 1;
	}
	if (nMode == SCR_4096) {
		tmpDx1 <<= 1;
		tmpDx2 <<= 1;
	}

	if (bWindowOpen != window_open) {
		if (window_open) {
			/* EBhEJꍇ */
			tmpLeft = tmpDx1;
			tmpRight = tmpDx2;
			tmpTop = tmpDy1;
			tmpBottom = tmpDy2;
		}
		else {
			/* EBhEꍇ */
			tmpLeft = nWindowDx1;
			tmpRight = nWindowDx2;
			tmpTop = nWindowDy1;
			tmpBottom = nWindowDy2;
		}
	}
	else {
		if (window_open) {
			/* XV̈TCŶ݂̂ɐݒ */
			tmpTop = nDrawTop;
			tmpBottom = nDrawBottom;
			tmpLeft = nDrawLeft;
			tmpRight = nDrawRight;

			/* WύX`FbN */
			if (!((nWindowDx1 == tmpDx1) &&
				  (nWindowDy1 == tmpDy1) &&
				  (nWindowDx2 == tmpDx2) &&
				  (nWindowDy2 == tmpDy2))) {
				/* X */
				if (nWindowDx1 < tmpDx1) {
					tmpLeft = nWindowDx1;
				}
				else {
					tmpLeft = tmpDx1;
				}

				/* EX */
				if (nWindowDx2 > tmpDx2) {
					tmpRight = nWindowDx2;
				}
				else {
					tmpRight = tmpDx2;
				}

				/* Y */
				if (nWindowDy1 < tmpDy1) {
					tmpTop = nWindowDy1;
				}
				else {
					tmpTop = tmpDy1;
				}

				/* EY */
				if (nWindowDy2 > tmpDy2) {
					tmpBottom = nWindowDy2;
				}
				else {
					tmpBottom = tmpDy2;
				}
			}
		}
		else {
			/* EBhEJĂȂ̂ŉȂ */
			return;
		}
	}

	/* O̍ĕ`̈ƔrčLΗ̈XV */
	if (tmpLeft < nDrawLeft) {
		nDrawLeft = tmpLeft;
	}
	if (tmpRight > nDrawRight) {
		nDrawRight = tmpRight;
	}
	if (tmpTop < nDrawTop) {
		nDrawTop = tmpTop;
	}
	if (tmpBottom > nDrawBottom) {
		nDrawBottom = tmpBottom;
	}

	/* ĕ`tOXV */
	if ((nDrawLeft < nDrawRight) && (nDrawTop < nDrawBottom)) {
		if (nMode == SCR_400LINE) {
			SetDirtyFlag(nDrawTop, nDrawBottom + 1, TRUE);
		}
		else {
			SetDirtyFlag(nDrawTop >> 1, (nDrawBottom >> 1) + 1, TRUE);
		}
		p = &GDIDrawFlag[(nDrawTop >> 3) * 80 + (nDrawLeft >> 3)];
		for (i = (nDrawTop >> 3); i < ((nDrawBottom + 7) >> 3) ; i++) {
			memset(p, 1, (nDrawRight - nDrawLeft) >> 3);
			p += 80;
		}
	}

	/* EBhEI[vԂۑ */
	bWindowOpen = window_open;
	nWindowDx1 = tmpDx1;
	nWindowDy1 = tmpDy1;
	nWindowDx2 = tmpDx2;
	nWindowDy2 = tmpDy2;
}
#endif

/*
 *	ZNgLZ
 */
void SelectCancelGDI(void)
{
	if (!bFullScreen) {
		bSelectCancelFlag = TRUE;
		SetDirtyFlag(0, 400, TRUE);
		SetDrawFlag(TRUE);
	}
}

#endif	/* _WIN32 */
