/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API L[{[hEWCXeBbNE}EX ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#ifdef DINPUT8
#define DIRECTINPUT_VERSION		0x0800		/* DirectX8w */
#else
#define DIRECTINPUT_VERSION		0x0300		/* DirectX3w */
#endif
#include <dinput.h>
#include <mmsystem.h>
#include <assert.h>
#include "xm7.h"
#include "mainetc.h"
#include "keyboard.h"
#include "device.h"
#include "mouse.h"
#include "event.h"
#include "w32.h"
#include "w32_sch.h"
#include "w32_kbd.h"
#include "w32_bar.h"
#include "w32_res.h"
#include "w32_draw.h"

/*
 *	GUID`
 */
#ifdef DEFINE_GUID
#undef DEFINE_GUID
#define DEFINE_GUID(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8) \
        EXTERN_C const GUID DECLSPEC_SELECTANY name \
                = { l, w1, w2, { b1, b2,  b3,  b4,  b5,  b6,  b7,  b8 } }
#endif
DEFINE_GUID(IID_IDirectInput8A, 0xBF798030, 0x483A, 0x4DA2, 0xAA, 0x99, 0x5D, 0x64, 0xED, 0x36, 0x97, 0x00);
DEFINE_GUID(GUID_SysKeyboard, 0x6F1D2B61, 0xD5A0, 0x11CF, 0xBF, 0xC7, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00);
DEFINE_GUID(GUID_SysMouse, 0x6F1D2B60, 0xD5A0, 0x11CF, 0xBF, 0xC7, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00);
DEFINE_GUID(GUID_Key, 0x55728220, 0xD33C, 0x11CF, 0xBF, 0xC7, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00);
DEFINE_GUID(GUID_XAxis, 0xA36D02E0, 0xC9F3, 0x11CF, 0xBF, 0xC7, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00);
DEFINE_GUID(GUID_YAxis, 0xA36D02E1, 0xC9F3, 0x11CF, 0xBF, 0xC7, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00);
DEFINE_GUID(GUID_ZAxis, 0xA36D02E2, 0xC9F3, 0x11CF, 0xBF, 0xC7, 0x44, 0x45, 0x53, 0x54, 0x00, 0x00);

/*
 *	O[o [N
 */
BYTE kbd_map[256];							/* L[{[h }bv */
BYTE kbd_table[256];						/* ΉFM-7R[h */
int nJoyType[2];							/* WCXeBbN^Cv */
int nJoyRapid[2][2];						/* A˃^Cv */
int nJoyCode[2][7];							/* R[h */
BOOL bKbdReal;								/* ^A^CL[XL */
BOOL bTenCursor;							/* eL[ϊ */
BOOL bArrow8Dir;							/* eL[ϊ 8[h */
BOOL bNTkeyPushFlag[4];						/* NT΍pL[tO */
BOOL bNTkeyMakeFlag[128];					/* NT΍pMaketO */
BOOL bNTkbMode;								/* NT΍tO */
BOOL bAutoDetectKeyMap;						/* L[}bvݒtO */
#ifdef MOUSE
BYTE uMidBtnMode;							/* {^Ԏ擾[h */
BOOL bDetectMouse;							/* }EXmFtO */
BOOL bMouseCapture;							/* }EXLv`tO */
#endif
#ifdef KBDPASTE
UINT uPasteWait;							/* y[Xg҂(ms) */
UINT uPasteWaitCntl;						/* y[Xg҂ControlCode(ms) */
#endif

/*
 *	X^eBbN [N
 */
#ifdef DINPUT8
static LPDIRECTINPUT8 lpdi;					/* DirectInput */

static LPDIRECTINPUTDEVICE8 lpkbd;			/* L[{[hfoCX */
#else
static LPDIRECTINPUT lpdi;					/* DirectInput */

static LPDIRECTINPUTDEVICE lpkbd;			/* L[{[hfoCX */
#endif
static DWORD keytime;						/* L[{[h|[O */
static DWORD menuexittime;					/* j[oԃ^C} */
static DWORD dmyreadtime;					/* L[R[h_~[[h^C} */
static BOOL bDummyRead;						/* L[R[h_~[[htO */

static BYTE nKeyBuffer[KEYBUFFER_SIZE];		/* L[obt@ */
static BYTE nKeyReadPtr;					/* L[obt@Ǐo|C^ */
static BYTE nKeyWritePtr;					/* L[obt@|C^ */
static BYTE nLastKey;						/* ŌɉꂽL[(FULL) */
static BYTE nLastKey2;						/* ŌɉꂽL[(10KEY) */
static BYTE nTenDir;						/* eL[ϊ f[^ */
static BYTE nTenDir2;						/* eL[ϊ f[^2 */

static BYTE joydat[6];						/* WCXeBbNf[^ */
static BYTE joybk[2];						/* WCXeBbNobNAbv */
static int joyrapid[2][2];					/* WCXeBbNA˃JE^ */
static DWORD joytime;						/* WCXeBbN|[O */
static DWORD joytime2;						/* WCXeBbN|[O */
static BOOL joyplugged[2];					/* WCXeBbNڑtO */
static JOYCAPS joycaps[2];					/* WCXeBbN@\ */

#ifdef MOUSE
#ifdef DINPUT8
static LPDIRECTINPUTDEVICE8 lpmos;			/* }EXfoCX */
#else
static LPDIRECTINPUTDEVICE lpmos;			/* }EXfoCX */
#endif
static DWORD mostime;						/* }EX|[O */
static BOOL bCapture;						/* }EXLv`tO(Win32) */
static int nMouseX;							/* }EX Xړ */
static int nMouseY;							/* }EX Yړ */
static BYTE nMouseButton;					/* }EX {^ */
static BYTE nMouseButtons;					/* }EX {^ */
static BOOL bMouseCursor;					/* }EX J[\\ */
static int nMouseSX;						/* }EX XWۑ */
static int nMouseSY;						/* }EX YWۑ */
#endif

#ifdef KBDPASTE
static PTSTR strPaste;						/* y[Xgf[^ */
static PTSTR strPtr;						/* y[Xg|C^ */
static char cPreKey;						/* ÕR[h */
static UINT uWaitCnt;						/* y[Xg҂JE^ */
#endif


/*
 *	DirectInputR[hFM-7 R[h
 *	R[hΏƕ\(106L[{[hp)
 */
static BYTE kbd_106_table[] = {
	DIK_ESCAPE,			0x5c,			/* BREAK(ESC) */
	DIK_F1,				0x5d,			/* PF1 */
	DIK_F2,				0x5e,			/* PF2 */
	DIK_F3,				0x5f,			/* PF3 */
	DIK_F4,				0x60,			/* PF4 */
	DIK_F5,				0x61,			/* PF5 */
	DIK_F6,				0x62,			/* PF6 */
	DIK_F7,				0x63,			/* PF7 */
	DIK_F8,				0x64,			/* PF8 */
	DIK_F9,				0x65,			/* PF9 */
	DIK_F10,			0x66,			/* PF10 */

	DIK_KANJI,			0x01,			/* ESC(p/Sp) */
	DIK_1,				0x02,			/* 1 */
	DIK_2,				0x03,			/* 2 */
	DIK_3,				0x04,			/* 3 */
	DIK_4,				0x05,			/* 4 */
	DIK_5,				0x06,			/* 5 */
	DIK_6,				0x07,			/* 6 */
	DIK_7,				0x08,			/* 7 */
	DIK_8,				0x09,			/* 8 */
	DIK_9,				0x0a,			/* 9 */
	DIK_0,				0x0b,			/* 0 */
	DIK_MINUS,			0x0c,			/* - */
	DIK_CIRCUMFLEX,		0x0d,			/* ^ */
	DIK_YEN,			0x0e,			/* \ */
	DIK_BACK,			0x0f,			/* BS */

	DIK_TAB,			0x10,			/* TAB */
	DIK_Q,				0x11,			/* Q */
	DIK_W,				0x12,			/* W */
	DIK_E,				0x13,			/* E */
	DIK_R,				0x14,			/* R */
	DIK_T,				0x15,			/* T */
	DIK_Y,				0x16,			/* Y */
	DIK_U,				0x17,			/* U */
	DIK_I,				0x18,			/* I */
	DIK_O,				0x19,			/* O */
	DIK_P,				0x1a,			/* P */
	DIK_AT,				0x1b,			/* @ */
	DIK_LBRACKET,		0x1c,			/* [ */
	DIK_RETURN,			0x1d,			/* CR */

	DIK_LCONTROL,		0x52,			/* CTRL(Ctrl) */
	DIK_A,				0x1e,			/* A */
	DIK_S,				0x1f,			/* S */
	DIK_D,				0x20,			/* D */
	DIK_F,				0x21,			/* F */
	DIK_G,				0x22,			/* G */
	DIK_H,				0x23,			/* H */
	DIK_J,				0x24,			/* J */
	DIK_K,				0x25,			/* K */
	DIK_L,				0x26,			/* L */
	DIK_SEMICOLON,		0x27,			/* ; */
	DIK_COLON,			0x28,			/* : */
	DIK_RBRACKET,		0x29,			/* ] */

	DIK_LSHIFT,			0x53,			/* SHIFT */
	DIK_Z,				0x2a,			/* Z */
	DIK_X,				0x2b,			/* X */
	DIK_C,				0x2c,			/* C */
	DIK_V,				0x2d,			/* V */
	DIK_B,				0x2e,			/* B */
	DIK_N,				0x2f,			/* N */
	DIK_M,				0x30,			/* M */
	DIK_COMMA,			0x31,			/* , */
	DIK_PERIOD,			0x32,			/* . */
	DIK_SLASH,			0x33,			/* / */
	DIK_BACKSLASH,		0x34,			/* _ */
	DIK_RSHIFT,			0x54,			/* ESHIFT */

	DIK_CAPITAL,		0x55,			/* CAP(Caps Lock) */
	DIK_NOCONVERT,		0x56,			/* GRAPH(ϊ) */
	DIK_CONVERT,		0x57,			/* SPACE(ϊ) */
	DIK_KANA,			0x58,			/* SPACE(J^Ji) */
	DIK_SPACE,			0x35,			/* ESPACE(SPACE) */
	DIK_RCONTROL,		0x5a,			/* (ECtrl) */

	DIK_INSERT,			0x48,			/* INS(Insert) */
	DIK_DELETE,			0x4b,			/* DEL(Delete) */
	DIK_UP,				0x4d,			/*  */
	DIK_LEFT,			0x4f,			/*  */
	DIK_DOWN,			0x50,			/*  */
	DIK_RIGHT,			0x51,			/*  */

	DIK_HOME,			0x49,			/* EL(Home) */
	DIK_PRIOR,			0x4a,			/* CLS(Page Up) */
	DIK_END,			0x4c,			/* DUP(End) */
	DIK_NEXT,			0x4e,			/* HOME(Page Down) */

	DIK_MULTIPLY,		0x36,			/* Tenkey * */
	DIK_DIVIDE,			0x37,			/* Tenkey / */
	DIK_ADD,			0x38,			/* Tenkey + */
	DIK_SUBTRACT,		0x39,			/* Tenkey - */
	DIK_NUMPAD7,		0x3a,			/* Tenkey 7 */
	DIK_NUMPAD8,		0x3b,			/* Tenkey 8 */
	DIK_NUMPAD9,		0x3c,			/* Tenkey 9 */
	DIK_NUMPAD4,		0x3e,			/* Tenkey 4 */
	DIK_NUMPAD5,		0x3f,			/* Tenkey 5 */
	DIK_NUMPAD6,		0x40,			/* Tenkey 6 */
	DIK_NUMPAD1,		0x42,			/* Tenkey 1 */
	DIK_NUMPAD2,		0x43,			/* Tenkey 2 */
	DIK_NUMPAD3,		0x44,			/* Tenkey 3 */
	DIK_NUMPAD0,		0x46,			/* Tenkey 0 */
	DIK_DECIMAL,		0x47,			/* Tenkey . */
	DIK_NUMPADENTER,	0x45			/* Tenkey CR */
};

/*
 *	DirectInputR[hFM-7 R[h
 *	R[hΏƕ\(NEC PC-98L[{[hp)
 */
static BYTE kbd_98_table[] = {
	DIK_SYSRQ,			0x5c,			/* BREAK(COPY) */
	DIK_F1,				0x5d,			/* PF1 */
	DIK_F2,				0x5e,			/* PF2 */
	DIK_F3,				0x5f,			/* PF3 */
	DIK_F4,				0x60,			/* PF4 */
	DIK_F5,				0x61,			/* PF5 */
	DIK_F6,				0x62,			/* PF6 */
	DIK_F7,				0x63,			/* PF7 */
	DIK_F8,				0x64,			/* PF8 */
	DIK_F9,				0x65,			/* PF9 */
	DIK_F10,			0x66,			/* PF10 */

	DIK_ESCAPE,			0x01,			/* ESC */
	DIK_1,				0x02,			/* 1 */
	DIK_2,				0x03,			/* 2 */
	DIK_3,				0x04,			/* 3 */
	DIK_4,				0x05,			/* 4 */
	DIK_5,				0x06,			/* 5 */
	DIK_6,				0x07,			/* 6 */
	DIK_7,				0x08,			/* 7 */
	DIK_8,				0x09,			/* 8 */
	DIK_9,				0x0a,			/* 9 */
	DIK_0,				0x0b,			/* 0 */
	DIK_MINUS,			0x0c,			/* - */
	DIK_CIRCUMFLEX,		0x0d,			/* ^ */
	DIK_YEN,			0x0e,			/* \ */
	DIK_BACK,			0x0f,			/* BS */

	DIK_TAB,			0x10,			/* TAB */
	DIK_Q,				0x11,			/* Q */
	DIK_W,				0x12,			/* W */
	DIK_E,				0x13,			/* E */
	DIK_R,				0x14,			/* R */
	DIK_T,				0x15,			/* T */
	DIK_Y,				0x16,			/* Y */
	DIK_U,				0x17,			/* U */
	DIK_I,				0x18,			/* I */
	DIK_O,				0x19,			/* O */
	DIK_P,				0x1a,			/* P */
	DIK_AT,				0x1b,			/* @ */
	DIK_LBRACKET,		0x1c,			/* [ */
	DIK_RETURN,			0x1d,			/* CR */

	DIK_LCONTROL,		0x52,			/* CTRL(Ctrl) */
	DIK_A,				0x1e,			/* A */
	DIK_S,				0x1f,			/* S */
	DIK_D,				0x20,			/* D */
	DIK_F,				0x21,			/* F */
	DIK_G,				0x22,			/* G */
	DIK_H,				0x23,			/* H */
	DIK_J,				0x24,			/* J */
	DIK_K,				0x25,			/* K */
	DIK_L,				0x26,			/* L */
	DIK_SEMICOLON,		0x27,			/* ; */
	DIK_COLON,			0x28,			/* : */
	DIK_RBRACKET,		0x29,			/* ] */

	DIK_LSHIFT,			0x53,			/* SHIFT(SHIFT) */
	DIK_Z,				0x2a,			/* Z */
	DIK_X,				0x2b,			/* X */
	DIK_C,				0x2c,			/* C */
	DIK_V,				0x2d,			/* V */
	DIK_B,				0x2e,			/* B */
	DIK_N,				0x2f,			/* N */
	DIK_M,				0x30,			/* M */
	DIK_COMMA,			0x31,			/* , */
	DIK_PERIOD,			0x32,			/* . */
	DIK_SLASH,			0x33,			/* / */
	DIK_UNDERLINE,		0x34,			/* _ */
	DIK_RSHIFT,			0x54,			/* ESHIFT(SHIFT) */

	DIK_CAPITAL,		0x55,			/* CAP(CAPS) */
	DIK_NOCONVERT,		0x56,			/* GRAPH(NFER) */
	/* SPACE 蓖ĂȂ */
	DIK_SPACE,			0x35,			/* ESPACE(SPACE) */
	DIK_KANJI,			0x58,			/* SPACE(XFER) */
	DIK_KANA,			0x5a,			/* (Ji) */

	DIK_INSERT,			0x48,			/* INS(Insert) */
	DIK_DELETE,			0x4b,			/* DEL(Delete) */
	DIK_UP,				0x4d,			/*  */
	DIK_LEFT,			0x4f,			/*  */
	DIK_DOWN,			0x50,			/*  */
	DIK_RIGHT,			0x51,			/*  */

	DIK_HOME,			0x49,			/* EL(HOME CLR) */
	DIK_PRIOR,			0x4a,			/* CLS(ROLL DOWN) */
	DIK_END,			0x4c,			/* DUP(HELP) */
	DIK_NEXT,			0x4e,			/* HOME(ROLL UP) */

	DIK_MULTIPLY,		0x36,			/* Tenkey * */
	DIK_DIVIDE,			0x37,			/* Tenkey / */
	DIK_ADD,			0x38,			/* Tenkey + */
	DIK_SUBTRACT,		0x39,			/* Tenkey - */
	DIK_NUMPAD7,		0x3a,			/* Tenkey 7 */
	DIK_NUMPAD8,		0x3b,			/* Tenkey 8 */
	DIK_NUMPAD9,		0x3c,			/* Tenkey 9 */
	DIK_NUMPADEQUALS,	0x3d,			/* Tenkey = */
	DIK_NUMPAD4,		0x3e,			/* Tenkey 4 */
	DIK_NUMPAD5,		0x3f,			/* Tenkey 5 */
	DIK_NUMPAD6,		0x40,			/* Tenkey 6 */
	DIK_NUMPADCOMMA,	0x41,			/* Tenkey , */
	DIK_NUMPAD1,		0x42,			/* Tenkey 1 */
	DIK_NUMPAD2,		0x43,			/* Tenkey 2 */
	DIK_NUMPAD3,		0x44,			/* Tenkey 3 */
	DIK_NUMPAD0,		0x46,			/* Tenkey 0 */
	DIK_DECIMAL,		0x47,			/* Tenkey . */
	DIK_NUMPADENTER,	0x45			/* Tenkey CR */
};

/*
 *	DirectInputR[hFM-7 R[h
 *	R[hΏƕ\(101L[{[hp)
 */
static BYTE kbd_101_table[] = {
	DIK_ESCAPE,			0x5c,			/* BREAK(ESC) */
	DIK_F1,				0x5d,			/* PF1 */
	DIK_F2,				0x5e,			/* PF2 */
	DIK_F3,				0x5f,			/* PF3 */
	DIK_F4,				0x60,			/* PF4 */
	DIK_F5,				0x61,			/* PF5 */
	DIK_F6,				0x62,			/* PF6 */
	DIK_F7,				0x63,			/* PF7 */
	DIK_F8,				0x64,			/* PF8 */
	DIK_F9,				0x65,			/* PF9 */
	DIK_F10,			0x66,			/* PF10 */

	DIK_GRAVE,			0x01,			/* ESC(`) */
	DIK_1,				0x02,			/* 1 */
	DIK_2,				0x03,			/* 2 */
	DIK_3,				0x04,			/* 3 */
	DIK_4,				0x05,			/* 4 */
	DIK_5,				0x06,			/* 5 */
	DIK_6,				0x07,			/* 6 */
	DIK_7,				0x08,			/* 7 */
	DIK_8,				0x09,			/* 8 */
	DIK_9,				0x0a,			/* 9 */
	DIK_0,				0x0b,			/* 0 */
	DIK_MINUS,			0x0c,			/* - */
	DIK_EQUALS,			0x0d,			/* ^(=) */
	DIK_BACKSLASH,		0x0e,			/* \ */
	DIK_BACK,			0x0f,			/* BS */

	DIK_TAB,			0x10,			/* TAB */
	DIK_Q,				0x11,			/* Q */
	DIK_W,				0x12,			/* W */
	DIK_E,				0x13,			/* E */
	DIK_R,				0x14,			/* R */
	DIK_T,				0x15,			/* T */
	DIK_Y,				0x16,			/* Y */
	DIK_U,				0x17,			/* U */
	DIK_I,				0x18,			/* I */
	DIK_O,				0x19,			/* O */
	DIK_P,				0x1a,			/* P */
	DIK_LBRACKET,		0x1b,			/* @([) */
	DIK_RBRACKET,		0x1c,			/* [(]) */
	DIK_RETURN,			0x1d,			/* CR */

	DIK_LCONTROL,		0x52,			/* CTRL(Ctrl) */
	DIK_A,				0x1e,			/* A */
	DIK_S,				0x1f,			/* S */
	DIK_D,				0x20,			/* D */
	DIK_F,				0x21,			/* F */
	DIK_G,				0x22,			/* G */
	DIK_H,				0x23,			/* H */
	DIK_J,				0x24,			/* J */
	DIK_K,				0x25,			/* K */
	DIK_L,				0x26,			/* L */
	DIK_SEMICOLON,		0x27,			/* ; */
	DIK_APOSTROPHE,		0x28,			/* :(') */
	/* ] 蓖ĂȂ */

	DIK_LSHIFT,			0x53,			/* SHIFT */
	DIK_Z,				0x2a,			/* Z */
	DIK_X,				0x2b,			/* X */
	DIK_C,				0x2c,			/* C */
	DIK_V,				0x2d,			/* V */
	DIK_B,				0x2e,			/* B */
	DIK_N,				0x2f,			/* N */
	DIK_M,				0x30,			/* M */
	DIK_COMMA,			0x31,			/* , */
	DIK_PERIOD,			0x32,			/* . */
	DIK_SLASH,			0x33,			/* / */
	/* _ 蓖ĂȂ */
	DIK_RSHIFT,			0x54,			/* ESHIFT */

	DIK_CAPITAL,		0x55,			/* CAP(Caps Lock) */
	DIK_NUMLOCK,		0x56,			/* GRAPH(Num Lock) */
	/* Xy[X蓖ĂȂ */
	/* Xy[X蓖ĂȂ */
	DIK_SPACE,			0x35,			/* ESPACE(SPACE) */
	DIK_RCONTROL,		0x5a,			/* (ECtrl) */

	DIK_INSERT,			0x48,			/* INS(Insert) */
	DIK_DELETE,			0x4b,			/* DEL(Delete) */
	DIK_UP,				0x4d,			/*  */
	DIK_LEFT,			0x4f,			/*  */
	DIK_DOWN,			0x50,			/*  */
	DIK_RIGHT,			0x51,			/*  */

	DIK_HOME,			0x49,			/* EL(Home) */
	DIK_PRIOR,			0x4a,			/* CLS(Page Up) */
	DIK_END,			0x4c,			/* DUP(End) */
	DIK_NEXT,			0x4e,			/* HOME(Page Down) */

	DIK_MULTIPLY,		0x36,			/* Tenkey * */
	DIK_DIVIDE,			0x37,			/* Tenkey / */
	DIK_ADD,			0x38,			/* Tenkey + */
	DIK_SUBTRACT,		0x39,			/* Tenkey - */
	DIK_NUMPAD7,		0x3a,			/* Tenkey 7 */
	DIK_NUMPAD8,		0x3b,			/* Tenkey 8 */
	DIK_NUMPAD9,		0x3c,			/* Tenkey 9 */
	DIK_NUMPAD4,		0x3e,			/* Tenkey 4 */
	DIK_NUMPAD5,		0x3f,			/* Tenkey 5 */
	DIK_NUMPAD6,		0x40,			/* Tenkey 6 */
	DIK_NUMPAD1,		0x42,			/* Tenkey 1 */
	DIK_NUMPAD2,		0x43,			/* Tenkey 2 */
	DIK_NUMPAD3,		0x44,			/* Tenkey 3 */
	DIK_NUMPAD0,		0x46,			/* Tenkey 0 */
	DIK_DECIMAL,		0x47,			/* Tenkey . */
	DIK_NUMPADENTER,	0x45			/* Tenkey CR */
};

/*
 *	WinNTnL[͑΍
 *	R[he[u
 */
static const BYTE DI_keyno[] = { DIK_KANJI, DIK_KANA, DIK_CAPITAL };

/*
 *	J[\L[  eL[ϊ
 *	R[hΏƕ\
 */
static BYTE TenDirTable[16] = {
	0x00, 0x3b, 0x43, 0xff, 0x3e, 0x3a, 0x42, 0x3e,
	0x40, 0x3c, 0x44, 0x40, 0xff, 0x3b, 0x43, 0xff
};


/*
 *	L[{[h
 *	ftHg}bv擾
 */
void FASTCALL GetDefMapKbd(BYTE *pMap, int mode)
{
	int i;
	int type;

	ASSERT(pMap);
	ASSERT((mode >= 0) && (mode <= 3));

	/* xAׂăNA */
	memset(pMap, 0, 256);

	/* L[{[h^Cv擾 */
	switch (mode) {
		/*  */
		case 0:
			type = GetKeyboardType(1);
			if ((type & 0xf00) == 0xd00) {
				type = 0xd00;
			}
			else {
				if (type != 1) {
					if (GetKeyboardType(0) == 7) {
				type = 0;
			}
					else {
						type = 1;
					}
				}
			}
			break;
		/* 106 */
		case 1:
			type = 0;
			break;
		/* PC-98 */
		case 2:
			type = 0xd00;
			break;
		/* 101 */
		case 3:
			type = 1;
			break;
		/* ̑ */
		default:
			ASSERT(FALSE);
			break;
	}

	/* ϊe[uɃftHglZbg */
	if (type & 0xd00) {
		/* PC-98 */
		for (i=0; i<sizeof(kbd_98_table)/2; i++) {
			pMap[kbd_98_table[i * 2]] = kbd_98_table[i * 2 + 1];
		}
		return;
	}

	if (type == 0) {
		/* 106 */
		for (i=0; i<sizeof(kbd_106_table)/2; i++) {
			pMap[kbd_106_table[i * 2]] = kbd_106_table[i * 2 + 1];
		}
	}
	else {
		/* 101 */
		for (i=0; i<sizeof(kbd_101_table)/2; i++) {
			pMap[kbd_101_table[i * 2]] = kbd_101_table[i * 2 + 1];
		}
	}
}

/*
 *	L[{[h
 *	}bvݒ
 */
void FASTCALL SetMapKbd(BYTE *pMap)
{
	ASSERT(pMap);

	/* Rs[邾 */
	memcpy(kbd_table, pMap, sizeof(kbd_table));

	/* NT/PC-9801΍ : ΍tO(bit7)Lɂ */
	if (bNTkbMode) {
		kbd_table[DIK_KANJI] |= 0x80;
		kbd_table[DIK_CAPITAL] |= 0x80;
		kbd_table[DIK_KANA] |= 0x80;
	}
}

/*
 *	
 */
void FASTCALL InitKbd(void)
{
	BYTE tmp_kbdmap[256];

	/* NT΍􏈗 */
	if (bNTflag || (GetKeyboardType(1) & 0xd00)) {
		bNTkbMode = TRUE;
	}
	else {
		bNTkbMode = FALSE;
	}

	/* [NGA(L[{[h) */
	lpdi = NULL;
	lpkbd = NULL;
	memset(kbd_map, 0, sizeof(kbd_map));
	memset(kbd_table, 0, sizeof(kbd_table));
	keytime = 0;

	/* [NGA(L[{[h NT/PC-9801΍) */
	memset(bNTkeyPushFlag, 0, sizeof(bNTkeyPushFlag));
	memset(bNTkeyMakeFlag, 0, sizeof(bNTkeyMakeFlag));
	bDummyRead = TRUE;
	dmyreadtime = 0;

	/* [NGA(L[obt@/^A^CL[XL) */
	memset(nKeyBuffer, 0, sizeof(nKeyBuffer));
	nKeyReadPtr = 0;
	nKeyWritePtr = 0;
	nLastKey = 0;
	nLastKey2 = 0;
	bKbdReal = FALSE;

	/* [NGA(eL[ϊ) */
	bArrow8Dir = TRUE;
	nTenDir = 0;
	nTenDir2 = 0;

	/* [NGA(WCXeBbN) */
	joytime = 0;
	joytime2 = 0;
	memset(joydat, 0, sizeof(joydat));
	memset(joybk, 0, sizeof(joybk));
	memset(joyrapid, 0, sizeof(joyrapid));
	joyplugged[0] = FALSE;	/* Iɋ@\擾邽ߕύX */
	joyplugged[1] = FALSE;	/*  */
	memset(joycaps, 0, sizeof(joycaps));

#ifdef MOUSE
	/* [NGA(}EX) */
	mostime = 0;
	bCapture = FALSE;
	nMouseX = 0;
	nMouseY = 0;
	bMouseCursor = TRUE;
	nMouseButton = 0xf0;
	uMidBtnMode = MOSCAP_NONE;
	bMouseCapture = FALSE;
#endif

	/* ftHg}bvݒ */
	GetDefMapKbd(tmp_kbdmap, 0);
	SetMapKbd(tmp_kbdmap);
	bAutoDetectKeyMap = FALSE;

	/* eL[G~[V */
	bTenCursor = FALSE;

#ifdef KBDPASTE
	/* L[y[Xgf[^ */
	strPaste = NULL;
	strPtr = NULL;
	cPreKey = 0;
	uWaitCnt = 0;
	uPasteWait = 20;
	uPasteWaitCntl = 200;
#endif
}

/*
 *	N[Abv
 */
void FASTCALL CleanKbd(void)
{
	/* DirectInputDevice(L[{[h) */
	if (lpkbd) {
		lpkbd->Unacquire();
		lpkbd->Release();
		lpkbd = NULL;
	}

#ifdef MOUSE
	/* DirectInputDevice(}EX) */
	if (lpmos) {
		lpmos->Unacquire();
		lpmos->Release();
		lpmos = NULL;
	}
#endif

	/* DirectInput */
	if (lpdi) {
		lpdi->Release();
		lpdi = NULL;
	}

#ifdef KBDPASTE
	if (strPaste) {
		free(strPaste);
		strPaste = NULL;
		strPtr = NULL;
		uWaitCnt = 0;
	}
#endif
}

/*
 *	ZNg
 */
BOOL FASTCALL SelectKbd(HWND hWnd)
{
#ifdef MOUSE
	DIDEVCAPS caps;
#endif

	ASSERT(hWnd);

#ifdef DINPUT8
	/* DirectInputIuWFNg쐬 */
	if (FAILED(DirectInput8Create(hAppInstance, DIRECTINPUT_VERSION,
							IID_IDirectInput8A, (void**)&lpdi, NULL))) {
		return FALSE;
	}
#else
	/* DirectInputIuWFNg쐬 */
	if (FAILED(DirectInputCreate(hAppInstance, DIRECTINPUT_VERSION,
							&lpdi, NULL))) {
		return FALSE;
	}
#endif

	/* L[{[hfoCX쐬 */
	if (FAILED(lpdi->CreateDevice(GUID_SysKeyboard, &lpkbd, NULL))) {
		return FALSE;
	}

	/* L[{[hf[^`ݒ */
	if (FAILED(lpkbd->SetDataFormat(&c_dfDIKeyboard))) {
		return FALSE;
	}

	/* xݒ */
	if (FAILED(lpkbd->SetCooperativeLevel(hWnd,
						DISCL_BACKGROUND | DISCL_NONEXCLUSIVE))) {
		return FALSE;
	}

#ifdef MOUSE
	/* }EXfoCX쐬 */
	bDetectMouse = FALSE;
	if (SUCCEEDED(lpdi->CreateDevice(GUID_SysMouse, &lpmos, NULL))) {
		/* }EXf[^`ݒ */
		if (SUCCEEDED(lpmos->SetDataFormat(&c_dfDIMouse))) {
			/* xݒ */
			if (SUCCEEDED(lpmos->SetCooperativeLevel(hWnd,
						DISCL_BACKGROUND | DISCL_NONEXCLUSIVE))) {
				/* }EX{^擾 */
				caps.dwSize = sizeof(DIDEVCAPS);
				if (SUCCEEDED(lpmos->GetCapabilities(&caps))) {
					nMouseButtons = (BYTE)caps.dwButtons;
					bDetectMouse = TRUE;
				}
			}
		}
	}
#endif

	return TRUE;
}

/*
 *	L[{[h
 *	L[obt@o^
 */
static void FASTCALL PushKeyCode(BYTE code)
{
	int i;

	if (nKeyWritePtr < KEYBUFFER_SIZE) {
		nKeyBuffer[nKeyWritePtr++] = code;
	}
	else {
		for (i=0; i<KEYBUFFER_SIZE; i++) {
			nKeyBuffer[i] = nKeyBuffer[i + 1];
		}
		nKeyBuffer[KEYBUFFER_SIZE - 1] = code;

		if (nKeyReadPtr > 0) {
			nKeyReadPtr --;
		}
	}
}

/*
 *	L[{[h eL[ϊ
 *	R[hϊ
 */
static BYTE FASTCALL Cur2Ten_DirCode(BYTE code)
{
	switch (code) {
		case 0x4d:	/*  */
			return 0x01;
		case 0x4f:	/*  */
			return 0x04;
		case 0x50:	/*  */
			return 0x02;
		case 0x51:	/*  */
			return 0x08;
	}

	return 0;
}

/*
 *	L[{[h eL[ϊ
 *	}XNR[hϊ
 */
static BYTE FASTCALL Cur2Ten_MaskCode(BYTE code)
{
	switch (code) {
		case 0x4d:	/*  */
		case 0x50:	/*  */
			return 0xfc;
		case 0x4f:	/*  */
		case 0x51:	/*  */
			return 0xf3;
	}

	return 0xff;
}

/*
 *	L[{[h
 *	J[\L[eL[ϊ Make
 */
static BOOL FASTCALL Cur2Ten_Make(BYTE code)
{
	BYTE nTenDirOld;
	BYTE dircode;

	/* ȑO̕R[hۑ */
	nTenDirOld = nTenDir;

	/* L[R[hϊ */
	dircode = Cur2Ten_DirCode(code);
	if (dircode) {
		if (bArrow8Dir) {
			/* 8 */
			nTenDir2 = (BYTE)(nTenDir & (BYTE)~Cur2Ten_MaskCode(code));
			nTenDir &= Cur2Ten_MaskCode(code);
			nTenDir |= dircode;
		}
		else {
			if (key_format == KEY_FORMAT_SCAN) {
				if (TenDirTable[dircode]) {
					keyboard_make(TenDirTable[dircode]);
				}
				return TRUE;
			}

			/* 4 */
			nTenDir2 = (BYTE)(nTenDir & (BYTE)Cur2Ten_MaskCode(code));
			nTenDir = dircode;
		}

		/* e[ue0xff̏ꍇAȑgݍ킹 */
		if (TenDirTable[nTenDir] == 0xff) {
			nTenDir = nTenDirOld;
			return TRUE;
		}

		/* L[R[hs */
		if (nTenDir != nTenDirOld) {
			if (TenDirTable[nTenDirOld]) {
				keyboard_break(TenDirTable[nTenDirOld]);
				PushKeyCode(TenDirTable[nTenDir]);
			}
			else {
				keyboard_make(TenDirTable[nTenDir]);
			}
		}
		return TRUE;
	}

	return FALSE;
}

/*
 *	L[{[h
 *	J[\L[eL[ϊ Break
 */
static BOOL FASTCALL Cur2Ten_Break(BYTE code)
{
	BYTE nTenDirOld;
	BYTE dircode;

	/* ȑO̕R[hۑ */
	nTenDirOld = nTenDir;

	/* L[R[hϊ */
	dircode = Cur2Ten_DirCode(code);
	if (dircode) {
		if ((key_format == KEY_FORMAT_SCAN) && !bArrow8Dir) {
			if (TenDirTable[dircode]) {
				keyboard_break(TenDirTable[dircode]);
			}
			return TRUE;
		}

		nTenDir2 &= (BYTE)~dircode;
		if (nTenDir2) {
			nTenDir |= nTenDir2;
			nTenDir2 = 0;
		}
		nTenDir &= (BYTE)~dircode;

		/* e[ue0xff̏ꍇAȑgݍ킹 */
		if (TenDirTable[nTenDir] == 0xff) {
			nTenDir = nTenDirOld;
			return TRUE;
		}

		/* L[R[hs */
		if (nTenDir != nTenDirOld) {
			keyboard_break(TenDirTable[nTenDirOld]);
			if (TenDirTable[nTenDir]) {
				PushKeyCode(TenDirTable[nTenDir]);
			}
			else if (key_format != KEY_FORMAT_SCAN) {
				/* ~"5"𔭍s */
				PushKeyCode(0x3f);
				PushKeyCode(0xbf);
			}
		}
		return TRUE;
	}

	return FALSE;
}

/*
 *	L[{[h
 *	j[o^C}ݒ
 */
void FASTCALL SetMenuExitTimer(void)
{
	/* sYĂ鏈Ƃ̂߂ɔÔߓĂ */
	keyboard_inputmask(FALSE);

	bMenuExit = TRUE;
	menuexittime = dwExecTotal;
}

/*
 *	L[{[h
 *	|[O
 */
void FASTCALL PollKbd(void)
{
	HRESULT hr;
	BYTE buf[256];
	int i;
	BYTE fm7;
	BOOL bFlag;
	static BYTE key_last = 0;

#ifdef KBDPASTE
	/* ŃJEg */
	if (uWaitCnt < (unsigned int)-1) {
		uWaitCnt ++;
	}
#endif

	/* DirectInput`FbN */
	if (!lpkbd) {
		return;
	}

	/* j[o`FbN */
	if (bMenuExit) {
		if ((dwExecTotal - menuexittime) < 300000) {
			return;
		}
		bMenuExit = FALSE;
	}

	/* ŏ`FbN */
	if (bMinimize) {
		return;
	}

	/* ANeBx[g`FbN */
	if (!bActivate) {
		/* ANeBu̓_~[[h^C}XV */
		bDummyRead = TRUE;
		dmyreadtime = dwExecTotal;
		return;
	}

	/* _~[[hԃ`FbN */
	if (bDummyRead && ((dwExecTotal - dmyreadtime) > 100000)) {
		bDummyRead = FALSE;
	}

	/* ԃ`FbN */
	if ((dwExecTotal - keytime) < 20000) {
		return;
	}
	keytime = dwExecTotal;

#ifdef KBDPASTE
	/* y[Xgobt@Ƀf[^܂ĂꍇAɏ */
	if (strPtr) {
		/* XL[h͑ΏۊOƂ */
		if (key_format == KEY_FORMAT_SCAN) {
			/* ̂Ă܂I */
			free(strPaste);
			strPaste = NULL;
			strPtr = NULL;
			uWaitCnt = 0;
			if (bKeyStrokeModeless) {
				SetForegroundWindow(hKeyStrokeDialog);
			}
		}
		else if (*strPtr) {
			if (((cPreKey & 0xe0) == 0) || (cPreKey == 0x7f)) {
				if (uWaitCnt - 1 >= uPasteWaitCntl) {
					if (keyboard_paste(*strPtr)) {
						/* y[XĝŃ|C^߂ */
						cPreKey = (*strPtr);
						strPtr++;
						uWaitCnt = 0;
					}
				}
			}
			else {
				if (uWaitCnt - 1 >= uPasteWait) {
					if (keyboard_paste(*strPtr)) {
						/* y[XĝŃ|C^߂ */
						cPreKey = (*strPtr);
						strPtr++;
						uWaitCnt = 0;
					}
				}
			}
			/* y[Xg͑̃L[͂Ȃ */
			return;
		}
		else {
			/* y[XgÎŉ */
			free(strPaste);
			strPaste = NULL;
			strPtr = NULL;
			uWaitCnt = 0;
			if (bKeyStrokeModeless) {
				SetForegroundWindow(hKeyStrokeDialog);
			}
		}
	}
#endif

	/* L[obt@ɃR[h܂ĂꍇAɏ */
	if (nKeyReadPtr < nKeyWritePtr) {
		fm7 = nKeyBuffer[nKeyReadPtr++];
		if (fm7 == 0xff) {
			/* 0xff : L[s[g^C}ύX */
			keyboard_repeat();
		}
		else if (fm7 & 0x80) {
			/* 0x80-0xfe : BreakR[h */
			keyboard_break((BYTE)(fm7 & 0x7f));
		}
		else {
			/* 0x00-0x7f : MakeR[h */
			keyboard_make(fm7);
		}

		return;
	}
	else {
		nKeyReadPtr = 0;
		nKeyWritePtr = 0;
	}

	/* foCX̃ANZX擾(擾Ă悢) */
	hr = lpkbd->Acquire();
	if ((hr != DI_OK) && (hr != S_FALSE)) {
		return;
	}

	/* foCXԂ擾 */
	if (lpkbd->GetDeviceState(sizeof(buf), buf) != DI_OK) {
		return;
	}

	/* NT/PC-9801΍􏈗 : L[̃tO] */
	if (bNTkbMode) {
		for (i = 0; i < 3; i ++) {
			buf[DI_keyno[i]] = kbd_map[DI_keyno[i]];
			if (bNTkeyPushFlag[i]) {
				buf[DI_keyno[i]] ^= (BYTE)0x80;
				bNTkeyPushFlag[i] = FALSE;
			}
		}
	}

	/* Vista/Windows7 ESHIFT΍􏈗 */
#ifndef DINPUT8
	if (bVistaflag) {
		if (bNTkeyPushFlag[KNT_RSHIFT]) {
			buf[DIK_RSHIFT] |= 0x80;
		}
		else {
			buf[DIK_RSHIFT] &= (BYTE)~0x80;
		}
	}
#endif

	/* tOoff */
	bFlag = FALSE;

	/* ܂ł̏ԂƔrāAɕϊ */
	for (i=0; i<sizeof(buf); i++) {
		if (((buf[i] & 0x80) != (kbd_map[i] & 0x80)) && (!bFlag)) {
			if ((buf[i] & 0x80) || (kbd_table[i] & 0x80)) {
				/* L[ */
				fm7 = (BYTE)(kbd_table[i] & 0x7f);

				/* ALTL[Ƃ͖̓ */
				if ((!kbd_table[DIK_LMENU]) && (buf[DIK_LMENU] & 0x80)) {
					fm7 = 0;
				}
				if ((!kbd_table[DIK_RMENU]) && (buf[DIK_RMENU] & 0x80)) {
					fm7 = 0;
				}

				if (fm7 > 0) {
					if (!bMenuLoop) {
						if (!(bDummyRead && (kbd_table[i] & 0x80))) {

							if ((fm7 >= 0x4d) && (fm7 <= 0x51) && bTenCursor) {
								/* XL[h͖{̃R[hs */
								if ((key_format == KEY_FORMAT_SCAN) && (fm7 != 0x4e)) {
									PushKeyCode(fm7);
								}

								/* XLR[hϊ */
								if (Cur2Ten_Make(fm7)) {
									fm7 = 0;
								}
							}

							/* NT/PC-9801΍ : L[Breakp^C}o^ */
							if (kbd_table[i] & 0x80) {
								if (bNTkeyMakeFlag[kbd_table[i] & 0x7f]) {
									KillTimer(hMainWnd, kbd_table[i]);
									if ((key_format != KEY_FORMAT_SCAN) && (fm7 != 0x5c)) {
										keyboard_break(fm7);
										PushKeyCode(fm7);
									}
								}
								else {
									keyboard_make(fm7);
								}
								bNTkeyMakeFlag[kbd_table[i] & 0x7f] = TRUE;
								SetTimer(hMainWnd, kbd_table[i], 100, NULL);
							}
							else if (fm7) {
								keyboard_make(fm7);
							}
							bDummyRead = FALSE;

							/* ^A^CL[XL */
							if (bKbdReal && (key_format != KEY_FORMAT_SCAN)) {
								if ((fm7 >= 0x3a) && (fm7 <= 0x46)) {
									/* ȑOĂL[΍Ĕs */
									if (nLastKey && key_repeat_flag) {
										PushKeyCode(nLastKey);
									}
									PushKeyCode(0xff);

									/* ĔspɃL[R[hL */
									if (fm7 == 0x3f) {
										nLastKey2 = 0;
									}
									else {
										nLastKey2 = fm7;
									}
								}
								else {
									/* ĔspɃL[R[hL */
									if (fm7 != 0x5c) {
										nLastKey = fm7;
									}
									else {
										nLastKey = 0;
									}
								}
							}
						}

						if (fm7 != 0x5c) {
							key_last = fm7;
						}

						bFlag = TRUE;
					}
				}
			}
			else {
				/* L[ */
				fm7 = (BYTE)(kbd_table[i] & 0x7f);
				if (fm7 > 0) {
					if ((fm7 >= 0x4d) && (fm7 <= 0x51) && bTenCursor) {
						/* XL[h͖{̃R[hs */
						if ((key_format == KEY_FORMAT_SCAN) && (fm7 != 0x4e)) {
							PushKeyCode((BYTE)(fm7 | 0x80));
						}

						/* XLR[hϊ */
						if (Cur2Ten_Break(fm7)) {
							fm7 = 0;
						}
					}

					/* ^A^CL[XL */
					if (bKbdReal && (key_format != KEY_FORMAT_SCAN)) {
						if ((fm7 >= 0x3a) && (fm7 <= 0x46)) {
							/* eL[̏ꍇ */
							PushKeyCode((BYTE)(fm7 | 0x80));

							if (nLastKey2 == fm7) {
								PushKeyCode(0x3f);
								PushKeyCode(0xbf);
								nLastKey2 = 0;
							}

							/* ȑOeL[ȊOĂꍇAĔs */
							if (nLastKey && key_repeat_flag) {
								PushKeyCode(nLastKey);
								PushKeyCode(0xff);
								key_last = nLastKey;
							}
						}
						else {
							/* eL[ȊȌꍇ */
							keyboard_break(fm7);
							if (nLastKey == fm7) {
								nLastKey = 0;
							}

							/* ȑOeL[ĂꍇAĔs */
							if (nLastKey2 && (nLastKey2 != key_last)) {
								PushKeyCode(nLastKey2);
								PushKeyCode(0xff);
								key_last = nLastKey2;
							}
						}
					}
					else if (fm7) {
						keyboard_break(fm7);
					}

					bFlag = TRUE;
				}
			}

			/* f[^Rs[ */
			kbd_map[i] = buf[i];
		}
	}
}

/*
 *	L[{[h
 *	|[OL[擾
 *	VM̃bN͍sĂȂ̂Œ
 */
BOOL FASTCALL GetKbd(BYTE *pBuf)
{
	HRESULT hr;
	int i;

	ASSERT(pBuf);

	/* NA */
	memset(pBuf, 0, 256);

	/* DirectInput`FbN */
	if (!lpkbd) {
		return FALSE;
	}

	/* foCX̃ANZX擾(擾Ă悢) */
	hr = lpkbd->Acquire();
	if ((hr != DI_OK) && (hr != S_FALSE)) {
		return FALSE;
	}

	/* foCXԂ擾 */
	if (lpkbd->GetDeviceState(256, pBuf) != DI_OK) {
		return FALSE;
	}

	/* NT΍ */
	if (bNTkbMode) {
		for (i=0; i<3; i++) {
			pBuf[DI_keyno[i]] = 0;
		}
		if ((GetAsyncKeyState(VKNT_KANJI1) |
			 GetAsyncKeyState(VKNT_KANJI2)) & 1) {
			pBuf[DIK_KANJI] = 0x80;
		}
		if ((GetAsyncKeyState(VKNT_CAPITAL) |
			 GetAsyncKeyState(VK_CAPITAL)) & 1) {
			pBuf[DIK_CAPITAL] = 0x80;
		}
		if ((GetAsyncKeyState(VKNT_KANA) |
			 GetAsyncKeyState(VK_KANA)) & 1) {
			pBuf[DIK_KANA] = 0x80;
		}
	}

#ifndef DINPUT8
	/* Vista/Windows7 ESHIFT΍􏈗 */
	if (bVistaflag) {
		/* Vista/Window7ł0Ԃ͏ɖ */
		pBuf[0x00] = 0x00;

		/* EVtgL[`FbN */
		if (GetAsyncKeyState(VK_RSHIFT)) {
			pBuf[DIK_RSHIFT] = 0x80;
		}
	}
#endif

	return TRUE;
}

/*
 *	L[{[h
 *	L[͏Ԃ̃NAv
 */
void FASTCALL keyclear_request(void)
{
	int i;
	int j;
	BOOL flag;

	if (!bNTkbMode) {
		memset(kbd_map, 0, 256);
	}
	else {
		for (i=0; i<256; i++) {
			flag = TRUE;

			/* EChEbZ[WŎML[ȊȎSẲԂ */
			for (j=0; j<4; j++) {
				if (DI_keyno[j] == i) {
					flag = FALSE;
					break;
				}
			}
			if (flag) {
				kbd_map[i] = 0;
			}
		}
	}

	/* L[ǂݎԊu^C}IɃNA */
	/* (ɂAPollKbdŋIɃL[͏Ԃ𓾂) */
	keytime = 0;
}

/*
 *	WCXeBbN
 *	foCXǂݍ
 */
static BYTE FASTCALL GetJoy(int index, BOOL flag)
{
	int num;
	JOYINFOEX jiex;
	MMRESULT result;
	BYTE ret;

	/* assert */
	ASSERT((index == 0) || (index == 1));

	/* T|[gWCXeBbN擾 */
	num = joyGetNumDevs();
	if (index >= num) {
		joyplugged[index] = FALSE;
		return 0;
	}

	/* f[^NA */
	ret = 0;

	/* f[^擾 */
	memset(&jiex, 0, sizeof(jiex));
	jiex.dwSize = sizeof(jiex);
	jiex.dwFlags = JOY_RETURNX | JOY_RETURNY | JOY_RETURNBUTTONS |
				   JOY_RETURNPOV | JOY_RETURNCENTERED;

	result = joyGetPosEx(index, &jiex);
	if (result == JOYERR_NOERROR) {
		if (!joyplugged[index]) {
			/* foCXoꂽWCXeBbN̋@\擾 */
			joyGetDevCaps(index, &joycaps[index], sizeof(JOYCAPS));
		}
		joyplugged[index] = TRUE;

		/* f[^]() */
		if (jiex.dwXpos < 0x4000) {
			ret |= 0x04;
		}
		if (jiex.dwXpos > 0xc000) {
			ret |= 0x08;
		}
		if (jiex.dwYpos < 0x4000) {
			ret |= 0x01;
		}
		if (jiex.dwYpos > 0xc000) {
			ret |= 0x02;
		}
		if (joycaps[index].wCaps & JOYCAPS_HASPOV) {
			switch (jiex.dwPOV) {
				case 0:
					ret |= 0x01;
					break;
				case 4500:
					ret |= 0x09;
					break;
				case 9000:
					ret |= 0x08;
					break;
				case 13500:
					ret |= 0x0a;
					break;
				case 18000:
					ret |= 0x02;
					break;
				case 22500:
					ret |= 0x06;
					break;
				case 27000:
					ret |= 0x04;
					break;
				case 31500:
					ret |= 0x05;
					break;
			}
		}

		/* f[^]({^) */
		if (jiex.dwButtons & JOY_BUTTON1) {
			ret |= 0x20;
		}
		if (jiex.dwButtons & JOY_BUTTON2) {
			ret |= 0x10;
		}
	}
	else if (flag) {
		joyplugged[index] = FALSE;
		ret = 0;
	}

	return ret;
}

/*
 *	WCXeBbN
 *	A˃JE^e[u(БA20msP)
 */
static const BYTE JoyRapidCounter[] = {
	0,			/* Ȃ */
	25,			/* 1Vbg */
	12,			/* 2Vbg */
	8,			/* 3Vbg */
	6,			/* 4Vbg */
	5,			/* 5Vbg */
	4,			/* 6Vbg */
	3,			/* 8Vbg */
	2,			/* 12Vbg */
	1			/* 25Vbg */
};

/*
 *	WCXeBbN
 *	foCXǂݍ(A˂)
 */
static BYTE FASTCALL GetRapidJoy(int index, BOOL flag)
{
	int i;
	BYTE bit;
	BYTE dat;

	/* assert */
	ASSERT((index == 0) || (index == 1));

	/* ڑ`FbN1 (ڑ`FbN͒ʂ) */
	if ((!flag) && (!joyplugged[index])) {
		return 0x00;
	}

	/* f[^擾 */
	dat = GetJoy(index, flag);

	/* ڑ`FbN2 */
	if (!joyplugged[index]) {
		return 0x00;
	}

	/* {^`FbN */
	bit = 0x10;
	for (i=0; i<2; i++) {
		if ((dat & bit) && (nJoyRapid[index][i] > 0)) {
			/* A˂ŉĂBJE^`FbN */
			if (joyrapid[index][i] == 0) {
				/* JE^ */
				joyrapid[index][i] = JoyRapidCounter[nJoyRapid[index][i]];
			}
			else {
				/* JE^fNg */
				joyrapid[index][i]--;
				if ((joyrapid[index][i] & 0xff) == 0) {
					/* ]^C~OȂ̂ŁAԂZĔ] */
					joyrapid[index][i] += JoyRapidCounter[nJoyRapid[index][i]];
					joyrapid[index][i] ^= 0x100;
				}
			}
			/* {^ĂȂ悤ɐU镑 */
			if (joyrapid[index][i] >= 0x100) {
				dat &= (BYTE)(~bit);
			}
		}
		else {
			/* {^ĂȂ̂ŁAA˃JE^NA */
			joyrapid[index][i] = 0;
		}
		/* ̃rbg */
		bit <<= 1;
	}

	return dat;
}

/*
 *	WCXeBbN
 *	R[hϊ
 */
static BYTE FASTCALL PollJoyCode(int code)
{
	/* 0x70͖ */
	if (code < 0x70) {
		return 0;
	}

	/* 0x70㉺E */
	switch (code) {
		/*  */
		case 0x70:
			return 0x01;
		/*  */
		case 0x71:
			return 0x02;
		/*  */
		case 0x72:
			return 0x04;
		/* E */
		case 0x73:
			return 0x08;
		/* A{^ */
		case 0x74:
			return 0x10;
		/* B{^ */
		case 0x75:
			return 0x20;
		/* ȊO */
		default:
			ASSERT(FALSE);
			break;
	}

	return 0;
}

/*
 *	WCXeBbN
 *	|[O(WCXeBbN)
 */
static BYTE FASTCALL PollJoySub(int index, BYTE dat)
{
	int i;
	BYTE ret;
	BYTE bit;

	/* assert */
	ASSERT((index == 0) || (index == 1));

	/* If[^NA */
	ret = 0;

	/*  */
	bit = 0x01;
	for (i=0; i<4; i++) {
		/* {^Ă邩 */
		if (dat & bit) {
			/* R[hϊ */
			ret |= PollJoyCode(nJoyCode[index][i]);
		}
		bit <<= 1;
	}

	/* Z^[`FbN */
	if ((dat & 0x0f) == 0) {
		if ((joybk[index] & 0x0f) != 0) {
			ret |= PollJoyCode(nJoyCode[index][4]);
		}
	}

	/* {^ */
	if (dat & 0x10) {
		ret |= PollJoyCode(nJoyCode[index][5]);
	}
	if (dat & 0x20) {
		ret |= PollJoyCode(nJoyCode[index][6]);
	}

	return ret;
}

/*
 *	WCXeBbN
 *	|[O(L[{[h)
 */
static void FASTCALL PollJoyKbd(int index, BYTE dat)
{
	BYTE bit;
	int i;

	/* ㉺E */
	bit = 0x01;
	for (i=0; i<4; i++) {
		if (dat & bit) {
			/* ߂ĉꂽAmakes */
			if ((joybk[index] & bit) == 0) {
				if ((nJoyCode[index][i] > 0) && (nJoyCode[index][i] <= 0x66)) {
					keyboard_make((BYTE)nJoyCode[index][i]);
				}
			}
		}
		else {
			/* ߂ėꂽAbreaks */
			if ((joybk[index] & bit) != 0) {
				if ((nJoyCode[index][i] > 0) && (nJoyCode[index][i] <= 0x66)) {
					keyboard_break((BYTE)nJoyCode[index][i]);
				}
			}
		}
		bit <<= 1;
	}

	/* Z^[`FbN */
	if ((dat & 0x0f) == 0) {
		if ((joybk[index] & 0x0f) != 0) {
			/* make/break𑱂ďo */
			if ((nJoyCode[index][4] > 0) && (nJoyCode[index][4] <= 0x66)) {
				keyboard_make((BYTE)nJoyCode[index][4]);
				keyboard_break((BYTE)nJoyCode[index][4]);
			}
		}
	}

	/* {^ */
	bit = 0x10;
	for (i=0; i<2; i++) {
		if (dat & bit) {
			/* ߂ĉAmakes */
			if ((joybk[index] & bit) == 0) {
				if ((nJoyCode[index][i + 5] > 0) && (nJoyCode[index][i + 5] <= 0x66)) {
					keyboard_make((BYTE)nJoyCode[index][i + 5]);
				}
			}
		}
		else {
			/* ߂ėꂽAbreaks */
			if ((joybk[index] & bit) != 0) {
				if ((nJoyCode[index][i + 5] > 0) && (nJoyCode[index][i + 5] <= 0x66)) {
					keyboard_break((BYTE)nJoyCode[index][i + 5]);
				}
			}
		}
		bit <<= 1;
	}
}

/*
 *	WCXeBbN
 *	|[O
 */
void FASTCALL PollJoy(void)
{
	BYTE dat;
	BOOL check;
	int i;

	/* Ԋu`FbN(|[Op) */
	if ((dwExecTotal - joytime) < 10000) {
		return;
	}
	joytime = dwExecTotal;

	/* f[^NA */
	memset(joydat, 0, sizeof(joydat));

	/* `FbN */
	if ((nJoyType[0] == 0) && (nJoyType[1] == 0)) {
		return;
	}

	/* Ԋu`FbN(ڑ`FbNp) */
	if ((dwExecTotal - joytime2) >= 500000) {
		joytime2 = dwExecTotal;
		check = TRUE;
	}
	else {
		check = FALSE;
	}

	/* foCX[v */
	for (i=0; i<2; i++) {
		/* f[^擾(A˂) */
		dat = GetRapidJoy(i, check);
		if (!joyplugged[i]) {
			continue;
		}

		/* ^Cv */
		switch (nJoyType[i]) {
			/* WCXeBbN|[g1 */
			case 1:
				joydat[0] = PollJoySub(i, dat);
				break;
			/* WCXeBbN|[g2 */
			case 2:
				joydat[1] = PollJoySub(i, dat);
				break;
			/* L[{[h */
			case 3:
				PollJoyKbd(i, dat);
				break;
			/* dgVЃWCXeBbN */
			case 4:
				joydat[2] = PollJoySub(i, dat);
				break;
			/* ASCII QuickShot JOY I/F */
			case 5:
				joydat[3] = PollJoySub(i, dat);
				break;
			/* h!FM JOY I/F #1 */
			case 6:
				joydat[4] = PollJoySub(i, dat);
				break;
			/* h!FM JOY I/F #2 */
			case 7:
				joydat[5] = PollJoySub(i, dat);
				break;
		}

		/* f[^XV */
		joybk[i] = dat;
	}
}

/*
 *	WCXeBbN
 *	f[^NGXg
 */
BYTE FASTCALL joy_request(BYTE no)
{
/*	ASSERT((no >= 0) && (no < 6)); */
	ASSERT(no < 6);

	return joydat[no];
}

#ifdef MOUSE
/*
 *	}EX
 *	|[O
 */
void FASTCALL PollMos(void)
{
	DIMOUSESTATE buf;
	HRESULT hr;

	/* DirectInput`FbN */
	if (!lpmos) {
		return;
	}

	/* }EX݃`FbN */
	if (!bDetectMouse) {
		return;
	}

	/* ANeBx[g`FbN */
	if (!bActivate) {
		return;
	}

	/* ԃ`FbN */
	if ((dwExecTotal - mostime) < 10000) {
		return;
	}
	mostime = dwExecTotal;

	/* foCX̃ANZX擾(擾Ă悢) */
	hr = lpmos->Acquire();
	if ((hr != DI_OK) && (hr != S_FALSE)) {
		return;
	}

	/* foCXԂ擾 */
	if (lpmos->GetDeviceState(sizeof(DIMOUSESTATE), &buf) != DI_OK) {
		return;
	}

	/* ړE{^Ԃ̐ݒ */
	if (bCapture && bMouseCapture) {
		/* }EXړ~ */
		nMouseX -= buf.lX;
		nMouseY -= buf.lY;

		/* {^Ԃݒ */
		nMouseButton = 0xf0;
		if (buf.rgbButtons[0]) {
			/* {^ */
			nMouseButton &= ~0x10;
		}
		if (buf.rgbButtons[1]) {
			/* E{^ */
			nMouseButton &= ~0x20;
		}
	}
}


/*
 *	}EX
 *	Lv`Ԑݒ
 */
void FASTCALL SetMouseCapture(BOOL en)
{
	DIMOUSESTATE buf;
	HRESULT hr;
	RECT rect;
	POINT center;

	ASSERT(hDrawWnd);

	/* Lv`~/VM~/ANeBu͋Iɖ */
	if (!bMouseCapture || stopreq_flag || !run_flag || !bActivate || !lpmos) {
		en = FALSE;
	}

	/* J[\\/ */
	if (bMouseCursor == en) {
		ShowCursor(!en);
		bMouseCursor = !en;
	}

	/* Lv`ԂɕωȂ΋A */
	if (bCapture == en) {
		return;
	}

	if (en) {
		if (hDrawWnd) {
			/* J[\ʒuۑ */
			GetCursorPos(&center);
			nMouseSX = center.x;
			nMouseSY = center.y;

			/* `EBhE̎l̍W߂ */
			GetWindowRect(hDrawWnd, &rect);

			/* l̍W璆SW߂ */
			center.x = (rect.right + rect.left) / 2;
			center.y = (rect.bottom + rect.top) / 2;

			/* J[\EBhEɌŒ */
			rect.left   = center.x;
			rect.right  = center.x;
			rect.top    = center.y;
			rect.bottom = center.y;
			SetCursorPos(center.x, center.y);
			ClipCursor(&rect);
		}
	}
	else {
		/* Nbv */
		ClipCursor(0);

		/* J[\ʒu𕜌 */
		SetCursorPos(nMouseSX, nMouseSY);
	}

	/* Lv`Ԃۑ */
	bCapture = en;

	/* }EXړNA */
	nMouseX = 0;
	nMouseY = 0;

	/* foCXԂǂ */
	if (lpmos) {
		hr = lpmos->Acquire();
		if ((hr == DI_OK) || (hr == S_FALSE)) {
			lpmos->GetDeviceState(sizeof(DIMOUSESTATE), &buf);
		}
	}
}

/*
 *	}EX
 *	ړf[^NGXg
 */
void FASTCALL mospos_request(BYTE *move_x, BYTE *move_y)
{
	if (bCapture && bMouseCapture) {
		/* ړ𕄍tWrbg͈̔͂Ɏ߂ */
		if (nMouseX > 127) {
			nMouseX = 127;
		}
		else if (nMouseX < -127) {
			nMouseX = -127;
		}

		if (nMouseY > 127) {
			nMouseY = 127;
		}
		else if (nMouseY < -127) {
			nMouseY = -127;
		}

		*move_x = (BYTE)nMouseX;
		*move_y = (BYTE)nMouseY;
	}
	else {
		/* Lv`ꎞ~͈ړĂȂƂɂ */
		*move_x = 0;
		*move_y = 0;
	}

	/* }EXړNA */
	nMouseX = 0;
	nMouseY = 0;
}

/*
 *	}EX
 *	{^f[^NGXg
 */
BYTE FASTCALL mosbtn_request(void)
{
	/* {^Ԃ */
	return nMouseButton;
}

#endif	/* MOUSE */

#ifdef KBDPASTE
/*
 *	L[{[h
 *	y[XgiNbv{[hj
 */
void FASTCALL PasteClipboardKbd(HWND hWnd)
{
	BOOL b;
	HGLOBAL hg;
	PTSTR strClip;

	ASSERT(hWnd);

	b = OpenClipboard(hWnd);
	hg = GetClipboardData(CF_TEXT);
	if (b && hg) {
		strClip = (PTSTR)GlobalLock(hg);
		if (strPaste) {
			free(strPaste);
			strPaste = NULL;
			strPtr = NULL;
			uWaitCnt = 0;
		}
		strPtr = strPaste = (PTSTR)malloc(GlobalSize(hg));
		if (strPaste) {
			lstrcpy(strPaste, strClip);
		}
		GlobalUnlock(hg);
		CloseClipboard();
	}
}

/*
 *	L[{[h
 *	y[Xgij
 */
void FASTCALL PasteKbd(char *pstring)
{
	PTSTR strKbd;

	strKbd = (PTSTR)pstring;
	if (strPaste) {
		free(strPaste);
		strPaste = NULL;
		strPtr = NULL;
		uWaitCnt = 0;
	}
	strPtr = strPaste = (PTSTR)malloc(strlen(pstring));
	if (strPaste) {
		lstrcpy(strPaste, strKbd);
	}
}
#endif

#endif	/* _WIN32 */
