/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API TuEBhEQ ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>
#include <stdlib.h>
#include "xm7.h"
#include "mainetc.h"
#include "fdc.h"
#include "opn.h"
#include "subctrl.h"
#include "dmac.h"
#include "display.h"
#include "aluline.h"
#include "bubble.h"
#include "mouse.h"
#include "w32.h"
#include "w32_res.h"
#include "w32_snd.h"
#include "w32_sub.h"
#include "w32_kbd.h"
#include "w32_draw.h"

#ifdef OPNAMODE
#define	OPN_CHNUM	(3 * 16 + 3)
#else
#define	OPN_CHNUM	(3 * 6 + 3)
#endif

/*
 *	O[o [N
 */
#if XM7_VER == 1 && defined(BUBBLE)
BOOL bBMC;								/* BMC \[h */
#endif
int nOPNRegMode;						/* OPNWX^ w背x */
int nOPNDispMode;						/* OPNfBXvC w背x */

/*
 *	X^eBbN [N
 */
static BYTE *pFDC;						/* FDC Drawobt@ */
static BYTE *pOPNReg;					/* OPNWX^ Drawobt@ */
static UINT nOPNReg;					/* OPNWX^ x */
static HMENU hOPNRegMenu;				/* OPNWX^ j[nh */
static BYTE *pOPNDisp;					/* OPNfBXvC rbg}bvobt@ */
static HBITMAP hOPNDisp;				/* OPNfBXvC rbg}bvnh */
static RECT rOPNDisp;					/* OPNfBXvC ` */
static UINT nOPNDisp;					/* OPNfBXvC x */
static HMENU hOPNDispMenu;				/* OPNfBXvC j[nh */
static int knOPNDisp[OPN_CHNUM];		/* OPNfBXvC Ճ[N */
static int ktOPNDisp[OPN_CHNUM];		/* OPNfBXvC Ճ[N */
static BYTE cnOPNDisp[OPN_CHNUM][49 * 2];/* OPNfBXvC [N */
static BYTE ctOPNDisp[OPN_CHNUM][49 * 2];/* OPNfBXvC [N */
static int lnOPNDisp[OPN_CHNUM];		/* OPNfBXvC x[N */
static int ltOPNDisp[OPN_CHNUM];		/* OPNfBXvC x[N */
static BYTE *pSubCtrl;					/* TuRg[ Drawobt@ */
#if XM7_VER >= 2
static BYTE *pALULine;					/* _Z/ Drawobt@ */
#endif
#if XM7_VER == 1 && defined(SFDC)
static UINT nFDC;						/* FDC x */
#endif	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER == 1 && defined(BUBBLE)
static BYTE *pBMC;						/* BMC Drawobt@ */
static UINT nBMC;						/* BMC x */
static HMENU hBMC;						/* BMC j[nh */
#endif

/*
 *	OPNx`
 */
enum {
	LEVEL_AUTO,
	LEVEL_PSG,
	LEVEL_OPN,
	LEVEL_WHG,
	LEVEL_THG,
} OPN_LEVEL;

/*-[ TuCPURg[EChE ]-----------------------------------------*/

/*
 *	TuCPURg[EChE
 *	ZbgAbv(LRAM)
 */
static void FASTCALL SetupSubCtrlShared(BYTE *p, int cx, int y)
{
	char string[128];
	char temp[4];
	int i, j;

	ASSERT(p);
	ASSERT(cx > 0);

	/* ^Cg */
	strncpy(string, "Shared RAM:", sizeof(string));
	memcpy(&p[cx * y], string, strlen(string));
	y++;

	/* [v */
	for (i=0; i<8; i++) {
		/* 쐬 */
		_snprintf(string, sizeof(string), "+%02X:", i * 16);
		for (j=0; j<16; j++) {
			_snprintf(temp, sizeof(temp), " %02X", shared_ram[i * 16 + j]);
			strncat(string, temp, sizeof(string) - strlen(string) - 1);
		}

		/* Zbg */
		memcpy(&p[cx * y], string, strlen(string));
		y++;
	}
}

/*
 *	TuCPURg[EChE
 *	ZbgAbv(tO)
 */
static void FASTCALL SetupSubCtrlFlag(BYTE *p, int cx, int x, int y,
									char *title, BOOL flag)
{
	int i;

	ASSERT(p);
	ASSERT(cx > 0);

	/*  */
	memset(&p[cx * y + x], 0x20, 22);

	/* ^CgZbg */
	for (i=0; i<(int)strlen(title); i++) {
		if (title[i] == '\0') {
			break;
		}
		p[cx * y + x + i] = title[i];
	}

	/* On܂Off */
	if (flag) {
		memcpy(&p[cx * y + x + 20], "On", 2);
	}
	else {
		memcpy(&p[cx * y + x + 19], "Off", 3);
	}
}


#if XM7_VER == 1 && defined(L4CARD)
/*
 *	TuCPURg[EChE
 *	ZbgAbv(CRTCWX^)
 */
static void FASTCALL SetupSubCtrlCRTC(BYTE *p, int cx, int y)
{
	char string[128];
	char temp[4];
	int i, j;

	ASSERT(p);
	ASSERT(cx > 0);

	/* ^Cg */
	strncpy(string, "CRTC Register:", sizeof(string));
	memcpy(&p[cx * y], string, strlen(string));
	y++;

	/* [v */
	for (i=0; i<2; i++) {
		/* 쐬 */
		_snprintf(string, sizeof(string), "+%02X:", i * 16);
		for (j=0; j<16; j++) {
			_snprintf(temp, sizeof(temp), " %02X", crtc_register[i * 16 + j]);
			if (((i * 16 + j) == crtc_regnum) && enable_400line) {
				temp[1] |= 0x80;
				temp[2] |= 0x80;
			}
			strncat(string, temp, sizeof(string) - strlen(string) - 1);
		}

		/* Zbg */
		memcpy(&p[cx * y], string, strlen(string));
		y++;
	}
}
#endif

/*
 *	TuCPURg[EChE
 *	ZbgAbv
 */
static void FASTCALL SetupSubCtrl(BYTE *p, int x, int y)
{
#if XM7_VER >= 3
#define	YOFS	7
#else
#define	YOFS	6
#endif

	char string[128];

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* FM-7݊tO */
	SetupSubCtrlFlag(p, x,  0, 0, "Sub Halt", subhalt_flag);
	SetupSubCtrlFlag(p, x,  0, 1, "Busy Flag", subbusy_flag);
	SetupSubCtrlFlag(p, x,  0, 2, "CRT Display", crt_flag);
	SetupSubCtrlFlag(p, x, 30, 0, "Cancel IRQ", subcancel_flag);
	SetupSubCtrlFlag(p, x, 30, 1, "Attention FIRQ", subattn_flag);
	SetupSubCtrlFlag(p, x, 30, 2, "VRAM Access Flag", vrama_flag);

#if XM7_VER >= 2
	/* Tuj^ROM */
	strncpy(string, "Sub Monitor", sizeof(string));
	memcpy(&p[x * 4 + 0], string, strlen(string));
	switch (subrom_bank) {
		case 0:
			strncpy(string, "Type-C", sizeof(string));
			break;
		case 1:
			strncpy(string, "Type-A", sizeof(string));
			break;
		case 2:
			strncpy(string, "Type-B", sizeof(string));
			break;
		case 3:
			strncpy(string, "CG ROM", sizeof(string));
			break;
#if XM7_VER >= 3
		case 4:
			strncpy(string, "   RAM", sizeof(string));

			/* Type-D/EF */
			if ((subramde[0x1fe0] == 'S') &&
				(subramde[0x1fe1] == 'U') &&
				(subramde[0x1fe2] == 'B') &&
				(subramde[0x1fe3] == '8')) {
				if (subramde[0x1fe4] == 'A') {
					strncpy(string,"Type-D", sizeof(string));
				}
				else if (subramde[0x1fe4] == 'B') {
					strncpy(string,"Type-E", sizeof(string));
				}
			}
			break;
#endif
		default:
			ASSERT(FALSE);
			break;
	}
	memcpy(&p[x * 4 + 16], string, strlen(string));

	/* 320[h */
	strncpy(string, "Display Mode", sizeof(string));
	memcpy(&p[x * 4 + 30], string, strlen(string));
#if XM7_VER >= 3
	if (mode400l) {
		strncpy(string, "640x400", sizeof(string));
	}
	else if (mode256k) {
		strncpy(string, "262,144", sizeof(string));
	}
	else if (mode320) {
#else
	if (mode320) {
#endif
		strncpy(string, "320x200", sizeof(string));
	}
	else {
		strncpy(string, "640x200", sizeof(string));
	}
	memcpy(&p[x * 4 + 45], string, strlen(string));

	/* CGoN */
#if XM7_VER >= 3
	if (subrom_bank == 4) {
		strncpy(string, "CG RAM", sizeof(string));
		memcpy(&p[x * 5 + 0], string, strlen(string));
		_snprintf(string, sizeof(string), "Bank %1d", cgram_bank);
	}
	else {
		strncpy(string, "CG ROM", sizeof(string));
		memcpy(&p[x * 5 + 0], string, strlen(string));
		_snprintf(string, sizeof(string), "Bank %1d", cgrom_bank);
	}
#else
	strncpy(string, "CG ROM", sizeof(string));
	memcpy(&p[x * 5 + 0], string, strlen(string));
	_snprintf(string, sizeof(string), "Bank %1d", cgrom_bank);
#endif
	memcpy(&p[x * 5 + 16], string, strlen(string));

	/* TuZbg */
	strncpy(string, "Sub Reset", sizeof(string));
	memcpy(&p[x * 5 + 30], string, strlen(string));
	if (subreset_flag) {
		strncpy(string, "Software", sizeof(string));
	}
	else {
		strncpy(string, "Hardware", sizeof(string));
	}
	memcpy(&p[x * 5 + 44], string, strlen(string));

	/* VRAMItZbgWX^0 */
	strncpy(string, "VRAM Offset 0", sizeof(string));
	memcpy(&p[x * (YOFS + 0) + 30], string, strlen(string));
	_snprintf(string, sizeof(string), "$%04X", vram_offset[0]);
	memcpy(&p[x * (YOFS + 0) + 47], string, strlen(string));

	/* VRAMItZbgWX^1 */
	strncpy(string, "VRAM Offset 1", sizeof(string));
	memcpy(&p[x * (YOFS + 1) + 30], string, strlen(string));
	_snprintf(string, sizeof(string), "$%04X", vram_offset[1]);
	memcpy(&p[x * (YOFS + 1) + 47], string, strlen(string));

	/* ANeBuy[W */
	strncpy(string, "Active Page", sizeof(string));
	memcpy(&p[x * (YOFS + 0) + 0], string, strlen(string));
#if XM7_VER >= 3
	if (screen_mode & SCR_AV40) {
		_snprintf(string, sizeof(string), "Page %1d", subram_vrambank);
	}
	else {
		_snprintf(string, sizeof(string), "Page %1d", vram_active);
	}
#else
	_snprintf(string, sizeof(string), "Page %1d", vram_active);
#endif
	memcpy(&p[x * (YOFS + 0) + 16], string, strlen(string));

	/* \y[W */
	strncpy(string, "Display Page", sizeof(string));
	memcpy(&p[x * (YOFS + 1) + 0], string, strlen(string));
	_snprintf(string, sizeof(string), "Page %1d", vram_display);
	memcpy(&p[x * (YOFS + 1) + 16], string, strlen(string));

#if XM7_VER >= 3
	/* R\[RAMoN */
	strncpy(string, "Console RAM", sizeof(string));
	memcpy(&p[x * 6 + 0], string, strlen(string));
	if (subrom_bank == 4) {
		_snprintf(string, sizeof(string), "Bank %1d", consram_bank);
	}
	else {
		_snprintf(string, sizeof(string), "Bank 0");
	}
	memcpy(&p[x * 6 + 16], string, strlen(string));

	/* Tuj^RAMveNg */
	strncpy(string, "Sub Protect", sizeof(string));
	memcpy(&p[x * 6 + 30], string, strlen(string));
	if (subram_protect) {
		strncpy(string, " Enable", sizeof(string));
	}
	else {
		strncpy(string, "Disable", sizeof(string));
	}
	memcpy(&p[x * 6 + 45], string, strlen(string));

	/* ANeBuubN */
	strncpy(string, "Active Block", sizeof(string));
	memcpy(&p[x * 9 + 0], string, strlen(string));
	_snprintf(string, sizeof(string), "Block %1d", block_active);
	memcpy(&p[x * 9 + 15], string, strlen(string));

	/* \ubN */
	strncpy(string, "Display Block", sizeof(string));
	memcpy(&p[x * 10 + 0], string, strlen(string));
	_snprintf(string, sizeof(string), "Block %1d", block_display);
	memcpy(&p[x * 10 + 15], string, strlen(string));

	/* n[hEFAEBhE */
	strncpy(string, "Window Start", sizeof(string));
	memcpy(&p[x * 9 + 30], string, strlen(string));
	strncpy(string, "Window End", sizeof(string));
	memcpy(&p[x * 10 + 30], string, strlen(string));
	if (window_open) {
		_snprintf(string, sizeof(string), "(%3d,%3d)", window_dx1, window_dy1);
		memcpy(&p[x * 9 + 43], string, strlen(string));
		_snprintf(string, sizeof(string), "(%3d,%3d)",
			window_dx2 - 1, window_dy2 - 1);
	}
	else {
		strncpy(string, "(  0,  0)", sizeof(string));
		memcpy(&p[x * 9 + 43], string, strlen(string));
	}
	memcpy(&p[x * 10 + 43], string, strlen(string));

	/* LRAM */
	SetupSubCtrlShared(p, x, 12);
#else
	/* LRAM */
	SetupSubCtrlShared(p, x, 9);
#endif
#elif defined(L4CARD)
	/* ʃ[h */
	strncpy(string, "Display Mode", sizeof(string));
	memcpy(&p[x * 4 + 0], string, strlen(string));
	if (enable_400line) {
		strncpy(string, "640x400", sizeof(string));
	}
	else {
		strncpy(string, "640x200", sizeof(string));
	}
	memcpy(&p[x * 4 + 15], string, strlen(string));

	/* VRAMItZbgWX^ */
	strncpy(string, "VRAM Offset", sizeof(string));
	memcpy(&p[x * 4 + 30], string, strlen(string));
	_snprintf(string, sizeof(string), "$%04X", vram_offset[0]);
	memcpy(&p[x * 4 + 47], string, strlen(string));

	/* eLXgX^[gAhX */
	strncpy(string, "Text Start Addr", sizeof(string));
	memcpy(&p[x * 5 + 0], string, strlen(string));
	_snprintf(string, sizeof(string), "$%04X", text_start_addr);
	memcpy(&p[x * 5 + 17], string, strlen(string));

	/* J[\AhXWX^ */
	strncpy(string, "Cursor Address", sizeof(string));
	memcpy(&p[x * 5 + 30], string, strlen(string));
	_snprintf(string, sizeof(string), "$%04X", cursor_addr);
	memcpy(&p[x * 5 + 47], string, strlen(string));

	/* CRTCWX^ */
	SetupSubCtrlCRTC(p, x, 7);

	/* LRAM(V1,L4) */
	SetupSubCtrlShared(p, x, 11);
#else
	/* LRAM(V1,L2) */
	SetupSubCtrlShared(p, x, 4);
#endif
#undef	YOFS
}


/*
 *	TuCPURg[EChE
 *	`
 */
static void FASTCALL DrawSubCtrl(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pSubCtrl) {
		return;
	}
	SetupSubCtrl(pSubCtrl, x, y);

	/* ` */
#if XM7_VER == 1 && defined(L4CARD)
	DrawWindowText2(hDC, pSubCtrl, x, y);
#else
	DrawWindowText(hDC, pSubCtrl, x, y);
#endif
}

/*
 *	TuCPURg[EChE
 *	tbV
 */
void FASTCALL RefreshSubCtrl(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_SUBCTRL] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_SUBCTRL];
	hDC = GetDC(hWnd);
	DrawSubCtrl(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	TuCPURg[EChE
 *	ĕ`
 */
static void FASTCALL PaintSubCtrl(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pSubCtrl;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawSubCtrl(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	TuCPURg[EChE
 *	EChEvV[W
 */
static LRESULT CALLBACK SubCtrlProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintSubCtrl(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pSubCtrl, NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	TuCPURg[EChE
 *	EChE쐬
 */
HWND FASTCALL CreateSubCtrl(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_SubCtrl";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 52;
#if XM7_VER >= 3
	rect.bottom = lCharHeight * 21;
#elif XM7_VER >= 2
	rect.bottom = lCharHeight * 18;
#elif defined(L4CARD)
	rect.bottom = lCharHeight * 20;
#else
	rect.bottom = lCharHeight * 13;
#endif

	/* EChE^CgAobt@m */
	GetMsg(hResInstance, IDS_SWND_SUBCTRL,
				szWndName, sizeof(szWndName));
	pSubCtrl = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = SubCtrlProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ _Z/ԃEChE ]-------------------------------------------*/

#if XM7_VER >= 2
/*
 *	_Z/ԃEChE
 *	ZbgAbv
 */
static void FASTCALL SetupALULine(BYTE *p, int x, int y)
{
	static const char *ALUString[8] = {
		"PSET", "----", " OR ", " AND", " XOR", " NOT", "TILE", " CMP"
	};

	char string[128];
	int i;

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* _ZR}h */
	strncpy(string, "ALU", sizeof(string));
	memcpy(&p[x * 0 + 0], string, strlen(string));
	if (alu_command & 0x80) {
		strncpy(string, ALUString[alu_command & 0x07], sizeof(string));
	}
	else {
		strncpy(string, " Off", sizeof(string));
	}
	memcpy(&p[x * 0 + 17], string, strlen(string));

	/* ݃[h */
	strncpy(string, "Write Mode", sizeof(string));
	memcpy(&p[x * 1 + 0], string, strlen(string));
	if (alu_command & 0x40) {
		if (alu_command & 0x20) {
			strncpy(string, "Not Eq", sizeof(string));
		}
		else {
			strncpy(string, " Equal", sizeof(string));
		}
	}
	else {
		strncpy(string, "Always", sizeof(string));
	}
	memcpy(&p[x * 1 + 15], string, strlen(string));

	/* ZJ[f[^ */
	_snprintf(string, sizeof(string), "Color               %01d",
		alu_color & 0x07);
	memcpy(&p[x * 2 + 0], string, strlen(string));

	/* }XNrbg */
	_snprintf(string, sizeof(string), "Mask               %02X", alu_mask);
	memcpy(&p[x * 3 + 0], string, strlen(string));

	/* RyArbg */
	strncpy(string, "Compare      --------", sizeof(string));
	for (i=0; i<8; i++) {
		if (~alu_cmpdat[i] & 0x80) {
			string[13 + i] = (char)(0x30 | (alu_cmpdat[i] & 0x07));
		}
	}
	memcpy(&p[x * 4 + 0], string, strlen(string));

	/* fBZ[uoN */
	strncpy(string, "Disable           BRG", sizeof(string));
	for (i=0; i<3; i++) {
		if (~alu_disable & (1 << i)) {
			string[18 + i] = '-';
		}
	}
	memcpy(&p[x * 5 + 0], string, strlen(string));

	/* ^CyCgf[^ */
	_snprintf(string, sizeof(string), "TILE   B:%02X R:%02X G:%02X",
		alu_tiledat[0], alu_tiledat[1], alu_tiledat[2]);
	memcpy(&p[x * 6 + 0], string, strlen(string));

	/* ԃXe[^X */
	strncpy(string, "Line LSI", sizeof(string));
	memcpy(&p[x * 0 + 25], string, strlen(string));
	if (line_busy) {
		strncpy(string, " Busy", sizeof(string));
	}
	else {
		strncpy(string, "Ready", sizeof(string));
	}
	memcpy(&p[x * 0 + 41], string, strlen(string));

	/* AhXItZbg */
	_snprintf(string, sizeof(string), "Offset           %04X", line_offset);
	memcpy(&p[x * 1 + 25], string, strlen(string));

	/* CX^C */
	_snprintf(string, sizeof(string), "Line Style       %04X", line_style);
	memcpy(&p[x * 2 + 25], string, strlen(string));

	/* n_EI_ */
	_snprintf(string, sizeof(string), "Line X0           %3d", line_x0);
	memcpy(&p[x * 3 + 25], string, strlen(string));
	_snprintf(string, sizeof(string), "Line Y0           %3d", line_y0);
	memcpy(&p[x * 4 + 25], string, strlen(string));
	_snprintf(string, sizeof(string), "Line X1           %3d", line_x1);
	memcpy(&p[x * 5 + 25], string, strlen(string));
	_snprintf(string, sizeof(string), "Line Y1           %3d", line_y1);
	memcpy(&p[x * 6 + 25], string, strlen(string));
}

/*
 *	_Z/ԃEChE
 *	`
 */
static void FASTCALL DrawALULine(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pALULine) {
		return;
	}
	SetupALULine(pALULine, x, y);

	/* ` */
	DrawWindowText(hDC, pALULine, x, y);
}

/*
 *	_Z/ԃEChE
 *	tbV
 */
void FASTCALL RefreshALULine(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_ALULINE] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_ALULINE];
	hDC = GetDC(hWnd);
	DrawALULine(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	_Z/ԃEChE
 *	ĕ`
 */
static void FASTCALL PaintALULine(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pALULine;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawALULine(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	_Z/ԃEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK ALULineProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintALULine(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pALULine, NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	_Z/ԃEChE
 *	EChE쐬
 */
HWND FASTCALL CreateALULine(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_ALULine";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 46;
	rect.bottom = lCharHeight * 7;

	/* EChE^CgAobt@m */
	GetMsg(hResInstance, IDS_SWND_ALULINE,
				szWndName, sizeof(szWndName));
	pALULine = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = ALULineProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}
#endif

/*-[ OPNfBXvCEChE ]---------------------------------------------*/

/*
 *	OPNfBXvCEChE
 *	hbg`
 */
static void FASTCALL PSetOPNDisp(int x, int y, int color)
{
	int w;
	BYTE *p;
	BYTE dat;

	ASSERT((x >= 0) && (x < rOPNDisp.right));
	ASSERT((y >= 0) && (y < rOPNDisp.bottom));
	ASSERT((color >= 0) && (color <= 15));

	/* rOPNDispArbg}bv̉TCYvZ(4oCgACg) */
	w = (((rOPNDisp.right / 2) + 3) >> 2) << 2;

	/* AhXAf[^擾 */
	if (!pOPNDisp) {
		return;
	}
	p = &pOPNDisp[w * y + (x >> 1)];
	dat = *p;

	/* Qɕ */
	if (x & 1) {
		dat &= 0xf0;
		dat |= (BYTE)color;
	}
	else {
		dat &= 0x0f;
		dat |= (BYTE)(color << 4);
	}

	/*  */
	*p = dat;
}

/*
 *	OPNfBXvCEChE
 *	{bNXtB`
 */
static void FASTCALL BfOPNDisp(int x, int y, int cx, int cy, int color)
{
	int i;
	int j;

	ASSERT((color >= 0) && (color < 16));

	for (i = 0; i < cy; i++) {
		for (j = 0; j < cx; j++) {
			PSetOPNDisp(x + j, y, color);
		}
		y++;
	}
}

/*
 *	OPNfBXvCEChE
 *	x[^`
 */
static void FASTCALL LvlOPNDisp(int x, int y, int cx, int cy,
								int color1, int color2, int color3)
{
	int i;
	int j;

	ASSERT((color1 >= 0) && (color1 < 16));
	ASSERT((color2 >= 0) && (color2 < 16));
	ASSERT((color3 >= 0) && (color3 < 16));

	for (i = 0; i < cy; i++) {
		for (j = 0; j < cx; j++) {
			if ((j % 4) == 3) {
				PSetOPNDisp(x + j, y, 8);
			}
			else {
				if (j < 48) {
					PSetOPNDisp(x + j, y, color1);
				}
				else if (j < 80) {
					PSetOPNDisp(x + j, y, color2);
				}
				else {
					PSetOPNDisp(x + j, y, color3);
				}
			}
		}
		y++;
	}
}

/*
 *	OPNfBXvCEChE
 *	LN^`
 */
static void FASTCALL ChrOPNDisp(char c, int x, int y, int color)
{
	int i;
	int j;
	BYTE *p;
	BYTE dat;

	ASSERT((color >= 0) && (color < 16));

	/* TuROM(C)̃tHgAhX𓾂 */
	p = &subrom_c[c * 8];

	/* y[v */
	for (i = 0; i < 8; i++) {
		/* f[^擾 */
		dat = *p;
		p++;

		/* x[v */
		for (j = 0; j < 8; j++) {
			if (dat & 0x80) {
				PSetOPNDisp(x, y, color);
			}
			else {
				PSetOPNDisp(x, y, 0);
			}
			dat <<= 1;
			x++;
		}

		/* y */
		x -= 8;
		y++;
	}
}

/*
 *	OPNfBXvCEChE
 *	Օ`
 *
 *	ʃju	IN^[u 0`7
 *	ʃju	C, C#, D, D#, E, F, F#, G, G#, A, A#, B
 *	x, y̓gbN̊_Bcolor=-1̓ftHgJ[
 */
static void FASTCALL KbdOPNDisp(int code, int x, int y, int color)
{
	int i;
	int j;
	int ys;

	/* xWB=5dotA=4dotBƔ̌Ԃ1dot} */
	static const int x_table[] = {
		0, 0+4, 6, 6+4, 12, 18, 18+4, 24, 24+4, 30, 30+4, 36
	};
	/* 0:E 1:E 2: 3: */
	static const int type_table[] = {
		0, 3, 1 ,3, 2, 0, 3, 1, 3, 1, 3, 2
	};

	ASSERT(code <= 0x7f);
	ASSERT((code & 0x0f) <= 0x0b);

	/* cTCY */
#ifdef OPNAMODE
	if (((nOPNDisp & 32) || ((nOPNDisp & 7) == 3)) && !bPopupSwnd) {
#else
	if (((nOPNDisp & 7) == 3) && !bPopupSwnd) {
#endif
		ys = 6;
	}
	else {
		ys = 8;
	}

	/* xʒu */
	x += (code >> 4) * 42;
	code &= 0x0f;
	x += x_table[code];

	/* Fݒ */
	if (color < 0) {
		if (type_table[code] < 3) {
			color = 15;
		}
		else {
			color = 8;
		}
	}

	/* ^Cv */
	switch (type_table[code]) {
		/* BEɍ */
		case 0:
			/* Ԃ */
			for (i = 0; i < ys * 2; i++) {
				PSetOPNDisp(x, y + i, 0);
			}
			/*  */
			for (i = 0; i < ys; i++) {
				for (j=1; j<=3; j++) {
					PSetOPNDisp(x + j, y + i, color);
				}
			}
			/*  */
			for (i = 0; i < ys; i++) {
				for (j=1; j<=5; j++) {
					PSetOPNDisp(x + j, y + ys + i, color);
				}
			}
			break;

		/* BEɍ */
		case 1:
			/* Ԃ */
			for (i = 0; i < ys; i++) {
				PSetOPNDisp(x, y + ys + i, 0);
			}
			/*  */
			for (i = 0; i < ys; i++) {
				for (j=3; j<=3; j++) {
					PSetOPNDisp(x + j, y + i, color);
				}
			}
			/*  */
			for (i = 0; i < ys; i++) {
				for (j=1; j<=5; j++) {
					PSetOPNDisp(x + j, y + ys + i, color);
				}
			}
			break;

		/* Bɍ */
		case 2:
			/* Ԃ */
			for (i = 0; i < ys; i++) {
				PSetOPNDisp(x, y + ys + i, 0);
			}
			/*  */
			for (i = 0; i < ys; i++) {
				for (j=3; j<=5; j++) {
					PSetOPNDisp(x + j, y + i, color);
				}
			}
			/*  */
			for (i = 0; i < ys; i++) {
				for (j=1; j<=5; j++) {
					PSetOPNDisp(x + j, y + ys + i, color);
				}
			}
			break;

		/*  */
		case 3:
			for (i = 0; i < ys; i++) {
				for (j = 0; j < 5; j++) {
					PSetOPNDisp(x + j, y + i, color);
				}
			}
			break;

		default:
			ASSERT(FALSE);
	}
}

/*
 *	OPNfBXvCEChE
 *	ՑS`
 *
 *	x, y̓gbN̊_
 */
void FASTCALL AllKbdOPNDisp(int x, int y)
{
	int i;
	int j;

	/* Qd[v */
	for (i = 0; i < 8; i++) {
		for (j = 0; j < 12; j++) {
			KbdOPNDisp((i * 16) + j, x, y, -1);
		}
	}
}

/*
 *	OPNfBXvCEChE
 *	ZbgAbv
 */
static BOOL FASTCALL SetupOPNDisp(int *first, int *end)
{
	int ch;
	int no;
	int i;
	int j;
	int x;
	int y;
	int ys;
	BOOL mute;
	BOOL flag[OPN_CHNUM];
#ifdef OPNAMODE
	int ttl, psg;
#else
	const int ttl = 6;
	const int psg = 3;
#endif

	/*  */
	memset(flag, 0, sizeof(flag));

#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		ttl = 16;
		psg = 6;
	}
	else {
		ttl = 6;
		psg = 3;
	}
#endif

#ifdef OPNAMODE
	if (((nOPNDisp & 32) || ((nOPNDisp & 7) == 3)) && !bPopupSwnd) {
#else
	if (((nOPNDisp & 7) == 3) && !bPopupSwnd) {
#endif
		ys = (8 + 12 + 1);
	}
	else {
		ys = (8 * 3 + 1);
	}

	for (ch = 0, x = 0, y = 0, no = OPN_STD; no <= OPN_PSG; no++) {
		if (no < OPN_PSG) {
			/* LOPN̂ݎ{ */
			if ((int)(nOPNDisp & 7) <= no) {
				continue;
			}
		}
		else {
			/* LPSĜݎ{ */
			if (!(nOPNDisp & 8)) {
				continue;
			}
		}

		for (i = 0; i < ttl; i++) {
			/* `l */
			if (no == OPN_PSG) {
				/* PSG3`l܂ */
				if (i >= 3) {
					break;
				}
			}

			/*  */
			for (j = 0; j < 42; j++) {
				/* v`FbN */
				if (cnOPNDisp[ch][j * 2] == ctOPNDisp[ch][j * 2]) {
					if (cnOPNDisp[ch][j*2+1] == ctOPNDisp[ch][j*2+1]) {
						continue;
					}
				}

				/* ` */
				ChrOPNDisp(ctOPNDisp[ch][j * 2], j * 8 + x, y, ctOPNDisp[ch][j * 2 + 1]);

				/* Rs[ */
				cnOPNDisp[ch][j * 2] = ctOPNDisp[ch][j * 2];
				cnOPNDisp[ch][j * 2 + 1] = ctOPNDisp[ch][j * 2 + 1];

				/* tO */
				flag[ch] = TRUE;
			}

			/*  */
			if (knOPNDisp[ch] == -2) {
				/* ԂȂAL[ItԂɂ̂挈 */
				AllKbdOPNDisp(x, y + 8);
				knOPNDisp[ch] = -1;
				flag[ch] = TRUE;
			}
			/* `FbN */
			if (knOPNDisp[ch] != ktOPNDisp[ch]) {
				/* L[ItԂȂA */
				if (knOPNDisp[ch] >= 0) {
					KbdOPNDisp(knOPNDisp[ch], x, y + 8, -1);
				}
				/* L[IɂȂA */
				if (ktOPNDisp[ch] >= 0) {
#ifdef OPNAMODE
					if (nOPNDisp & 32) {
						mute = GetMute2(no, i);
					}
					else {
						mute = GetMute(no * ttl + i);
					}
#else
					mute = GetMute(no * ttl + i);
#endif
					if (!mute) {
						if (no == OPN_PSG) {
							/* PSG */
							KbdOPNDisp(ktOPNDisp[ch], x, y + 8, 13);
						}
						else {
							/* OPN */
							if (i < psg) {
								/* FM (F) */
								KbdOPNDisp(ktOPNDisp[ch], x, y + 8, 10);
							}
#ifdef OPNAMODE
							else if (i < psg + 3) {
#else
							else {
#endif
								/* SSG (F) */
								KbdOPNDisp(ktOPNDisp[ch], x, y + 8, 9);
							}
#ifdef OPNAMODE
							else if (nOPNDisp & 32) {
								if (i < 9) {
									/* RHYDZM (F) */
									KbdOPNDisp(ktOPNDisp[ch], x, y + 8, 11);
								}
								else /*if (i < 15)*/ {
									/* ADPCM (F) */
									KbdOPNDisp(ktOPNDisp[ch], x, y + 8, 11);
								}
							}
#endif
						}
					}
				}

				/* XV */
				knOPNDisp[ch] = ktOPNDisp[ch];
				flag[ch] = TRUE;
			}

			/* x */
			/* v`FbN */
			if (lnOPNDisp[ch] != ltOPNDisp[ch]) {
				/* ؂ւ_ƂȂxW߂ */
				if (ltOPNDisp[ch] >= 445) {
					ltOPNDisp[ch] = 444;
				}
				else if (ltOPNDisp[ch] < 0) {
					ltOPNDisp[ch] = 0;
				}
				j = 100 * ltOPNDisp[ch];
				j /= 445;

				/* ` */
				BfOPNDisp(236 + j + x, y, 100 - j, 7, 8);
				LvlOPNDisp(236 + x, y, j, 7, 14, 13, 12);

				/* XV */
				lnOPNDisp[ch] = ltOPNDisp[ch];
				flag[ch] = TRUE;
			}

			/*  */
			y += ys;
			ch++;
		}
#ifdef OPNAMODE
		if (nOPNDisp & 32) {
			x += (42 * 8) + 8;
			y = 0;
		}
#endif
	}

	/* flag */
	y = -1;
	j = ch;
	for (i = 0; i < ch; i++) {
		if (flag[i]) {
			/*  */
			if (i < j) {
				j = i;
			}
			/*  */
			if (y < i) {
				y = i;
			}
		}
	}

	/* jy܂ŁA΂悢 */
	if (y >= 0) {
		*first = j;
		*end = y;
		return TRUE;
	}

	/* `̕KvȂ */
	return FALSE;
}

/*
 *	e[u
 *	Ƃ̂肵łȂA̒Ԃ̂l\
 */
static const double pitch_table[] = {
	31.772, 33.661, 35.663, 37.784, 40.030, 42.411, 44.933, 47.605, 50.435, 53.434, 56.612, 59.978,
	63.544, 67.323, 71.326, 75.567, 80.061, 84.822, 89.865, 95.209, 100.870, 106.869, 113.223, 119.956,
	127.089, 134.646, 142.652, 151.135, 160.122, 169.643, 179.731, 190.418, 201.741, 213.737, 226.446, 239.912,
	254.178, 269.292, 285.305, 302.270, 320.244, 339.287, 359.461, 380.836, 403.482, 427.474, 452.893, 479.823,
	508.356, 538.584, 570.610, 604.540, 640.488, 678.573, 718.923, 761.672, 806.964, 854.948, 905.786, 959.647,
	1016.711, 1077.168, 1141.220, 1209.080, 1280.975, 1357.146, 1437.846, 1523.345, 1613.927, 1709.896, 1811.572, 1919.293,
	2033.422, 2154.336, 2282.439, 2418.160, 2561.951, 2714.292, 2875.692, 3046.689, 3227.855, 3419.792, 3623.144, 3838.587,
	4066.845, 4308.672, 4564.878, 4836.319, 5123.901, 5428.584, 5751.384, 6093.378, 6455.709, 6839.585, 7246.287, 7677.173,
	8133.681
};

/*
 *	R[hϊ
 *	-1͔͈͊O
 */
static int FASTCALL ConvOPNDisp(double freq)
{
	int i;
	int ret;

	/* ͈͊O`FbN */
	if (freq < pitch_table[0]) {
		return -1;
	}
	if (pitch_table[96] <= freq) {
		return -1;
	}

	/* [vA */
	ret = -1;
	for (i = 0; i < 96; i++) {
		if ((pitch_table[i] <= freq) && (freq <pitch_table[i + 1])) {
			/* iϊ */
			ret = i / 12;
			ret <<= 4;
			ret += (i % 12);
			break;
		}
	}

	return ret;
}

/*
 *	OPNfBXvCEChE
 *	FMR[hϊ
 */
static int FASTCALL FMConvOPNDisp(BYTE a4, BYTE a0, int scale)
{
	int oct;
	int fnum;
	double freq;
	double d;

	/* Octave, F-Number߂ */
	oct = (int)a4;
	oct >>= 3;
	fnum = (int)a4;
	fnum &= 0x07;
	fnum <<= 8;
	fnum |= (int)a0;

	/* g߂ */
	freq = OPN_CLOCK * 100;
	while (oct != 0) {
		freq *= 2;
		oct--;
	}
	freq *= fnum;
	d = (double)(1 << 20);
	d *= 12;
	d *= scale;
	freq /= d;

	/* ϊ */
	return ConvOPNDisp(freq);
}

/*
 *	OPNfBXvCEChE
 *	PSGR[hϊ
 */
static int FASTCALL PSGConvOPNDisp(BYTE low, BYTE high, int scale)
{
	int pitch;
	double freq;
	double d;

	/* sb`Zo */
	pitch = (int)high;
	pitch &= 0x0f;
	pitch <<= 8;
	pitch |= (int)low;

	/* g߂ */
	freq = OPN_CLOCK * 100;
	d = 8;
	d *= pitch;
	switch (scale) {
		case 3:
			d *= 2;
			break;
		case 6:
			d *= 4;
			break;
		case 2:
			d *= 1;
			break;
	}
	if (d == 0) {
		freq = 0;
	}
	else {
		freq /= d;
	}

	/* ϊ */
	return ConvOPNDisp(freq);
}

/*
 *	OPNfBXvCEChE
 *	L[{[hXe[^X
 */
static void FASTCALL StatKbdOPNDisp(void)
{
	int no, i, psg_vol, tmp;
#ifdef OPNAMODE
	int ttl, psg;
#else
	const int ttl = 6;
	const int psg = 3;
#endif

#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		ttl = 16;
		psg = 6;
	}
	else {
		ttl = 6;
		psg = 3;
	}
#endif

	for (no = OPN_STD; no <= OPN_THG; no++) {
		/* LOPN̂ݎ{ */
		if ((int)(nOPNDisp & 7) <= no) {
			break;
		}

		/* FM */
		for (i = 0; i < psg; i++) {
			if (!opn_key[no][i]) {
				ktOPNDisp[no * ttl + i] = -1;
			}
			else {
#ifdef OPNAMODE
				if (i < 3) {
					ktOPNDisp[no * ttl + i] = FMConvOPNDisp(
												opn_reg[no][0xa4 + i],
												opn_reg[no][0xa0 + i],
												opn_scale[no]);
				}
				else {
					ktOPNDisp[no * ttl + i] = FMConvOPNDisp(
												opn_reg[no][0x1a4 + i - 3],
												opn_reg[no][0x1a0 + i - 3],
												opn_scale[no]);
				}
#else
				ktOPNDisp[no * ttl + i] = FMConvOPNDisp(
											opn_reg[no][0xa4 + i],
											opn_reg[no][0xa0 + i],
											opn_scale[no]);
#endif
			}
		}

		/* SSG */
		for (i = 0; i < 3; i++) {
			psg_vol = ltOPNDisp[no * ttl + i + psg];
			if (psg_vol == 0) {
				psg_vol = (int)(opn_reg[no][i + 8] & 0x1f);
#if 0
				if (psg_vol & 0x10) {
					psg_vol = 0;
				}
#endif
			}
			if (psg_vol > 0) {
				ktOPNDisp[no * ttl + i + psg] = PSGConvOPNDisp(
										opn_reg[no][i * 2 + 0],
										opn_reg[no][i * 2 + 1],
										opn_scale[no]);
			}
			else {
				ktOPNDisp[no * ttl + i + psg] = -1;
			}
		}

#ifdef OPNAMODE
		if (nOPNDisp & 32) {
			/* RHYDZM */
			for (i = 0; i < 6; i++) {
				/* L[I */
				if (opn_reg[no][0x10] & 0x80) {
					ktOPNDisp[no * ttl + i + 9] = -1;
				} else {
					if (opn_reg[no][0x10] & (0x01 << i)) {
						ktOPNDisp[no * ttl + i + 9] = 0x30;
					} else {
						ktOPNDisp[no * ttl + i + 9] = -1;
					}
				}
			}

			/* ADPCM */
			if (opn_reg[no][0x100] & 0x80) {
				ktOPNDisp[no * ttl + 15] = 0x30;
			} else {
				ktOPNDisp[no * ttl + 15] = -1;
			}
		}
#endif
	}

	/* PSG */
	if (nOPNDisp & 8) {
		tmp = (nOPNDisp & 7) * ttl;
		for (i = 0; i < 3; i++) {
			psg_vol = ltOPNDisp[i + tmp];
			if (psg_vol == 0) {
				psg_vol = (int)(psg_reg[i + 8] & 0x1f);
#if 0
				if (psg_vol & 0x10) {
					psg_vol = 0;
				}
#endif
			}
			if (psg_vol > 0) {
				ktOPNDisp[i + tmp] = PSGConvOPNDisp(
										psg_reg[i * 2 + 0],
										psg_reg[i * 2 + 1],
										3);
			}
			else {
				ktOPNDisp[i + tmp] = -1;
			}
		}
	}
}

/*
 *	OPNfBXvCEChE
 *	xXe[^X
 */
static void FASTCALL StatLevOPNDisp(void)
{
	int no, i, tmp;
	BYTE m;
	BYTE p;
#ifdef OPNAMODE
	int ttl, psg;
#else
	const int ttl = 6;
	const int psg = 3;
#endif

#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		ttl = 16;
		psg = 6;
	}
	else {
		ttl = 6;
		psg = 3;
	}
#endif

	/* U擾 */
	for (i = 0; i < ttl * 3 + 3; i++) {
#ifdef OPNAMODE
		if (nOPNDisp & 32) {
			if (GetMute2(i / ttl, i % ttl)) {
				ltOPNDisp[i] = 0;
			}
			else {
				ltOPNDisp[i] = GetLevelSnd2(i / ttl, i % ttl);
			}
		}
		else {
			if (GetMute(i)) {
				ltOPNDisp[i] = 0;
			}
			else {
				ltOPNDisp[i] = GetLevelSnd(i);
			}
		}
#else
		if (GetMute(i)) {
			ltOPNDisp[i] = 0;
		}
		else {
			ltOPNDisp[i] = GetLevelSnd(i);
		}
#endif
	}

	/* SSG}XN`FbN */
	for (no = OPN_STD; no <= OPN_THG; no++) {
		m = opn_reg[no][7];
		p = 9;
		for (i = 0; i < 3; i++) {
			if ((m & p) == p) {
				ltOPNDisp[i + ttl * no + psg] = 0;
			}
			p <<= 1;
		}
	}

	/* PSG}XN`FbN */
	m = psg_reg[7];
	p = 9;
	for (i = 0; i < 3; i++) {
		if ((m & p) == p) {
			ltOPNDisp[i + ttl * 3] = 0;
		}
		p <<= 1;
	}

	/* PSGԂړ */
	if ((nOPNDisp & 8) && ((nOPNDisp & 7) < 3)) {
		tmp = (nOPNDisp & 7) * ttl;
		for (i = 0; i < 3; i++) {
			ltOPNDisp[i + tmp] = ltOPNDisp[i + ttl * 3];
		}
	}
}

/*
 *	OPNfBXvCEChE
 *	Zbg
 */
static void FASTCALL StrOPNDisp(char *string, int x, int y, int color)
{
	char ch;

	ASSERT(string);
	ASSERT((x >= 0) && (x < 49));
	ASSERT((y >= 0) && (y < OPN_CHNUM));
	ASSERT((color >= 0) && (color < 16));

	/* [v */
	for (;;) {
		/* 擾 */
		ch = *string++;
		if (ch == '\0') {
			break;
		}

		/* x`FbN */
		if (x >= 49) {
			continue;
		}

		/* Zbg */
		ctOPNDisp[y][x * 2 + 0] = ch;
		ctOPNDisp[y][x * 2 + 1] = (BYTE)color;
		x++;
	}
}

/*
 *	OPNfBXvCEChE
 *	f[^Zbg
 */
static void FASTCALL StatChrOPNDisp(void)
{
	enum {
		part_fm,
		part_psg,
#ifdef OPNAMODE
		part_rhydzm,
		part_adpcm,
#endif
	} part;
	const BYTE cslot[4] = {2, 3, 3, 4};
#ifdef OPNAMODE
	const char *rhythm[] = { "BD ", "SD ", "TOP", "HH ", "TOM", "RIM" };
#endif
	const char * const OPNDisp[] = {
		"FL",
		"L ",
		"LR",
		" R",
		"FR",
	};
	int ch, no, i, j;
	char string[128];
	BYTE alg;
	BYTE tl;
	int vols;
	BYTE mix;
#ifdef OPNAMODE
	int sft;
	BOOL mute;
	BOOL opnamode;
#endif
#ifdef OPNAMODE
	int ttl, psg;
#else
	const int ttl = 6;
	const int psg = 3;
#endif

#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		ttl = 16;
		psg = 6;
	}
	else {
		ttl = 6;
		psg = 3;
	}
#endif

	for (ch = 0, no = OPN_STD; no <= OPN_PSG; no++) {
		if (no < OPN_PSG) {
			/* LOPN̂ݎ{ */
			if ((int)(nOPNDisp & 7) <= no) {
				continue;
			}
		}
		else {
			/* LPSĜݎ{ */
			if (!(nOPNDisp & 8)) {
				continue;
			}
		}

		for (i = 0; i < ttl; i++) {
			/* `l */
			if (no == OPN_PSG) {
				/* PSG3`l܂ */
				if (i >= 3) {
					break;
				}
				part = part_psg;
			}
			else if (i < psg) {
				part = part_fm;
			}
#ifdef OPNAMODE
			else if (i < psg + 3) {
				part = part_psg;
			}
			else if (i < psg + 3 + 6) {
				part = part_rhydzm;
			}
			else if (i < psg + 3 + 6 + 1) {
				part = part_adpcm;
			}
#else
			else {
				part = part_psg;
			}
#endif
#ifdef OPNAMODE
			if (no == OPN_PSG) {
				_snprintf(string, sizeof(string), "PSG%1d", i + 1);
			}
			else if (nOPNDisp & 32) {
				if (part == part_fm) {
					if (no == OPN_STD) {
						_snprintf(string, sizeof(string), "OPN%1d", i + 1);
					}
					else if (no == OPN_WHG) {
						_snprintf(string, sizeof(string), "WHG%1d", i + 1);
					}
					else {
						_snprintf(string, sizeof(string), "THG%1d", i + 1);
					}
				}
				else if (part == part_psg) {
					_snprintf(string, sizeof(string), "SSG%1d", i - 5);
				}
				else if (part == part_rhydzm) {
					_snprintf(string, sizeof(string), "RHY%1d", i - 8);
				}
				else if (part == part_adpcm) {
					strncpy(string, "ADPCM", sizeof(string));
				}
			}
			else if (no == OPN_STD) {
				_snprintf(string, sizeof(string), "OPN%1d", i + 1);
			}
			else if (no == OPN_WHG) {
				_snprintf(string, sizeof(string), "WHG%1d", i + 1);
			}
			else {
				_snprintf(string, sizeof(string), "THG%1d", i + 1);
			}
#else
			if (no == OPN_STD) {
				_snprintf(string, sizeof(string), "OPN%1d", i + 1);
			}
			else if (no == OPN_WHG) {
				_snprintf(string, sizeof(string), "WHG%1d", i + 1);
			}
			else if (no == OPN_THG) {
				_snprintf(string, sizeof(string), "THG%1d", i + 1);
			}
			else {
				_snprintf(string, sizeof(string), "PSG%1d", i + 1);
			}
#endif
			StrOPNDisp(string, 0, ch, 7);

#ifdef OPNAMODE
			if (nOPNDisp & 32) {
				sft = ((i >= 3) ? (0x0100) : (0x0000));
			}
			else {
				sft = 0;
			}
#endif

			/* g */
			if (part == part_fm) {
#ifdef OPNAMODE
				j = opn_reg[no][0xa4 + (i % 3) + sft];
				j <<= 8;
				j |= opn_reg[no][0xa0 + (i % 3) + sft];
#else
				j = opn_reg[no][0xa4 + (i % 3)];
				j <<= 8;
				j |= opn_reg[no][0xa0 + (i % 3)];
#endif
				_snprintf(string, sizeof(string), "F:$%04X", j);
			}
#ifdef OPNAMODE
			else if (part == part_rhydzm) {
				/* Y */
				strncpy(string, rhythm[i - 9], sizeof(string));
			}
			else if (part == part_adpcm) {
				/*  */
				strncpy(string, "    ", sizeof(string));
			}
#endif
			else {
				if (no < OPN_PSG) {
					j = opn_reg[no][(i % 3) * 2 + 1];
					j &= 0x0f;
					j <<= 8;
					j |= opn_reg[no][(i % 3) * 2 + 0];
				}
				else {
					j = psg_reg[(i % 3) * 2 + 1];
					j &= 0x0f;
					j <<= 8;
					j |= psg_reg[(i % 3) * 2 + 0];
				}
				_snprintf(string, sizeof(string), "P:$%04X", j);
			}
			StrOPNDisp(string, 6, ch, 12);

			/* {[ */
#ifdef OPNAMODE
			if (nOPNDisp & 32) {
				mute = GetMute2(ch / 16, ch % 16);
			}
			else {
				mute = GetMute(ch);
			}
			if (mute) {
#else
			if (GetMute(ch)) {
#endif
				strncpy(string, "MUTE          ", sizeof(string));
				StrOPNDisp(string, 14, ch, 15);
			}
			else {
				if (part == part_fm) {
#ifdef OPNAMODE
					alg = (BYTE)(opn_reg[no][0xb0 + (i % 3) + sft] & 0x07);
#else
					alg = (BYTE)(opn_reg[no][0xb0 + (i % 3)] & 0x07);
#endif
					if (alg >= 4) {
						alg -= (BYTE)4;
						vols = 127;
						for (j = 0; j < cslot[alg]; j++) {
#ifdef OPNAMODE
							tl = (BYTE)(opn_reg[no][0x4c - (j << 2) + (i % 3) + sft] & 0x7f);
#else
							tl = (BYTE)(opn_reg[no][0x4c - (j << 2) + (i % 3)] & 0x7f);
#endif
							if (vols > tl) {
								vols = tl;
							}
						}
					}
					else {
#ifdef OPNAMODE
						vols = opn_reg[no][0x4c + (i % 3) + sft];
#else
						vols = opn_reg[no][0x4c + (i % 3)];
#endif
					}

					/* 7bit̂ݗL */
					vols &= 0x7f;
					_snprintf(string, sizeof(string), "V:%03d", 127 - vols);
				}
#ifdef OPNAMODE
				else if (part == part_rhydzm) {
					j = opn_reg[no][0x18 + i - 9];
					_snprintf(string, sizeof(string), "IL:%02d", j & 0x1f);
				}
				else if (part == part_adpcm) {
					j = opn_reg[no][0x10b];
					_snprintf(string, sizeof(string), "L:%03d", j & 0xff);
				}
#endif
				else {
					if (no < OPN_PSG) {
						j = opn_reg[no][8 + (i % 3)];
					}
					else {
						j = psg_reg[8 + (i % 3)];
					}
					/* 0`15ƁA16ȏɕ */
					j &= 0x1f;
					if (j >= 0x10) {
						j = 0x10;
					}
					_snprintf(string, sizeof(string), "V:%03d", j);
				}
				StrOPNDisp(string, 14, ch, 12);

#ifdef OPNAMODE
				if (no < OPN_PSG) {
					opnamode = opna_use[no];
				}
				else {
					opnamode = FALSE;
				}
#endif
				if (part == part_fm) {
					/* L[I */
					if (opn_key[no][i]) {
						strncpy(string, "KEYON   ", sizeof(string));
					}
					else {
						strncpy(string, "        ", sizeof(string));
					}
				}
#ifdef OPNAMODE
				else if (part == part_rhydzm) {
					/* L[I */
					j = opn_reg[no][0x10];
					if (j & 0x80) {
						strncpy(string, "     ", sizeof(string));
					} else {
						j >>= (i - 9);
						if (j & 0x01) {
							strncpy(string, "KEYON", sizeof(string));
						} else {
							strncpy(string, "     ", sizeof(string));
						}
					}
				}
				else if (part == part_adpcm) {
					memset(string, 0, sizeof(string));
				}
#endif
				else {
					/* ~LT */
					if (no < OPN_PSG) {
						j = opn_reg[no][7];
					}
					else {
						j = psg_reg[7];
					}
					if ((i % 3) == 1) {
						j >>= 1;
					}
					if ((i % 3) == 2) {
						j >>= 2;
					}
					j &= 0x09;

					if (no < OPN_PSG) {
						mix = opn_reg[no][6];
					}
					else {
						mix = psg_reg[6];
					}
					switch (j) {
						case 0:
							if ((uStereoOut == 0) ||								/* mݒ肩	*/
								(!whg_use && !thg_use &&							/* WHG/THGgp	*/
								 ((!opn_use) ||										/*  PSĜ݂		*/
#if XM7_VER >= 2
								  (model_type >= MODEL_FM77AV) ||					/*  AVV[Y	*/
#endif
								  (pan_set[uStereoOut][OPN_PSG] == PAN_C)))) {		/*  PSG		*/
								/* Monaural */
								_snprintf(string, sizeof(string), "T + N:%2d", (BYTE)(mix & 0x1f));
							}
							else {
								/* Stereo */
								_snprintf(string, sizeof(string), "TN:%2d   ", (BYTE)(mix & 0x1f));
							}
							break;
						case 1:
							if ((uStereoOut == 0) ||								/* mݒ肩	*/
								(!whg_use && !thg_use &&							/* WHG/THGgp	*/
								 ((!opn_use) ||										/*  PSĜ݂		*/
#if XM7_VER >= 2
								  (model_type >= MODEL_FM77AV) ||					/*  AVV[Y	*/
#endif
								  (pan_set[uStereoOut][OPN_PSG] == PAN_C)))) {		/*  PSG		*/
								/* Monaural */
								_snprintf(string, sizeof(string), "NOISE:%2d", (BYTE)(mix & 0x1f));
							}
							else {
								/* Stereo */
								_snprintf(string, sizeof(string), "N :%2d   ", (BYTE)(mix & 0x1f));
							}
							break;
						case 8:
							strncpy(string, "TONE    ", sizeof(string));
							break;
						case 9:
							strncpy(string, "        ", sizeof(string));
							break;
					}
				}
				StrOPNDisp(string, 20, ch, 12);

#ifdef OPNAMODE
				if ((opnamode) &&
					(pan_set[uStereoOut][no] == PAN_C)) {
					if (part == part_fm) {
						switch (opn_reg[no][0xb4 + (i % 3) + sft] & 0xC0) {
							case 0x00:
								strncpy(string, "  ", sizeof(string));
								break;
							case 0x40:
								strncpy(string, OPNDisp[PAN_R], sizeof(string));
								break;
							case 0x80:
								strncpy(string, OPNDisp[PAN_L], sizeof(string));
								break;
							case 0xC0:
							default:
								strncpy(string, OPNDisp[PAN_C], sizeof(string));
								break;
						}
					}
					else if (part == part_rhydzm) {
						switch (opn_reg[no][0x18 + i - 9] & 0xC0) {
							case 0x00:
								strncpy(string, "  ", sizeof(string));
								break;
							case 0x40:
								strncpy(string, OPNDisp[PAN_R], sizeof(string));
								break;
							case 0x80:
								strncpy(string, OPNDisp[PAN_L], sizeof(string));
								break;
							case 0xC0:
							default:
								strncpy(string, OPNDisp[PAN_C], sizeof(string));
								break;
						}
					}
					else if (part == part_adpcm) {
						switch (opn_reg[no][0x101] & 0xC0) {
							case 0x00:
								strncpy(string, "  ", sizeof(string));
								break;
							case 0x40:
								strncpy(string, OPNDisp[PAN_R], sizeof(string));
								break;
							case 0x80:
								strncpy(string, OPNDisp[PAN_L], sizeof(string));
								break;
							case 0xC0:
							default:
								strncpy(string, OPNDisp[PAN_C], sizeof(string));
								break;
						}
					}
					else {
						strncpy(string, OPNDisp[PAN_C], sizeof(string));
					}
					StrOPNDisp(string, 26, ch, 12);
				}
				else if ((uStereoOut == 0) ||							/* mݒ肩	*/
						 (!whg_use && !thg_use &&						/* WHG/THGgp	*/
						  ((!opn_use) ||								/*  PSĜ݂		*/
#if XM7_VER >= 2
						   (model_type >= MODEL_FM77AV) ||				/*  AVV[Y	*/
#endif
						   (pan_set[uStereoOut][OPN_PSG] == PAN_C)))) {	/*  PSG		*/
					/* Monaural */
					;
				}
#else
				if ((uStereoOut == 0) ||								/* mݒ肩	*/
					(!whg_use && !thg_use &&							/* WHG/THGgp	*/
					 ((!opn_use) ||										/*  PSĜ݂		*/
#if XM7_VER >= 2
					  (model_type >= MODEL_FM77AV) ||					/*  AVV[Y	*/
#endif
					  (pan_set[uStereoOut][OPN_PSG] == PAN_C)))) {		/*  PSG		*/
					/* Monaural */
					;
				}
#endif
				else {
					strncpy(string, OPNDisp[pan_set[uStereoOut][no]], sizeof(string));
					StrOPNDisp(string, 26, ch, 12);
				}
			}

			ch++;
		}
	}
}

/*
 *	OPNfBXvCEChE
 *	`
 */
static void FASTCALL DrawOPNDisp(HDC hDC, BOOL flag)
{
	HDC hMemDC;
	HBITMAP hBitmap;
	int first, end;
#ifdef OPNAMODE
	int xs;
#endif
	int ys;
#ifdef OPNAMODE
	int ttl;
#else
	const int ttl = 6;
#endif

	ASSERT(hDC);

	/* rbg}bvnhE|C^ȂAȂ */
	if (!hOPNDisp || !pOPNDisp) {
		return;
	}

	/* Xe[^X`FbN */
	StatLevOPNDisp();
	StatKbdOPNDisp();
	StatChrOPNDisp();

#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		ttl = 16;
	}
	else {
		ttl = 6;
	}
#endif

	/* ZbgAbv``FbN */
	if (!SetupOPNDisp(&first, &end) && !flag) {
		return;
	}
	if (flag) {
		first = 0;
		end = (nOPNDisp & 7) * ttl;
		if (nOPNDisp & 8) {
#ifdef OPNAMODE
		if (nOPNDisp & 32) {
			end += ttl;
		}
		else {
			end += 3;
		}
#else
			end += 3;
#endif
		}
	}

#ifdef OPNAMODE
	xs = (42 * 8) + 8;
#endif

#ifdef OPNAMODE
	if (((nOPNDisp & 32) || ((nOPNDisp & 7) == 3)) && !bPopupSwnd) {
#else
	if (((nOPNDisp & 7) == 3) && !bPopupSwnd) {
#endif
		ys = (8 + 12 + 1);
	}
	else {
		ys = (8 * 3 + 1);
	}

	/* rbg}bvnhE|C^Lx`FbN */
	if (!hOPNDisp || !pOPNDisp) {
		return;
	}

	/* DC쐬Arbg}bvZNg */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hBitmap = (HBITMAP)SelectObject(hMemDC, hOPNDisp);

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hBitmap) {
		/* pbgݒABitBlt */
		SetDIBColorTable(hMemDC, 0, 16, rgbQuadTable16);
#ifdef OPNAMODE
		if (nOPNDisp & 32) {
			/* OPN/PSG */
			if (first < 16 * 1) {
				if (end < 16 * 1) {
					/* OPN/PSG only */
					BitBlt(	hDC,
							xs * 0, ys * (first - 16 * 0),
							xs, ys * (end - first + 1),
							hMemDC,
							xs * 0, ys * (first - 16 * 0),
							SRCCOPY);
				}
				else {
					/* with NEXT */
					BitBlt(	hDC,
							xs * 0, ys * (first - 16 * 0),
							xs, ys * (16 * 1 - first + 1),
							hMemDC,
							xs * 0, ys * (first - 16 * 0),
							SRCCOPY);
				}
			}
			/* WHG/PSG */
			if ((end >= 16 * 1) && (first < 16 * 2)) {
				if (end < 16 * 2) {
					if (first < 16 * 1) {
						/* with PREV */
						BitBlt(	hDC,
								xs * 1, ys * 0,
								xs, ys * (end - 16 * 1 + 1),
								hMemDC,
								xs * 1, ys * 0,
								SRCCOPY);
					}
					else {
						/* WHG/PSG only */
						BitBlt(	hDC,
								xs * 1, ys * (first - 16 * 1),
								xs, ys * (end - first + 1),
								hMemDC,
								xs * 1, ys * (first - 16 * 1),
								SRCCOPY);
					}
				}
				else {
					if (first < 16 * 1) {
						/* with PREV/NEXT */
						BitBlt(	hDC,
								xs * 1, ys * 0,
								xs, ys * (16 * 1 + 1),
								hMemDC,
								xs * 1, ys * 0,
								SRCCOPY);
					}
					else {
						/* with NEXT */
						BitBlt(	hDC,
								xs * 1, ys * (first - 16 * 1),
								xs, ys * (16 * 2 - first + 1),
								hMemDC,
								xs * 1, ys * (first - 16 * 1),
								SRCCOPY);
					}
				}
			}
			/* THG/PSG */
			if ((end >= 16 * 2) && (first < 16 * 3)) {
				if (end < 16 * 3) {
					if (first < 16 * 2) {
						/* with PREV */
						BitBlt(	hDC,
								xs * 2, ys * 0,
								xs, ys * (end - 16 * 2 + 1),
								hMemDC,
								xs * 2, ys * 0,
								SRCCOPY);
					}
					else {
						/* THG/PSG only */
						BitBlt(	hDC,
								xs * 2, ys * (first - 16 * 2),
								xs, ys * (end - first + 1),
								hMemDC,
								xs * 2, ys * (first - 16 * 2),
								SRCCOPY);
					}
				}
				else {
					if (first < 16 * 2) {
						/* with PREV/NEXT */
						BitBlt(	hDC,
								xs * 2, ys * 0,
								xs, ys * (16 * 1 + 1),
								hMemDC,
								xs * 2, ys * 0,
								SRCCOPY);
					}
					else {
						/* with NEXT */
						BitBlt(	hDC,
								xs * 2, ys * (first - 16 * 2),
								xs, ys * (16 * 3 - first + 1),
								hMemDC,
								xs * 2, ys * (first - 16 * 2),
								SRCCOPY);
					}
				}
			}
			/* PSG */
			if (end >= 16 * 3) {
				if (first < 16 * 3) {
					/* with PREV */
					BitBlt(	hDC,
							xs * 3, ys * 0,
							xs, ys * (end - 16 * 3 + 1),
							hMemDC,
							xs * 3, ys * 0,
							SRCCOPY);
				}
				else {
					/* PSG only */
					BitBlt(	hDC,
							xs * 3, ys * (first - 16 * 3),
							xs, ys * (end - first + 1),
							hMemDC,
							xs * 3, ys * (first - 16 * 3),
							SRCCOPY);
				}
			}
		}
		else {
			BitBlt(hDC,
					0, ys * first,
					rOPNDisp.right, ys * (end - first + 1), hMemDC,
					0, ys * first, SRCCOPY);
		}
#else
		BitBlt(hDC,
				0, ys * first,
				rOPNDisp.right, ys * (end - first + 1), hMemDC,
				0, ys * first, SRCCOPY);
#endif

		/* IuWFNgăZNg */
		SelectObject(hMemDC, hBitmap);
	}

	/* DC폜 */
	DeleteDC(hMemDC);
}

/*
 *	OPNfBXvCEChE
 *	tbV
 */
void FASTCALL RefreshOPNDisp(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_OPNDISP] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_OPNDISP];
	hDC = GetDC(hWnd);
	DrawOPNDisp(hDC, TRUE);
	ReleaseDC(hWnd, hDC);
}

/*
 *	OPNfBXvCEChE
 *	ĕ`
 */
static void FASTCALL PaintOPNDisp(HWND hWnd, BOOL flag)
{
	HDC hDC;
	PAINTSTRUCT ps;
	int i;

	ASSERT(hWnd);

	/* [NGAׂĖ */
	for (i = 0; i < OPN_CHNUM; i++) {
		knOPNDisp[i] = -2;
		lnOPNDisp[i] = -1;
	}
	memset(cnOPNDisp, 0xff, sizeof(cnOPNDisp));

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawOPNDisp(hDC, flag);
	EndPaint(hWnd, &ps);
}

/*
 *	OPNfBXvCEChE
 *	~[g@`lvZ
 */
static int CalcCh(LPARAM lParam)
{
#ifdef OPNAMODE
	int x;
#endif
	int y;
	int ch;

#ifdef OPNAMODE
	x = LOWORD(lParam);
#endif
	y = HIWORD(lParam);

#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		ch = y / (8 * 3 + 1) + x / ((42 * 8) + 8) * 16;
	}
	else if (((nOPNDisp & 7) == 3) && !bPopupSwnd) {
		ch = y / (8 + 12 + 1);
	}
#else
	if (((nOPNDisp & 7) == 3) && !bPopupSwnd) {
		ch = y / (8 + 12 + 1);
	}
#endif
	else {
		ch = y / (8 * 3 + 1);
	}

	return ch;
}

/*
 *	OPNfBXvCEChE
 *	R}h
 */
static BOOL FASTCALL CmdOPNDisp(HWND hWnd, WORD wID)
{
	int i;
#ifdef OPNAMODE
	const int ttl = 16;
#endif

	/* ^[QbgAhX */
	switch (wID) {
		/* ~[gS */
		case IDM_OPNDISP_UNMUTEALL:
			LockVM();
#ifdef OPNAMODE
			for (i = 0; i < 3 * ttl + 3; i++) {
				SetMute2(i / ttl, i % ttl, FALSE);
			}
#else
			for (i = 0; i < OPN_CHNUM; i++) {
				SetMute(i, FALSE);
			}
#endif
			PaintOPNDisp(hWnd, FALSE);
			UnlockVM();
			return TRUE;

		case IDM_OPNDISP_AUTO:
			nOPNDispMode = LEVEL_AUTO;
			break;

		case IDM_OPNDISP_PSG:
			nOPNDispMode = LEVEL_PSG;
			break;

		case IDM_OPNDISP_OPN:
			nOPNDispMode = LEVEL_OPN;
			break;

		case IDM_OPNDISP_WHG:
			nOPNDispMode = LEVEL_WHG;
			break;

		case IDM_OPNDISP_THG:
			nOPNDispMode = LEVEL_THG;
			break;

		default:
			break;
	}

	return FALSE;
}

/*
 *	OPNfBXvCEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK OPNDispProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	int i;
	HMENU hMenu;
	POINT point;
#ifdef OPNAMODE
	const int ttl = 16;
	int no, ch;
#endif
	static BOOL bDblClk = FALSE;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintOPNDisp(hWnd, TRUE);
			UnlockVM();
			return 0;

		/* wi` */
		case WM_ERASEBKGND:
			return TRUE;

		/* ~[g */
		case WM_LBUTTONUP:
			if (bDblClk) {
				bDblClk = FALSE;
			}
			else {
				LockVM();
#ifdef OPNAMODE
				if (nOPNDisp & 32) {
					no = CalcCh(lParam) / ttl;
					ch = CalcCh(lParam) % ttl;
					if ((no < OPN_PSG) || (ch < 3)) {
						SetMute2(no, ch, !GetMute2(no, ch));
					}
				}
				else {
					SetMute(CalcCh(lParam), !GetMute(CalcCh(lParam)));
				}
#else
				SetMute(CalcCh(lParam), !GetMute(CalcCh(lParam)));
#endif
				PaintOPNDisp(hWnd, FALSE);
				UnlockVM();
			}
			return 0;

		/* \t */
		case WM_LBUTTONDBLCLK:
			LockVM();
#ifdef OPNAMODE
			if (nOPNDisp & 32) {
				no = CalcCh(lParam) / ttl;
				ch = CalcCh(lParam) % ttl;
				for (i = 0; i < 3 * ttl + 3; i++) {
					SetMute2(i / ttl, i % ttl, (BOOL)(i != (no * ttl + ch)));
				}
			}
			else {
				for (i = 0; i < 3 * 6 + 3; i++) {
					SetMute(i, (BOOL)(i != CalcCh(lParam)));
				}
			}
#else
			for (i = 0; i < OPN_CHNUM; i++) {
				SetMute(i, (BOOL)(i != CalcCh(lParam)));
			}
#endif
			PaintOPNDisp(hWnd, FALSE);
			UnlockVM();
			bDblClk = TRUE;
			return 0;

#if 0
		/* ~[gS */
		case WM_RBUTTONUP:
			LockVM();
#ifdef OPNAMODE
			for (i = 0; i < 3 * ttl + 3; i++) {
				SetMute2(i / ttl, i % ttl, FALSE);
			}
#else
			for (i = 0; i < OPN_CHNUM; i++) {
				SetMute(i, FALSE);
			}
#endif
			PaintOPNDisp(hWnd, FALSE);
			UnlockVM();
			return 0;
#endif

		/* ReLXgj[ */
		case WM_CONTEXTMENU:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (bMouseCapture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* ReLXgj[[h */
			hMenu = GetSubMenu(hOPNDispMenu, 0);

			/* [h`FbN */
#if XM7_VER >= 2
#ifdef OPNAMODE
			EnableMenuSub(hMenu, IDM_OPNDISP_PSG, (model_type == MODEL_FM7) && ((nOPNDisp & 32) == 0));
#else
			EnableMenuSub(hMenu, IDM_OPNDISP_PSG, model_type == MODEL_FM7);
#endif
			EnableMenuSub(hMenu, IDM_OPNDISP_OPN, (model_type != MODEL_FM7) || (opn_enable));
#else
#ifdef OPNAMODE
			EnableMenuSub(hMenu, IDM_OPNDISP_PSG, (nOPNDisp & 32) == 0);
#endif
			EnableMenuSub(hMenu, IDM_OPNDISP_OPN, opn_enable);
#endif
			EnableMenuSub(hMenu, IDM_OPNDISP_WHG, whg_enable);
			EnableMenuSub(hMenu, IDM_OPNDISP_THG, thg_enable);
			CheckMenuRadioItem(hMenu, 2, 6, 2 + nOPNDispMode, MF_BYPOSITION);

			/* ReLXgj[s */
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			TrackPopupMenu(hMenu, 0, point.x, point.y, 0, hWnd, NULL);
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* j[폜 */
			DestroyMenu(hOPNDispMenu);

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, NULL, NULL);

			/* rbg}bv */
			if (hOPNDisp) {
				DeleteObject(hOPNDisp);
				hOPNDisp = NULL;
				pOPNDisp = NULL;
			}

			UnlockVM();
			break;

		/* R}h */
		case WM_COMMAND:
			if (CmdOPNDisp(hWnd, LOWORD(wParam))) {
				return 0;
			}
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	OPNfBXvCEChE
 *	
 */
void FASTCALL InitOPNDisp(HWND hWnd)
{
	BITMAPINFOHEADER *pbmi;
	HDC hDC;
	int i;

	/* S̃[N */
	pOPNDisp = NULL;
	hOPNDisp = NULL;

	/* \Ǘ[N */
	for (i = 0; i < OPN_CHNUM; i++) {
		knOPNDisp[i] = -2;
		ktOPNDisp[i] = -1;
		lnOPNDisp[i] = -1;
		ltOPNDisp[i] = 0;
	}
	memset(cnOPNDisp, 0xff, sizeof(cnOPNDisp));
	memset(ctOPNDisp, 0, sizeof(ctOPNDisp));

	/* rbg}bvwb_ */
	pbmi = (BITMAPINFOHEADER*)malloc(sizeof(BITMAPINFOHEADER)
										 + sizeof(RGBQUAD) * 16);
	if (pbmi) {
		memset(pbmi, 0, sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * 16);
		pbmi->biSize = sizeof(BITMAPINFOHEADER);
		pbmi->biWidth = rOPNDisp.right;
		pbmi->biHeight = -rOPNDisp.bottom;
		pbmi->biPlanes = 1;
		pbmi->biBitCount = 4;
		pbmi->biCompression = BI_RGB;

		/* DC擾ADIBZNV쐬 */
		hDC = GetDC(hWnd);
		hOPNDisp = CreateDIBSection(hDC, (BITMAPINFO*)pbmi, DIB_RGB_COLORS,
								(void**)&pOPNDisp, NULL, 0);
		ReleaseDC(hWnd, hDC);
		free(pbmi);
	}
}

/*
 *	OPNfBXvCEChE
 *	TCYύX
 */
void FASTCALL ReSizeOPNDisp(void)
{
	HWND hWnd;
	UINT uMode;
	RECT wrect;
	RECT crect;
	char szWndName[128];
	int n;
	int ys;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_OPNDISP] == NULL) {
		return;
	}

	/* WIgύX`FbN */
	hWnd = hSubWnd[SWND_OPNDISP];
	uMode = 1;							/* OPN only */
	if (whg_use) {
		uMode = 2;						/* OPN + WHG */
	}
	if (thg_use) {
		uMode = 3;						/* OPN + WHG + THG */
	}
	n = 6 * uMode;
#if XM7_VER >= 2
	if (model_type < MODEL_FM77AV) {
		if (opn_use) {
			uMode |= 8;					/* with PSG */
			n += 3;
		}
		else {
			uMode = 8;					/* PSG only */
			n = 3;
		}
	}
#else
	if (opn_use) {
		uMode |= 8;					/* with PSG */
		n += 3;
	}
	else {
		uMode = 8;					/* PSG only */
		n = 3;
	}
#endif
#ifdef OPNAMODE
	if ((opna_use[OPN_STD] && opn_use) ||
		(opna_use[OPN_WHG] && whg_use) ||
		(opna_use[OPN_THG] && thg_use)) {
		uMode |= 32;					/* OPNA */
		if (opn_use) {
			n = 1;
		}
		if (whg_use) {
			n = 2;
		}
		if (thg_use) {
			n = 3;
		}
#if XM7_VER >= 2
		if (model_type < MODEL_FM77AV) {
			n += 1;
		}
#else
		n += 1;
#endif
	}
#endif
#if XM7_VER >= 2
#ifdef OPNAMODE
	if ((nOPNDispMode == LEVEL_PSG) && ((model_type != MODEL_FM7) || (uMode & 32))) {
#else
	if ((nOPNDispMode == LEVEL_PSG) && (model_type != MODEL_FM7)) {
#endif
		nOPNDispMode = LEVEL_AUTO;
	}
#else
#ifdef OPNAMODE
	if ((nOPNDispMode > LEVEL_PSG) && ((model_type == MODEL_FM8) || (uMode & 32))) {
#else
	if ((nOPNDispMode > LEVEL_PSG) && (model_type == MODEL_FM8)) {
#endif
		nOPNDispMode = LEVEL_AUTO;
	}
#endif	/* XM7_VER >= 2 */
	if (nOPNDispMode != LEVEL_AUTO) {
		switch (nOPNDispMode) {
			case LEVEL_PSG:	uMode = (uMode & 0x28);		break;
			case LEVEL_OPN:	uMode = (uMode & 0x28) | 1;	break;
			case LEVEL_WHG:	uMode = (uMode & 0x28) | 2;	break;
			case LEVEL_THG:	uMode = (uMode & 0x28) | 3;	break;
			default:									break;
		}
		n = 6 * (uMode & 0x03);
#if XM7_VER >= 2
		if (model_type < MODEL_FM77AV) {
			n += 3;
		}
#else
		n += 3;
#endif
#ifdef OPNAMODE
		if (uMode & 32) {
			switch (nOPNDispMode) {
				case LEVEL_OPN:	n = 1;	break;
				case LEVEL_WHG:	n = 2;	break;
				case LEVEL_THG:	n = 3;	break;
				default:				break;
			}
#if XM7_VER >= 2
			if (model_type < MODEL_FM77AV) {
				n += 1;
			}
#else
			n += 1;
#endif
		}
#endif
	}
	if (IsIconic(hWnd)) {
		/* ACR`FbN */
		uMode |= 0x80000000;
	}
	if (uMode == nOPNDisp) {
		return;
	}

	/* bN */
	LockVM();

	/* EChE^CgύX */
	if (((uMode & 7) == 0) ^ ((nOPNDisp & 7) == 0)) {
		if ((opn_use && (nOPNDispMode == LEVEL_AUTO)) || (nOPNDispMode > LEVEL_PSG)) {
			GetMsg(hResInstance, IDS_SWND_OPNDISP,
						szWndName, sizeof(szWndName));
		}
		else {
			GetMsg(hResInstance, IDS_SWND_PSGDISP,
						szWndName, sizeof(szWndName));
		}
		SetWindowText(hWnd, szWndName);
	}

	/* [h`FW */
	nOPNDisp = uMode;

	/* rbg}bv */
	if (hOPNDisp) {
		DeleteObject(hOPNDisp);
		hOPNDisp = NULL;
		pOPNDisp = NULL;
	}

	/* ̑ */
#ifdef OPNAMODE
	if (((nOPNDisp & 32) || ((nOPNDisp & 7) == 3)) && !bPopupSwnd) {
#else
	if (((nOPNDisp & 7) == 3) && !bPopupSwnd) {
#endif
		ys = (8 + 12 + 1);
	}
	else {
		ys = (8 * 3 + 1);
	}

#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		rOPNDisp.right = ((42 * 8) + 8) * n;
		rOPNDisp.bottom = ys * 16;
	}
	else {
		rOPNDisp.right = 42 * 8;
		rOPNDisp.bottom = ys * n;
	}
#else
	rOPNDisp.bottom = ys * n;
#endif
	InitOPNDisp(hWnd);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		wrect.right = ((42 * 8) + 8) * n;
	}
	else {
		wrect.right += (42 * 8);
	}
#else
	wrect.right += (42 * 8);
#endif
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
#ifdef OPNAMODE
	if (nOPNDisp & 32) {
		wrect.bottom += (ys * 16);
	}
	else {
		wrect.bottom += (ys * n);
	}
#else
	wrect.bottom += (ys * n);
#endif
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);

	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* AbN */
	UnlockVM();
}

/*
 *	OPNfBXvCEChE
 *	EChE쐬
 */
HWND FASTCALL CreateOPNDisp(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_OPNDisp";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;
	UINT uResourceID;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);

	/* EChÉA42hbg * 8 IN^[u */
	rect.right = 42 * 8;
	if (opn_use) {
		/* EChEćA8 * 3 + 1hbg * 6 `l */
		rect.bottom = (8 * 3 + 1) * 6;
	}
	else {
		/* EChEćA8 * 3 + 1hbg * 3 `l */
		rect.bottom = (8 * 3 + 1) * 3;
	}

	/* EChE^Cg */
	if (opn_use) {
		GetMsg(hResInstance, IDS_SWND_OPNDISP,
					szWndName, sizeof(szWndName));
		nOPNDisp = 1;					/* OPN only */
	}
	else {
		GetMsg(hResInstance, IDS_SWND_PSGDISP,
					szWndName, sizeof(szWndName));
		nOPNDisp = 8;					/* PSG only */
	}

	/* j[[h */
	uResourceID = IDR_OPNDISPMENU;
#ifdef ENGLISH_SUPPORT
	if (!IsJapanese(FALSE)) {
		/* pւ̑Ή */
		uResourceID = IDR_US_OPNDISPMENU;
	}
#endif
	hOPNDispMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(uResourceID));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_DBLCLKS | CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = OPNDispProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
#if 0
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
#else
	wcex.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
#endif
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);

		/* `ۑ */
		rOPNDisp = rect;

		/* ̑ */
		InitOPNDisp(hWnd);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ OPNWX^EChE ]-------------------------------------------------*/

/*
 *	OPNWX^EChE
 *	ZbgAbv(OPNWX^Zbg)
 */
static void FASTCALL SetupOPNRegSub(BYTE *p, int cx, BYTE *reg, int x, int y, BOOL fm_flag)
{
	int i;
	int j;
	char string[128];

	/* XKCh\ */
	strncpy(string, "+0+1+2+3+4+5+6+7+8+9+A+B+C+D+E+F", sizeof(string));
	memcpy(&p[cx * y + x], string, strlen(string));

	/* [v */
	for (i = 0; i < 16; i++) {
		if ((i != 0) && ((nOPNReg & 7) == 0)) {
			continue;
		}

		for (j = 0; j < 16; j++) {
			if (((i == 0) || fm_flag) && reg) {
				_snprintf(string, sizeof(string), "%02X", reg[i * 16 + j]);
			}
			else {
				strncpy(string, "--", sizeof(string));
			}
			memcpy(&p[cx * (y + i + 2) + x + j * 2], string, strlen(string));
		}
	}
}

/*
 *	OPNWX^EChE
 *	ZbgAbv(vXP[)
 */
static void FASTCALL SetupOPNRegPs(BYTE *p, int cx, int scale, int x, int y)
{
	char string[128];
	int i;

	/* vXP[ */
	if (scale == 2) {
		i = 1;
	}
	else {
		i = (scale * 2) / 3;
	}
	if ((nOPNReg & 7) == 0) {
		strncpy(string, "Prescaler   FM : -/-   PSG : 1/2", sizeof(string));
	}
	else {
		_snprintf(string, sizeof(string), "Prescaler   FM : 1/%d   PSG : 1/%d",
			scale, i);
	}
	memcpy(&p[cx * y + x], string, strlen(string));
}

/*
 *	OPNWX^EChE
 *	ZbgAbv
 */
static void FASTCALL SetupOPNReg(BYTE *p, int x, int y)
{
	static const char *fmgen[] = {
		"fmgen",
		"ROMEO",
#if defined(OPNAMODE)
		"-",
#else
		"ROMEO+fmgen",
#endif
#ifdef FMDLL
		"fmdll",
#endif
	};
	static const char *chiptype[] = {
		"Unknown", "OPN", "OPNA", "OPNB-B", "PSG"
	};
	char string[128];
	char *pfmgen;
	int no, i, nx, ny;

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* YKCh\ */
	for (i = 0; i < 16; i++) {
		if ((i != 0) && ((nOPNReg & 7) == 0)) {
			continue;
		}

		_snprintf(string, sizeof(string), "+%02X", i * 16);
		if ((nOPNReg & 7) == 0) {
			memcpy(&p[x * (i + 4) + 0], string, strlen(string));
		}
		else {
			memcpy(&p[x * (i + 6) + 0], string, strlen(string));
#ifdef OPNAMODE
			if (nOPNReg & 32) {
				memcpy(&p[x * (i + 25) + 0], string, strlen(string));
			}
#endif
		}
	}

	/* OPN / WHG / THG */
	for (no = OPN_STD; no <= OPN_THG; no++) {
		nx = 4 + (32 + 2) * no;
		if (((int)(nOPNReg & 7) > no) && (x > nx)) {
			pfmgen = (char *)fmgen[0];
			if (no == OPN_STD) {
				if (bUseRomeo) {
					pfmgen = (char *)fmgen[1];
				}
			}
#if !defined(OPNAMODE)
			else if (no == OPN_WHG) {
				if (bUseRomeo) {
					pfmgen = (char *)fmgen[2];
				}
			}
#endif
			if (bSCCI_available && bUseSCCI && nSCCI_ChipInfo[no]) {
				pfmgen = (char *)chiptype[nSCCI_ChipInfo[no]];
			}
#if defined(FMDLL)
			if (GetUseFmDllFlag(no)) {
				pfmgen = (char *)fmgen[3];
			}
#endif
			if (no == OPN_STD) {
				_snprintf(string, sizeof(string), "OPN (Standard / %s)", pfmgen);
			}
			else if (no == OPN_WHG) {
				_snprintf(string, sizeof(string), "WHG (Extension / %s)", pfmgen);
			}
			else /*if (no == OPN_THG) */{
				_snprintf(string, sizeof(string), "THG (Extension / %s)", pfmgen);
			}
			memcpy(&p[x * 0 + nx], string, strlen(string));
			SetupOPNRegPs(p, x, opn_scale[no], nx, 2);
			SetupOPNRegSub(p, x, opn_reg[no], nx, 4, TRUE);
#ifdef OPNAMODE
			if (nOPNReg & 32) {
				if (opna_use[no]) {
					SetupOPNRegSub(p, x, opn_reg[no] + 256, nx, 23, TRUE);
				}
				else {
					SetupOPNRegSub(p, x, NULL, nx, 23, TRUE);
				}
			}
#endif
		}
	}

	/* PSG */
	if (nOPNReg & 8) {
		nx = 4;
		ny = 2;
		if ((nOPNReg & 7) > 0) {
			ny = 4;
			if (x > 4 + (32 + 2) * (int)(nOPNReg & 7)) {
				nx = 4 + (32 + 2) * (int)(nOPNReg & 7);
			}
		}
		pfmgen = (char *)fmgen[0];
		if (bSCCI_available && bUseSCCI && nSCCI_ChipInfo[OPN_PSG]) {
			pfmgen = (char *)chiptype[nSCCI_ChipInfo[OPN_PSG]];
		}
#if defined(FMDLL)
		if (GetUseFmDllFlag(OPN_PSG)) {
			pfmgen = (char *)fmgen[3];
		}
#endif
		_snprintf(string, sizeof(string), "PSG (Standard / %s)", pfmgen);
		memcpy(&p[x * 0 + nx], string, strlen(string));
		SetupOPNRegSub(p, x, psg_reg, nx, ny, FALSE);
#ifdef OPNAMODE
		if (nOPNReg & 32) {
			SetupOPNRegSub(p, x, NULL, nx, 23, TRUE);
		}
#endif
	}

#ifdef OPNAMODE
	if (nOPNReg & 32) {
		strncpy(string, "STD", sizeof(string));
		memcpy(&p[x * 4], string, strlen(string));
		strncpy(string, "EXT", sizeof(string));
		memcpy(&p[x * 23], string, strlen(string));
	}
#endif
}

/*
 *	OPNWX^EChE
 *	`
 */
static void FASTCALL DrawOPNReg(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pOPNReg) {
		return;
	}
	SetupOPNReg(pOPNReg, x, y);

	/* ` */
	DrawWindowText(hDC, pOPNReg, x, y);
}

/*
 *	OPNWX^EChE
 *	tbV
 */
void FASTCALL RefreshOPNReg(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_OPNREG] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_OPNREG];
	hDC = GetDC(hWnd);
	DrawOPNReg(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	OPNWX^EChE
 *	TCYύX
 */
void FASTCALL ReSizeOPNReg(void)
{
	HWND hWnd;
	UINT uMode;
	RECT wrect;
	RECT crect;
	char szWndName[128];
	int cx;
	int cy;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_OPNREG] == NULL) {
		return;
	}

	/* WIgύX`FbN */
	hWnd = hSubWnd[SWND_OPNREG];
	uMode = 1;							/* OPN only */
	cx = 36;
	cy = 22;
	if (whg_use) {
		uMode = 2;						/* OPN + WHG */
		cx = 36 + 32 + 2;
	}
	if (thg_use) {
		uMode = 3;						/* OPN + WHG + THG */
		cx = 36 + (32 + 2) * 2;
	}
#ifdef OPNAMODE
	if ((opna_use[OPN_STD] && opn_use) ||
		(opna_use[OPN_WHG] && whg_use) ||
		(opna_use[OPN_THG] && thg_use)) {
		uMode |= 32;					/* OPNA */
		cy = 22 + 19;
	}
#endif
#if XM7_VER >= 2
	if (model_type < MODEL_FM77AV) {
		if (opn_use) {
			uMode |= 8;					/* with PSG */
			cx += 32 + 2;
		}
		else {
			uMode = 8;					/* PSG only */
			cx = 36;
			cy = 5;
		}
	}
#else
	if (opn_use) {
		uMode |= 8;					/* with PSG */
		cx += 32 + 2;
	}
	else {
		uMode = 8;					/* PSG only */
		cx = 36;
		cy = 5;
	}
#endif
#if XM7_VER >= 2
#ifdef OPNAMODE
	if ((nOPNRegMode == LEVEL_PSG) && ((model_type != MODEL_FM7) || (uMode & 32))) {
#else
	if ((nOPNRegMode == LEVEL_PSG) && (model_type != MODEL_FM7)) {
#endif
		nOPNRegMode = LEVEL_AUTO;
	}
#else
#ifdef OPNAMODE
	if ((nOPNRegMode > LEVEL_PSG) && ((model_type == MODEL_FM8) || (uMode & 32))) {
#else
	if ((nOPNRegMode > LEVEL_PSG) && (model_type == MODEL_FM8)) {
#endif
		nOPNRegMode = LEVEL_AUTO;
	}
#endif	/* XM7_VER >= 2 */
	if (nOPNRegMode != LEVEL_AUTO) {
		cx = 36;
		cy = 22;
		switch (nOPNRegMode) {
			case LEVEL_PSG:	uMode = (uMode & 0x28);		cy = 5;				break;
			case LEVEL_OPN:	uMode = (uMode & 0x28) | 1;						break;
			case LEVEL_WHG:	uMode = (uMode & 0x28) | 2;	cx += (32 + 2) * 1;	break;
			case LEVEL_THG:	uMode = (uMode & 0x28) | 3;	cx += (32 + 2) * 2;	break;
			default:									break;
		}
#ifdef OPNAMODE
		if (uMode & 32) {
			cy = 22 + 19;
		}
#endif
#if XM7_VER >= 2
		if (model_type < MODEL_FM77AV) {
			if (nOPNRegMode != LEVEL_PSG) {
				cx += 32 + 2;
			}
		}
#else
		if (nOPNRegMode != LEVEL_PSG) {
			cx += 32 + 2;
		}
#endif
	}
	if (IsIconic(hWnd)) {
		/* ACR`FbN */
		uMode |= 0x80000000;
	}
	if (uMode == nOPNReg) {
		return;
	}

	/* bN */
	LockVM();

	/* EChE^CgύX */
#ifdef OPNAMODE
	if ((uMode & 32) && !(nOPNReg & 32)) {				/* OPNA? */
		GetMsg(hResInstance, IDS_SWND_OPNAREG,
					szWndName, sizeof(szWndName));
		SetWindowText(hWnd, szWndName);
	}
	else
#endif
	if (((uMode & 7) == 0) ^ ((nOPNReg & 7) == 0)) {	/* PSG only? */
		if ((opn_use && (nOPNRegMode == LEVEL_AUTO)) || (nOPNRegMode > LEVEL_PSG)) {
			GetMsg(hResInstance, IDS_SWND_OPNREG,
						szWndName, sizeof(szWndName));
		}
		else {
			GetMsg(hResInstance, IDS_SWND_PSGREG,
						szWndName, sizeof(szWndName));
		}
		SetWindowText(hWnd, szWndName);
	}

	/* [h`FW */
	nOPNReg = uMode;

	/* obt@Ď擾A0xFFŖ߂ */
	free(pOPNReg);
	pOPNReg = malloc(2 * cx * cy);

	ASSERT(pOPNReg);
	memset(pOPNReg, 0xff, 2 * cx * cy);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += cx * lCharWidth;
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += cy * lCharHeight;
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);

	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* AbN */
	UnlockVM();
}

/*
 *	OPNWX^EChE
 *	ĕ`
 */
static void FASTCALL PaintOPNReg(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pOPNReg;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawOPNReg(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	OPNWX^EChE
 *	R}h
 */
static void FASTCALL CmdOPNReg(HWND hWnd, WORD wID)
{
	UNUSED(hWnd);
	/* ^[QbgAhX */
	switch (wID) {
		case IDM_OPNREG_AUTO:
			nOPNRegMode = LEVEL_AUTO;
			break;

		case IDM_OPNREG_PSG:
			nOPNRegMode = LEVEL_PSG;
			break;

		case IDM_OPNREG_OPN:
			nOPNRegMode = LEVEL_OPN;
			break;

		case IDM_OPNREG_WHG:
			nOPNRegMode = LEVEL_WHG;
			break;

		case IDM_OPNREG_THG:
			nOPNRegMode = LEVEL_THG;
			break;

		default:
			break;
	}
}

/*
 *	OPNWX^EChE
 *	EChEvV[W
 */
static LRESULT CALLBACK OPNRegProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	HMENU hMenu;
	POINT point;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintOPNReg(hWnd);
			UnlockVM();
			return 0;

		/* ReLXgj[ */
		case WM_CONTEXTMENU:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (bMouseCapture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* ReLXgj[[h */
			hMenu = GetSubMenu(hOPNRegMenu, 0);

			/* [h`FbN */
#if XM7_VER >= 2
#ifdef OPNAMODE
			EnableMenuSub(hMenu, IDM_OPNREG_PSG, (model_type == MODEL_FM7) && ((nOPNReg & 32) == 0));
#else
			EnableMenuSub(hMenu, IDM_OPNREG_PSG, model_type == MODEL_FM7);
#endif
			EnableMenuSub(hMenu, IDM_OPNREG_OPN, (model_type != MODEL_FM7) || (opn_enable));
#else
#ifdef OPNAMODE
			EnableMenuSub(hMenu, IDM_OPNREG_PSG, (nOPNReg & 32) == 0);
#endif
			EnableMenuSub(hMenu, IDM_OPNREG_OPN, opn_enable);
#endif
			EnableMenuSub(hMenu, IDM_OPNREG_WHG, whg_enable);
			EnableMenuSub(hMenu, IDM_OPNREG_THG, thg_enable);
			CheckMenuRadioItem(hMenu, 0, 4, nOPNRegMode, MF_BYPOSITION);

			/* ReLXgj[s */
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			TrackPopupMenu(hMenu, 0, point.x, point.y, 0, hWnd, NULL);
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* j[폜 */
			DestroyMenu(hOPNRegMenu);

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pOPNReg, NULL);

			UnlockVM();
			break;

		/* R}h */
		case WM_COMMAND:
			CmdOPNReg(hWnd, LOWORD(wParam));
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	OPNWX^EChE
 *	EChE쐬
 */
HWND FASTCALL CreateOPNReg(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_OPNReg";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;
	UINT uResourceID;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 36;
	if (opn_use) {
		rect.bottom = lCharHeight * 22;
	}
	else {
		rect.bottom = lCharHeight * 5;
	}

	/* EChE^CgAobt@m */
	if (opn_use) {
		GetMsg(hResInstance, IDS_SWND_OPNREG,
					szWndName, sizeof(szWndName));
		nOPNReg = 1;				/* OPN only */
	}
	else {
		GetMsg(hResInstance, IDS_SWND_PSGREG,
					szWndName, sizeof(szWndName));
		nOPNReg = 8;				/* PSG only */
	}
	pOPNReg = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* j[[h */
	uResourceID = IDR_OPNREGMENU;
#ifdef ENGLISH_SUPPORT
	if (!IsJapanese(FALSE)) {
		/* pւ̑Ή */
		uResourceID = IDR_US_OPNREGMENU;
	}
#endif
	hOPNRegMenu = LoadMenu(hResInstance, MAKEINTRESOURCE(uResourceID));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = OPNRegProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ FDCEChE ]---------------------------------------------------------*/

/*
 *	FDCEChE
 *	ZbgAbv(tO)
 */
static void FASTCALL SetupFDCFlag(BYTE *p, BYTE length, char *title, BYTE flag)
{
	int i;

	/*  */
	memset(p, 0x20, length);

	/* Rs[ */
	for (i=0; i<length; i++) {
		if (title[i] == '\0') {
			break;
		}
		p[i] = title[i];
	}

	/* tOɉĐݒ */
	if (flag) {
		memcpy(&p[length - 2], "On", 2);
	}
	else {
		memcpy(&p[length - 3], "Off", 3);
	}
}

/*
 *	FDCEChE
 *	ZbgAbv(R}h)
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL SetupFDCCmd(BYTE *p, int x, int cx, int no)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL SetupFDCCmd(BYTE *p, int x, int cx)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	const BYTE steprate[4] = { 6, 12, 20, 30 };

	BYTE high, low;
	char buffer[128];
	int y;

	ASSERT(p);
	ASSERT(cx > 0);

	/*  */
	y = 0;

#if XM7_VER >= 2
	/* 2DD[h */
	if (fdc_2ddmode) {
		strncpy(buffer, "2DD Mode", sizeof(buffer));
		memcpy(&p[x + 2 * cx], buffer, strlen(buffer));
	}

	/* DMA */
	if (dma_pcr & 0x01) {
		_snprintf(buffer, sizeof(buffer),  "DMA %04X", dma_bcr[0]);
		memcpy(&p[10 + 2 * cx], buffer, strlen(buffer));
	}
#elif XM7_VER == 1 && defined(SFDC)
	/* FDC */
	if (nFDC != 0) {
		if (model_type == MODEL_FM77) {
			if (no) {
				strncpy(buffer, "1MB FDC        ", sizeof(buffer));
			}
			else {
				strncpy(buffer, "320KB FDC      ", sizeof(buffer));
			}
			if ((no && fdc_mode) || (!no && !fdc_mode)) {
				strncat(buffer, "[*]", sizeof(buffer)-strlen(buffer)-1);
			}
			else {
				strncat(buffer, "[ ]", sizeof(buffer)-strlen(buffer)-1);
			}
		}
		else {
			if (no) {
				if (dma_pcr & 0x08) {
					_snprintf(buffer, sizeof(buffer)-1,"SFDC      DMA %04X",
																dma_bcr[3]);
				}
				else {
					strncpy(buffer, "SFDC", sizeof(buffer));
				}
			}
			else {
				strncpy(buffer, "MFDC", sizeof(buffer));
			}
		}
		memcpy(&p[x + 2 * cx], buffer, strlen(buffer));
	}
#endif

#if XM7_VER == 1 && defined(SFDC)
	if (!fdc_enable[no] && (model_type != MODEL_FM77)) {
#elif  XM7_VER == 1
	if (!fdc_enable && (model_type != MODEL_FM77)) {
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (!fdc_enable && (model_type == MODEL_FM7)) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
		/* R}h */
		strncpy(buffer, "Disabled", sizeof(buffer));
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		return;
	}

#if XM7_VER == 1 && defined(SFDC)
	if (fdc_use[no] != TRUE) {
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (fdc_use != TRUE) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
		/* R}h */
		strncpy(buffer, "NO COMMAND", sizeof(buffer));
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		return;
	}

#if XM7_VER == 1 && defined(SFDC)
	high = (BYTE)(fdc_command[no] >> 4);
	low = (BYTE)(fdc_command[no] & 0x0f);
#else	/* XM7_VER == 1 && defined(SFDC) */
	high = (BYTE)(fdc_command >> 4);
	low = (BYTE)(fdc_command & 0x0f);
#endif	/* XM7_VER == 1 && defined(SFDC) */
	if (high < 8) {
		/* TYPE I */
		switch (high) {
			/* RESTORE */
			case 0:
				strncpy(buffer, "RESTORE", sizeof(buffer));
				break;
			/* SEEK */
			case 1:
				strncpy(buffer, "SEEK", sizeof(buffer));
				break;
			/* STEP */
			case 2:
			case 3:
				strncpy(buffer, "STEP", sizeof(buffer));
				break;
			/* STEP IN */
			case 4:
			case 5:
				strncpy(buffer, "STEP IN", sizeof(buffer));
				break;
			/* STEP IN */
			case 6:
			case 7:
				strncpy(buffer, "STEP OUT", sizeof(buffer));
				break;
		}
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		strncpy(buffer, "(TYPE I)", sizeof(buffer));
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y += 3;

		_snprintf(buffer, sizeof(buffer),  "Step Rate     %2dms",
			steprate[low & 0x03]);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		SetupFDCFlag(&p[x + y * cx], 18, "Verify", (BYTE)(low & 0x04));
		y++;

		strncpy(buffer, "Head        ", sizeof(buffer));
		if (low & 0x08) {
			strncat(buffer, "  Load", sizeof(buffer) - strlen(buffer) - 1);
		}
		else {
			strncat(buffer, "Unload", sizeof(buffer) - strlen(buffer) - 1);
		}
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		if (high >= 2) {
			SetupFDCFlag(&p[x + y * cx], 18, "Update Track", (BYTE)(low & 0x10));
		}
#if !(XM7_VER == 1 && defined(SFDC))
		return;
#endif	/* !(XM7_VER == 1 && defined(SFDC)) */
	}
	else if (high == 0x0d) {
		/* TYPE IV */
		strncpy(buffer, "FORCE INTERRUPT", sizeof(buffer));
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		strncpy(buffer, "(TYPE IV)", sizeof(buffer));
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y += 3;

		SetupFDCFlag(&p[x + (y + 0) * cx], 18, "READY  In", (BYTE)(low & 0x01));
		SetupFDCFlag(&p[x + (y + 1) * cx], 18, "READY Out", (BYTE)(low & 0x02));
		SetupFDCFlag(&p[x + (y + 2) * cx], 18, "INDEX", (BYTE)(low & 0x04));
		SetupFDCFlag(&p[x + (y + 3) * cx], 18, "One Shot", (BYTE)(low & 0x08));
	}
	else {
		/* TYPE II/III */
		switch (high) {
			case 0x08:
			case 0x09:
				strncpy(buffer, "READ DATA", sizeof(buffer));
				break;
			case 0x0a:
			case 0x0b:
				strncpy(buffer, "WRITE DATA", sizeof(buffer));
				break;
			case 0x0c:
				strncpy(buffer, "READ ADDRESS", sizeof(buffer));
				break;
			case 0x0e:
				strncpy(buffer, "READ TRACK", sizeof(buffer));
				break;
			case 0x0f:
				strncpy(buffer, "WRITE TRACK", sizeof(buffer));
				break;
		}
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		if (high < 0x0c) {
			strncpy(buffer, "(TYPE II)", sizeof(buffer));
			memcpy(&p[x + y * cx], buffer, strlen(buffer));
			y += 2;

			SetupFDCFlag(&p[x + y * cx], 18, "Multi Sector", (BYTE)(high & 0x01));
			y++;

			strncpy(buffer, "Compare Side   ", sizeof(buffer));
			if (low & 0x02) {
				if (low & 0x08) {
					strncat(buffer, "  0", sizeof(buffer) - strlen(buffer) - 1);
				}
				else {
					strncat(buffer, "  1", sizeof(buffer) - strlen(buffer) - 1);
				}
			} else {
				strncat(buffer, "Off", sizeof(buffer) - strlen(buffer) - 1);
			}
			memcpy(&p[x + y * cx], buffer, strlen(buffer));
			y++;

			strncpy(buffer, "Addr. Mark ", sizeof(buffer));
			if (low & 0x01) {
				strncat(buffer, "Deleted", sizeof(buffer) - strlen(buffer) - 1);
			}
			else {
				strncat(buffer, " Normal", sizeof(buffer) - strlen(buffer) - 1);
			}
			memcpy(&p[x + y * cx], buffer, strlen(buffer));
			y++;
		}
		else {
			strncpy(buffer, "(TYPE III)", sizeof(buffer));
			memcpy(&p[x + y * cx], buffer, strlen(buffer));
			y += 5;
		}

#if XM7_VER == 1 && defined(SFDC)
		_snprintf(buffer, sizeof(buffer),  "Total Bytes   %04X", fdc_totalcnt[no]);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		_snprintf(buffer, sizeof(buffer),  "Xfer. Bytes   %04X", fdc_nowcnt[no]);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
#else	/* XM7_VER == 1 && defined(SFDC) */
		_snprintf(buffer, sizeof(buffer),  "Total Bytes   %04X", fdc_totalcnt);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		_snprintf(buffer, sizeof(buffer),  "Xfer. Bytes   %04X", fdc_nowcnt);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
#endif	/* XM7_VER == 1 && defined(SFDC) */
	}

#if XM7_VER == 1 && defined(SFDC)
	y++;

	/* SFD荞݃[h */
	if ((model_type == MODEL_FM77) && no) {
		strncpy(buffer, "Interrupt   ", sizeof(buffer));
		if (sfd_firq_mode) {
			strncat(buffer, "   SFD", sizeof(buffer) - strlen(buffer) - 1);
		}
		else {
			strncat(buffer, "Normal", sizeof(buffer) - strlen(buffer) - 1);
		}
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
	}
	y++;

	if (no) {
		/* Door Lock */
		strncpy(buffer, "Door        ", sizeof(buffer));
		if (fdc_doorlock[fdc_drvreg[no]]) {
			strncat(buffer, "  Lock", sizeof(buffer) - strlen(buffer) - 1);
		}
		else {
			strncat(buffer, "Unlock", sizeof(buffer) - strlen(buffer) - 1);
		}
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
}

/*
 *	FDCEChE
 *	ZbgAbv(WX^)
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL SetupFDCReg(BYTE *p, int x, int cx, int no)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL SetupFDCReg(BYTE *p, int x, int cx)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	int y;
	char buffer[128];

	ASSERT(p);
	ASSERT(cx > 0);

	/*  */
	y = 0;

#if XM7_VER == 1 && defined(SFDC)
	_snprintf(buffer, sizeof(buffer),  "Drive   %1d", fdc_drvreg[no]);
	if (fdc_motor[no]) {
		strncat(buffer, "( On)", sizeof(buffer)-strlen(buffer)-1);
	}
	else {
		strncat(buffer, "(Off)", sizeof(buffer)-strlen(buffer)-1);
	}
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	if (fdc_drvreg[no] < FDC_DRIVES) {
		_snprintf(buffer, sizeof(buffer), "Track       %02X", fdc_track[no][fdc_drvreg[no]]);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
	}
	y++;

	_snprintf(buffer, sizeof(buffer), "Track  Reg. %02X", fdc_trkreg[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer), "Sector Reg. %02X", fdc_secreg[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer), "Side   Reg. %02X", fdc_sidereg[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer), "Data   Reg. %02X", fdc_datareg[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	SetupFDCFlag(&p[x + (y + 0) * cx], 14, "DRQ", (BYTE)(fdc_drqirq[no] & 0x80));
	SetupFDCFlag(&p[x + (y + 1) * cx], 14, "IRQ", (BYTE)(fdc_drqirq[no] & 0x40));
#else	/* XM7_VER == 1 && defined(SFDC) */
#if XM7_VER >= 2
	if (fdc_dsenable) {
		strncpy(buffer, "Drive  Disable", sizeof(buffer));
	}
	else {
#if XM7_VER >= 2
		if (model_type > MODEL_FM77AV) {
			_snprintf(buffer, sizeof(buffer), "Drive %1d/%1d",
				fdc_drvregP, fdc_drvreg);
		}
		else {
			_snprintf(buffer, sizeof(buffer), "Drive   %1d", fdc_drvreg);
		}
#else
		_snprintf(buffer, sizeof(buffer), "Drive   %1d", fdc_drvreg);
#endif
		if (fdc_motor) {
			strncat(buffer, "( On)", sizeof(buffer) - strlen(buffer) - 1);
		}
		else {
			strncat(buffer, "(Off)", sizeof(buffer) - strlen(buffer) - 1);
		}
	}
#endif
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	if (fdc_drvreg < FDC_DRIVES) {
		_snprintf(buffer, sizeof(buffer),  "Track       %02X",
			fdc_track[fdc_drvreg]);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
	}
	y++;

	_snprintf(buffer, sizeof(buffer),  "Track  Reg. %02X", fdc_trkreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Sector Reg. %02X", fdc_secreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Side   Reg. %02X", fdc_sidereg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Data   Reg. %02X", fdc_datareg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	SetupFDCFlag(&p[x + (y + 0) * cx], 14, "DRQ", (BYTE)(fdc_drqirq & 0x80));
	SetupFDCFlag(&p[x + (y + 1) * cx], 14, "IRQ", (BYTE)(fdc_drqirq & 0x40));
	if (mfd_irq_mask) {
		strncpy(buffer, "Mask/", sizeof(buffer));
		memcpy(&p[(x + 6) + (y + 1) * cx], buffer, strlen(buffer));
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

#if XM7_VER == 1 && defined(SFDC)
	if (no) {
		y+=2;

		/* Px/{x */
		if (fdc_dden[no]) {
			strncpy(buffer, "Double", sizeof(buffer));
		}
		else {
			strncpy(buffer, "Single", sizeof(buffer));
		}
		strncat(buffer, " Density", sizeof(buffer)-strlen(buffer)-1);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;

		/* Ж/ */
		if (fdc_sided[fdc_drvreg[no]]) {
			strncpy(buffer, "Double", sizeof(buffer));
		}
		else {
			strncpy(buffer, "Single", sizeof(buffer));
		}
		strncat(buffer, "   Sided", sizeof(buffer)-strlen(buffer)-1);
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
}

/*
 *	FDCEChE
 *	ZbgAbv(Xe[^X)
 */
#if XM7_VER == 1 && defined(SFDC)
static void FASTCALL SetupFDCStat(BYTE *p, int x, int cx, int no)
#else	/* XM7_VER == 1 && defined(SFDC) */
static void FASTCALL SetupFDCStat(BYTE *p, int x, int cx)
#endif	/* XM7_VER == 1 && defined(SFDC) */
{
	int y;
	int type;
	int i;
	BYTE dat;
	BYTE bit;
	char buffer[128];

	ASSERT(p);
	ASSERT(cx > 0);

	/*  */
	y = 0;

	/* ^Cv */
	type = 0;
#if XM7_VER == 1 && defined(SFDC)
	switch (fdc_cmdtype[no]) {
#else	/* XM7_VER == 1 && defined(SFDC) */
	switch (fdc_cmdtype) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
		case 2:
			/* READ DATA */
			type = 1;
			break;
		case 3:
			/* WRITE DATA */
			type = 2;
			break;
		case 4:
			/* READ ADDRESS */
			type = 1;
			break;
		case 5:
			/* WRITE TRACK */
			type = 2;
			break;
		case 6:
			/* READ TRACK */
			type = 1;
			break;
		default:
			break;
	}

	/* ݒ */
#if XM7_VER == 1 && defined(SFDC)
	dat = fdc_status[no];
#else	/* XM7_VER == 1 && defined(SFDC) */
	dat = fdc_status;
#endif	/* XM7_VER == 1 && defined(SFDC) */
	bit = 0x80;

	/* Wrbg[v */
	for (i=7; i>=0; i--) {
		_snprintf(buffer, sizeof(buffer),  "bit%d ", i);
		if (dat & bit) {
			switch (i) {
				/* BUSY */
				case 0:
					strncat(buffer, "BUSY", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* INDEX or DATA REQUEST */
				case 1:
					if (type == 0) {
						strncat(buffer, "INDEX", sizeof(buffer) - strlen(buffer) - 1);
					}
					else {
						strncat(buffer, "DATA REQUEST", sizeof(buffer) - strlen(buffer) - 1);
					}
					break;
				/* TRACK00 or LOST DATA */
				case 2:
					if (type == 0) {
						strncat(buffer, "TRACK00", sizeof(buffer) - strlen(buffer) - 1);
					}
					else {
						strncat(buffer, "LOST DATA", sizeof(buffer) - strlen(buffer) - 1);
					}
					break;
				/* CRC ERROR */
				case 3:
					strncat(buffer, "CRC ERROR", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* SEEK ERROR or RECORD NOT FOUND */
				case 4:
					if (type == 0) {
						strncat(buffer, "SEEK ERROR", sizeof(buffer) - strlen(buffer) - 1);
					}
					else {
						strncat(buffer, "RECORD NOT FOUND", sizeof(buffer) - strlen(buffer) - 1);
					}
					break;
				/* HEAD ENGAGED or RECORD TYPE or WRITE FAULT */
				case 5:
					switch (type) {
						case 0:
							strncat(buffer, "HEAD ENGAGED", sizeof(buffer) - strlen(buffer) - 1);
							break;
						case 1:
							strncat(buffer, "RECORD TYPE", sizeof(buffer) - strlen(buffer) - 1);
							break;
						case 2:
							strncat(buffer, "WRITE FAULT", sizeof(buffer) - strlen(buffer) - 1);
							break;
					}
					break;
				/* WRITE PROTECT */
				case 6:
					strncat(buffer, "WRITE PROTECT", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* NOT READY */
				case 7:
					strncat(buffer, "NOT READY", sizeof(buffer) - strlen(buffer) - 1);
					break;
			}
		}
		else {
			strncat(buffer, "----------------", sizeof(buffer) - strlen(buffer) - 1);
		}
		bit >>= 1;
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;
	}
}

#if XM7_VER == 1 && defined(SFDC)
/*
 *	IFDCEChE
 */
static void FASTCALL SetupIFDC(BYTE *p, int xx, int cx)
{
	int x, y;
	char buffer[128];

	ASSERT(p);
	ASSERT(cx > 0);

	x = xx;
	y = 0;

	strncpy(buffer, "IFDC", sizeof(buffer));
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x = xx + 12;

	/* IFDC WX^U[uWX^(RRR) */
	_snprintf(buffer, sizeof(buffer), "RRR:%02X", ifdc_rrr);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x = xx + 12 + strlen(buffer);
	switch (ifdc_rrr) {
		/* CMDREG */
		case 0x00:
			strncpy(buffer, "CMDREG", sizeof(buffer));
			break;
		/* TRKREG */
		case 0x01:
			strncpy(buffer, "TRKREG", sizeof(buffer));
			break;
		/* SSREG */
		case 0x02:
			strncpy(buffer, "SSREG", sizeof(buffer));
			break;
		/* SEREG */
		case 0x03:
			strncpy(buffer, "SEREG", sizeof(buffer));
			break;
		/* SIDREG */
		case 0x04:
			strncpy(buffer, "SIDREG", sizeof(buffer));
			break;
		/* DNSREG */
		case 0x05:
			strncpy(buffer, "DNSREG", sizeof(buffer));
			break;
		/* DRVREG */
		case 0x06:
			strncpy(buffer, "DRVREG", sizeof(buffer));
			break;
		/* DLCREG */
		case 0x07:
			strncpy(buffer, "DLCREG", sizeof(buffer));
			break;
		/* ERESET */
		case 0x08:
			strncpy(buffer, "ERESET", sizeof(buffer));
			break;
		/* SFORMT */
		case 0x09:
			strncpy(buffer, "SFORMT", sizeof(buffer));
			break;
		/* SECDAT */
		case 0x0a:
			strncpy(buffer, "SECDAT", sizeof(buffer));
			break;
		/* MODREG */
		case 0x0b:
			strncpy(buffer, "MODREG", sizeof(buffer));
			break;
		/* OTHERS */
		default:
			strncpy(buffer, "OTHERS", sizeof(buffer));
			break;
	}
	memcpy(&p[x + y * cx], "[", 1);
	x += 1;
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += strlen(buffer);
	memcpy(&p[x + y * cx], "]", 1);
	x = xx + 12 + 18;

	/* IFDC WX^f[^WX^(RDR) */
	_snprintf(buffer, sizeof(buffer), "RDR:%02X", ifdc_rdr);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x = xx + 12 + 18 + 9;

	/* IFDC VXeXe[^XWX^(SSR) */
	_snprintf(buffer, sizeof(buffer), "SSR:%02X", ifdc_ssr);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x = xx + 12 + 18 + 9 + 9;

	/* IFDC FDCXe[^XWX^(FSR) */
	_snprintf(buffer, sizeof(buffer), "FSR:%02X", ifdc_fsr);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));

	x = xx;
	y = 1;

	/* IFDC R}hWX^(CMDREG) */
	_snprintf(buffer, sizeof(buffer), "CMDREG:%02X", ifdc_cmdreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += strlen(buffer) + 3;
	memcpy(&p[x + y * cx], "[", 1);
	x += 1;
	switch (ifdc_cmdreg) {
		/* IFDC INITIALIZE */
		case 0x00:
			strncpy(buffer, "IFDC INITIALIZE", sizeof(buffer));
			break;
		/* MAKE SOFT SECTOR FORMAT */
		case 0x01:
			strncpy(buffer, "MAKE SOFT SECTOR FORMAT", sizeof(buffer));
			break;
		/* SEEK TRACK */
		case 0x02:
			strncpy(buffer, "SEEK TRACK", sizeof(buffer));
			break;
		/* READ SECTOR (FLOPPY --> IFDC BUFFER) */
		case 0x03:
			strncpy(buffer, "READ SECTOR (FLOPPY --> IFDC BUFFER)", sizeof(buffer));
			break;
		/* WRITE SECTOR(FLOPPY <-- IFDC BUFFER) */
		case 0x04:
			strncpy(buffer, "WRITE SECTOR(FLOPPY <-- IFDC BUFFER)", sizeof(buffer));
			break;
		/* DATA GET (PROGRAM TRANSFER) FM8 <-- IFDC */
		case 0x05:
			strncpy(buffer, "DATA GET (PROGRAM TRANSFER) FM8 <-- IFDC", sizeof(buffer));
			break;
		/* DATA PUT (PROGRAM TRANSFER) FM8 --> IFDC */
		case 0x06:
			strncpy(buffer, "DATA PUT (PROGRAM TRANSFER) FM8 --> IFDC", sizeof(buffer));
			break;
		/* DATA GET (DMA TRANSFER)     FM8 <-- IFDC */
		case 0x07:
			strncpy(buffer, "DATA GET (DMA TRANSFER)     FM8 <-- IFDC", sizeof(buffer));
			break;
		/* DATA PUT (DMA TRANSFER)     FM8 --> IFDC */
		case 0x08:
			strncpy(buffer, "DATA PUT (DMA TRANSFER)     FM8 --> IFDC", sizeof(buffer));
			break;
		/* DIRECT READ SECTOR (PROGRAM) FM8 <-- FLOPPY */
		case 0x09:
			strncpy(buffer, "DIRECT READ SECTOR (PROGRAM) FM8 <-- FLOPPY", sizeof(buffer));
			break;
		/* DIRECT WRITE SECTOR(PROGRAM) FM8 --> FLOPPY */
		case 0x0a:
			strncpy(buffer, "DIRECT WRITE SECTOR(PROGRAM) FM8 --> FLOPPY", sizeof(buffer));
			break;
		/* DIRECT READ SECTOR (DMA)     FM8 <-- FLOPPY) */
		case 0x0b:
			strncpy(buffer, "DIRECT READ SECTOR (DMA)     FM8 <-- FLOPPY", sizeof(buffer));
			break;
		/* DIRECT WRITE SECTOR(DMA)     FM8 --> FLOPPY) */
		case 0x0c:
			strncpy(buffer, "DIRECT WRITE SECTOR(DMA)     FM8 --> FLOPPY", sizeof(buffer));
			break;
		/* FLOPPY DISK SEEK TRACK OPERATION */
		case 0x0d:
			strncpy(buffer, "FLOPPY DISK SEEK TRACK OPERATION", sizeof(buffer));
			break;
		/* OTHERS */
		default:
			strncpy(buffer, "OTHERS", sizeof(buffer));
			break;
	}
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += strlen(buffer);
	memcpy(&p[x + y * cx], "]", 1);

	x = xx;
	y = 2;

	/* IFDC gbNWX^(TRKREG) */
	_snprintf(buffer, sizeof(buffer), "TRKREG:%02X", ifdc_trkreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC ZN^X^[gWX^(SSREG) */
	_snprintf(buffer, sizeof(buffer), "SSREG: %02X", ifdc_ssreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC ZN^GhWX^(SEREG) */
	_snprintf(buffer, sizeof(buffer), "SEREG: %02X", ifdc_sereg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC TChWX^(SIDREG) */
	_snprintf(buffer, sizeof(buffer), "SIDREG:%02X", ifdc_sidreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC xWX^(DNSREG) */
	_snprintf(buffer, sizeof(buffer), "DNSREG:%02X", ifdc_dnsreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));

	x = xx;
	y++;

	/* IFDC hCuWX^(DRVREG) */
	_snprintf(buffer, sizeof(buffer), "DRVREG:%02X", ifdc_drvreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC hAbNWX^(DLCREG) */
	_snprintf(buffer, sizeof(buffer), "DLCREG:%02X", ifdc_dlcreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC ZN^tH[}bg(SFORMT) */
	_snprintf(buffer, sizeof(buffer), "SFORMT:%02X", ifdc_sformt);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC ZN^OX(SECDAT) */
	_snprintf(buffer, sizeof(buffer), "SECDAT:%02X", ifdc_secdat);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	x += 12;

	/* IFDC [hZbgWX^(MODREG) */
	_snprintf(buffer, sizeof(buffer), "MODREG:%02X", ifdc_modreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
}
#endif	/* XM7_VER == 1 && defined(SFDC) */

/*
 *	FDCEChE
 *	ZbgAbv
 */
static void FASTCALL SetupFDC(BYTE *p, int x, int y)
{
	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* Tu֐Ă */
#if XM7_VER == 1 && defined(SFDC)
	SetupFDCCmd(p, 0, x, FDC_M);
	SetupFDCReg(p, 20, x, FDC_M);
	SetupFDCStat(p, 36, x, FDC_M);

	if (nFDC != 0) {
		memset(&p[8 * x], '-', x);
		SetupFDCCmd(p, 9 * x + 0, x, FDC_S);
		SetupFDCReg(p, 9 * x + 20, x, FDC_S);
		SetupFDCStat(p, 9 * x + 36, x, FDC_S);
	}

	if ((nFDC != 0) && (model_type == MODEL_FM8)) {
		memset(&p[19 * x], '-', x);
		SetupIFDC(p, 20 * x + 0, x);
	}
#else	/* XM7_VER == 1 && defined(SFDC) */
	SetupFDCCmd(p, 0, x);
	SetupFDCReg(p, 20, x);
	SetupFDCStat(p, 36, x);
#endif	/* XM7_VER == 1 && defined(SFDC) */
}

/*
 *	FDCEChE
 *	`
 */
static void FASTCALL DrawFDC(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pFDC) {
		return;
	}
	SetupFDC(pFDC, x, y);

	/* ` */
	DrawWindowText(hDC, pFDC, x, y);
}

/*
 *	FDCEChE
 *	tbV
 */
void FASTCALL RefreshFDC(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_FDC] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_FDC];
	hDC = GetDC(hWnd);
	DrawFDC(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

#if XM7_VER == 1 && defined(SFDC)
/*
 *	FDCEChE
 *	TCYύX
 */
void FASTCALL ReSizeFDC(void)
{
	HWND hWnd;
	UINT uMode;
	RECT wrect;
	RECT crect;
	int cx;
	int cy;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_FDC] == NULL) {
		return;
	}

	/* ŏȂAȂ */
	hWnd = hSubWnd[SWND_FDC];
	if (IsIconic(hWnd)) {
		return;
	}

	/* WIgύX`FbN */
	uMode = 0;
	cx = 57;
	cy = 8;
	if (fdc_use[FDC_S] == TRUE) {
		if (model_type == MODEL_FM8) {
			uMode = 2;
			cy = 24;
		}
		else {
			uMode = 1;
			cy = 19;
		}
	}

	if (uMode == nFDC) {
		return;
	}

	/* bN */
	LockVM();

	/* [h`FW */
	nFDC = uMode;

	/* obt@ */
	free(pFDC);
	pFDC = malloc(2 * cx * cy);
	memset(pFDC, 0xff, 2 * cx * cy);
	ASSERT(pFDC);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += cx * lCharWidth;
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += cy * lCharHeight;
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* AbN */
	UnlockVM();
}
#endif	/* XM7_VER == 1 && defined(SFDC) */

/*
 *	FDCEChE
 *	ĕ`
 */
static void FASTCALL PaintFDC(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pFDC;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawFDC(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	FDCEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK FDCProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintFDC(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pFDC, NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	FDCEChE
 *	EChE쐬
 */
HWND FASTCALL CreateFDC(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_FDC";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 57;
	rect.bottom = lCharHeight * 8;

	/* EChE^CgAobt@m */
	GetMsg(hResInstance, IDS_SWND_FDC,
				szWndName, sizeof(szWndName));
	pFDC = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));
#if XM7_VER == 1 && defined(SFDC)
	nFDC = 0;
#endif	/* XM7_VER == 1 && defined(SFDC) */

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = FDCProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ ouRg[EChE ]------------------------------------*/

#if XM7_VER == 1 && defined(BUBBLE)
/*
 *	ouRg[EChE
 *	ZbgAbv(R}h/WX^)
 */
static void FASTCALL SetupBMCReg(BYTE *p, int x, int cx, int no)
{
	char buffer[128];
	int y;
	BYTE ncmd;
	WORD npagreg;
	WORD ncntreg;

	ASSERT(p);
	ASSERT(cx > 0);

	/*  */
	y = 0;

	/* BMC */
	if (no) {
		strncpy(buffer, "128KB BMC", sizeof(buffer));
	}
	else {
		strncpy(buffer, "32KB BMC", sizeof(buffer));
	}
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	if (!bmc_enable[no] ||
		((no == BMC_32) &&
		 (model_type != MODEL_FM8) &&
		 (opn_enable || (mos_port == 3) || !(bmc_secret_flag || bmc_use[BMC_32])))) {
		/* ouz_jbg */
		strncpy(buffer, "Disabled", sizeof(buffer));
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		return;
	}

	if (bBMC) {
		ncmd = nBubbleCommand[no];
		npagreg = wBubblePageReg[no];
		ncntreg = wBubbleCountReg[no];
	}
	else {
		ncmd = bmc_command[no];
		npagreg = bmc_pagereg[no];
		ncntreg = bmc_countreg[no];
	}

	switch (ncmd & 0x0f) {
		case 0x00:
			strncpy(buffer, "NO COMMAND", sizeof(buffer));
			break;
		case 0x01:
			strncpy(buffer, "READ", sizeof(buffer));
			break;
		case 0x02:
			strncpy(buffer, "WRITE", sizeof(buffer));
			break;
		case 0x04:
		case 0x0f:
			strncpy(buffer, "INITIALIZE", sizeof(buffer));
			break;
		default:
			_snprintf(buffer, sizeof(buffer),  "UNDEFINED     %02X", ncmd & 0x0f);
			break;
	}
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Data Reg.     %02X", bmc_datareg[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Unit          %2d", bmc_unit[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Page Reg.   %04X", npagreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Count Reg.  %04X", ncntreg);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Total Bytes %04X", bmc_totalcnt[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
	y++;

	_snprintf(buffer, sizeof(buffer),  "Xfer. Bytes %04X", bmc_nowcnt[no]);
	memcpy(&p[x + y * cx], buffer, strlen(buffer));
}

/*
 *	ouRg[EChE
 *	ZbgAbv(Xe[^X)
 */
static void FASTCALL SetupBMCStat(BYTE *p, int x, int cx, int no)
{
	int y;
	int i;
	BYTE dat;
	BYTE bit;
	char buffer[128];

	ASSERT(p);
	ASSERT(cx > 0);

	/*  */
	y = 0;

	/* ݒ */
	dat = bmc_status[no];
	if (!bmc_enable[no] ||
		((no == BMC_32) &&
		 (model_type != MODEL_FM8) &&
		 (opn_enable || (mos_port == 3) || !(bmc_secret_flag || bmc_use[BMC_32])))) {
		dat = 0;
	}
	bit = 0x80;

	/* Wrbg[v */
	for (i=7; i>=0; i--) {
		_snprintf(buffer, sizeof(buffer),  "bit%d ", i);
		if (dat & bit) {
			switch (i) {
				/* BUSY */
				case 0:
					strncat(buffer, "BUSY", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* ERROR ANALYSIS */
				case 1:
					strncat(buffer, "ERROR ANALYSIS", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* WRITE PROTECT */
				case 2:
					strncat(buffer, "WRITE PROTECT", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* DEVICE READY */
				case 3:
					strncat(buffer, "DEVICE READY", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* ---------------- */
				case 4:
					strncat(buffer, "-----------------", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* RDA */
				case 5:
					strncat(buffer, "RDA", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* TDRA */
				case 6:
					strncat(buffer, "TDRA", sizeof(buffer) - strlen(buffer) - 1);
					break;
				/* CME */
				case 7:
					strncat(buffer, "CME", sizeof(buffer) - strlen(buffer) - 1);
					break;
			}
		}
		else {
			strncat(buffer, "-----------------", sizeof(buffer) - strlen(buffer) - 1);
		}
		bit >>= 1;
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;
	}
}

/*
 *	ouRg[EChE
 *	ZbgAbv(G[Xe[^X)
 */
static void FASTCALL SetupBMCErr(BYTE *p, int x, int cx, int no)
{
	int y;
	int i;
	BYTE dat;
	BYTE bit;
	char buffer[128];

	ASSERT(p);
	ASSERT(cx > 0);

	/*  */
	y = 0;

	/* ݒ */
	dat = bmc_errorreg[no];
	bit = 0x80;

	/* Wrbg[v */
	for (i=7; i>=0; i--) {
		memset(buffer, 0, sizeof(buffer));
		if (dat & bit) {
			switch (i) {
				/* UNDEFINED COMMAND ERROR */
				case 0:
					strncpy(buffer, "UNDEFINED CMD ERR", sizeof(buffer));
					break;
				/* NO MARKER */
				case 1:
					strncpy(buffer, "NO MARKER", sizeof(buffer));
					break;
				/* MANY BAD LOOP */
				case 2:
					strncpy(buffer, "MANY BAD LOOP", sizeof(buffer));
					break;
				/* TRANSFER MISSING */
				case 3:
					strncpy(buffer, "TRANSFER MISSING", sizeof(buffer));
					break;
				/* CRC ERROR */
				case 4:
					strncpy(buffer, "CRC ERROR", sizeof(buffer));
					break;
				/* PAGE ADDRESS ERROR */
				case 5:
					strncpy(buffer, "PAGE ADDRESS ERR.", sizeof(buffer));
					break;
				/* DEVICE NOT READY */
				case 6:
				case 7:
					strncpy(buffer, "EJECT ERROR", sizeof(buffer));
					break;
			}
		}
		else {
			strncpy(buffer, "-----------------", sizeof(buffer));
		}
		bit >>= 1;
		memcpy(&p[x + y * cx], buffer, strlen(buffer));
		y++;
	}
}

/*
 *	ouRg[EChE
 *	ZbgAbv
 */
static void FASTCALL SetupBMC(BYTE *p, int x, int y)
{
	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* Tu֐Ă */
	SetupBMCReg( p,  0, x, BMC_32);
	SetupBMCStat(p, 17, x, BMC_32);
	SetupBMCErr( p, 40, x, BMC_32);

	if (nBMC == 1) {
		memset(&p[8 * x], '-', x);
		SetupBMCReg( p, 9 * x +  0, x, BMC_128);
		SetupBMCStat(p, 9 * x + 17, x, BMC_128);
		SetupBMCErr( p, 9 * x + 40, x, BMC_128);
	}
}

/*
 *	ouRg[EChE
 *	`
 */
static void FASTCALL DrawBMC(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pBMC) {
		return;
	}
	SetupBMC(pBMC, x, y);

	/* ` */
	DrawWindowText(hDC, pBMC, x, y);
}

/*
 *	ouRg[EChE
 *	tbV
 */
void FASTCALL RefreshBMC(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_BMC] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_BMC];
	hDC = GetDC(hWnd);
	DrawBMC(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	ouRg[EChE
 *	TCYύX
 */
void FASTCALL ReSizeBMC(void)
{
	HWND hWnd;
	UINT uMode;
	RECT wrect;
	RECT crect;
	int cx;
	int cy;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_BMC] == NULL) {
		return;
	}

	/* ŏȂAȂ */
	hWnd = hSubWnd[SWND_BMC];
	if (IsIconic(hWnd)) {
		return;
	}

	/* WIgύX`FbN */
	uMode = 0;
	cx = 57;
	cy = 8;
	if (nBubbleCommand[BMC_128]) {
		uMode = 1;
		cy = 17;
	}

	if (uMode == nBMC) {
		return;
	}

	/* bN */
	LockVM();

	/* [h`FW */
	nBMC = uMode;

	/* obt@ */
	free(pBMC);
	pBMC = malloc(2 * cx * cy);
	memset(pBMC, 0xff, 2 * cx * cy);
	ASSERT(pBMC);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += cx * lCharWidth;
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += cy * lCharHeight;
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* AbN */
	UnlockVM();
}

/*
 *	ouRg[EChE
 *	ĕ`
 */
static void FASTCALL PaintBMC(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pBMC;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawBMC(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	ouRg[EChE
 *	EChEvV[W
 */
static LRESULT CALLBACK BMCProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	HMENU hMenu;
	POINT point;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintBMC(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pBMC, hBMC);

			UnlockVM();
			break;

		/* ReLXgj[ */
		case WM_CONTEXTMENU:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (bMouseCapture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif

			/* ReLXgj[[h */
			hMenu = GetSubMenu(hBMC, 0);

			/* [h`FbN */
			CheckMenuSub(hMenu, IDM_BMC_FLAG, bBMC);

			/* ReLXgj[s */
			point.x = LOWORD(lParam);
			point.y = HIWORD(lParam);
			TrackPopupMenu(hMenu, 0, point.x, point.y, 0, hWnd, NULL);
			return 0;

		/* R}h */
		case WM_COMMAND:
			bBMC = !bBMC;
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	ouRg[EChE
 *	EChE쐬
 */
HWND FASTCALL CreateBMC(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_BMC";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;
	UINT uResourceID;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 57;
	rect.bottom = lCharHeight * 8;

	/* EChE^CgAobt@m */
	GetMsg(hResInstance, IDS_SWND_BMC,
				szWndName, sizeof(szWndName));
	pBMC = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));
	nBMC = 0;

	/* j[[h */
	uResourceID = IDR_BMCMENU;
#ifdef ENGLISH_SUPPORT
	if (!IsJapanese(FALSE)) {
		/* pւ̑Ή */
		uResourceID = IDR_US_BMCMENU;
	}
#endif
	hBMC = LoadMenu(hResInstance, MAKEINTRESOURCE(uResourceID));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = BMCProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}
#endif

#endif	/* _WIN32 */
