/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2025 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2025 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2025 GIMONS
 *	Copyright (C) 2010-2025 Toma
 *
 *	[ Win32API TuEBhER ]
 */

#ifdef _WIN32

#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <assert.h>
#include <stdlib.h>
#include "xm7.h"
#include "keyboard.h"
#include "rtc.h"
#include "mmr.h"
#include "jcard.h"
#include "dmac.h"
#include "subctrl.h"
#include "display.h"
#include "multipag.h"
#include "ttlpalet.h"
#include "apalet.h"
#include "mouse.h"
#include "kanji.h"
#include "jsubsys.h"
#include "tapelp.h"
#include "mainetc.h"
#include "w32.h"
#include "w32_res.h"
#include "w32_sub.h"
#include "w32_kbd.h"
#include "w32_draw.h"

/*
 *	X^eBbN [N
 */
static BYTE *pKeyboard;					/* L[{[h Drawobt@ */
#if XM7_VER == 1
static BYTE uKeyboardType;				/* L[{[h n[hEFA^Cv */
#endif
static BYTE *pMMR;						/* MMR Drawobt@ */
static UINT nMMR;						/* MMR x */
#if XM7_VER == 1
static BYTE nSegDsp;					/* MMRZOg\Jnʒu */
static BYTE nSegDat;					/* ݂MMRZOg */
#endif
static BYTE *pDMAC;						/* DMAC Drawobt@ */
static BYTE *pPaletteReg;				/* pbgWX^ rbg}bvobt@ */
static HBITMAP hPaletteReg;				/* pbgWX^ rbg}bvnh */
static RECT rPaletteReg;				/* pbgWX^ ` */
#if XM7_VER >= 2
static UINT uPaletteReg;				/* pbgWX^ x */
#endif
static BYTE *pIRQ[MAXCPU];				/* IRQ Drawobt@ */
static BYTE *pKanji;					/* ROM rbg}bvobt@ */
static HBITMAP hKanji;					/* ROM rbg}bvnh */
static RECT rKanji;						/* ROM ` */
static UINT nKanji;						/* ROM x */
static BYTE cnKanji[6 * 3][16 * 3];		/* ROM [N */
static BYTE ctKanji[6 * 3][16 * 3];		/* ROM [N */
static BYTE uKanjiType;					/* ROM n[hEFA^Cv */
static WORD wAddrKanji;					/* ROM AhX[N */
#ifdef JSUB
static WORD wAddrJSubKanji;				/* ROM AhX[Ni{ʐMJ[hj */
#endif
static BYTE *pTapeLP;					/* JZbge[vv^ Drawobt@ */

/*
 *	pbgWX^EChEpFe[u
 */
static const DWORD Palet16Tc[] = {
	0x00000000,	0x000000ff,	0x00ff0000,	0x00ff00ff,
	0x0000ff00,	0x0000ffff,	0x00ffff00,	0x00ffffff,
#if XM7_VER == 1 && defined(L4CARD)
	0x00000000,	0x000000bb,	0x00bb0000,	0x00bb00bb,
	0x0000bb00,	0x0000bbbb,	0x00bbbb00,	0x00bbbbbb,
	0x00444444,	0x004444ff,	0x00ff4444,	0x00ff44ff,
	0x0044ff44,	0x0044ffff,	0x00ffff44,	0x00ffffff,
#endif
};

#define	KANJI_SET		(1)		/* $FD20-3 JISxx1 */
#define	KANJI_JCARD		(2)		/* $FD2C-F JIS832 */
#define	KANJI_JSUB		(4)		/* $FD28-B JIS78N1 */
#define	KANJI_JSUBINFO	(8)		/* {ʐMJ[h */
#define	KANJI_DICROM	(16)	/* $2E000- JIS78N1 */

/*-[ MMREChE ]---------------------------------------------------------*/

/*
 *	MMREChE
 *	MMR}bv ZbgAbv
 */
static void FASTCALL SetupMMRMap(BYTE *p, int y, int x)
{
	char string[128];
	char temp[16];
	int i, j;
	BYTE tmp;
	int segnum;
#if XM7_VER == 1
	int segtop;
#endif

	ASSERT(p);
	ASSERT(y > 0);
	ASSERT(x > 0);

	/* ^Cg */
	strncpy(string, "MMR Map:", sizeof(string));
	memcpy(&p[x * y], string, strlen(string));
	y++;

	if (nMMR) {
		segnum = 8;
#if XM7_VER == 1
		segtop = (mmr_seg & ~(segnum - 1));
#endif
	}
	else {
		segnum = 4;
#if XM7_VER == 1
		segtop = 0;
#endif
	}
#if XM7_VER == 1
	if (nSegDat != mmr_seg) {
		nSegDat = mmr_seg;
		nSegDsp = segtop;
	}
#endif

	/* eyƂ */
	for (i=0; i<segnum; i++) {
		/* 16i_v쐬 */
#if XM7_VER == 1
		_snprintf(string, sizeof(string), "Segment%2d-", nSegDsp + i);
#else
		_snprintf(string, sizeof(string), "Segment %d-", i);
#endif
		for (j=0; j<16; j++) {
#if XM7_VER == 1
			tmp = mmr_reg[(nSegDsp + i) * 0x10 + j];
#else
			tmp = mmr_reg[i * 0x10 + j];
#endif
#if XM7_VER >= 2
			if (!mmr_ext) {
				tmp &= 0x3f;
			}
#else
			if (!nMMR) {
				tmp &= 0x3f;
			}
#endif
			_snprintf(temp, sizeof(temp), "%02X", tmp);
			strncat(string, temp, sizeof(string) - strlen(string) - 1);
		}

		/* MMRgpŃZOgv΁A] */
#if XM7_VER == 1
		if (mmr_flag && (mmr_seg == (nSegDsp + i))) {
#else
		if (mmr_flag && (mmr_seg == i)) {
#endif
			for (j=10;;j++) {
				if (string[j] == '\0') {
					break;
				}
				string[j] |= 0x80;
			}
		}

		/* Zbg */
		memcpy(&p[x * y], string, strlen(string));
		y++;
	}
}

/*
 *	MMREChE
 *	}bv ZbgAbv
 */
static void FASTCALL SetupMMRMemory(BYTE *p, int y, int x)
{
	char string[128];
	char temp[16];
	BYTE mem[16];
	int i, j, k;
	int nsize;

	ASSERT(p);
	ASSERT(y > 0);
	ASSERT(x > 0);

	/* mem16oCgɁAԂݒ */
	for (i=0; i<16; i++) {
		mem[i] = (BYTE)(0x30 + i);
	}
	/* MMR */
	if (mmr_flag) {
		for (i=0; i<16; i++) {
#if XM7_VER >= 2
			if (mmr_ext) {
				mem[i] = mmr_reg[mmr_seg * 0x10 + i];
			}
			else {
				mem[i] = (BYTE)(mmr_reg[(mmr_seg & 3) * 0x10 + i] & 0x3f);
			}
#else
			if (nMMR) {
				mem[i] = mmr_reg[mmr_seg * 0x10 + i];
			}
			else {
				mem[i] = (BYTE)(mmr_reg[(mmr_seg & 3) * 0x10 + i] & 0x3f);
			}
#endif
		}
	}

	/* ^Cg */
	strncpy(string, "Physical Memory Map:", sizeof(string));
	memcpy(&p[x * y], string, strlen(string));
	y++;

	/* eyƂ */
#if XM7_VER >= 3
	if (nMMR == 1) {
		/* 1MB Memory */
		nsize = 16;
	}
	else {
		/* 256KB Memory */
		nsize = 4;
	}
#elif XM7_VER == 1
	if (nMMR) {
		/* 1MB Memory */
		nsize = 16;
	}
	else {
		/* 256KB Memory */
		nsize = 4;
	}
#else
	/* 256KB Memory */
	nsize = 4;
#endif
	for (i=0; i<nsize; i++) {
		/* 16i_v쐬 */
		switch (i) {
#if XM7_VER >= 2
			case 0:
#if XM7_VER >= 3
				strncpy(string, "Extend 0  ", sizeof(string));
#else
				strncpy(string, "Extend    ", sizeof(string));
#endif
				break;
			case 1:
				strncpy(string, "Sub CPU   ", sizeof(string));
				break;
			case 2:
				switch (model_type) {
					case MODEL_FM77AV20EX:
					case MODEL_FM77AV20:
					case MODEL_FM77AV:
						if (jcard_mode != JCARD_ON) {
							strncpy(string, "Extend 2  ", sizeof(string));
						}
						else {
							strncpy(string, "Japanese  ", sizeof(string));
						}
						break;
					default:
						strncpy(string, "Japanese  ", sizeof(string));
						break;
				}
				break;
#endif
			case 3:
				strncpy(string, "Standard  ", sizeof(string));
				break;
			default:
				_snprintf(string, sizeof(string),  "Extend %X  ", i);
				break;
		}
		for (j=0; j<16; j++) {
			_snprintf(temp, sizeof(temp), "%02X", i * 0x10 + j);
			/* ] */
			for (k=0; k<16; k++) {
				if (mem[k] == (i * 0x10 + j)) {
					temp[0] |= 0x80;
					temp[1] |= 0x80;
				}
			}
			/* ǉ */
			strncat(string, temp, sizeof(string) - strlen(string) - 1);
		}

		/* Zbg */
		memcpy(&p[x * y], string, strlen(string));
		y++;
	}
}

/*
 *	MMREChE
 *	ZbgAbv
 */
static void FASTCALL SetupMMR(BYTE *p, int x, int y)
{
	char string[128];

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* MMR */
	if (mmr_flag) {
		_snprintf(string, sizeof(string),  "MMR  Enable(Seg:%1d)", mmr_seg);
	}
	else {
		strncpy(string, "MMR        Disable", sizeof(string));
	}
	memcpy(&p[x * 0 + 0], string, strlen(string));

	/* TWR */
	if (twr_flag) {
		_snprintf(string, sizeof(string),  "TWR  Enable($%02X00)",
			(twr_reg + 0x7c) & 0xff);
	}
	else {
		strncpy(string, "TWR        Disable", sizeof(string));
	}
	memcpy(&p[x * 1 + 0], string, strlen(string));

#if XM7_VER >= 2
	/* ROM */
	if (!dicrom_en) {
		strncpy(string, "DIC        Disable", sizeof(string));
	}
	else {
#if XM7_VER >= 3
		if (extrom_sel) {
			_snprintf(string, sizeof(string),  "DIC  ExtROM(Bank %2d)",
				dicrom_bank);
		}
		else {
			_snprintf(string, sizeof(string),  "DIC  Enable(Bank %2d)",
				dicrom_bank);
		}
#else
		_snprintf(string, sizeof(string),  "DIC  Enable(Bank %2d)", dicrom_bank);
#endif
	}
	memcpy(&p[x * 2 + 0], string, strlen(string));

	/* RAM */
	if (!dicram_en) {
		strncpy(string, "DICRAM     Disable", sizeof(string));
	}
	else {
		strncpy(string, "DICRAM      Enable", sizeof(string));
	}
	memcpy(&p[x * 3 + 0], string, strlen(string));
#endif

#if XM7_VER == 1
	/* gRAMoNZNg */
	if ((model_type != MODEL_FM77) || (mmr_extram == EXRAM_OFF)) {
		strncpy(string, "Extend RAM Disable", sizeof(string));
	}
	else if (mmr_extram == EXRAM_64KB) {
		_snprintf(string, sizeof(string), "Extend RAM Bank %2d", bsr_reg);
	}
	else if (mmr_extram == EXRAM_192KB) {
		strncpy(string, "Extend RAM   192KB", sizeof(string));
	}
	else {
		strncpy(string, "Extend RAM   960KB", sizeof(string));
	}
	memcpy(&p[x * 2 + 0], string, strlen(string));
#endif

#if XM7_VER >= 2
	/* u[gRAM */
	strncpy(string, "Boot RAM         R/", sizeof(string));
	if (bootram_rw) {
		strncat(string, "W", sizeof(string) - strlen(string) - 1);
	}
	else {
		strncat(string, "O", sizeof(string) - strlen(string) - 1);
	}
	memcpy(&p[x * 0 + 22], string, strlen(string));

	/* CjVG[^ROM */
	strncpy(string, "Initiate ROM ", sizeof(string));
	if (initrom_en) {
		strncat(string, " Enable", sizeof(string) - strlen(string) - 1);
	}
	else {
		strncat(string, "Disable", sizeof(string) - strlen(string) - 1);
	}
	memcpy(&p[x * 1 + 22], string, strlen(string));
#else
	/* F-BASIC ROM */
	strncpy(string, "F-BASIC ROM  ", sizeof(string));
	if (!basicrom_en) {
		strncat(string, "Disable", sizeof(string) - strlen(string) - 1);
	}
	else {
		strncat(string, " Enable", sizeof(string) - strlen(string) - 1);
	}
	memcpy(&p[x * 0 + 22], string, strlen(string));

	/* u[gROM/RAM */
	strncpy(string, "Boot ROM     ", sizeof(string));
	if (bootram_rw) {
		strncat(string, "Disable", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, " Enable", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[x * 1 + 22], string, strlen(string));

	/* gRAM[h */
	strncpy(string, "Extend MMR   ", sizeof(string));
	if (mmr_extram == EXRAM_768KB) {
		strncat(string, "  768KB", sizeof(string)-strlen(string)-1);
	}
	else if (mmr_extram == EXRAM_256KB) {
		strncat(string, "  256KB", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "Disable", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[x * 2 + 22], string, strlen(string));
#endif

#if XM7_VER >= 2
	/* gMMR */
	strncpy(string, "Extend MMR   ", sizeof(string));
	if (mmr_ext) {
		strncat(string, " Enable", sizeof(string) - strlen(string) - 1);
	}
	else {
		strncat(string, "Disable", sizeof(string) - strlen(string) - 1);
	}
	memcpy(&p[x * 2 + 22], string, strlen(string));

	/* MMR */
	strncpy(string, "MMR Fastmode ", sizeof(string));
	if (mmr_fastmode) {
		strncat(string, " Enable", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "Disable", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[x * 3 + 22], string, strlen(string));

	/* MMR}bv */
	SetupMMRMap(p, 5, x);

	/* }bv */
	if (nMMR) {
		SetupMMRMemory(p, 15, x);
	}
	else {
		SetupMMRMemory(p, 11, x);
	}
#else
	/* MMR}bv */
	SetupMMRMap(p, 4, x);

	/* }bv */
	if (nMMR) {
		SetupMMRMemory(p, 14, x);
	}
	else {
		SetupMMRMemory(p, 10, x);
	}
#endif
}

/*
 *	MMREChE
 *	`
 */
static void FASTCALL DrawMMR(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pMMR) {
		return;
	}
	SetupMMR(pMMR, x, y);

	/* ` */
	DrawWindowText2(hDC, pMMR, x, y);
}

/*
 *	MMREChE
 *	tbV
 */
void FASTCALL RefreshMMR(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_MMR] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_MMR];
	hDC = GetDC(hWnd);
	DrawMMR(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	MMREChE
 *	TCYύX
 */
void FASTCALL ReSizeMMR(void)
{
	HWND hWnd;
	UINT uMode;
	RECT wrect;
	RECT crect;
	int cx;
	int cy;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_MMR] == NULL) {
		return;
	}

	/* ŏȂAȂ */
	hWnd = hSubWnd[SWND_MMR];
	if (IsIconic(hWnd)) {
		return;
	}

	/* WIgύX`FbN */
	uMode = 0;
	cx = 42;
#if XM7_VER >= 2
	switch (model_type) {
		case MODEL_FM77AV20EX:
			cy = 20;
			uMode = 2;
			break;

#if XM7_VER >= 3
		case MODEL_FM77AV40:
		case MODEL_FM77AV40EX:
			cy = 32;
			uMode = 1;
			break;
#endif

		default:
			cy = 16;
			break;
	}
#else
	if ((model_type == MODEL_FM77) &&
		(mmr_extram == EXRAM_960KB)) {
		cy = 31;
		uMode = 1;
	}
	else {
		cy = 15;
	}
#endif

	if (uMode == nMMR) {
		return;
	}

	/* bN */
	LockVM();

	/* [h`FW */
	nMMR = uMode;

	/* obt@ */
	free(pMMR);
	pMMR = malloc(2 * cx * cy);
	memset(pMMR, 0xff, 2 * cx * cy);
	ASSERT(pMMR);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += cx * lCharWidth;
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += cy * lCharHeight;
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* AbN */
	UnlockVM();
}

/*
 *	MMREChE
 *	ĕ`
 */
static void FASTCALL PaintMMR(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pMMR;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawMMR(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

#if XM7_VER == 1
/*
 *	MMREChE
 *	MSRy[WRg[
 */
static void FASTCALL PageMMR(BYTE vkey)
{
	/* MMRĝ */
	if (!nMMR) {
		return;
	}

	switch (vkey) {
		/* [] */
		case VK_UP:
			if (nSegDsp > 0x00) {
				nSegDsp -= 1;
			}
			break;

		/* [] */
		case VK_DOWN:
			if (nSegDsp < 0x38) {
				nSegDsp += 1;
			}
			break;

		/* [][PageUp] */
		case VK_LEFT:
		case VK_PRIOR:
			if (nSegDsp > 0x08) {
				nSegDsp -= 8;
			}
			else {
				nSegDsp = 0;
			}
			break;

		/* [][PageDown] */
		case VK_RIGHT:
		case VK_NEXT:
			if (nSegDsp < 0x30) {
				nSegDsp += 8;
			}
			else {
				nSegDsp = 0x38;
			}
			break;

		/* [Home] */
		case VK_HOME:
			nSegDsp = 0;
			break;

		/* [End] */
		case VK_END:
			nSegDsp = 0x38;
			break;

		default:
			break;
	}
}
#endif	/* XM7_VER == 1 */

/*
 *	MMREChE
 *	EChEvV[W
 */
static LRESULT CALLBACK MMRProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	int ntmp;

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintMMR(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pMMR, NULL);

			UnlockVM();
			break;

#if XM7_VER == 1
		case WM_KEYDOWN:
			PageMMR(LOBYTE(wParam));
			break;

		/* NbN */
		case WM_LBUTTONDOWN:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (bMouseCapture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif
			PageMMR(VK_LEFT);
			break;

		/* ENbN */
		case WM_RBUTTONDOWN:
#ifdef MOUSE
			/* }EXG~[Vgp͖ */
			if (bMouseCapture && !stopreq_flag && run_flag) {
				return FALSE;
			}
#endif
			PageMMR(VK_RIGHT);
			break;

		/* }EXzC[ */
#ifdef WM_MOUSEWHEEL
		case WM_MOUSEWHEEL:
#ifdef MOUSE
			if ((uMidBtnMode == MOSCAP_WHEELWM)) {
				break;
			}
#endif
			ntmp = GET_WHEEL_DELTA_WPARAM(wParam);
			if (ntmp > 0) {
				PageMMR(VK_UP);
			}
			else if (ntmp < 0) {
				PageMMR(VK_DOWN);
			}
			break;
#endif
#endif

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	MMREChE
 *	EChE쐬
 */
HWND FASTCALL CreateMMR(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_MMR";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 42;
#if XM7_VER >= 2
	rect.bottom = lCharHeight * 16;
#else
	rect.bottom = lCharHeight * 15;
#endif

	/* EChE^CgAobt@m */
	GetMsg(hResInstance, IDS_SWND_MMR,
				szWndName, sizeof(szWndName));
	pMMR = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));
	nMMR = 0;
#if XM7_VER == 1
	nSegDsp = 0;
	nSegDat = mmr_seg;
#endif

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = MMRProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ L[{[hEChE ]--------------------------------------------------*/

/*
 *	L[{[hEChE
 *	ZbgAbv(tO)
 */
static void FASTCALL SetupKeyboardFlag(BYTE *p, char *title, BOOL flag)
{
	int i;

#if XM7_VER >= 2
	/*  */
	memset(p, 0x20, 17);

	/* Rs[ */
	for (i=0; i<17; i++) {
		if (title[i] == '\0') {
			break;
		}
		p[i] = title[i];
	}

	/* tOɉĐݒ */
	if (flag) {
		memcpy(&p[15], "On", 2);
	}
	else {
		memcpy(&p[14], "Off", 3);
	}
#else
	/*  */
	memset(p, 0x20, 18);

	/* Rs[ */
	for (i=0; i<18; i++) {
		if (title[i] == '\0') {
			break;
		}
		p[i] = title[i];
	}

	/* tOɉĐݒ */
	if (flag) {
		memcpy(&p[16], "On", 2);
	}
	else {
		memcpy(&p[15], "Off", 3);
	}
#endif
}


/*
 *	L[{[hEChE
 *	ZbgAbv
 */
static void FASTCALL SetupKeyboard(BYTE *p, int x, int y)
{
	char string[128];
#if XM7_VER == 1
	int i, j;
#endif

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

#if XM7_VER >= 2
	/* R[h */
	_snprintf(string, sizeof(string),  "Logical Code     %04X", key_fm7);
	memcpy(&p[x * 0 + 0], string, strlen(string));
	_snprintf(string, sizeof(string),  "Scan Code          %02X", key_scan);
	memcpy(&p[x * 1 + 0], string, strlen(string));

	/* R[htH[}bg */
	strncpy(string, "Code Format ", sizeof(string));
	switch (key_format) {
		case KEY_FORMAT_9BIT:
			strncat(string, "     FM-7", sizeof(string) - strlen(string) - 1);
			break;
		case KEY_FORMAT_FM16B:
			strncat(string, "FM-16Beta", sizeof(string) - strlen(string) - 1);
			break;
		case KEY_FORMAT_SCAN:
			strncat(string, "     Scan", sizeof(string) - strlen(string) - 1);
			break;
		default:
			ASSERT(FALSE);
	}
	memcpy(&p[x * 2 + 0], string, strlen(string));

	/* L[s[g */
	strncpy(string, "Key Repeat", sizeof(string));
	memcpy(&p[x * 4 + 0], string, strlen(string));
	if (key_repeat_flag) {
		strncpy(string, " Enable", sizeof(string));
	}
	else {
		strncpy(string, "Disable", sizeof(string));
	}
	memcpy(&p[x * 4 + 14], string, strlen(string));
	_snprintf(string, sizeof(string),  "Repeat Time 1  %4dms",
		key_repeat_time1 / 1000);
	memcpy(&p[x * 5 + 0], string, strlen(string));
	_snprintf(string, sizeof(string),  "Repeat Time 2  %4dms",
		key_repeat_time2 / 1000);
	memcpy(&p[x * 6 + 0], string, strlen(string));

	/* tO */
	SetupKeyboardFlag(&p[x * 0 + 24], "CAP  LED" , caps_flag);
	SetupKeyboardFlag(&p[x * 1 + 24], "KANA LED", kana_flag);
	SetupKeyboardFlag(&p[x * 2 + 24], "INS  LED", ins_flag);
	SetupKeyboardFlag(&p[x * 3 + 24], "SHIFT", shift_flag);
	SetupKeyboardFlag(&p[x * 4 + 24], "CTRL", ctrl_flag);
	SetupKeyboardFlag(&p[x * 5 + 24], "GRAPH", graph_flag);
	SetupKeyboardFlag(&p[x * 6 + 24], "BREAK", break_flag);

	/* v */
	_snprintf(string, sizeof(string),  "RTC Date     %02d/%02d/%02d",
		rtc_year, rtc_month, rtc_day);
	memcpy(&p[x * 8 + 0], string, strlen(string));
	_snprintf(string, sizeof(string),  "RTC Time     %02d:%02d:%02d",
		rtc_hour, rtc_minute, rtc_second);
	memcpy(&p[x * 9 + 0], string, strlen(string));
	_snprintf(string, sizeof(string),  "RTC 12h/24h");
	memcpy(&p[x * 10 + 0], string, strlen(string));
	if (rtc_24h) {
		strncpy(string, "24h", sizeof(string));
	}
	else {
		strncpy(string, "12h", sizeof(string));
	}
	memcpy(&p[x * 10 + 18], string, strlen(string));
	strncpy(string, "RTC AM/PM", sizeof(string));
	memcpy(&p[x * 11 + 0], string, strlen(string));
	if (rtc_pm) {
		strncpy(string, "PM", 128);
	}
	else {
		strncpy(string, "AM", 128);
	}
	memcpy(&p[x * 11 + 19], string, strlen(string));
	strncpy(string, "RTC Leap Mode", sizeof(string));
	memcpy(&p[x * 12 + 0], string, strlen(string));
	_snprintf(string, sizeof(string),  "%1d", rtc_leap);
	memcpy(&p[x * 12 + 20], string, strlen(string));

	/* X[p[C|[Y */
	strncpy(string, "Super Impose   ", sizeof(string));
	switch (simpose_mode) {
		case 0:
		case 3:	/* fBW^CY[hPC\ */
			strncat(string, "PC", sizeof(string) - strlen(string) - 1);
			break;
		case 1:
			strncat(string, "On", sizeof(string) - strlen(string) - 1);
			break;
		case 2:
			strncat(string, "TV", sizeof(string) - strlen(string) - 1);
			break;
	}
	memcpy(&p[x * 8 + 24], string, strlen(string));
	strncpy(string, "Brightness   ", sizeof(string));
	if (simpose_half) {
		strncat(string, "Half", sizeof(string) - strlen(string) - 1);
	}
	else {
		strncat(string, "Full", sizeof(string) - strlen(string) - 1);
	}
	memcpy(&p[x * 9 + 24], string, strlen(string));

	/* fBW^CY */
	strncpy(string, "Digitize  ", sizeof(string));
	if ((digitize_enable) && (simpose_mode == 3)) {
		strncat(string, " Enable", sizeof(string) - strlen(string) - 1);
	}
	else {
		strncat(string, "Disable", sizeof(string) - strlen(string) - 1);
	}
	memcpy(&p[x * 11 + 24], string, strlen(string));
	strncpy(string, "Key Wait    ", sizeof(string));
	if (digitize_keywait) {
		strncat(string, " Wait", sizeof(string) - strlen(string) - 1);
	}
	else {
		strncat(string, "Ready", sizeof(string) - strlen(string) - 1);
	}
	memcpy(&p[x * 12 + 24], string, strlen(string));
#else
	/* R[h */
	_snprintf(string, sizeof(string),  "Logical Code  %04X", key_fm7);
	memcpy(&p[x * 0 + 0], string, strlen(string));
	_snprintf(string, sizeof(string),  "Scan Code       %02X", key_scan);
	memcpy(&p[x * 1 + 0], string, strlen(string));

	/* L[s[g */
	strncpy(string, "Key Repeat", sizeof(string));
	memcpy(&p[x * 2 + 0], string, strlen(string));
	if (key_repeat_flag) {
		strncpy(string, " Enable", sizeof(string));
	}
	else {
		strncpy(string, "Disable", sizeof(string));
	}
	memcpy(&p[x * 2 + 11], string, strlen(string));

	/* L[ */
	for (i = 0; i < 4; i ++) {
		_snprintf(string, sizeof(string),  "Key %d", i);
		memcpy(&p[x * (3 + i)], string, strlen(string));
		for (j = 0; j < 4; j ++) {
			_snprintf(string, sizeof(string),  "%02X", key_push_data[i * 4 + j]);
			memcpy(&p[x * (3 + i) + j * 2 + 10], string, strlen(string));
		}
	}

	/* tO */
	SetupKeyboardFlag(&p[x * 0 + 21], "CAP  LED" , caps_flag);
	SetupKeyboardFlag(&p[x * 1 + 21], "KANA LED", kana_flag);
	SetupKeyboardFlag(&p[x * 2 + 21], "INS  LED", ins_flag);
	SetupKeyboardFlag(&p[x * 3 + 21], "SHIFT", shift_flag);
	SetupKeyboardFlag(&p[x * 4 + 21], "CTRL", ctrl_flag);
	SetupKeyboardFlag(&p[x * 5 + 21], "GRAPH", graph_flag);
	SetupKeyboardFlag(&p[x * 6 + 21], "BREAK", break_flag);
#endif
}

/*
 *	L[{[hEChE
 *	`
 */
static void FASTCALL DrawKeyboard(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pKeyboard) {
		return;
	}
	SetupKeyboard(pKeyboard, x, y);

	/* ` */
	DrawWindowText(hDC, pKeyboard, x, y);
}

/*
 *	L[{[hEChE
 *	tbV
 */
void FASTCALL RefreshKeyboard(void)
{
	HWND hWnd;
	HDC hDC;
#if XM7_VER == 1
	RECT wrect;
	RECT crect;
	char szWndName[128];
	int cx;
	int cy;
	BYTE *p;
#endif

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_KEYBOARD] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_KEYBOARD];
	hDC = GetDC(hWnd);
	DrawKeyboard(hWnd, hDC);
	ReleaseDC(hWnd, hDC);

#if XM7_VER == 1
	if (uKeyboardType == model_type) {
		return;
	}

	/* bN */
	LockVM();

	/* [h`FW */
	uKeyboardType = model_type;

	/* EChE^CgύX */
	if (uKeyboardType == MODEL_FM8) {
		GetMsg(hResInstance, IDS_SWND_KEYBOARD8,
					szWndName, sizeof(szWndName));
	}
	else {
		GetMsg(hResInstance, IDS_SWND_KEYBOARD,
					szWndName, sizeof(szWndName));
	}
	SetWindowText(hWnd, szWndName);

	/* AbN */
	UnlockVM();
#endif
}

/*
 *	L[{[hEChE
 *	ĕ`
 */
static void FASTCALL PaintKeyboard(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pKeyboard;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawKeyboard(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	L[{[hEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK KeyboardProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintKeyboard(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pKeyboard, NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	L[{[hEChE
 *	EChE쐬
 */
HWND FASTCALL CreateKeyboard(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_Keyboard";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
#if XM7_VER >= 2
	rect.right = lCharWidth * 41;
	rect.bottom = lCharHeight * 13;
#else
	rect.right = lCharWidth * 39;
	rect.bottom = lCharHeight * 7;
#endif

	/* EChE^CgAobt@m */
#if XM7_VER == 1
	uKeyboardType = model_type;
	if (uKeyboardType == MODEL_FM8) {
		GetMsg(hResInstance, IDS_SWND_KEYBOARD8,
					szWndName, sizeof(szWndName));
	}
	else {
		GetMsg(hResInstance, IDS_SWND_KEYBOARD,
					szWndName, sizeof(szWndName));
	}
#else
	GetMsg(hResInstance, IDS_SWND_KEYBOARD,
				szWndName, sizeof(szWndName));
#endif
	pKeyboard = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = KeyboardProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ DMACEChE ]--------------------------------------------------------*/

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
/*
 *	DMACEChE
 *	ZbgAbv
 */
static void FASTCALL SetupDMAC(BYTE *p, int x, int y)
{
	char string[128];

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* CHCR - BUSY */
	strncpy(string, "Status", sizeof(string));
	memcpy(&p[x * 0 + 0], string, strlen(string));

	if (dma_chcr[dma_ch] & 0x40) {
		strncpy(string, " Busy", sizeof(string));
	}
	else {
		strncpy(string, "Ready", sizeof(string));
	}
	memcpy(&p[x * 0 + 25], string, strlen(string));

	/* CHCR - MODE */
	strncpy(string, "Request Mode", sizeof(string));
	memcpy(&p[x * 1 + 0], string, strlen(string));

	if (dma_chcr[dma_ch] & 0x02) {
		strncpy(string, "      Burst", sizeof(string));
	}
	else {
		strncpy(string, "Cycle Steal", sizeof(string));
	}
	memcpy(&p[x * 1 + 19], string, strlen(string));

	/* CHCR - R/W */
	strncpy(string, "Direction", sizeof(string));
	memcpy(&p[x * 2 + 0], string, strlen(string));

	if (dma_chcr[dma_ch] & 0x01) {
		strncpy(string, "Memory->Device", sizeof(string));
	}
	else {
		strncpy(string, "Device->Memory", sizeof(string));
	}
	memcpy(&p[x * 2 + 16], string, strlen(string));

	/* CHCR - DIRECTION */
	strncpy(string, "Address Inc/Dec", sizeof(string));
	memcpy(&p[x * 3 + 0], string, strlen(string));

	if (dma_chcr[dma_ch] & 0x08) {
		strncpy(string, "Decrement", sizeof(string));
	}
	else {
		strncpy(string, "Increment", sizeof(string));
	}
	memcpy(&p[x * 3 + 21], string, strlen(string));

	/* ADR */
	_snprintf(string, sizeof(string),  "Address                  $%04X",
		dma_adr[dma_ch]);
	memcpy(&p[x * 4 + 0], string, strlen(string));

	/* BCR */
	_snprintf(string, sizeof(string),  "Byte Count               $%04X",
		dma_bcr[dma_ch]);
	memcpy(&p[x * 5 + 0], string, strlen(string));

	/* CHCR - DEND */
	strncpy(string, "DEND", sizeof(string));
	memcpy(&p[x * 6 + 0], string, strlen(string));

	if (dma_chcr[dma_ch] & 0x80) {
		strncpy(string, " Enable", sizeof(string));
	}
	else {
		strncpy(string, "Disable", sizeof(string));
	}
	memcpy(&p[x * 6 + 23], string, strlen(string));

	/* PCR - TxRQ */
	strncpy(string, "TxRQ", sizeof(string));
	memcpy(&p[x * 7 + 0], string, strlen(string));

	if (dma_pcr & dma_txrq) {
		strncpy(string, " Enable", sizeof(string));
	}
	else {
		strncpy(string, "Disable", sizeof(string));
	}
	memcpy(&p[x * 7 + 23], string, strlen(string));

	/* ICR - IRQ ENABLE */
	strncpy(string, "IRQ", sizeof(string));
	memcpy(&p[x * 8 + 0], string, strlen(string));

	if (dma_icr & dma_txrq) {
		if (dma_icr & 0x80) {
			strncpy(string, " Enable", sizeof(string));
		}
		else {
			strncpy(string, "Disable", sizeof(string));
		}
	}
	else {
		strncpy(string, "   Mask", sizeof(string));
	}
	memcpy(&p[x * 8 + 23], string, strlen(string));
}

/*
 *	DMACEChE
 *	`
 */
static void FASTCALL DrawDMAC(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pDMAC) {
		return;
	}
	SetupDMAC(pDMAC, x, y);

	/* ` */
	DrawWindowText(hDC, pDMAC, x, y);
}

/*
 *	DMACEChE
 *	tbV
 */
void FASTCALL RefreshDMAC(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_DMAC] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_DMAC];
	hDC = GetDC(hWnd);
	DrawDMAC(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	DMACEChE
 *	ĕ`
 */
static void FASTCALL PaintDMAC(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pDMAC;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawDMAC(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	DMACEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK DMACProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintDMAC(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pDMAC, NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	DMACEChE
 *	EChE쐬
 */
HWND FASTCALL CreateDMAC(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_DMAC";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 30;
	rect.bottom = lCharHeight * 9;

	/* EChE^CgAobt@m */
	GetMsg(hResInstance, IDS_SWND_DMAC,
				szWndName, sizeof(szWndName));
	pDMAC = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = DMACProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

#endif	/* XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC)) */

/*-[ pbgWX^EChE ]--------------------------------------------*/

/*
 *	pbgWX^EChE
 *	hbg`
 */
static void FASTCALL PSetPaletteReg(int x, int y, int color)
{
	ASSERT((x >= 0) && (x < rPaletteReg.right));
	ASSERT((y >= 0) && (y < rPaletteReg.bottom));

	/*  */
	if (!pPaletteReg) {
		return;
	}
	*(DWORD *)(&pPaletteReg[((y * rPaletteReg.right) + x) * 4]) = color;
}

/*
 *	pbgWX^EChE
 *	{bNXtB`
 */
static void FASTCALL BfPaletteReg(int x, int y, int cx, int cy, int color)
{
	int i;
	int j;

	for (i=0; i<cy; i++) {
		for (j=0; j<cx; j++) {
			PSetPaletteReg(x + j, y, color);
		}
		y++;
	}
}

/*
 *	pbgWX^EChE
 *	LN^`
 */
static void FASTCALL ChrPaletteReg(char c, int x, int y, int fc, int bc)
{
	int i;
	int j;
	BYTE *p;
	BYTE dat;

	/* TuROM(C)̃tHgAhX𓾂 */
	p = &subrom_c[c * 8];

	/* y[v */
	for (i=0; i<8; i++) {
		/* f[^擾 */
		dat = *p;
		p++;

		/* x[v */
		for (j=0; j<8; j++) {
			if (dat & 0x80) {
				PSetPaletteReg(x, y, fc);
			}
			else {
				PSetPaletteReg(x, y, bc);
			}
			dat <<= 1;
			x++;
		}

		/* y */
		x -= 8;
		y++;
	}
}

/*
 *	pbgWX^EChE
 *	LN^`
 */
static void FASTCALL StrPaletteReg(char *str, int x, int y, int fc, int bc)
{
	/* Ȃ蔲B */
	while (*str) {
		ChrPaletteReg(*str++, x, y, fc, bc);
		x += 8;
	}

}

/*
 *	pbgWX^EChE
 *	ZbgAbv
 */
static BOOL FASTCALL SetupPaletteReg(void)
{
	int color, col;
	BYTE vpage;
#if XM7_VER >= 2
	int x, y;
	int r, g, b;
	int amask;
#endif

	/* ĕ`ԃ`FbN */
	if (!bPaletteRefresh && !bMultiPageRefresh) {
		/* ĕ`悪KvłȂȂ牽ȂŖ߂ */
		return FALSE;
	}

	/* }`y[W */
	if (bMultiPageRefresh) {
		bMultiPageRefresh = FALSE;

		StrPaletteReg("ACTIVE PAGE", 0, 0 * 8, 0xFFFFFF, 0x000000);
		StrPaletteReg("---", 192 - 8 * 3, 0 * 8, 0x777777, 0x000000);
		if (!(multi_page & 0x04)) {
			ChrPaletteReg('G', 192 - 8 * 3, 0 * 8, 0x00FF00, 0x000000);
		}
		if (!(multi_page & 0x02)) {
			ChrPaletteReg('R', 192 - 8 * 2, 0 * 8, 0xFF0000, 0x000000);
		}
		if (!(multi_page & 0x01)) {
			ChrPaletteReg('B', 192 - 8 * 1, 0 * 8, 0x0000FF, 0x000000);
		}
		StrPaletteReg("DISPLAY PAGE", 0, 1 * 8, 0xFFFFFF, 0x000000);
		StrPaletteReg("---", 192 - 8 * 3, 1 * 8, 0x777777, 0x000000);
		if (!(multi_page & 0x40)) {
			ChrPaletteReg('G', 192 - 8 * 3, 1 * 8, 0x00FF00, 0x000000);
		}
		if (!(multi_page & 0x20)) {
			ChrPaletteReg('R', 192 - 8 * 2, 1 * 8, 0xFF0000, 0x000000);
		}
		if (!(multi_page & 0x10)) {
			ChrPaletteReg('B', 192 - 8 * 1, 1 * 8, 0x0000FF, 0x000000);
		}
	}

	/* TTLpbg */
	if (bPaletteRefresh) {
		bPaletteRefresh = FALSE;

		BfPaletteReg(0, 3 * 8 + 6, 192, 1, 0xFFFFFF);
		StrPaletteReg("TTL PALETTE", 0, 3 * 8 - 2, 0xFFFFFF, 0x000000);
		for (color = 0; color < 8; color ++) {
			/* }`y[WA\v[𓾂 */
			vpage = (BYTE)((~(multi_page >> 4)) & 0x07);
#if XM7_VER == 1 && defined(L4CARD)
			if (enable_400line && enable_400linecard) {
				col = Palet16Tc[ttl_palet[(color & vpage)] + 8];
			}
			else {
				col = Palet16Tc[ttl_palet[(color & vpage)] & 7];
			}
#else
			col = Palet16Tc[ttl_palet[(color & vpage)] & 7];
#endif
			BfPaletteReg(color * 24, 4 * 8, 24, 24, col);
			ChrPaletteReg((BYTE)(color + 0x30), color * 24 + 16, 6 * 8,
				0xFFFFFF, 0x000000);
		}

#if XM7_VER >= 2
		if (!uPaletteReg) {
			return TRUE;
		}
#endif

#if XM7_VER >= 2
		/* AiOpbg */
		BfPaletteReg(0, 8 * 8 + 6, 192, 1, 0xFFFFFF);
		StrPaletteReg("ANALOG PALETTE", 0, 8 * 8 - 2, 0xFFFFFF, 0x000000);
		for (r = 0; r < 16; r++) {
			for (g = 0; g < 16; g++) {
				for (b = 0; b < 16; b++) {
					color = 0;

					/* AiO}XN쐬 */
					amask = 0;
					if (!(multi_page & 0x10)) {
						amask |= 0x000f;
					}
					if (!(multi_page & 0x20)) {
						amask |= 0x00f0;
					}
					if (!(multi_page & 0x40)) {
						amask |= 0x0f00;
					}

					/* pbgf[^擾 */
					col = ((g * 0x100) + (r * 0x10) + b) & amask;

					/* R */
					color |= (WORD)apalet_r[col] * 0x11;
					color <<= 8;

					/* G */
					color |= (WORD)apalet_g[col] * 0x11;
					color <<= 8;

					/* B */
					color |= (WORD)apalet_b[col] * 0x11;

					/* ` */
					x = (((g & 3) << 4) + b);
					y = (((g & 12) << 2) + r);
					BfPaletteReg(x * 3, y * 3 + (9 * 8), 3, 3, color);
				}
			}
		}
#endif
	}

	return TRUE;
}

/*
 *	pbgWX^EChE
 *	`
 */
static void FASTCALL DrawPaletteReg(HDC hDC, BOOL flag)
{
	HDC hMemDC;
	HBITMAP hBitmap;

	ASSERT(hDC);

	/* rbg}bvnhE|C^ȂAȂ */
	if (!hPaletteReg || !pPaletteReg) {
		return;
	}

	/* ZbgAbv``FbN */
	if (!SetupPaletteReg() && !flag) {
		return;
	}

	/* rbg}bvnhE|C^Lx`FbN */
	if (!hPaletteReg || !pPaletteReg) {
		return;
	}

	/* DC쐬Arbg}bvZNg */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hBitmap = (HBITMAP)SelectObject(hMemDC, hPaletteReg);

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hBitmap) {
		/* BitBlt */
		BitBlt(hDC, 0, 0,
				rPaletteReg.right, rPaletteReg.bottom, hMemDC,
				0, 0, SRCCOPY);

		/* IuWFNgăZNg */
		SelectObject(hMemDC, hBitmap);
	}

	/* DC폜 */
	DeleteDC(hMemDC);
}

/*
 *	pbgWX^EChE
 *	tbV
 */
void FASTCALL RefreshPaletteReg(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_PALETTE] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_PALETTE];
	hDC = GetDC(hWnd);
	DrawPaletteReg(hDC, TRUE);
	ReleaseDC(hWnd, hDC);
}

/*
 *	pbgWX^EChE
 *	ĕ`
 */
static void FASTCALL PaintPaletteReg(HWND hWnd, BOOL flag)
{
	HDC hDC;
	PAINTSTRUCT ps;

	ASSERT(hWnd);

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawPaletteReg(hDC, flag);
	EndPaint(hWnd, &ps);
}

/*
 *	pbgWX^EChE
 *	EChEvV[W
 */
static LRESULT CALLBACK PaletteRegProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintPaletteReg(hWnd, TRUE);
			UnlockVM();
			return 0;

		/* wi` */
		case WM_ERASEBKGND:
			return TRUE;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, NULL, NULL);

			/* rbg}bv */
			if (hPaletteReg) {
				DeleteObject(hPaletteReg);
				hPaletteReg = NULL;
				pPaletteReg = NULL;
			}

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	pbgWX^EChE
 *	
 */
static void FASTCALL InitPaletteReg(HWND hWnd)
{
	BITMAPINFOHEADER *pbmi;
	HDC hDC;

	/* S̃[N */
	pPaletteReg = NULL;
	hPaletteReg = NULL;

	/* rbg}bvwb_ */
	pbmi = (BITMAPINFOHEADER*)malloc(sizeof(BITMAPINFOHEADER));
	if (pbmi) {
		memset(pbmi, 0, sizeof(BITMAPINFOHEADER));
		pbmi->biSize = sizeof(BITMAPINFOHEADER);
		pbmi->biWidth = rPaletteReg.right;
		pbmi->biHeight = -rPaletteReg.bottom;
		pbmi->biPlanes = 1;
		pbmi->biBitCount = 32;
		pbmi->biCompression = BI_RGB;

		/* DC擾ADIBZNV쐬 */
		hDC = GetDC(hWnd);
		hPaletteReg = CreateDIBSection(hDC, (BITMAPINFO*)pbmi, DIB_RGB_COLORS,
								(void**)&pPaletteReg, NULL, 0);
		ReleaseDC(hWnd, hDC);
		free(pbmi);
	}
}

#if XM7_VER >= 2
/*
 *	pbgWX^EChE
 *	TCYύX
 */
void FASTCALL ReSizePaletteReg(void)
{
	HWND hWnd;
	UINT uMode;
	RECT rect;
	RECT crect, wrect;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_PALETTE] == NULL) {
		return;
	}

	/* WIgύX`FbN */
	hWnd = hSubWnd[SWND_PALETTE];
	uMode = 0;
	rect = rPaletteReg;
	rect.bottom = (4 + 3) * 8;
	if (model_type != MODEL_FM7) {
		uMode = 1;
		rect.bottom = 192 + (4 + 3 + 2) * 8;
	}

	/* ŏȂAȂ */
	if (IsIconic(hWnd)) {
		/* ACR`FbN */
		uMode |= 0x80000000;
	}
	if (uMode == uPaletteReg) {
		return;
	}

	/* bN */
	LockVM();

	/* [h`FW */
	uPaletteReg = uMode;

	/* rbg}bv */
	if (hPaletteReg) {
		DeleteObject(hPaletteReg);
		hPaletteReg = NULL;
		pPaletteReg = NULL;
	}

	/* `ۑ */
	rPaletteReg = rect;

	/* ̑ */
	InitPaletteReg(hWnd);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += rect.right;
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += rect.bottom;
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ĕ` */
	bPaletteRefresh = TRUE;

	/* AbN */
	UnlockVM();
}
#endif

/*
 *	pbgWX^EChE
 *	EChE쐬
 */
HWND FASTCALL CreatePaletteReg(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_Palette";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* tOZbg(EChE쐬ɂ͕Kĕ`悷) */
	bPaletteRefresh = TRUE;
	bMultiPageRefresh = TRUE;

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);

	/* EChETCY */
	rect.right = 192;
#if XM7_VER == 1
	rect.bottom = (4 + 3) * 8;
#else
	rect.bottom = 192 + (4 + 3 + 2) * 8;
	uPaletteReg = 1;
#endif

	/* EChE^Cg */
	GetMsg(hResInstance, IDS_SWND_PALETTE, szWndName, sizeof(szWndName));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = PaletteRegProc;;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);

		/* `ۑ */
		rPaletteReg = rect;

		/* ̑ */
		InitPaletteReg(hWnd);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ IRQEChE ]---------------------------------------------------------*/

/*
 *	IRQEChE
 *	ZbgAbv(tO)
 */
static void FASTCALL SetupIRQFlag(BYTE *p, char *title, BOOL flag, BOOL mask, BYTE irq)
{
	char string[28];
	int i;

	/*  */
	memset(string, 0x20, sizeof(string));
	string[sizeof(string) - 1] = '\0';

	/* Rs[ */
	for (i=0; i<sizeof(string) - 2; i++) {
		if (title[i] == '\0') {
			break;
		}
		string[i] = title[i];
	}

	if (mask) {
		strncpy(&string[19], "   Mask", sizeof(string)-19);
	}
	else if (!flag) {
		strncpy(&string[19], "       ", sizeof(string)-19);
	}
	else if (irq) {
		strncpy(&string[19], "   Pend", sizeof(string)-19);
	}
	else {
		strncpy(&string[19], "Request", sizeof(string)-19);
	}

	/* Zbg */
	memcpy(p, string, 26);
}


/*
 *	IRQEChE
 *	ZbgAbv
 */
static void FASTCALL SetupIRQ(BYTE nCPU, BYTE *p, int x, int y)
{
	int n;
	BOOL firq_mask;
#if XM7_VER == 1 && defined(SFDC)
	BOOL sfd_drq_mask;
#endif

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* [NGA */
	n = 0;
#if XM7_VER == 1 && defined(SFDC)
	firq_mask = sfd_firq_mode;
	sfd_drq_mask = FALSE;
#else
	firq_mask = FALSE;
#endif

	/* tO */
	if (nCPU == MAINCPU) {
#if XM7_VER == 1 && defined(SFDC)
		if (model_type != MODEL_FM77) {
			sfd_drq_mask = TRUE;
		}
#endif
#if XM7_VER == 1 && defined(SFDC)
		SetupIRQFlag(&p[n     ], "NMI  (SFDC IRQ)    ", maincpu.intr & INTR_NMI, FALSE, 0);
		SetupIRQFlag(&p[n += x], "FIRQ (SFDC DRQ)    ", sfd_drq_flag, sfd_drq_mask, maincpu.cc & 0x40);
		SetupIRQFlag(&p[n += x], "FIRQ (Break Key)   ", break_flag, firq_mask, maincpu.cc & 0x40);
#else
		SetupIRQFlag(&p[n     ], "FIRQ (Break Key)   ", break_flag, firq_mask, maincpu.cc & 0x40);
#endif
		SetupIRQFlag(&p[n += x], "FIRQ (Attention)   ", subattn_flag, firq_mask, maincpu.cc & 0x40);
		SetupIRQFlag(&p[n += x], "IRQ  (Keyboard)    ", key_irq_flag, key_irq_mask, maincpu.cc & 0x10);
		SetupIRQFlag(&p[n += x], "IRQ  (2.03ms Timer)", timer_irq_flag, timer_irq_mask, maincpu.cc & 0x10);
		SetupIRQFlag(&p[n += x], "IRQ  (Printer)     ", lp_irq_flag, lp_irq_mask, maincpu.cc & 0x10);
#if XM7_VER == 1 && defined(SFDC)
		SetupIRQFlag(&p[n += x], "IRQ  (MFDC)        ", mfd_irq_flag, mfd_irq_mask, maincpu.cc & 0x10);
		SetupIRQFlag(&p[n += x], "IRQ  (SFDC)        ", sfd_irq_flag, sfd_irq_mask, maincpu.cc & 0x10);
#else
		SetupIRQFlag(&p[n += x], "IRQ  (FDC)         ", mfd_irq_flag, mfd_irq_mask, maincpu.cc & 0x10);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		SetupIRQFlag(&p[n += x], "IRQ  (USART TxRdy) ", txrdy_irq_flag, txrdy_irq_mask, maincpu.cc & 0x10);
		SetupIRQFlag(&p[n += x], "IRQ  (USART RxRdy) ", rxrdy_irq_flag, rxrdy_irq_mask, maincpu.cc & 0x10);
		SetupIRQFlag(&p[n += x], "IRQ  (USART Sync)  ", syndet_irq_flag, syndet_irq_mask, maincpu.cc & 0x10);
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
		SetupIRQFlag(&p[n += x], "IRQ  (DMAC)        ", dma_irq_flag, !(dma_icr & dma_txrq), maincpu.cc & 0x10);
#endif	/* XM7_VER == 1 && defined(SFDC) */
		SetupIRQFlag(&p[n += x], "IRQ  (OPN Timer)   ", opn_irq_flag, FALSE, maincpu.cc & 0x10);
		SetupIRQFlag(&p[n += x], "IRQ  (WHG Timer)   ", whg_irq_flag, FALSE, maincpu.cc & 0x10);
		SetupIRQFlag(&p[n += x], "IRQ  (THG Timer)   ", thg_irq_flag, FALSE, maincpu.cc & 0x10);
	}
	else {
		SetupIRQFlag(&p[n     ], "NMI  (20ms Timer)  ", subcpu.intr & INTR_NMI, FALSE, 0);
		SetupIRQFlag(&p[n += x], "FIRQ (Keyboard)    ", ((!key_irq_mask && key_irq_flag) || (subcpu.cc & 0x40)),
														(key_irq_mask || !key_irq_flag) && (subcpu.cc & 0x40),
														!key_irq_mask && key_irq_flag && subcpu.cc & 0x40);
		SetupIRQFlag(&p[n += x], "IRQ  (Cancel)      ", subcancel_flag, !subcancel_flag && (subcpu.cc & 0x10),
														subcpu.cc & 0x40);
	}
}

/*
 *	IRQEChE
 *	`
 */
static void FASTCALL DrawIRQ(HWND hWnd, HDC hDC)
{
	int i;
	BYTE nCPU;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			nCPU = (BYTE)((i - SWND_IRQ_MAIN) % (SUBCPU + 1));
			p = pIRQ[nCPU];
			break;
		}
	}
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	SetupIRQ(nCPU, p, x, y);

	/* ` */
	DrawWindowText(hDC, p, x, y);
}

/*
 *	IRQEChE
 *	tbV
 */
void FASTCALL RefreshIRQ(void)
{
	HWND hWnd;
	HDC hDC;

	/* CCPU */
	if (hSubWnd[SWND_IRQ_MAIN]) {
		hWnd = hSubWnd[SWND_IRQ_MAIN];
		hDC = GetDC(hWnd);
		DrawIRQ(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}

	/* TuCPU */
	if (hSubWnd[SWND_IRQ_SUB]) {
		hWnd = hSubWnd[SWND_IRQ_SUB];
		hDC = GetDC(hWnd);
		DrawIRQ(hWnd, hDC);
		ReleaseDC(hWnd, hDC);
	}
}

/*
 *	IRQEChE
 *	ĕ`
 */
static void FASTCALL PaintIRQ(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	int i;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* Drawobt@𓾂(݂ȂΉȂ) */
	p = NULL;
	for (i=0; i<SWND_MAXNUM; i++) {
		if (hSubWnd[i] == hWnd) {
			p = pIRQ[((i - SWND_IRQ_MAIN) % (SUBCPU + 1))];
			break;
		}
	}
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawIRQ(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	IRQEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK IRQProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	BYTE nCPU;

	if (hSubWnd[SWND_IRQ_MAIN] == hWnd) {
		nCPU = MAINCPU;
	}
	else {
		nCPU = SUBCPU;
	}

	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintIRQ(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pIRQ[nCPU], NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	IRQEChE
 *	EChE쐬
 */
HWND FASTCALL CreateIRQ(HWND hParent, BYTE nCPU, int index)
{
	WNDCLASSEX wcex;
	char szClassNameMain[] = "XM7_IRQ_Main";
	char szClassNameSub[] = "XM7_IRQ_Sub";
	char *szClassName;
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	UINT id;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 26;

	/* EChE^CgAobt@m */
	switch (nCPU) {
		case MAINCPU :	id = IDS_SWND_IRQ_MAIN;
						szClassName = szClassNameMain;
#if XM7_VER >= 3
						rect.bottom = lCharHeight * 13;
#elif XM7_VER == 1 && defined(SFDC)
						rect.bottom = lCharHeight * 16;
//						rect.bottom = lCharHeight * 15;
#else
						rect.bottom = lCharHeight * 13;
//						rect.bottom = lCharHeight * 12;
#endif
						break;
		case SUBCPU :	id = IDS_SWND_IRQ_SUB;
						szClassName = szClassNameSub;
						rect.bottom = lCharHeight * 3;
						break;
	}
	GetMsg(hResInstance, id, szWndName, sizeof(szWndName));
	pIRQ[nCPU] = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = IRQProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

HWND FASTCALL CreateIRQMain(HWND hParent, int index)
{
	return	CreateIRQ(hParent, MAINCPU, index);
}

HWND FASTCALL CreateIRQSub(HWND hParent, int index)
{
	return	CreateIRQ(hParent, SUBCPU, index);
}

/*-[ ROMEChE ]-----------------------------------------------------*/

/*
 *	ROMEChE
 *	hbg`
 */
static void FASTCALL PSetKanji(int x, int y, int color)
{
	BYTE *p;
	BYTE dat;

	ASSERT((x >= 0) && (x < rKanji.right));
	ASSERT((y >= 0) && (y < rKanji.bottom));
	ASSERT((color >= 0) && (color <= 15));

	if (!pKanji) {
		return;
	}

	/* AhXAf[^擾 */
	/* rKanjiArbg}bv̉TCYvZ(4oCgACg) */
	p = &pKanji[((((rKanji.right / 2) + 3) >> 2) << 2) * y + (x >> 1)];
	dat = *p;

	/* Qɕ */
	if (x & 1) {
		dat &= 0xf0;
		dat |= (BYTE)color;
	}
	else {
		dat &= 0x0f;
		dat |= (BYTE)(color << 4);
	}

	/*  */
	*p = dat;
}

/*
 *	ROMEChE
 *	{bNXtB`
 */
static void FASTCALL BfKanji(int x, int y, int cx, int cy, int color)
{
	int i;
	int j;
	int col;

	ASSERT((color >= 0) && (color <= 15));

	for (i=0; i<cy; i++) {
		for (j=0; j<cx; j++) {
			if ((cx >= 3) && (cy >= 3) &&
				((i == cy - 1) ||
				 (j == cx - 1))) {
				col = 1;
			}
			else {
				col = color;
			}
			PSetKanji(x + j, y, col);
		}
		y++;
	}
}

/*
 *	ROMEChE
 *	LN^`
 */
static void FASTCALL ChrKanji(char c, int x, int y, int fc, int bc)
{
	int i;
	int j;
	BYTE *p;
	BYTE dat;

	ASSERT((fc >= 0) && (fc < 16));
	ASSERT((bc >= 0) && (bc < 16));

	/* TuROM(C)̃tHgAhX𓾂 */
	p = &subrom_c[c * 8];

	/* y[v */
	for (i=0; i<8; i++) {
		/* f[^擾 */
		dat = *p;
		p++;

		/* x[v */
		for (j=0; j<8; j++) {
			if (dat & 0x80) {
				PSetKanji(x, y, fc);
			}
			else {
				PSetKanji(x, y, bc);
			}
			dat <<= 1;
			x++;
		}

		/* y */
		x -= 8;
		y++;
	}
}

/*
 *	ROMEChE
 *	`
 */
static void FASTCALL FontKanji(BYTE *paddr, int x, int y, int fc, int bc, BOOL flag)
{
	int i, j;
	WORD dat;
	int color;

	/* [v */
	for (i=0; i<16; i++) {
		if (flag) {
			dat = 0xffff;
		}
		else {
			dat = *(paddr + 0);
			dat <<= 8;
			dat |= *(paddr + 1);
		}
		for (j=0; j<16; j++) {
			if (dat & 0x8000) {
				color = fc;
			}
			else {
				color = bc;
			}
			BfKanji(j * 4 + x, i * 4 + y, 4, 4, color);
			dat <<= 1;
		}
		paddr += 2;
	}
}

/*
 *	ROMEChE
 *	ZbgAbv
 */
static void FASTCALL SetupStrKanji(int n, int y, int ys, BOOL *flag)
{
	int i;
	int j;

	/*  */
	for (i=0; i<6; i++) {
		for (j=0; j<16; j++) {
			/* v`FbN */
			if (cnKanji[n * 6 + i][j * 3 + 0] == ctKanji[n * 6 + i][j * 3 + 0]) {
				if (cnKanji[n * 6 + i][j * 3 + 1] == ctKanji[n * 6 + i][j * 3 + 1]) {
					if (cnKanji[n * 6 + i][j * 3 + 2] == ctKanji[n * 6 + i][j * 3 + 2]) {
						continue;
					}
				}
			}

			/* ` */
			ChrKanji(	ctKanji[n * 6 + i][j * 3 + 0], 16 * 4 + 2 + j * 8, y + 2,
						ctKanji[n * 6 + i][j * 3 + 1],
						ctKanji[n * 6 + i][j * 3 + 2]);

			/* Rs[ */
			cnKanji[n * 6 + i][j * 3 + 0] = ctKanji[n * 6 + i][j * 3 + 0];
			cnKanji[n * 6 + i][j * 3 + 1] = ctKanji[n * 6 + i][j * 3 + 1];
			cnKanji[n * 6 + i][j * 3 + 2] = ctKanji[n * 6 + i][j * 3 + 2];

			/* tO */
			*flag = TRUE;
		}

		/*  */
		y += ys;
	}
}

/*
 *	ROMEChE
 *	ZbgAbv
 */
static BOOL FASTCALL SetupKanji(void)
{
	int n;
	BYTE *krom;
	BOOL flag;

	n = 0;
	flag = FALSE;

	/* JIS1 */
	krom = kanji_rom;
	if ((model_type == MODEL_FM7) &&
		(nKanji == KANJI_JSUBINFO)) {
		krom = kanji_romn;
	}

	/* ROMAhX̕ύX΍XVȂ */
	if ((wAddrKanji & 0xfff0) != (kanji_addr & 0xfff0)) {
		/* tHg */
#ifdef JSUB
		FontKanji(	&krom[(kanji_addr & 0xfff0) << 1],
					0, n * 17 * 4, 15, 0,
					(nKanji != KANJI_JSUBINFO) && !(nKanji & KANJI_SET));
#else
		FontKanji(	&krom[(kanji_addr & 0xfff0) << 1],
					0, n * 17 * 4, 15, 0,
					!(nKanji & KANJI_SET));
#endif
	}

	/* 1 */
	SetupStrKanji(n, 0, 10, &flag);

#if XM7_VER >= 2
	/* {J[h JIS2 */
	if (nKanji & KANJI_JCARD) {

		n += 1;

		/* ROMAhX̕ύX΍XVȂ */
		if ((wAddrKanji & 0xfff0) != (kanji_addr & 0xfff0)) {
			if (n > 0) {
				BfKanji(0, n * 17 * 4 - 1, 16 * 4 - 1, 1, 1);
			}

			/* tHg */
			FontKanji(	&kanji_rom2[(kanji_addr & 0xfff0) << 1],
						0, n * 17 * 4, 15, 0, FALSE);
		}

		/* 2 */
		SetupStrKanji(n, 17 * 4 * n, 10, &flag);
	}
#endif

#ifdef JSUB
	/* {ʐMJ[h JIS1 */
	if (nKanji & KANJI_JSUB) {

		n += 1;

		/* ROMAhX̕ύX΍XVȂ */
		if ((wAddrJSubKanji & 0xfff0) != (jsub_kanji_addr & 0xfff0)) {
			if (n > 0) {
				BfKanji(0, n * 17 * 4 - 1, 16 * 4 - 1, 1, 1);
			}

			/* tHg */
			FontKanji(	&kanji_romn[(jsub_kanji_addr & 0xfff0) << 1],
						0, n * 17 * 4, 15, 0, FALSE);
		}

		/* 1 */
		SetupStrKanji(n, 17 * 4 * n, 10, &flag);
	}
#endif

	/* \ڂ̕ύX΍XVȂ */
	if ((uKanjiType == model_type) &&
		(wAddrKanji == kanji_addr) &&
#ifdef JSUB
		(wAddrJSubKanji == jsub_kanji_addr) &&
#endif
		!flag) {
		return FALSE;
	}

	/* \̍XV */
	uKanjiType = model_type;
	wAddrKanji = kanji_addr;
#ifdef JSUB
	wAddrJSubKanji = jsub_kanji_addr;
#endif

	return TRUE;
}

/*
 *	ROMEChE
 *	Zbg
 */
static void FASTCALL StrKanji(char *string, int x, int y, int fc, int bc)
{
	char ch;

	ASSERT(string);
	ASSERT((x >= 0) && (x < 16));
	ASSERT((y >= 0) && (y < 3 * 6));
	ASSERT((fc >= 0) && (fc < 16));
	ASSERT((bc >= 0) && (bc < 16));

	/* [v */
	for (;;) {
		/* 擾 */
		ch = *string++;
		if (ch == '\0') {
			break;
		}

		/* x`FbN */
		if (x >= 49) {
			continue;
		}

		/* Zbg */
		ctKanji[y][x * 3 + 0] = ch;
		ctKanji[y][x * 3 + 1] = (BYTE)fc;
		ctKanji[y][x * 3 + 2] = (BYTE)bc;
		x++;
	}
}

/*
 *	ROMEChE
 *	f[^Zbg
 */
static void FASTCALL StatChrKanji(void)
{
	const char * const cpKanjiMain =	"MAIN";
#if XM7_VER >= 3 || (XM7_VER == 1 && defined(L4CARD))
	const char * const cpKanjiSub =		" SUB";
#endif
#ifdef JSUB
	const char * const cpKanjiJSub =	"JSUB";
#endif
	const char * const pform[] = {		"    NONE",
										" JIS78-1",
										"JIS78E-1",
										"JIS78A-1",
										" JIS83-1",
										" JIS83-2",
										"JIS78N-1",
										" UNKNOWN",
										"    TEMP",
								};
	int n;
	char string[128];
	int nform;
	char *ptype;

	n = 0;

	/* JIS1 */
	switch (model_type) {
#if XM7_VER >= 2
#if XM7_VER >= 3
		case MODEL_FM77AV40EX:
		case MODEL_FM77AV40:
			if (subkanji_flag) {
				ptype = (char *)cpKanjiSub;
			}
			else {
				ptype = (char *)cpKanjiMain;
			}
			nform = kanji_rom_type;
			break;
#endif
		case MODEL_FM77AV20EX:
		case MODEL_FM77AV20:
		case MODEL_FM77AV:
			ptype = (char *)cpKanjiMain;
			nform = kanji_rom_type;
			break;
#else
		case MODEL_FM8:
			ptype = (char *)cpKanjiMain;
			if (nKanji & KANJI_SET) {
				nform = kanji_rom_type;
			}
			else {
				nform = KANJI_ROM_NONE;
			}
			break;
		case MODEL_FM77:
#ifdef L4CARD
			if (subkanji_flag) {
				ptype = (char *)cpKanjiSub;
			}
			else {
				ptype = (char *)cpKanjiMain;
			}
#else
			ptype = (char *)cpKanjiMain;
#endif
			nform = kanji_rom_type;
			break;
#endif
		case MODEL_FM7:
			ptype = (char *)cpKanjiMain;
#ifdef JSUB
			if (nKanji == KANJI_JSUBINFO) {
				nform = kanji_romn_type;
			}
			else if (nKanji & KANJI_SET) {
				nform = kanji_rom_type;
			}
#else
			if (nKanji & KANJI_SET) {
				nform = kanji_rom_type;
			}
#endif
			else {
				nform = KANJI_ROM_NONE;
			}
			break;
		default:
			break;
	}

	/* 1 */
	StrKanji("LEVEL 1", 0, n + 0, 15, 0);
	StrKanji(ptype, 12, n + 0, 15, 0);

	/* AhX */
	StrKanji("ADDRESS", 0, n + 1, 15, 0);
	_snprintf(string, sizeof(string), "%04X", kanji_addr);
	StrKanji(string, 12, n + 1, 15, 0);

	/* tHg^Cv */
	StrKanji("FONT", 0, n + 2, 15, 0);
	StrKanji((char *)pform[nform], 8, n + 2, 15, 0);

#ifdef JSUB
	/* {ʐMJ[h JIS1 */
	if ((nKanji & KANJI_JSUBINFO) && !(nKanji & KANJI_JSUB)) {

		/* HALT */
		StrKanji("JSUB HALT", 0, n + 3, 15, 0);
		if (jsub_haltflag) {
			StrKanji(" ON", 13, n + 3, 15, 0);
		}
		else {
			StrKanji("OFF", 13, n + 3, 15, 0);
		}

		/* DIC BANK */
		StrKanji("DIC BANK", 0, n + 4, 15, 0);
		_snprintf(string, sizeof(string), "%02x", jsub_dicbank);
		StrKanji(string, 14, n + 4, 15, 0);

		/* RCB ADDR */
		StrKanji("RCB ADDRESS", 0, n + 5, 15, 0);
		_snprintf(string, sizeof(string), "%04X", jsub_address);
		StrKanji(string, 12, n + 5, 15, 0);
	}
#endif

#if XM7_VER >= 3
	/* ROMBtHg */
	if (nKanji & KANJI_DICROM) {
		StrKanji("DICROM", 0, n + 5, 15, 0);
		StrKanji((char *)pform[kanji_romn_type], 8, n + 5, 15, 0);
	}
#endif

#if XM7_VER >= 2
	/* {J[h JIS2 */
	if (nKanji & KANJI_JCARD) {
		n += 6;

#if XM7_VER >= 3
		if (subkanji_flag &&
			((model_type == MODEL_FM77AV40EX) ||
			 (model_type == MODEL_FM77AV40))) {
			ptype = (char *)cpKanjiSub;
		}
		else {
			ptype = (char *)cpKanjiMain;
		}
#else
		ptype = (char *)cpKanjiMain;
#endif

		/* 2 */
		StrKanji("LEVEL 2", 0, n + 0, 15, 0);
		StrKanji(ptype, 12, n + 0, 15, 0);

		/* AhX */
		StrKanji("ADDRESS", 0, n + 1, 15, 0);
		_snprintf(string, sizeof(string), "%04X", kanji_addr);
		StrKanji(string, 12, n + 1, 15, 0);

		/* tHg^Cv */
		StrKanji("FONT", 0, n + 2, 15, 0);
		StrKanji((char *)pform[kanji_rom2_type], 8, n + 2, 15, 0);

#if XM7_VER >= 3
		/* TuVXe JIS1/2I */
		if ((model_type == MODEL_FM77AV40EX) ||
			(model_type == MODEL_FM77AV40)) {
			if (sub_kanji_bank) {
				StrKanji("SUB SELECT LEV.2", 0, n + 5, 15, 0);
			}
			else {
				StrKanji("SUB SELECT LEV.1", 0, n + 5, 15, 0);
			}
		}
#endif
	}
#endif

#ifdef JSUB
	/* {ʐMJ[h JIS1 */
	if (nKanji & KANJI_JSUB) {
		n += 6;

		/* 1 */
		StrKanji("LEVEL 1", 0, n + 0, 15, 0);
		StrKanji((char *)cpKanjiJSub, 12, n + 0, 15, 0);

		/* AhX */
		StrKanji("ADDRESS", 0, n + 1, 15, 0);
		_snprintf(string, sizeof(string), "%04X", jsub_kanji_addr);
		StrKanji(string, 12, n + 1, 15, 0);

		/* tHg^Cv */
		StrKanji("FONT", 0, n + 2, 15, 0);
		StrKanji((char *)pform[kanji_romn_type], 8, n + 2, 15, 0);

		/* HALT */
		StrKanji("JSUB HALT", 0, n + 3, 15, 0);
		if (jsub_haltflag) {
			StrKanji(" ON", 13, n + 3, 15, 0);
		}
		else {
			StrKanji("OFF", 13, n + 3, 15, 0);
		}

		/* DIC BANK */
		StrKanji("DIC BANK", 0, n + 4, 15, 0);
		_snprintf(string, sizeof(string), "%02x", jsub_dicbank);
		StrKanji(string, 14, n + 4, 15, 0);

		/* RCB ADDR */
		StrKanji("RCB ADDRESS", 0, n + 5, 15, 0);
		_snprintf(string, sizeof(string), "%04X", jsub_address);
		StrKanji(string, 12, n + 5, 15, 0);
	}
#endif
}

/*
 *	ROMEChE
 *	`
 */
static void FASTCALL DrawKanji(HDC hDC)
{
	HDC hMemDC;
	HBITMAP hBitmap;

	ASSERT(hDC);

	/* rbg}bvnhE|C^ȂAȂ */
	if (!hKanji || !pKanji) {
		return;
	}

	/* Xe[^X`FbN */
	StatChrKanji();

	/* ZbgAbv */
	if (!SetupKanji()) {
		return;
	}

	/* rbg}bvnhE|C^Lx`FbN */
	if (!hKanji || !pKanji) {
		return;
	}

	/* DC쐬Arbg}bvZNg */
	hMemDC = CreateCompatibleDC(hDC);
	ASSERT(hMemDC);
	hBitmap = (HBITMAP)SelectObject(hMemDC, hKanji);

	/* fXNgbv̏󋵂ł́Aۂ蓾l */
	if (hBitmap) {
		/* pbgݒABitBlt */
		SetDIBColorTable(hMemDC, 0, 16, rgbQuadTable16);
		BitBlt(hDC, 0, 0,
				rKanji.right, rKanji.bottom, hMemDC,
				0, 0, SRCCOPY);

		/* IuWFNgăZNg */
		SelectObject(hMemDC, hBitmap);
	}

	/* DC폜 */
	DeleteDC(hMemDC);
}

/*
 *	ROMEChE
 *	tbV
 */
void FASTCALL RefreshKanji(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_KANJI] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_KANJI];
	hDC = GetDC(hWnd);
	DrawKanji(hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	ROMEChE
 *	ĕ`
 */
static void FASTCALL PaintKanji(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	if (!pKanji) {
		return;
	}

	/* [NGAׂĖ */
	memset(cnKanji, 0xff, sizeof(cnKanji));
	uKanjiType = 0xff;
	wAddrKanji = 0xfffe;
#ifdef JSUB
	wAddrJSubKanji = 0xfffe;
#endif

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawKanji(hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	ROMEChE
 *	EChEvV[W
 */
static LRESULT CALLBACK KanjiProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintKanji(hWnd);
			UnlockVM();
			return 0;

		/* wi` */
		case WM_ERASEBKGND:
			return TRUE;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, NULL, NULL);

			/* rbg}bv */
			if (hKanji) {
				DeleteObject(hKanji);
				hKanji = NULL;
				pKanji = NULL;
			}

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	ROMEChE
 *	
 */
static void FASTCALL InitKanji(HWND hWnd)
{
	BITMAPINFOHEADER *pbmi;
	HDC hDC;

	/* S̃[N */
	pKanji = NULL;
	hKanji = NULL;

	/* \Ǘ[N */
	memset(cnKanji, 0xff, sizeof(cnKanji));
	memset(ctKanji, 0xff, sizeof(ctKanji));
	uKanjiType = 0xff;
	wAddrKanji = 0xfffe;
#ifdef JSUB
	wAddrJSubKanji = 0xfffe;
#endif

	/* rbg}bvwb_ */
	pbmi = (BITMAPINFOHEADER*)malloc(sizeof(BITMAPINFOHEADER)
										+ sizeof(RGBQUAD) * 16);
	if (pbmi) {
		memset(pbmi, 0, sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD) * 16);
		pbmi->biSize = sizeof(BITMAPINFOHEADER);
		pbmi->biWidth = rKanji.right;
		pbmi->biHeight = -rKanji.bottom;
		pbmi->biPlanes = 1;
		pbmi->biBitCount = 4;
		pbmi->biCompression = BI_RGB;

		/* DC擾ADIBZNV쐬 */
		hDC = GetDC(hWnd);
		hKanji = CreateDIBSection(hDC, (BITMAPINFO*)pbmi, DIB_RGB_COLORS,
								(void**)&pKanji, NULL, 0);
		ReleaseDC(hWnd, hDC);
		free(pbmi);
	}
}

/*
 *	ROMEChE
 *	TCYύX
 */
void FASTCALL ReSizeKanji(void)
{
	HWND hWnd;
	UINT uMode;
	RECT wrect;
	RECT crect;
	int xs;
	int ys;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_KANJI] == NULL) {
		return;
	}

	/* WIgύX`FbN */
	hWnd = hSubWnd[SWND_KANJI];
	uMode = 0;
	xs = 17 * 4 + 16 * 8;
	ys = 16 * 4;

	/*	KANJI_SET		JISxx 1\tO
	 *	KANJI_JCARD		JIS83 2\tO
	 *	KANJI_JSUB		JIS78N1\tO
	 *	KANJI_JSUBINFO	{ʐMJ[h\tO
	 *	KANJI_DICROM	FM77AV40EX ROMBtHg݃tO
	 */
	switch (model_type) {
#if XM7_VER >= 2
#if XM7_VER >= 3
		case MODEL_FM77AV40EX:
			uMode |= (KANJI_SET | KANJI_JCARD | KANJI_DICROM);
			ys += 17 * 4;
			break;
		case MODEL_FM77AV40:
			uMode |= (KANJI_SET | KANJI_JCARD);
			ys += 17 * 4;
			break;
#endif
		case MODEL_FM77AV20EX:
		case MODEL_FM77AV20:
			uMode |= KANJI_SET;
			if (jcard_mode == JCARD_ON) {
				uMode |= KANJI_JCARD;
				ys += 17 * 4;
			}
			break;
		case MODEL_FM77AV:
			uMode |= KANJI_SET;
			if (jcard_mode == JCARD_ON) {
				uMode |= KANJI_JCARD;
				ys += 17 * 4;
			}
#ifdef JSUB
			if (jsub_enable) {
				uMode |= KANJI_JSUBINFO;
				if (!jsub_kanji_dipsw) {
					uMode |= KANJI_JSUB;
					ys += 17 * 4;
				}
			}
#endif
			break;
#else
		case MODEL_FM8:
			if (kanji_enable) {
				uMode |= KANJI_SET;
			}
			break;
		case MODEL_FM77:
			uMode |= KANJI_SET;
#ifdef JSUB
			if (jsub_enable) {
				uMode |= KANJI_JSUBINFO;
				if (!jsub_kanji_dipsw) {
					uMode |= KANJI_JSUB;
					ys += 17 * 4;
				}
			}
#endif
			break;
#endif
		case MODEL_FM7:
			if (kanji_enable) {
				uMode |= KANJI_SET;
			}
#ifdef JSUB
			if (jsub_enable) {
				uMode |= KANJI_JSUBINFO;
				if (!jsub_kanji_dipsw) {
					uMode |= KANJI_JSUB;
					ys += 17 * 4;
				}
			}
#endif
			break;
		default:
			break;
	}
	if (IsIconic(hWnd)) {
		/* ACR`FbN */
		uMode |= 0x80000000;
	}
	if (uMode == nKanji) {
		return;
	}

	/* bN */
	LockVM();

	/* [h`FW */
	nKanji = uMode;

	/* rbg}bv */
	if (hKanji) {
		DeleteObject(hKanji);
		hKanji = NULL;
		pKanji = NULL;
	}

	/* ̑ */
	rKanji.bottom = ys;
	InitKanji(hWnd);

	/* EBhE`ύX */
	GetWindowRect(hWnd, &wrect);
	GetClientRect(hWnd, &crect);
	wrect.right -= wrect.left;
	wrect.right -= crect.right;
	wrect.right += xs;
	wrect.bottom -= wrect.top;
	wrect.bottom -= crect.bottom;
	wrect.bottom += ys;
	SetWindowPos(hWnd, HWND_TOP, 0, 0, wrect.right, wrect.bottom,
							SWP_NOZORDER | SWP_NOMOVE);
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* AbN */
	UnlockVM();
}

/*
 *	ROMEChE
 *	EChE쐬
 */
HWND FASTCALL CreateKanji(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_Kanji";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);

	/* EChETCY */
	rect.right = 17 * 4 + 16 * 8;
	rect.bottom = 16 * 4;

	/* EChE^Cg */
	GetMsg(hResInstance, IDS_SWND_KANJI, szWndName, sizeof(szWndName));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = KanjiProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);

		/* `ۑ */
		rKanji = rect;
		nKanji = 0;

		/* ̑ */
		InitKanji(hWnd);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

/*-[ JZbge[vv^EChE ]------------------------------------*/

/*
 *	JZbge[vv^EChE
 *	ZbgAbv
 */
static void FASTCALL SetupTapeLP(BYTE *p, int x, int y)
{
	char string[128];

	ASSERT(p);
	ASSERT(x > 0);
	ASSERT(y > 0);

	/* UXy[XŖ߂ */
	memset(p, 0x20, x * y);

	/* [^ */
	strncpy(string, "Motor        ", sizeof(string));
	if (tape_motor) {
		strncat(string, " On", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "Off", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[0 + x * 0], string, strlen(string));

	/* ^ */
	strncpy(string, "Record       ", sizeof(string));
	if (tape_rec) {
		strncat(string, " On", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "Off", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[0 + x * 1], string, strlen(string));

	/* ̓f[^ */
	strncpy(string, "Input  Data    ", sizeof(string));
	if (tape_in) {
		strncat(string, "1", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "0", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[0 + x * 2], string, strlen(string));

	/* o̓f[^ */
	strncpy(string, "Output Data    ", sizeof(string));
	if (tape_out) {
		strncat(string, "1", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "0", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[0 + x * 3], string, strlen(string));

	/* TCNJE^ */
	_snprintf(string, sizeof(string), "Cycle Count %04X", tape_count);
	memcpy(&p[0 + x * 4], string, strlen(string));

	/* t@CItZbg */
	_snprintf(string, sizeof(string), "Offset  %08X", tape_offset >> 4);
	memcpy(&p[0 + x * 5], string, strlen(string));

	/* Xg[utO */
	strncpy(string, "Strobe      ", sizeof(string));
	if (lp_strobe) {
		strncat(string, " On", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "Off", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[18 + x * 0], string, strlen(string));

	/* t@Co̓tO */
	strncpy(string, "File Output ", sizeof(string));
#ifdef LPRINT
	if ((lp_mode == LP_ENABLE) || (lp_mode == LP_LOG)) {
#else
	if (lp_fileh != -1) {
#endif
		strncat(string, " On", sizeof(string)-strlen(string)-1);
	}
	else {
		strncat(string, "Off", sizeof(string)-strlen(string)-1);
	}
	memcpy(&p[18 + x * 1], string, strlen(string));

	/* f[^ */
	_snprintf(string, sizeof(string), "Print Data   %02X", lp_data);
	memcpy(&p[18 + x * 2], string, strlen(string));
}

/*
 *	JZbge[vv^EChE
 *	`
 */
static void FASTCALL DrawTapeLP(HWND hWnd, HDC hDC)
{
	RECT rect;
	int x, y;

	ASSERT(hWnd);
	ASSERT(hDC);

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;
	if ((x == 0) || (y == 0)) {
		return;
	}

	/* ZbgAbv */
	if (!pTapeLP) {
		return;
	}
	SetupTapeLP(pTapeLP, x, y);

	/* ` */
	DrawWindowText(hDC, pTapeLP, x, y);
}

/*
 *	JZbge[vv^EChE
 *	tbV
 */
void FASTCALL RefreshTapeLP(void)
{
	HWND hWnd;
	HDC hDC;

	/* ɌĂ΂̂ŁA݃`FbN邱 */
	if (hSubWnd[SWND_TAPELP] == NULL) {
		return;
	}

	/* ` */
	hWnd = hSubWnd[SWND_TAPELP];
	hDC = GetDC(hWnd);
	DrawTapeLP(hWnd, hDC);
	ReleaseDC(hWnd, hDC);
}

/*
 *	JZbge[vv^EChE
 *	ĕ`
 */
static void FASTCALL PaintTapeLP(HWND hWnd)
{
	HDC hDC;
	PAINTSTRUCT ps;
	RECT rect;
	BYTE *p;
	int x, y;

	ASSERT(hWnd);

	/* |C^ݒ(݂ȂΉȂ) */
	p = pTapeLP;
	if (!p) {
		return;
	}

	/* EChEWIg𓾂 */
	GetClientRect(hWnd, &rect);
	x = rect.right / lCharWidth;
	y = rect.bottom / lCharHeight;

	/* 㔼GAFFŖ߂ */
	if ((x > 0) && (y > 0)) {
		memset(&p[x * y], 0xff, x * y);
	}

	/* ` */
	hDC = BeginPaint(hWnd, &ps);
	ASSERT(hDC);
	DrawTapeLP(hWnd, hDC);
	EndPaint(hWnd, &ps);
}

/*
 *	JZbge[vv^EChE
 *	EChEvV[W
 */
static LRESULT CALLBACK TapeLPProc(HWND hWnd, UINT message,
								 WPARAM wParam, LPARAM lParam)
{
	/* bZ[W */
	switch (message) {
		/* EChEĕ` */
		case WM_PAINT:
			/* bNKv */
			LockVM();
			PaintTapeLP(hWnd);
			UnlockVM();
			return 0;

		/* EChE폜 */
		case WM_DESTROY:
			LockVM();

			/* CEChE֎ʒm */
			DestroySubWindow(hWnd, &pTapeLP, NULL);

			UnlockVM();
			break;

		/* EBhEړ */
		case WM_MOVE:
			/* ŏ̃TuEBhEҔ */
			MoveSubWindow(hWnd);
			break;
	}

	/* ftHg EChEvV[W */
	return DefWindowProc(hWnd, message, wParam, lParam);
}

/*
 *	JZbge[vv^EChE
 *	EChE쐬
 */
HWND FASTCALL CreateTapeLP(HWND hParent, int index)
{
	WNDCLASSEX wcex;
	char szClassName[] = "XM7_TapeLP";
	char szWndName[128];
	RECT rect;
	RECT crect, wrect;
	HWND hWnd;
	DWORD dwStyle;

	ASSERT(hParent);

	/* EChE`vZ */
	PositioningSubWindow(hParent, &rect, index);
	rect.right = lCharWidth * 33;
	rect.bottom = lCharHeight * 6;

	/* EChE^CgAobt@m */
	GetMsg(hResInstance, IDS_SWND_TAPELP,
				szWndName, sizeof(szWndName));
	pTapeLP = malloc(2 * (rect.right / lCharWidth) *
								(rect.bottom / lCharHeight));

	/* EChENX̓o^ */
	memset(&wcex, 0, sizeof(wcex));
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_VREDRAW | CS_HREDRAW;
	wcex.lpfnWndProc = TapeLPProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hResInstance;
	wcex.hIcon = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = szClassName;
	wcex.hIconSm = LoadIcon(hAppInstance, MAKEINTRESOURCE(IDI_WNDICON));
	RegisterClassEx(&wcex);

	/* EChE쐬 */
	if (bPopupSwnd) {
		dwStyle =	WS_POPUP | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_BORDER;
	}
	else {
		dwStyle = WS_CHILD | WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION
					| WS_MINIMIZEBOX | WS_CLIPSIBLINGS;
	}
	if (bSwndVisible) {
		dwStyle |= WS_VISIBLE;
	}
	hWnd = CreateWindow(szClassName,
						szWndName,
						dwStyle,
						rect.left,
						rect.top,
						rect.right,
						rect.bottom,
						hParent,
						NULL,
						hResInstance,
						NULL);

	/* LȂATCY␳ĎOɒu */
	if (hWnd) {
		GetWindowRect(hWnd, &wrect);
		GetClientRect(hWnd, &crect);
		wrect.right += (rect.right - crect.right);
		wrect.bottom += (rect.bottom - crect.bottom);
		SetWindowPos(hWnd, HWND_TOP, wrect.left, wrect.top,
			wrect.right - wrect.left, wrect.bottom - wrect.top, SWP_NOMOVE);
	}

	/* |bvAbvEChE̓ANeBuEChEOʂɕύX */
	if (bPopupSwnd) {
		SetForegroundWindow(hMainWnd);
	}

	/* ʂA */
	return hWnd;
}

#endif	/* _WIN32 */
